import json
import logging
import os
from argparse import ArgumentParser
from pathlib import Path

from kamikaze3.auto_logrotate.utils import is_path_eco_banned
from kamikaze3.common.args import OUTPUT_FORMAT_JSON, add_default_args
from kamikaze3.common.filesystem import try_read_file, try_read_json_file
from kamikaze3.common.log import setup_logging

IGNORED_DIRECTORIES = (
    "/data/web/.composer",  # composer config / possible cache
    "/data/web/.cache",  # XDG_CACHE_HOME
    "/data/web/.config",  # XDG_CONFIG_HOME
    "/data/web/.local",  # prefix for XDG_DATA_HOME and XDG_STATE_HOME
    "/data/web/.hypernode",
    "/data/web/.vscode-server",
    "/data/web/incidents",  # incident reports
    "/data/web/nginx",  # nginx configs
)

MAGENTO2_COMPOSER_NAMES = (
    "magento/project-community-edition",
    "magento/project-enterprise-edition",
    "mage-os/project-community-edition",
)
MAGENTO2_COMPOSER_DEPS = (
    "magento/product-community-edition",
    "magento/product-enterprise-edition",
    "mage-os/product-community-edition",
)
# See https://getcomposer.org/doc/04-schema.md#type
# Also includes known Magento 2 and Shopware 6 composer types
COMPOSER_NON_PROJECT_TYPES = (
    "composer-plugin",
    "library",
    "magento2-language",
    "magento2-module",
    "magento2-theme",
    "metapackage",
    "php-ext",
    "php-ext-zend",
    "shopware-platform-plugin",
)

logger = logging.getLogger(__name__)


class ApplicationInfo:
    def __init__(
        self,
        path: os.PathLike,
        app_type: str,
        version: str = "unknown",
        features: list[str] | None = None,
    ):
        self.path = path
        self.app_type = app_type
        self.version = version.strip("v")
        self.features = features or []

    def serialize(self):
        return {
            "path": str(self.path),
            "app_type": self.app_type,
            "version": self.version,
            "features": self.features,
        }


class ComposerLock:
    def __init__(self, data: dict):
        self.packages = {}
        for package in data.get("packages", []):
            self.packages[package["name"]] = package

    def get_package_version(self, package_name: str) -> str | None:
        return self.packages.get(package_name, {}).get("version")


def parse_args(args: list[str] | None) -> tuple[bool, str]:
    parser = ArgumentParser(
        description="Scan /data/web for applications",
    )
    add_default_args(parser)
    result = parser.parse_args(args)
    return result.verbose, result.output_format


def is_magento2_project(composer_data: dict, composer_lock: ComposerLock) -> bool:
    if composer_data.get("name", "") in MAGENTO2_COMPOSER_NAMES:
        return True
    return any(
        package in MAGENTO2_COMPOSER_DEPS for package in composer_lock.packages.keys()
    )


def get_magento2_features(app_path: Path) -> list[str]:
    features = []

    config_php_path = app_path / "app/etc/config.php"
    config_contents = try_read_file(config_php_path)
    if config_contents:
        if "Hyva_Theme" in config_contents:
            features.append("hyva")

    return features


def detect_composer_app(app_path: Path) -> ApplicationInfo | None:
    composer_data = try_read_json_file(app_path / "composer.json") or {}
    composer_lock = ComposerLock(try_read_json_file(app_path / "composer.lock") or {})

    if composer_data.get("type") in COMPOSER_NON_PROJECT_TYPES:
        return None

    if is_magento2_project(composer_data, composer_lock):
        # Since there are multiple distributions for Magento 2, we need to check for
        # multiple names and dependencies.
        version = "unknown"
        for dep in MAGENTO2_COMPOSER_DEPS:
            found_version = composer_lock.get_package_version(dep)
            if found_version:
                version = found_version
                break
        return ApplicationInfo(
            path=app_path,
            app_type="magento2",
            version=version,
            features=get_magento2_features(app_path),
        )
    elif "shopware/core" in composer_data.get("require", {}):
        return ApplicationInfo(
            path=app_path,
            app_type="shopware6",
            version=(composer_lock.get_package_version("shopware/core") or "unknown"),
        )
    elif "sylius/sylius" in composer_data.get("require", {}):
        return ApplicationInfo(
            path=app_path,
            app_type="sylius",
            version=(composer_lock.get_package_version("sylius/sylius") or "unknown"),
        )
    elif "laravel/framework" in composer_data.get("require", {}):
        return ApplicationInfo(
            path=app_path,
            app_type="laravel",
            version=(
                composer_lock.get_package_version("laravel/framework") or "unknown"
            ),
        )
    elif "akeneo/pim-community-dev" in composer_data.get("require", {}):
        return ApplicationInfo(
            path=app_path,
            app_type="akeneo",
            version=(
                composer_lock.get_package_version("akeneo/pim-community-dev")
                or "unknown"
            ),
        )
    elif "pimcore/platform-version" in composer_data.get("require", {}):
        return ApplicationInfo(
            path=app_path,
            app_type="pimcore",
            version=(
                composer_lock.get_package_version("pimcore/platform-version")
                or "unknown"
            ),
        )


def detect_legacy_app(path: Path) -> ApplicationInfo | None:
    if os.path.isfile(path / "engine/Shopware/Shopware.php"):
        return ApplicationInfo(
            path=path,
            app_type="shopware5",
            version="unknown",
        )
    elif os.path.isfile(path / "app/Mage.php"):
        return ApplicationInfo(
            path=path,
            app_type="magento1",
            version="unknown",
        )
    elif os.path.isfile(path / "wp-config.php"):
        return ApplicationInfo(
            path=path,
            app_type=(
                "woocommerce"
                if os.path.isdir(path / "wp-content/plugins/woocommerce")
                else "wordpress"
            ),
            version="unknown",
        )
    elif os.path.isfile(path / "src/PrestaShopBundle/PrestaShopBundle.php"):
        return ApplicationInfo(
            path=path,
            app_type="prestashop",
            version="unknown",
        )


def find_applications(path: Path) -> list[ApplicationInfo]:
    result = []
    for dirpath, dirnames, _ in os.walk(path, topdown=True, followlinks=False):
        dirpath = Path(dirpath)
        if str(dirpath) in IGNORED_DIRECTORIES or is_path_eco_banned(dirpath):
            logger.debug(f"Ignoring directory {dirpath}")
            dirnames.clear()  # Stops further traversal of tree
            continue

        if "current" in dirnames and "releases" in dirnames:
            # In a Deployer-like setup, we want to resolve the current directory
            # instead of detecting each release dir. So, we select current and deselect
            # the releases directory.
            if os.path.islink(dirpath / "current"):
                releases_dir = dirpath / "releases"
                if releases_dir == (dirpath / "current").resolve().parent:
                    dirpath = dirpath / "current"
                    dirnames.remove("current")
                    dirnames.remove("releases")

        application_info = detect_composer_app(dirpath) or detect_legacy_app(dirpath)
        if application_info:
            result.append(application_info)
            # Continue traversing the tree, sometimes apps are stored in an app's subdirectory
    return result


def main(args: list[str] | None = None):
    verbose, output_format = parse_args(args)
    setup_logging(logger, verbose)

    result = find_applications(Path("/data/web"))
    if output_format == OUTPUT_FORMAT_JSON:
        result_json = json.dumps(list(map(lambda x: x.serialize(), result)))
        print(result_json)
    else:
        for app in result:
            print(
                f"Application {app.app_type} version {app.version} found at {app.path}"
            )
        if not result:
            print("No applications found.")

    return os.EX_OK


if __name__ == "__main__":
    main()
