"""
kamikaze3.analsysis.cronflock
=============================

20160624 willem@byte.nl
Parse crontab and add flock to commands, if required.

Caveats: flock requires encapsulation of the command.

# this works
* * * * * ! test -e /ikbestaniet && echo "uitroepteken werkt" > /tmp/succes1

# this doesn't
* * * * * flock -n ~/.testflock ! test -e /ikbestaniet && echo "uitroepteken werkt" > /tmp/succes2
# flock: !: No such file or directory

# but with -c, it does work:
* * * * * flock -n ~/.testflock -c '! test -e /ikbestaniet && echo "uitroepteken werkt" > /tmp/succes2'

# and so does variable substitution:
myvar=banaan
* * * * * flock -n ~/.testflock -c 'echo ik test of ik cron vars kan gebruiken; echo hier dan myvar: $myvar'
"""

import argparse
import os
import re
import sys
from hashlib import md5
from pipes import quote
from subprocess import PIPE, STDOUT, CalledProcessError, Popen, check_output

from kamikaze3.common.cron import split_crontab_cmd_line
from kamikaze3.common.env import get_app_name

CRONCMD = "/usr/bin/crontab"
IGNORE_KEYWORDS = (
    "flock",
    "lockrun",
    "hypernode-image-optimizer",
)
IGNORE_KEYWORDS_SLEEP = ("sleep",)


def get_cron(user: str) -> str:
    """Return crontab contents for the specified user.
    return empty string if no crontab exists.

    :raises: CalledProcessError on unknown errors
    """
    try:
        return check_output([CRONCMD, "-l", "-u", user], stderr=STDOUT).decode("utf-8")
    except CalledProcessError as err:
        if err.output is None:
            raise err
        err_msg = (
            err.output.decode("utf-8")
            if isinstance(err.output, bytes)
            else "{}".format(err.output)
        )
        if not err_msg.startswith("no crontab for"):
            raise err
    return ""


def write_cron(crontab: str, user: str) -> None:
    """Write contents of crontab for the specified user.
    :raises: RuntimeError on errors saving the crontab
    """
    proc = Popen([CRONCMD, "-u", user, "-"], stdout=PIPE, stdin=PIPE, stderr=PIPE)
    stdout, stderr = proc.communicate(input=crontab.encode("utf-8"))
    statuscode = proc.wait()
    if statuscode:
        raise RuntimeError(
            "Failed to saving cron. Status {} and output: {!r} {!r}".format(
                statuscode, stdout, stderr
            )
        )


def _can_flockerify(line: str) -> bool:
    for kw in IGNORE_KEYWORDS:
        if kw in line:
            return False
    return True


def _can_sleepify(line: str) -> bool:
    for kw in IGNORE_KEYWORDS_SLEEP:
        if kw in line:
            return False
    return True


def get_sleep_time(max_: int = 30) -> float:
    """Return the time to sleep in flockerify_line. This is deterministic
    per node, but different between nodes
    """
    return int(md5(get_app_name().encode("utf-8")).hexdigest(), 16) % max_


def flockerify_line(line: str, sleep: bool = False) -> str:
    """Modify a cron job, calling flock to make sure same cron job won't run
    concurrently, also adds a deterministic sleep call before calling the command.
    We see CPU steal issues on Combell OpenStack and suspect all crons running
    at the same time could contribute to that. The sleep would be
    different for Hypernodes, so spreads the load.

    Validates cron line and raises ValueErrors.
    Returns the cron job line prefixed with a flock call.

    :raises: ValueError on invalid cron lines
    """
    timespec, cmd, comment = split_crontab_cmd_line(line)
    if not cmd:
        # not a cmd line
        return line

    if comment:
        comment = " #" + comment

    new_cmd = cmd
    # In case of the Magento cron it makes sense to default to using the -o
    # flag for the flock. This is kind of a broad setting so only want to
    # apply it in the case of the bin/magento cron (like for example for
    # the Magento rabbitmq consumers) and not for all crons ever as that
    # would potentially cause too many crons to be unlocked while they
    # could be locked.
    close_flag = " -o" if "bin/magento cron" in line else ""
    if _can_flockerify(line):
        lockfile = "~/.{}.lock".format(md5(cmd.encode("utf-8")).hexdigest()[:8])
        new_cmd = "flock{} -n {} -c {}".format(close_flag, lockfile, quote(cmd))

    if sleep and _can_sleepify(line):
        new_cmd = "sleep {}; {}".format(get_sleep_time(), new_cmd)

    return "{} {}{}\n".format(timespec, new_cmd, comment)


def flockerify(crontab: str, sleep: bool = False) -> str:
    """Return a modified crontab, prefixing the command with flock to make sure same cron job won't run
    concurrently, also adds a deterministic sleep call before calling the command.
    We see CPU steal issues on Combell OpenStack and suspect all crons running
    at the same time could contribute to that. The sleep would be
    different for Hypernodes, so spreads the load.

    Also removes white blocks (multiple new lines larger than 10 lines) from crontabs.
    There is no reason to have such blocks, and we deployed a buggy version of this cronflock
    that created such blocks on crontabs on some Hypernodes, so this features also clears those
    blocks.
    """
    new_crontab = "".join(
        [flockerify_line(line, sleep) for line in crontab.splitlines(True)]
    )
    return re.sub(r"\n{10,}", "\n", new_crontab, re.MULTILINE)  # remove white blocks


def main(args=None) -> int:
    """parse args and run the program and return the exit code"""

    parser = argparse.ArgumentParser(
        description="Add 'flock' to cron commands so they cannot overlap"
    )
    parser.add_argument(
        "--user", help='Which user to modify cron for? Default "app"', default="app"
    )
    parser.add_argument(
        "--write",
        help="Save changes to disk. Otherwise, print.",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--sleep",
        help="Prefix the cronjob with a sleep",
        action="store_true",
        default=False,
    )
    # @TODO: this --file option seems unnecessary, we may remove it
    parser.add_argument("--file", help=argparse.SUPPRESS)
    args = parser.parse_args(args)

    if args.file:
        with open(args.file) as fh:
            oldcron = fh.read()
    else:
        oldcron = get_cron(args.user)

    if not oldcron:
        print("No cron found, doing nothing.")
        return os.EX_OK

    newcron = flockerify(oldcron, args.sleep)
    if args.file:
        print(newcron)
    elif args.write:
        if oldcron != newcron:
            print("Found updates, saving to disk.")
            write_cron(newcron, args.user)
        else:
            print("Couldn't flockerize anything, doing nothing.")
    else:
        print(newcron)
    return os.EX_OK


if __name__ == "__main__":
    sys.exit(main())
