import json
import os
import subprocess
import GeoIP

from argparse import ArgumentParser
from collections import defaultdict
from contextlib import suppress
from glob import glob
from logging import getLogger
from time import time
from typing import Any, Dict, List

from psutil import AccessDenied, NoSuchProcess, Process

from kamikaze3.common.env import get_hn_app_var
from kamikaze3.common.args import (
    OUTPUT_FORMAT_JSON,
    OUTPUT_FORMAT_TEXT,
    add_default_args,
)
from kamikaze3.common.cluster import get_child_apps_by_relation_type
from kamikaze3.common.env import get_app_name, get_hn_common_var, get_node_roles
from kamikaze3.common.filesystem import read_file, read_last_lines
from kamikaze3.common.psutil import get_tcp_connections, pconn_tcp
from kamikaze3.settings import ROLE_LOADBALANCER

gi = GeoIP.new(GeoIP.GEOIP_MEMORY_CACHE)
logger = getLogger(__name__)

DEFAULT_PHP_HANDLER = "phpfpm"


def determine_worker_status(proc: Process, connections: List[pconn_tcp]):
    """
    Get status of a FPM worker process

    The status is based on the state of the TCP connection it has to port 9000. If the process is
    idle, there will only be one connection in the LISTEN state. If the process is busy it has more
    connections. When the connection is CLOSE_WAIT, the client has already left but the process
    doesn't realise it yet or keeps working without closing its side of the connection.

    This function takes a list of pconn_tcp connections instead of querying proc.connections(),
    because proc.connections() is very slow when the server has many fpm workers and connections.
    So instead, before going through all the workers, we fetch the TCP connections.
    It's a bit less real time, but it's needed to make this module performant on high load environments.

    :param psutil.Process proc: Worker process object
    :param list[pconn_tcp] connections: List of connections
    :rtype: str
    """
    if not connections:
        return "-"

    try:
        conns = [
            conn
            for conn in connections
            if conn.pid == proc.pid and conn.laddr == ("127.0.0.1", 9000)
        ]
    except (NoSuchProcess, AccessDenied):
        # Proc is already gone
        return "-"

    if any(conn.status == "ESTABLISHED" for conn in conns):
        return "BUSY"
    if any(conn.status == "CLOSE_WAIT" for conn in conns):
        return "GONE"
    return "IDLE"


def worker_as_text(worker):
    """Return the status of the worker as a line of text

    :param dict worker: last request info dict of the worker
    :return: line of the worker status
    :rtype: str
    """
    # Python 3.6 str.format(**kwargs) is not playing well with defaultdict, still
    # rasing KeyErrors on missing keys.
    # So instead of a defaultdict, we'll use a normal dict and ensure required keys
    # exist and are empty string by default.
    safe_worker = dict()
    safe_worker.update(worker)

    for key in (
        "pid",
        "worker_status",
        "handler",
        "time",
        "remote_addr",
        "method",
        "http_host",
        "uri",
        "useragent",
        "country",
    ):
        if key not in safe_worker or safe_worker[key] is None:
            safe_worker[key] = ""

    with suppress(ValueError, TypeError):
        safe_worker["time"] = "{:5.1f}".format(safe_worker["time"])

    text = ""
    if safe_worker["app_name"] != get_app_name():
        text = "{:10s} ".format(
            safe_worker["app_name"][-min(10, len(safe_worker["app_name"])) :]
        )

    text = (
        text
        + "{pid:05d} {worker_status:4.4s} {time:5s}s {country:2.2s} {handler:10s}".format(
            **safe_worker
        )
    )
    return (
        text
        + "{remote_addr:15s} {method:4s} {http_host}{uri}   ({useragent})".format(
            **safe_worker
        )
    )


class FpmSource(object):
    """The source of information regarding FPM processes"""

    def __init__(self):
        self.pid_base_path = "/var/run"
        self.log_base_path = "/var/log/hypernode/fpm-inspector"

    def _pid_path(self, sub_path):
        """Return the absolute path to the provided subpath from PID base path.
        Note: Does not check if the path exists.

        :param sub_path: sub path under the base PID path
        :rtype: str
        """
        return os.path.abspath(os.path.join(self.pid_base_path, sub_path))

    def get_possible_fpm_pidfiles(self):
        return glob(self._pid_path("php/php[0-9].[0-9]-*fpm.pid"))

    def get_fpm_processes(self):
        """Get FPM processes
        :rtype: list psutil.Process
        """
        processes = []

        with suppress(IOError, NoSuchProcess, ValueError):
            pids = self.get_possible_fpm_pidfiles()
            for pid in pids:
                pid = int(read_file(pid))
                processes.append(Process(pid))

        return processes

    def get_process_last_logs(self, pid, max_logs=2, line_size=1024):
        """Get the last logs of an FPM process
        Return None if failed to read the log file

        :Note: the estimated line_size is 1KB, but it may happen to have larger
        useragents. So we may need to increase the line size in future.

        :param pid: The pid of the process
        :param max_logs: max number of logs
        :param line_size: estimated size of each line
        :return: The last logs
        :rtype: list[dict[str, V]] | None
        """
        filename = os.path.join(self.log_base_path, "{}.log".format(pid))
        block_size = (max_logs * line_size) + max_logs  # consider EOL chars

        last_lines = read_last_lines(filename, max_logs, block_size)
        if last_lines is None:
            return None

        def try_parse_json(line):
            with suppress(ValueError, UnicodeDecodeError):
                return json.loads(line)
            return None

        last_logs = [try_parse_json(line) for line in last_lines]
        return list(filter(None, last_logs))

    def get_last_request_info(self, pid) -> Dict[str, Any]:
        """Get details of the last request event of an FPM worker.

        :param int pid: Process ID of the FPM worker
        :return: Dictionary containing all kinds of request properties
        :rtype: dict[str, V]
        """
        logs = self.get_process_last_logs(pid, 2)
        request_info = defaultdict(lambda: "-")
        if not logs:
            return request_info

        """
        Fields come from: https://github.com/ByteInternet/hypernode-control/blob/master/ansible-playbooks/roles/magweb/files/fpm-inspector/prepend.php
            timestamp = $_SERVER['REQUEST_TIME_FLOAT']

        # last is an end-request line
        """

        last_log = logs[-1]
        last_log_timestamp = last_log.get("timestamp")
        if last_log["event"] == "start":
            # Last request has not finished yet
            last_log["time"] = (
                "-" if (last_log_timestamp is None) else (time() - last_log_timestamp)
            )
        elif len(logs) > 1:
            previous_log = logs[-2]
            assert previous_log["event"] == "start"
            previous_log_timestamp = previous_log.get("timestamp")
            last_log["time"] = (
                "-"
                if (previous_log_timestamp is None)
                else (last_log_timestamp - previous_log_timestamp)
            )
        else:
            # Apparently we only have the "end" event in the log file, so can't calculate the time
            last_log["time"] = "-"

        for key, value in last_log.items():
            request_info[key] = value if value is not None else "-"
        return request_info

    def get_local_scoreboard(self, fpm_pool=None):
        """Return the information on last requests of all worker process,
        sorted by fastest to slowest.

        :param str fpm_pool: Name of the fpm pool to filter on, default None for all pools
        :rtype: list[dict]
        """
        processes = self.get_fpm_processes()
        if not processes:
            logger.warning("Unable to get FPM status. Could not find master process!")
            return []

        result = []
        connections = get_tcp_connections()
        for master_process in processes:
            for worker in master_process.children():
                request = self.get_last_request_info(worker.pid)
                if fpm_pool and request["handler"] != fpm_pool:
                    continue

                request["app_name"] = get_app_name()
                request["pid"] = worker.pid
                request["worker_status"] = determine_worker_status(worker, connections)
                request["country"] = "-"
                if "remote_addr" in request and request["remote_addr"]:
                    with suppress(Exception):
                        country = gi.country_code_by_addr(request["remote_addr"])
                        if country:
                            request["country"] = country
                result.append(request)
        return result

    @staticmethod
    def get_cluster_scoreboard(fpm_pool=None):
        result = []
        child_apps = get_child_apps_by_relation_type(ROLE_LOADBALANCER)
        for app in child_apps:
            try:
                command = ["ssh", f"{app}.h", "hypernode-fpm-status-json"]
                if fpm_pool:
                    command.extend(["--fpm-pool", fpm_pool])
                data = subprocess.check_output(command, encoding="utf-8")
                result.extend(json.loads(data))
            except subprocess.CalledProcessError:
                logger.warning(f"Failed to get FPM status from {app}")
                continue
        return result

    def get_scoreboard(self, fpm_pool=None):
        if ROLE_LOADBALANCER in get_node_roles():
            result = self.get_cluster_scoreboard(fpm_pool)
        else:
            result = self.get_local_scoreboard(fpm_pool)

        # Python 3 does not support sorting of heterogeneous collections.
        # 'time' is numeric or '-' if unclear, for sorting let's assume '-' is 0
        result.sort(key=lambda x: 0 if x["time"] == "-" else x["time"], reverse=True)
        return result

    def get_status_report(self, fpm_pool=None):
        """
        :param str fpm_pool: Name of the fpm pool to filter on, default None for all pools
        """
        return "\n".join(map(worker_as_text, self.get_scoreboard(fpm_pool)))

    def get_status_report_json(self, fpm_pool=None):
        return json.dumps(self.get_scoreboard(fpm_pool), indent=4, sort_keys=True)


def print_fpm_report(fpm_pool=None, output_format=OUTPUT_FORMAT_TEXT):
    """
    :param str fpm_pool: Name of the fpm pool to filter on, default None for all pools
    :param str output_format: Output format (text or json)
    """
    if output_format == OUTPUT_FORMAT_JSON:
        print(FpmSource().get_status_report_json(fpm_pool))
    else:
        print(FpmSource().get_status_report(fpm_pool=fpm_pool))


def print_fpm_report_json():
    print(FpmSource().get_status_report_json())


def get_possible_php_handlers():
    """
    Return the possible PHP versions on the system
    :return list php_versions: The possible php versions.
    For example ['7.1', '7.2']
    """
    handlers = [DEFAULT_PHP_HANDLER]
    try:
        alternative_php_versions = get_hn_app_var(
            "alternative_php_versions", default_val=[]
        )
        for alternative_php_version in alternative_php_versions:
            alternative_php_version = "".join(alternative_php_version.split("."))
            handlers.append(f"phpfpm{alternative_php_version}alt")
    except RuntimeError:
        return handlers

    return handlers


def parse_arguments(args=None):
    """parse args and run the program and return the exit code"""
    parser = ArgumentParser(description="Show the current status of the FPM workers")
    parser.add_argument(
        "--fpm-pool",
        help="Specify the fpm-pool you want to see the status of. If not specified it will "
        "show all the pools by default",
        choices=get_possible_php_handlers(),
        required=False,
    )
    add_default_args(parser)

    return parser.parse_args(args)


def main(args=None) -> int:
    args = parse_arguments(args)

    print_fpm_report(fpm_pool=args.fpm_pool, output_format=args.output_format)

    return os.EX_OK
