import argparse
import logging
import os
import shutil
import sys
import tempfile
from hashlib import md5
from pathlib import Path
from subprocess import STDOUT, CalledProcessError, check_output

import psutil

from kamikaze3.onboarding import try_call_event_hook
from kamikaze3.utils import get_lock

"""
hypernode-image-optimizer

To be used by customers in cron, like this:

10 4 * * * hypernode-image-optimizer --write --newonly /data/web/public/media

Experimental mode uses pngquant and jpegoptim which achieve much higher
reduction with comparable quality.

Done:
- Do not use `convert` in --experimental mode when resizing is not required,
  as that saves some quality conversion loss + extra copy.
- Use pngquant in --experimental mode
- Use jpegoptim in --experimental
- Print partial summary when ^C
- KB -> MB
- Pretty warning when path doesn't exist
- Images should retain ownership
- Resize default 1000 -> 2000

"""

LAST_RUN_FILE = ".hypernode-image-optimizer-last-run"
TMP_DIR = "/run/shm"  # ramdisk to save on I/O writes
MINSIZE = 100  # bytes


class ConvertError(Exception):
    pass


class BaseOptimizer(object):
    def __init__(self, max_size, quality, write=False):
        self.max_size = max_size
        self.quality = quality
        self.write = write

    def _optimize(self, src, dst):
        raise RuntimeError("not implemented in base class")

    def _resize(self, src, dst):
        dimension = "%sx%s>" % (self.max_size, self.max_size)
        # http://stackoverflow.com/a/7262050/604515
        check_output(
            [
                "/usr/bin/convert",
                src,
                "-strip",
                "-interlace",
                "none",
                "-quality",
                str(self.quality),
                "-resize",
                str(dimension),
                dst,
            ],
            stderr=STDOUT,
        )

    def optimize(self, path):
        with tempfile.NamedTemporaryFile(dir=TMP_DIR) as tmp_img:
            tmp_img_path = tmp_img.name

            try:
                self._optimize(path, tmp_img_path)
            except CalledProcessError as e:
                raise ConvertError(
                    "Warning: could not optimize %s. Will skip. (%s)" % (path, e.output)
                )

            shutil.copystat(path, tmp_img_path)

            old_size = _get_filesize(path)
            new_size = _get_filesize(tmp_img_path)

            profit_abs = old_size - new_size

            if profit_abs < 0:
                profit_abs = 0

            profit_rel = _perc(profit_abs, old_size)

            print(
                "[{:2.0f}%] {} ({:.0f} KB smaller)".format(
                    profit_rel, repr(path)[1:-1], _to_kb(profit_abs)
                ),
                file=sys.stderr,
            )

            # only do something if profit > 1%, otherwise we might get endless loops of jpeg quality reduction
            if self.write and profit_rel > 1:
                # copy mtime and permissions as well
                shutil.copy2(tmp_img_path, path)

        return profit_abs, old_size


class SafeOptimizer(BaseOptimizer):
    # Just uses ImageMagick convert, but doesn't give very good results.

    def _optimize(self, src, dst):
        self._resize(src, dst)


class ExperimentalOptimizer(BaseOptimizer):
    # Uses pngquant and jpegoptim

    def is_resize_required(self, path):
        try:
            width, height = check_output(
                ["/usr/bin/identify", "-format", "%w %h", path]
            ).split()
            width, height = int(width), int(height)
            return width > self.max_size or height > self.max_size
        except ValueError:
            # Failed to parse identify output into width and height, skip this image
            return False

    def _optimize(self, src, dst):
        # 1. resize if required
        # 2. png -> pngquant
        # 3. jpg -> jpegoptim

        img_type = image_type(src)
        assert img_type in ("png", "jpg"), "unknown image: %s" % src

        # This hocus pocus is required, because the jpegoptim and pngquant
        # dont have rescale capabilities. To save one read/write cycle for
        # small images, we make rescale conditional. --WdG
        if self.is_resize_required(src):
            self._resize(src, dst)
            # overwrite temporary file (result from convert)
            getattr(self, img_type)(dst, dst)
        else:
            # convert original to temporary, call self.png() or self.jpg()
            getattr(self, img_type)(src, dst)

    def png(self, src, dst):
        check_output(
            ["/usr/bin/pngquant", src, "--output", dst, "--force"], stderr=STDOUT
        )

    def jpg(self, src, dst):
        # jpegoptim has no way to specify the destination filename but it can optimize in-place
        # so we copy to the given tmp file first
        if src != dst:
            shutil.copy2(src, dst)
        check_output(
            ["/usr/bin/jpegoptim", dst, "--max=%d" % self.quality, "--strip-all"],
            stderr=STDOUT,
        )


def parse_args(args=None):
    """
    :param list[str] args: a list or arguments. eg: ["/some/path", "--size", "2000"]
    """
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description="Automatically optimize PNG and JPG files under a directory tree. "
        "Make sure you have a backup of your images, and visually inspect "
        "the results before applying for real!",
    )
    parser.add_argument("path", help="File or dir to (recursively) process")
    parser.add_argument(
        "-w",
        "--write",
        help="Actually overwrite old files with new files",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "-q",
        "--quality",
        help="Quality percentage for lossy compression",
        type=int,
        default=90,
    )
    parser.add_argument(
        "-s",
        "--size",
        help="Downscale images that are bigger than x pixels",
        type=int,
        default=2000,
    )
    parser.add_argument(
        "-n",
        "--newonly",
        help="Only consider files that were modified/created since "
        "last successful run. To maintain state, I will create a "
        "{} file under the given path. Useful for cron.".format(LAST_RUN_FILE),
        action="store_true",
    )
    parser.add_argument(
        "-x",
        "--experimental",
        help="Use better but experimental compression",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--exclude",
        help="Specify one or more paths in which the files will be ignored. Space delimited.",
        nargs="+",
    )

    parsed_args = parser.parse_args(args)

    if parsed_args.newonly and not os.path.isdir(parsed_args.path):
        logging.error("I can only use --newonly on a directory!")
        sys.exit(1)

    if not os.path.exists(parsed_args.path):
        logging.error("{} does not exist!".format(parsed_args.path))
        sys.exit(1)

    return parsed_args


def image_type(fname):
    ext = fname.lower().split(".")[-1]
    ext = ext.replace("jpeg", "jpg")

    if ext in ["png", "jpg"]:
        return ext

    return None


def _to_mb(val):
    return float(val) / 1024 / 1024


def _to_kb(val):
    return float(val) / 1024


def _perc(x, y):
    if y == 0:
        return 0

    return float(x) / y * 100


def _get_filesize(path):
    # separate function, because otherwise mocking os.stat will mess up nose's internals --WdG
    return os.stat(path).st_size


def find_targets(root_path, ignore_older_than=None, exclude=None):
    """
     Produce an iterator for all the images recursively found under root_path.
     If root_path is a file, return an iterator with just that.
     Param ignore_older_than is an epoch timestamp from os.path.getmtime()

    :param str root_path: The root path to look in
    :param int ignore_older_than: Timestamp to ignore older than
    :param iter[str, ..] exclude: Iterable of paths to ignore all children from
    For example, if /data/web/public is root path and exclude contains /data/web/public/media
    then all images in media would be skipped.
    """

    if os.path.isfile(root_path):
        yield root_path

    elif os.path.isdir(root_path):
        for root, dirs, files in os.walk(root_path):
            for myfile in files:
                path = os.path.join(root, myfile)

                if not image_type(path):
                    continue

                if os.stat(path).st_size < MINSIZE:
                    continue

                if ignore_older_than is not None:
                    if os.path.getmtime(path) < ignore_older_than:
                        continue

                if exclude:
                    # skip any file that is in any of the directories specified in the exclude list
                    if any(
                        Path(excluded_path) in Path(path).parents
                        for excluded_path in exclude
                    ):
                        continue

                yield path

    else:
        raise RuntimeError("%s is not a file or directory" % root_path)


def find_last_run_timestamp(path):
    fname = os.path.join(path, LAST_RUN_FILE)

    if not os.path.isdir(path):
        return None

    if not os.path.isfile(fname):
        return None

    return os.path.getmtime(path)


def write_last_run_timestamp(path):
    fname = os.path.join(path, LAST_RUN_FILE)

    if os.path.exists(fname):
        os.unlink(fname)

    return open(fname, "a").close()


def send_onboarding_image_optimizer_event():
    try_call_event_hook(["imageoptimizer"])


def main():
    args = parse_args()
    # since the path is an arbitrary value and could be too long
    # we use a hash to keep the length of the lockname static and short
    lock_name = md5(
        "hypernode-image-optimizer: {}".format(args.path).encode("utf-8")
    ).hexdigest()

    if not get_lock(lock_name):
        print(
            "I am already running for {}, perhaps in cron? Refusing to duplicate myself.".format(
                args.path
            )
        )
        sys.exit(0)

    total_bytes = 0
    total_bytes_profit = 0
    num_files = 0

    if args.newonly:
        ignore_older_than = find_last_run_timestamp(args.path)
    else:
        ignore_older_than = None

    # don't swamp the machine, as resizing is pretty I/O demanding
    mylife = psutil.Process()
    mylife.ionice(psutil.IOPRIO_CLASS_IDLE)

    try:
        optimizer_class = ExperimentalOptimizer if args.experimental else SafeOptimizer
        optimizer = optimizer_class(
            max_size=args.size, quality=args.quality, write=args.write
        )

        for path in find_targets(
            args.path, ignore_older_than=ignore_older_than, exclude=args.exclude
        ):
            try:
                path_profit, path_size = optimizer.optimize(path)
            except ConvertError as e:
                print(str(e))
                continue

            num_files += 1
            total_bytes += path_size
            total_bytes_profit += path_profit

        if args.newonly and args.write:
            write_last_run_timestamp(args.path)
    except KeyboardInterrupt:
        pass

    if not num_files:
        print("No optimizable image files found under %s" % args.path)
        return 0

    print(
        "\n{} optimization profit over {:d} files: {:.0f} MB ({:.0f}%)".format(
            "Experimental" if args.experimental else "Safe",
            num_files,
            _to_mb(total_bytes_profit),
            _perc(total_bytes_profit, total_bytes),
        )
    )

    if args.write:
        print("I have written the optimized files to disk.")
    else:
        print("I have not written anything to disk, use --write for that.")
    send_onboarding_image_optimizer_event()
    return 0


if __name__ == "__main__":
    sys.exit(main())
