"""
kamikaze3.analysis.mail
=======================
functionality to analyze system mail logs, etc.
"""

import html
import sys
from argparse import ArgumentParser, Namespace
from datetime import datetime, timedelta
from logging import DEBUG, INFO, getLogger
from os import (
    EX_CONFIG,
    EX_NOINPUT,
    EX_OK,
    EX_OSERR,
    EX_SOFTWARE,
    EX_TEMPFAIL,
    EX_UNAVAILABLE,
)
from typing import List, Sequence

from kamikaze3.common.log import create_stream_handler, create_syslog_handler
from kamikaze3.common.postfix import (
    check_dns_for_domain,
    find_failed_transip_mails_from_logs,
    find_relevant_mail_domains_from_logs,
)
from kamikaze3.common.system import user_is_root
from kamikaze3.onboarding import try_call_event_hook

EX_MAIL_ISSUES = 1

MAX_ISSUES_IN_NOTIFICATION = (
    10  # we don't expect to report more per email, it won't be a helpful message
)
MAX_MAIL_ISSUES_PER_HOURS = (
    5000  # we don't expect to have more emails per hour on Hypernodes
)

logger = getLogger(__name__)


def parse_args(args: Sequence[str] = None) -> Namespace:
    parser = ArgumentParser(description="analyze system mail logs, status, etc.")
    parser.add_argument("-v", "--verbose", action="store_true")
    parser.add_argument("-q", "--quiet", action="store_true")

    subparsers = parser.add_subparsers(
        dest="action",  # namespace.action will have the name of the subcommand
        title="Available action subcommands",
        description="A subcommand with own arguments runs a specific action",
    )

    if user_is_root():
        failed_parser = subparsers.add_parser(
            "failed",
            help="look for all failed mail logs (spf, auth) and notify " "the client",
        )
        failed_parser.add_argument(
            "--no-notify", action="store_true", help="do not notify the client"
        )
        failed_parser.add_argument(
            "--since-hours",
            type=int,
            default=1,
            help="look for issues in logs only after these hours ago",
        )

    domains_parser = subparsers.add_parser(
        "domains", help="scan the mail logs for domains used to mail from"
    )
    domains_parser.add_argument(
        "--check-dns",
        action="store_true",
        help="check if dns is correct for the detected domains (SPF, DKIM, ...)",
    )

    return parser.parse_args(args)


def notify_client_of_mail_auth_issues(issues: List[str]) -> None:
    """Send an event to event API to notify client of mail authentication issues."""

    if not issues:
        logger.warning("skipping sending mail auth issue event. no issues specified")
        return

    issue = format_notification_issues(issues[:MAX_ISSUES_IN_NOTIFICATION])
    event = {"mail-auth-failure": issue}
    logger.debug("sending mail auth issue event: {}".format(issue))
    try_call_event_hook(event)
    logger.info("mail auth issues event sent")


def notify_client_of_mail_spf_issues(issues: List[str]) -> None:
    """Send an event to event API to notify client of mail SPF issues."""

    if not issues:
        logger.warning("skipping sending mail SPF issue event. no issues specified")
        return

    issue = format_notification_issues(issues[:MAX_ISSUES_IN_NOTIFICATION])
    event = {"mail-spf-failure": issue}
    logger.debug("sending mail SPF issue event: {}".format(issue))
    try_call_event_hook(event)
    logger.info("mail SPF issues event sent")


def notify_client_of_mail_dkim_issues(issues: List[str]) -> None:
    """Send an event to event API to notify client of mail DKIM issues."""

    if not issues:
        logger.warning("skipping sending mail DKIM issue event. no issues specified")
        return

    issue = format_notification_issues(issues[:MAX_ISSUES_IN_NOTIFICATION])
    event = {"mail-dkim-failure": issue}
    logger.debug("sending mail DKIM issue event: {}".format(issue))
    try_call_event_hook(event)
    logger.info("mail DKIM issues event sent")


def format_notification_issues(issues: List[str]) -> str:
    """Format issues for the notifications"""
    return "<ul>{}</ul>".format(
        "".join(["<li>{}</li>".format(html.escape(issue)) for issue in issues])
    )


def mail_cmd_failed(since_hours: int = 1, notify: bool = True) -> int:
    """Subcommand to check mail issues"""

    logger.info("checking failed mail issues ...")
    after = datetime.now() - timedelta(hours=since_hours)
    mail_issues = find_failed_transip_mails_from_logs(
        after, limit=MAX_MAIL_ISSUES_PER_HOURS * since_hours
    )
    logger.info("found {} failed mail issue logs".format(len(mail_issues)))

    # using sets instead of lists for faster contain asserts
    auth_issues = set()
    dkim_issues = set()
    spf_issues = set()
    exit_code = EX_OK
    for mail_issue in mail_issues:
        if "SPF records for hypernode.io could not be verified" in mail_issue:
            continue
        if "x-transip-mail-auth" in mail_issue:
            auth_issues.add(mail_issue)
        elif "failed DKIM verification" in mail_issue:
            dkim_issues.add(mail_issue)
        else:
            spf_issues.add(mail_issue)

    if len(auth_issues) > 0:
        logger.warning("found {} mail auth issues".format(len(auth_issues)))
        exit_code = EX_MAIL_ISSUES
        if notify:
            notify_client_of_mail_auth_issues(list(auth_issues))

    if len(dkim_issues) > 0:
        logger.warning("found {} mail DKIM issues".format(len(dkim_issues)))
        exit_code = EX_MAIL_ISSUES
        if notify:
            notify_client_of_mail_dkim_issues(list(dkim_issues))

    if len(spf_issues) > 0:
        logger.warning("found {} mail SPF issues".format(len(spf_issues)))
        exit_code = EX_MAIL_ISSUES
        if notify:
            notify_client_of_mail_spf_issues(list(spf_issues))

    return exit_code


def mail_cmd_domains(check_dns=False):
    """Subcommand to check domains used to mail from"""

    # TODO: refactor into smaller functions. We nest too deep here.
    logger.debug("checking logs for domains used to send mail from ...")
    domains = sorted(find_relevant_mail_domains_from_logs())
    exit_code = EX_OK
    if domains:
        logger.debug("domains used to mail from (excluding noreply@hypernode.io):")
        for domain in domains:
            output = domain
            if check_dns:
                try:
                    checks = check_dns_for_domain(domain)
                    if any([result != "ok" for result in checks.values()]):
                        # If any result returned an API error keep exit code EX_TEMPFAIL
                        exit_code = (
                            EX_TEMPFAIL if exit_code == EX_TEMPFAIL else EX_CONFIG
                        )
                    check_results = ", ".join([f"{k}: {v}" for k, v in checks.items()])
                    output += " - {}".format(check_results)
                except Exception as e:
                    logger.debug(
                        "Caught and passed on exception for domain '{}', but if we were not "
                        "running verbose this would have been suppressed. The exception was: "
                        "{}".format(domain, e)
                    )
                    output += " - Could not check domain"
                    exit_code = EX_TEMPFAIL
            logger.info(output)
    else:
        logger.debug("did not find any sent mails in the logs")
    return exit_code


def main(args: Sequence[str] = None) -> int:
    """Run the program with given arguments and return the exit code"""

    opts = parse_args(args)
    logger.setLevel(DEBUG)
    if not opts.quiet:
        log_level = DEBUG if opts.verbose else INFO
        logger.addHandler(
            create_stream_handler(level=log_level, log_format="%(message)s")
        )
    logger.addHandler(
        create_syslog_handler(
            level=DEBUG,
            name_override="kamikaze3.analysis.mail" if __name__ == "__main__" else None,
        )
    )

    if not opts.action:
        print("no action is specified. see --help for more info'", file=sys.stderr)
        return EX_NOINPUT

    try:
        if opts.action in ("failed", "failed-auth"):
            return mail_cmd_failed(opts.since_hours, not opts.no_notify)

        elif opts.action == "domains":
            return mail_cmd_domains(check_dns=opts.check_dns)
        print(
            "no handler is defined for action '{}'".format(opts.action), file=sys.stderr
        )
        return EX_UNAVAILABLE
    except Exception as err:
        logger.exception(err)
        if isinstance(err, OSError):
            return EX_OSERR
        if isinstance(err, RuntimeError):
            return EX_TEMPFAIL
        return EX_SOFTWARE


if __name__ == "__main__":
    sys.exit(main())
