import json
import logging
import os
import re
from datetime import datetime
from typing import List, Tuple

from kamikaze3.common.env import get_magento_2_install_dir
from kamikaze3.common.filesystem import read_json_file
from kamikaze3.common.system import user_is_root
from kamikaze3.onboarding import try_call_event_hook
from kamikaze3.settings import HYPERNODE_APP_HOMEDIR, WEBAPP_CACHE_STATUS_STATEFILE_PATH

logger = logging.getLogger(__name__)

MAX_CACHE_STATUSES = 30
MAX_CACHES_CHANGED = 15


def read_magento2_config_file(magento2_config_file: str) -> List[str]:
    """
    Read the Magento 2 config file
    :param magento2_config_file: The env.php
    :return list[str]: The config file lines if any
    """
    try:
        logger.debug(
            "Trying to open the Magento 2 configuration file "
            "in '{}'".format(magento2_config_file)
        )
        with open(magento2_config_file) as f:
            return f.readlines()
    except FileNotFoundError:
        raise RuntimeError(
            "Magento 2 config file '{}' does not exist!".format(magento2_config_file)
        )


def filter_cache_lines_from_magento2_config_lines(config_lines: List[str]) -> List[str]:
    """
    Filter the relevant config lines from the Magento 2 config file
    :param list[str] config_lines: The lines from the config
    :return list[str]: The relevant cache lines
    """

    def remove_whitespace(string_to_remove_whitespace_from: str) -> str:
        return re.sub(r"\s+", "", string_to_remove_whitespace_from)

    def check_in_cache_context(current_line):
        if remove_whitespace("'cache_types' => [") in current_line:
            return True
        if remove_whitespace('"cache_types" => [') in current_line:
            return True
        return False

    cache_lines = []
    in_cache_context = False
    for line in config_lines:
        current_line = remove_whitespace(line)
        if check_in_cache_context(current_line):
            in_cache_context = True
            continue
        if in_cache_context:
            if "]" in remove_whitespace(line):
                return cache_lines
            cache_lines.append(line)
    if in_cache_context:
        raise RuntimeError(
            "Failed to parse cache status from config, looks like the "
            "cache_types array was not closed correctly"
        )
    if not cache_lines:
        raise RuntimeError(
            "Magento 2 config did not appear to contain any cache config lines"
        )


def get_cache_status_from_magento2_cache_lines(cache_lines: List[str]) -> dict:
    """
    Get the cache status from the config lines that represent cache status
    in the Magento 2 config file.
    :param list[str] cache_lines: The relevant cache lines
    :return dict: The cache status
    """
    cache_status = {}
    for line in cache_lines:
        cache_line_parts = line.split("=>")
        if len(cache_line_parts) != 2:
            raise RuntimeError(
                "Unexpected cache line format " "encountered in line: '{}'".format(line)
            )
        cache_name, cache_value = [
            re.sub(r"['\",]+", "", c).strip() for c in cache_line_parts
        ]
        if cache_value not in ("1", "0"):
            raise RuntimeError(
                "Unexpected cache value encountered " "in line: '{}'".format(line)
            )
        cache_status[cache_name] = bool(int(cache_value))
    return cache_status


def determine_magento2_cache_status(installation_path: str) -> dict:
    """
    Check the cache status of a Magento 2 installation
    :param str installation_path: The Magento 2 installation path
    :return cache_status: The status of the cache
    """
    magento2_config_file = re.sub("/pub/$", "/app/etc/env.php", installation_path)
    lines = read_magento2_config_file(magento2_config_file)
    cache_lines = filter_cache_lines_from_magento2_config_lines(lines)
    return get_cache_status_from_magento2_cache_lines(cache_lines)


def read_cache_status_state() -> List[Tuple[str, dict]]:
    """
    Read the cache json and return
    the state as list
    """
    try:
        cache_statuses = read_json_file(WEBAPP_CACHE_STATUS_STATEFILE_PATH)
    except (FileNotFoundError, RuntimeError):
        cache_statuses = []
    return cache_statuses


def write_cache_status_state(cache_statuses: List[Tuple[str, dict]]) -> None:
    """
    Write the cache statuses (back) to disk
    """
    with open(WEBAPP_CACHE_STATUS_STATEFILE_PATH, "w") as f:
        json.dump(cache_statuses, f)


def notify_client_of_cache_status_change(caches_changed: List[str]) -> None:
    """
    Send an event to the event API so we can notify the client
    about their changed cache status.
    """
    """Send an event to event API to notify client of mail authentication issues."""

    if not caches_changed:
        logger.debug(
            "Skipping sending cache status changed event. No changes detected."
        )
        return

    amount_caches_changed = len(caches_changed)
    if amount_caches_changed > MAX_CACHES_CHANGED:
        logger.debug(
            "More than {0} cached changed. Total caches changed was {1}. To not overload our "
            "event bus we'll truncate the list so we keep only the first {0}".format(
                MAX_CACHES_CHANGED, amount_caches_changed
            )
        )
        caches_changed = caches_changed[:MAX_CACHES_CHANGED]

    event = {"webapp-cache-status-changed": caches_changed}
    logger.debug("Sending cache status changed event")
    try_call_event_hook(event)
    logger.debug("Finished sending cache status changed event")


def detect_cache_change(
    cache_status: dict, cache_statuses: List[Tuple[str, dict]]
) -> Tuple[str, list]:
    """
    Check if any of the caches have been changed since we last
    recorded what the cache state was. If any of the caches
    have been disabled (but not removed) then we'll want to
    know the names of those disabled caches because that
    might be suspicious and unintended.
    """

    cache_change = list()
    now_datetime = datetime.now()
    now_time = now_datetime.strftime("%Y-%m-%d %H:%M:%S")

    if cache_statuses:
        most_recent_cache_time, most_recent_cache_status = cache_statuses[-1]
        if most_recent_cache_status != cache_status:
            # Check which previously enabled caches have now been disabled
            # We don't care about newly enabled caches, and we also don't
            # care about caches that have been completely remove from the
            # config file. Only if a cache was previously enabled and now
            # is disabled that is suspicious.
            for cache_name, cache_state in most_recent_cache_status.items():
                if cache_name in cache_status:
                    if cache_state and not cache_status[cache_name]:
                        message = (
                            "Cache '{}' was recorded as having been enabled since at least "
                            "{} and now on {} we've detect it as disabled. That might "
                            "have an adverse impact on the performance of this Hypernode."
                            "".format(cache_name, most_recent_cache_time, now_time)
                        )
                        logger.info(message)
                        cache_change.append(cache_name)

            if not cache_change:
                logger.info(
                    "Cache status changed since we last detected a change "
                    "on {}. None previously known as enabled have been "
                    "disabled though.".format(most_recent_cache_time)
                )

        else:
            logger.info(
                "Cache status has not changed since "
                "{}".format(most_recent_cache_time)
            )

    return now_time, cache_change


def compare_cache_status(cache_status: dict) -> None:
    """
    Check if the cache status changed to something we
    should be worried about.
    :param dict cache_status: The current cache status
    :return None:
    """
    cache_statuses = read_cache_status_state()
    now_time, cache_change = detect_cache_change(cache_status, cache_statuses)
    if user_is_root():
        cache_statuses.append((now_time, cache_status))
        write_cache_status_state(cache_statuses[-MAX_CACHE_STATUSES:])
        notify_client_of_cache_status_change(cache_change)


def analyze_cache_status() -> int:
    """
    Attempt to check the cache status of the webapp on the Hypernode
    :return int exit_code: The exit code for the program
    """
    logger.debug("Looking for a Magento 2 installation")
    magento2_installation = get_magento_2_install_dir(HYPERNODE_APP_HOMEDIR)
    cache_status = None
    if magento2_installation:
        logger.debug(
            "Found a Magento installation in '{}'" "".format(magento2_installation)
        )
        cache_status = determine_magento2_cache_status(magento2_installation)

    if cache_status:
        compare_cache_status(cache_status)
        return os.EX_OK

    logger.debug(
        "This tool could not identify a webapp to check the "
        "cache status of at this moment"
    )
    return os.EX_NOINPUT
