import argparse
import logging
import os
import sys

from kamikaze3.analysis.webapp.cache import analyze_cache_status
from kamikaze3.log.log import configure_logging

logger = logging.getLogger(__name__)


def parse_arguments(args: list = None) -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Check the status of the webapp on various fronts"
    )
    actions = parser.add_mutually_exclusive_group()
    actions.add_argument(
        "--cache",
        action="store_true",
        help="Check if the cache has recently been disabled",
    )
    verbosity = parser.add_mutually_exclusive_group()
    verbosity.add_argument(
        "--verbose", "-v", help="Display debug output", action="store_true"
    )
    verbosity.add_argument(
        "--quiet", "-q", help="Do not show output", action="store_true", dest="quiet"
    )
    return parser.parse_args(args)


def perform_action(parsed_args: argparse.Namespace):
    """
    Perform any of the actions specified in the parsed
    command-line arguments.
    :param namespace parsed_args: The parsed arguments
    :return int exit_code: The return value of the program
    """
    if parsed_args.cache:
        return analyze_cache_status()
    else:
        print("no action is specified. see --help for more info'", file=sys.stderr)
        return os.EX_NOINPUT


def main(args: list = None) -> int:
    """
    Analyze the webapp for things such as whether
    or not the cache was recently disabled.
    :param list args: any preparsed arguments
    :return int exit_code: The return value of the program
    """
    try:
        parsed_args = parse_arguments(args)
        configure_logging(
            # Note that this is .webapp and not .webapp.main
            # That is because we also want to set up the handler
            # for .webapp.cache and other files in this module.
            "kamikaze3.analysis.webapp",
            parsed_args.verbose,
        )
        return perform_action(parsed_args)
    except Exception as err:
        logger.exception(err)
        if isinstance(err, OSError):
            return os.EX_OSERR
        if isinstance(err, RuntimeError):
            return os.EX_TEMPFAIL
        return os.EX_SOFTWARE


if __name__ == "__main__":
    sys.exit(main())
