import argparse
import logging
import os
import sys
from pathlib import Path
from typing import List, Tuple

from kamikaze3.analysis.cronflock import get_cron, write_cron
from kamikaze3.auto_logrotate.utils import find_big_logfiles
from kamikaze3.common.filesystem import (
    APP_USER_UID,
    human_readable_file_size,
    human_readable_size_to_bytes,
)
from kamikaze3.log.log import configure_logging

HOME_DIR = Path("/data/web")
CONFIG_FILE = HOME_DIR / "hypernode_logrotate.conf"
STATE_FILE = HOME_DIR / "hypernode_logrotate.state"

CONFIG_TEMPLATE = """
%s {
    rotate 4
    daily
    compress
    # do not raise an error if the logfile does not exist
    missingok
    # file size at which rotation is triggered
    size 50M
    # copy the file and truncate the original, in case
    # another process cannot be told to close the logfile
    copytruncate
}

"""

LOGROTATE_COMMAND = "/usr/sbin/logrotate"

logger = logging.getLogger(__name__)


def parse_arguments() -> Tuple[List[Path], bool, str, bool, bool]:
    parser = argparse.ArgumentParser(
        description="Configure log rotation for specified log file"
    )
    parser.add_argument("logfiles", nargs="*", help="Location of logfile to rotate")
    parser.add_argument(
        "--detect",
        action="store_true",
        help="find logfiles and rotate them "
        "(WARNING: on large filesystems this may take a while)",
    )
    parser.add_argument(
        "--threshold",
        help="The log size threshold in bytes, may contain human-readable suffix like MB, GB, TB, etc. Default: 500M",
        default="500M",
    )
    parser.add_argument(
        "--verbose",
        "-v",
        action="store_true",
        help="Display debug output",
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="Run auto_logrotate without modifying any files",
    )
    args = parser.parse_args()
    return (
        [Path(logfile).absolute() for logfile in args.logfiles],
        args.detect,
        args.threshold,
        args.verbose,
        args.dry_run,
    )


def logfile_is_valid(logfile: Path) -> bool:
    if not logfile.parent.is_dir():
        logger.info(f'Directory "{logfile.parent}" does not exist')
        return False
    if logfile.is_file() and not os.access(logfile, os.W_OK):
        logger.info(f'No write access to "{logfile}"')
        return False
    return True


def logfile_already_being_rotated(logfile: Path) -> bool:
    if not CONFIG_FILE.is_file():
        return False
    with open(CONFIG_FILE) as logrotate_config:
        return any(
            line.startswith(str(logfile)) for line in logrotate_config.readlines()
        )


def rotate_logfile(logfile: Path, dry_run: bool) -> None:
    if logfile_already_being_rotated(logfile):
        logger.info(f'Logfile "{logfile}" is already being rotated')
        return
    logger.info(f'Adding logrotate config entry for "{logfile}"')
    logger.debug(f'Appending the following lines to "{CONFIG_FILE}":')
    logger.debug(CONFIG_TEMPLATE % logfile)
    if not dry_run:
        with open(CONFIG_FILE, "a") as logrotate_config:
            logrotate_config.write(CONFIG_TEMPLATE % logfile)


def add_logrotate_cron_job(dry_run: bool) -> None:
    crontab = get_cron(user="app")
    if not dry_run:
        CONFIG_FILE.touch()
    log_rotate_job = (
        f"0 0 * * * {LOGROTATE_COMMAND} {CONFIG_FILE} --state {STATE_FILE} "
        "# added by hypernode-auto-logrotate\n"
    )
    if "# added by hypernode-auto-logrotate\n" not in crontab:
        logger.info("Adding logrotate cron job")
        if not dry_run:
            write_cron(crontab=crontab + log_rotate_job, user="app")


def main() -> None:
    logfiles, detect, threshold, verbose, dry_run = parse_arguments()

    configure_logging(__name__, verbose)

    if os.getuid() != APP_USER_UID:
        logger.error("This script should be run by the app user only")
        sys.exit(1)

    add_logrotate_cron_job(dry_run)

    if detect:
        logger.info("Searching for logfiles...")
        threshold_bytes = human_readable_size_to_bytes(threshold)
        logger.debug(f"Threshold human-readable: {threshold}, bytes: {threshold_bytes}")
        found_logfiles = find_big_logfiles(HOME_DIR, threshold_bytes)
        if not found_logfiles:
            logger.info(f"No logfiles found bigger than {threshold}")
        for logfile in found_logfiles:
            logger.info(
                f"Logfile found: {logfile} - {human_readable_file_size(logfile)}"
            )
        logfiles.extend(found_logfiles)

    for logfile in logfiles:
        if logfile_is_valid(logfile):
            rotate_logfile(logfile, dry_run)
