import argparse
import logging
from pathlib import Path

from kamikaze3.auto_logrotate.utils import find_big_logfiles
from kamikaze3.common.filesystem import (
    human_readable_file_size,
    human_readable_size_to_bytes,
)
from kamikaze3.log.log import configure_logging
from kamikaze3.onboarding import try_call_event_hook

logger = logging.getLogger(__name__)

HOME_DIR = "/data/web"


def parse_arguments() -> tuple[Path, str, bool]:
    parser = argparse.ArgumentParser(
        description="Configure log rotation for specified log file"
    )
    parser.add_argument(
        "--base-dir", help="Set base directory to start scanning from", default=HOME_DIR
    )
    parser.add_argument(
        "--threshold",
        help="The log size threshold in bytes, may contain human-readable suffix like MB, GB, TB, etc",
        default="1G",
    )
    parser.add_argument(
        "--verbose",
        "-v",
        action="store_true",
        help="Display debug output",
    )
    args = parser.parse_args()
    return Path(args.base_dir), args.threshold, args.verbose


def notify_client_of_big_logfiles(logfiles: list[Path]) -> None:
    for file in logfiles:
        logger.debug(f"Notifying user for {file}")
    event = {"big-logfiles-detected": [str(path) for path in logfiles]}
    logger.debug(f"Sending mail big logfile event: {event}")
    try_call_event_hook(event)
    logger.info("Mail big logfile event sent")


def main() -> None:
    base_dir, threshold, verbose = parse_arguments()
    configure_logging(__name__, verbose)

    logger.info(f"Scanning directory {base_dir} for log files >= {threshold}")

    threshold_bytes = human_readable_size_to_bytes(threshold)
    logger.debug(f"Threshold human-readable: {threshold}, bytes: {threshold_bytes}")

    big_logfiles = find_big_logfiles(base_dir, threshold_bytes)

    for file in big_logfiles:
        logger.debug(f"Found big logfile: {file} - {human_readable_file_size(file)}")

    if len(big_logfiles):
        notify_client_of_big_logfiles(big_logfiles)


if __name__ == "__main__":
    main()
