import os
from pathlib import Path

ECO_BANNED_PATHS = [
    ".local/share/docker",  # rootless docker
    ".tox",  # tox directory
    "dev/tests",  # magento2
    "dev/tools",  # magento2
    "generated/code",  # magento2
    "generated/metadata",  # magento2
    "pub/media",  # magento2
    "pub/static",  # magento2
    "public/cache",  # akeneo
    "public/css",  # shopware6, akeneo
    "public/dist",  # akeneo
    "public/js",  # shopware6, akeneo
    "public/media",  # shopware6, akeneo
    "public/thumbnail",  # shopware6
    "var/cache",  # common
    "var/file_storage",  # akeneo
    "node_modules",  # common
    "vendor",  # common
]

ECO_COMMUNITY_BANNED_PATHS = ["public/product-configurator"]


def _get_logfile_path_or_none(path: str, root: str) -> Path | None:
    if path.endswith(".log"):
        return Path(os.path.realpath(Path(root) / path))

    return None


def find_logfiles(path: Path) -> list[Path]:
    logfiles = []
    for root, _, files in os.walk(path):
        for file in files:
            logfile = _get_logfile_path_or_none(file, root)
            if logfile and logfile not in logfiles:
                logfiles.append(logfile)

    return logfiles


def is_path_eco_banned(path: os.PathLike | str) -> bool:
    for banned in ECO_BANNED_PATHS + ECO_COMMUNITY_BANNED_PATHS:
        if str(path).endswith(banned):
            return True

    return False


def find_logfiles_eco(
    scandir_path: Path, depth: int = 1, max_depth: int = 8
) -> list[Path]:
    """
    Find logfiles recursively while saving filesystem IO.
    The IO is saved by:
      - Skipping common directories we don't need to search in.
      - Limiting the traversal depth to max_depth

    :param Path scandir_path: Path to start searching from
    :param int max_depth: Max traversal depth
    :param int depth: Current traversal depth
    :return: List of paths to the logfiles found
    """
    if depth > max_depth:
        return []

    logfiles: list[Path] = []
    with os.scandir(scandir_path) as it:
        for entry in it:
            entry: os.DirEntry
            full_path: Path = os.path.join(scandir_path, entry.name)

            if entry.is_symlink():
                continue

            if entry.is_dir() and not is_path_eco_banned(full_path):
                logfiles.extend(find_logfiles_eco(full_path, depth + 1, max_depth))
            elif entry.is_file() and entry.name.endswith(".log"):
                logfiles.append(Path(full_path))

    return logfiles


def file_exceeds_threshold(file: Path, threshold: int) -> bool:
    size = os.stat(file).st_size

    return size >= threshold


def find_big_logfiles(path: Path, threshold: int) -> list[Path]:
    result = []
    logfiles = find_logfiles_eco(path)

    for file in logfiles:
        if file_exceeds_threshold(file, threshold):
            result.append(file)

    return result
