import re


def validate_acl(vcl_content):
    """
    Validate that the VCL contains an ACL for purging with at least 'localhost' or '127.0.0.1'
    and the '10.0.0.0/24' range. Also checks that 'vcl_recv' restricts PURGE requests to this ACL.

    Parameters:
        vcl_content (str): The VCL content as a string.

    Returns:
        tuple: (is_valid, message)
            is_valid (bool): True if the VCL is valid, False otherwise.
            message (str): Error message if invalid, or success message if valid.
    """
    # Regular expression to find the ACL block for 'purge'
    acl_pattern = re.compile(r"acl\s+purge\s*\{([^}]*)\}", re.MULTILINE | re.DOTALL)
    acl_match = acl_pattern.search(vcl_content)

    validation_results = []

    if not acl_match:
        validation_results.append((False, "Missing ACL block for purging."))

    # Extract ACL content and check if required IPs are present
    if acl_match:
        acl_content = acl_match.group(1)
    else:
        acl_content = ""

    if not ("localhost" in acl_content or "127.0.0.1" in acl_content):
        validation_results.append(
            (False, "Missing 'localhost' or '127.0.0.1' in purge ACL.")
        )

    if "10.0.0.0/24" not in acl_content:
        validation_results.append((False, "Missing '10.0.0.0/24' in purge ACL."))

    if not validation_results:
        validation_results.append((True, "VCL is valid for purging with proper ACL."))

    all_valid = all(result[0] for result in validation_results)
    messages = [result[1] for result in validation_results]

    return all_valid, messages
