import re


def validate_status(
    status_content, ping_content, backend_list_content, vcl_list_content
):
    """
    Validate Varnish status, ping response, backend status, and VCL configuration.

    Parameters:
        status_content (str): Output from 'varnishadm status'.
        ping_content (str): Output from 'varnishadm ping'.
        backend_list_content (str): Output from 'varnishadm backend.list'.
        vcl_list_content (str): Output from 'varnishadm vcl.list'.

    Returns:
        tuple: (is_valid, message)
            is_valid (bool): True if all checks pass, False otherwise.
            message (str): Error message if invalid, or success message if valid.
    """
    validation_results = []

    # Check Varnish status
    if "running" not in status_content.lower():
        validation_results.append((False, "Varnish is not running."))

    # Check Varnish ping response
    if "PONG" not in ping_content:
        validation_results.append((False, "Varnish ping failed."))

    # Check that at least one backend is healthy
    if not re.search(r"\bhealthy\b", backend_list_content):
        validation_results.append((False, "No healthy backends found."))

    # Check that there is at least one active VCL
    if not re.search(r"\bactive\b", vcl_list_content):
        validation_results.append((False, "No active VCL configurations found."))

    if not validation_results:
        validation_results.append(
            (True, "Varnish is running and healthy with active VCL and backends.")
        )

    all_valid = all(result[0] for result in validation_results)
    messages = [result[1] for result in validation_results]

    return all_valid, messages
