import argparse
import configparser
import json
import socket
import sys
import time
from logging import DEBUG, INFO, getLogger

import MySQLdb
from MySQLdb.cursors import DictCursor

from kamikaze3.cluster import program_name
from kamikaze3.common.env import get_hn_common_var
from kamikaze3.common.log import create_stream_handler, create_syslog_handler
from kamikaze3.settings import (
    HAPROXY_CONF_FILE,
    MYSQL_REPLICATION_HOST,
    MYSQL_ROOT_CONF_FILE,
    RELATIONS_INFO_PATH,
)

MYSQL_FAILOVER_ROLE = "mysqlfailover"
MYSQL_REPLICA_ROLE = "mysqlreplica"
MOUNT_CONFIG_PATH = "/etc/fstab"

logger = getLogger(program_name)


class HealthCheckException(Exception):
    pass


def get_ip_address(node_info: dict) -> str:
    src_private_network_ip = get_hn_common_var("private_network_ip")
    src_private_network = get_hn_common_var("private_network")

    dest_private_network_ip = node_info.get("private_network_ip")
    dest_private_network = node_info.get("private_network") or {}

    if (
        src_private_network_ip
        and dest_private_network_ip
        and dest_private_network.get("id") == src_private_network["id"]
    ):
        return dest_private_network_ip

    return node_info.get("floating_ip") or node_info["ip"]


def get_relations():
    with open(RELATIONS_INFO_PATH) as f:
        return json.load(f)


def has_parent_with_relation_type(relation_type: str) -> bool:
    for relation in get_relations()["parents"]:
        if relation["relation_type"] == relation_type:
            return True
    return False


def get_parent_info_for_relation_type(relation_type: str) -> dict:
    for relation in get_relations()["parents"]:
        if relation["relation_type"] == relation_type:
            return relation
    return {}


def get_nfs_mountpoints() -> list:
    data = []
    with open(MOUNT_CONFIG_PATH, encoding="utf-8") as f:
        for line in f:
            if "nfs" not in line:
                continue

            line_parts = line.split(" ")
            if len(line_parts) > 2:
                data.append(line_parts[1])

    f.close()
    return data


def get_haproxy_credentials() -> dict:
    """
    Read the haproxy.cfg file and extract the auth credentials from this file if there
    """
    authorization = {}
    with open(HAPROXY_CONF_FILE, encoding="utf-8") as f:
        for line in f:
            if "stats" not in line:
                continue

            # Remove duplicate whitespaces
            line = " ".join(line.split())

            # Split auth line and get haproxy credentials.
            # this line looks like "stats   auth hypernode:password"
            line_parts = line.split(" ")
            if len(line_parts) != 3 or line_parts[1] != "auth":
                continue

            auth_user, auth_password = line_parts[2].split(":")
            authorization = {"user": auth_user, "password": auth_password}

    return authorization


def mysql_health_check(node_info: dict) -> float:
    """Check if we can connect to the MySQL server on the given node.
    Return the amount of seconds it took to create a socket connection.

    :param node_info: Node info of the target node where MySQL is running
    :rtype: float
    """
    ip_address = get_ip_address(node_info)
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        sock.settimeout(3)  # Timeout after 3 seconds
        start = time.time()
        result = sock.connect_ex((ip_address, 3306))
        stop = time.time()
    if result != 0:
        raise HealthCheckException(
            f"MySQL health check failed to {ip_address} with exitcode {result}"
        )
    return stop - start


def mysql_replica_health_check() -> dict:
    """
    Get mysql connection, execute query and return results.
    :return dict: Return dict with query results
    """
    connection = get_mysql_connection()
    cursor = connection.cursor(DictCursor)
    cursor.execute("SHOW REPLICA status;")

    # result should have 1 row, so fetch the first one
    data = cursor.fetchone()
    cursor.close()
    connection.close()

    return data


def get_mysql_connection():
    """
    Read .my.cnf, parse file and setup Mysql connection
    """
    config = configparser.ConfigParser()
    config.read(MYSQL_ROOT_CONF_FILE)
    credentials = (
        MYSQL_REPLICATION_HOST,
        config["client"]["user"],
        config["client"]["password"],
    )

    return MySQLdb.connect(*credentials)


def health_check(relation: dict) -> float:
    """Check if a relation is healthy.

    :param relation: A relation dict as returned by get_relations()
    :return: The time it took to perform the health check
    """
    if relation["relation_type"] == "mysql":
        return mysql_health_check(relation["parent"])
    else:
        logger.debug("Unknown relation type: {}".format(relation["relation_type"]))
        return 0.0


def check_all_relations(write_metrics: bool = True):
    relations = get_relations()
    for relation in relations["parents"]:
        relation_type = relation["relation_type"]
        logger.debug(
            f"Checking {relation_type} relation health to {relation['parent']['app']}"
        )
        try:
            timing = health_check(relation)
        except HealthCheckException as e:
            logger.debug(e)
            # When we want to actively replica, this is the place to do it
            sys.exit(1)
        except Exception as e:
            logger.debug(f"Unknown exception: {e}")
            sys.exit(2)
        else:
            logger.debug(
                f"Successfully checked relation health to {relation['parent']['app']} in {timing:.5f} seconds"
            )
            # write output when we passed arg metrics
            if write_metrics:
                query_string = f"parent=\"{relation['parent']['app']}\",relation_type=\"{relation_type}\""
                print(
                    f"cluster_network_probe_duration_seconds{{{query_string}}} {timing}"
                )

            return timing


def child_relation_health_checks(write_metrics: bool = True):
    """
    Handle child relation health checks
    """
    relations = get_relations()
    for relation in relations["parents"]:
        relation_type = relation["relation_type"]
        logger.debug(
            f"Checking {relation_type} relation health to {relation['parent']['app']}"
        )
        if relation_type in [MYSQL_FAILOVER_ROLE, MYSQL_REPLICA_ROLE]:
            replica_status = mysql_replica_health_check()

            query_string = f"parent=\"{relation['parent']['app']}\",relation_type=\"{relation_type}\""
            io_thread_running = (
                1 if replica_status["Replica_IO_Running"] == "Yes" else 0
            )
            sql_thread_running = (
                1 if replica_status["Replica_SQL_Running"] == "Yes" else 0
            )
            # If mysqlreplica is not running, Seconds behind master is set to None. To avoid malformed
            # data in prometheus prom files we set it to -1
            seconds_behind_master = (
                replica_status["Seconds_Behind_Source"]
                if replica_status["Seconds_Behind_Source"] is not None
                else -1
            )

            if write_metrics:
                print(
                    f"cluster_replica_io_thread_running{{{query_string}}} {io_thread_running}"
                )
                print(
                    f"cluster_replica_sql_thread_running{{{query_string}}} {sql_thread_running}"
                )
                print(
                    f"cluster_replica_seconds_behind_master{{{query_string}}} {seconds_behind_master}"
                )

            # exit non zero if io_thread_running or sql_thread_running != 1
            if (io_thread_running + sql_thread_running) != 2:
                sys.exit(1)


def parse_args(args: list[str] | None = None) -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Check if all cluster relations are healthy"
    )
    parser.add_argument(
        "--verbose",
        "-v",
        action="store_true",
        help="Enable verbose logging",
    )

    parser.add_argument(
        "--metrics",
        "-m",
        action="store_true",
        help="Print metrics and its current values",
    )
    return parser.parse_args(args)


def setup_logging(verbose: bool = False):
    logger.setLevel(DEBUG)
    logger.addHandler(create_syslog_handler())
    logger.addHandler(create_stream_handler(level=DEBUG if verbose else INFO))


def main(args: list[str] | None = None):
    args = parse_args(args)
    setup_logging(args.verbose)
    logger.debug(f"Starting {program_name}")
    check_all_relations(args.metrics)
    child_relation_health_checks(args.metrics)
    logger.debug(f"Finished {program_name}")


if __name__ == "__main__":
    main()
