"""
kamikaze3.common.cron
~~~~~~~~~~~~~~~~~~~~~
functionality for parsing cron configs
"""

import re


def validate_timespec(timespec: str) -> bool:
    """Validate a cron job timespec string"""

    if timespec.lstrip("@") in (
        "reboot",
        "yearly",
        "annually",
        "monthly",
        "weekly",
        "daily",
        "midnight",
        "hourly",
    ):
        return True
    tokens = timespec.split()
    if len(tokens) != 5:
        return False
    # numbers, ranges (-), list (,), steps (/), any (*)
    regex_numeric_vals = re.compile(r"^[,\d\-\*\/]+$")

    # first 3 tokens can't have names,
    # last 2 tokens (month and day of week) also accept names (SAT and JAN)
    for token in tokens[:3]:
        if not regex_numeric_vals.match(token):
            return False
    regex_names = re.compile(r"^\w{2,5}$")
    for token in tokens[3:]:
        if not regex_numeric_vals.match(token) and not regex_names.match(token):
            return False
    return True


def _looks_like_cron_cmd(line: str) -> bool:
    """Return true if the line could be a cron job definition"""
    line = str(line).strip()
    if not line or line.startswith("#"):  # empty or comment
        return False

    if re.match(r"^\s*[\S]+\s*=", line):  # mailto = foo
        return False

    return True


def split_crontab_cmd_line(line: str):
    """Split a crontab line defining a job to its components and return
    a tuple of those components (timespec, cmd, comment).
    If the line is not a crontab cmd line, the timespec and cmd components
    are going to be empty strings, comments are still detected.

    The components are stripped of surrounding whitespaces, comments are only rstripped
    so left most spaces from comments are preserved.

    If the line looks like a crontab cmd line but with invalid format,
    raises ValueError.

    :raises: ValueError on lines with invalid format
    """
    timespec, cmd, comment = ("", "", "")

    croncmd, _, comment = str(line.rstrip()).partition("#")
    croncmd = croncmd.strip()
    if not _looks_like_cron_cmd(croncmd):
        return (timespec, cmd, comment)

    if croncmd[0] == "@":
        # timespec such as @daily
        timespec, cmd = croncmd.split(maxsplit=1)
    else:
        # timespec such as * * * * *
        tokens = croncmd.split(maxsplit=5)
        if len(tokens) != 6:
            raise ValueError("Cannot parse crontab line: {}".format(line))
        timespec = " ".join(tokens[:5])
        cmd = tokens[5]

    if not validate_timespec(timespec):
        raise ValueError("Invalid crontab timespec found: {}".format(timespec))

    if cmd.strip().startswith("*"):
        raise ValueError("Invalid crontab command found: {}".format(cmd))

    return (timespec, cmd, comment)
