"""
Common filesystem functionality.
"""

import hashlib
import io
import json
import os
import subprocess
from contextlib import closing, contextmanager, suppress
from json import JSONDecodeError
from logging import getLogger
from math import log
from platform import system
from subprocess import check_output

APP_USER_UID = 1000
APP_USER_GID = 1000

logger = getLogger(__name__)


def read_file(filepath):
    """
    Read a file a return its contents

    :param filepath: the path of the file to read
    :return: str: the file content or the exception
    """
    with open(filepath) as fh:
        content = fh.read()
    return content


def try_read_file(filepath) -> str | None:
    """
    Read a file and return its contents on success

    :param filepath: the path of the file to read
    :return: str: the file content or the exception
    """
    try:
        return read_file(filepath)
    except OSError as e:
        logger.debug(e)
        return None


def try_read_json_file(filepath):
    try:
        with open(filepath) as f:
            return json.load(f)
    except (OSError, JSONDecodeError):
        return None


def read_json_file(filepath):
    """Read a JSON text file and return parsed out contents.

    :param str filepath: the path to th file
    :raises: RuntimeError on inaccessible files, or invalid JSON data
    """
    try:
        logger.debug(f"reading JSON file '{filepath}' ...")
        with open(filepath) as fh:
            content = fh.read().strip()
        logger.debug(f"parsing JSON content from '{filepath}' ...")
        return json.loads(content)
    except (ValueError, RuntimeError, OSError) as err:
        raise RuntimeError(f"Failed to parse JSON from '{filepath}': {err}")


def read_first_existing_file(filenames):
    """Try specified files, return the contents of the first file that exists

    :param iterable filenames: iterable of paths to files that may or may not exist
    :return str content: Content of the first file that exists
    :raises ValueError: if filesnames is empty
    :raises IOError: on failure to read any of the files
    """
    file_handle = None
    checked_filenames = 0  # avoid unnecessary len() on the iterable
    for filename in filenames:
        checked_filenames += 1
        try:
            file_handle = open(filename)
            break
        except OSError:
            pass
    if checked_filenames < 1:
        raise ValueError("No files are specified to read from")
    if not file_handle:
        raise OSError("Couldn't read any of the files: " + ",".join(filenames))
    with closing(file_handle):
        contents = file_handle.read()
    return contents


def read_last_block(filename, block_size=4096, raise_error=False):
    """Return last block from the specified file, without reading the
    complete file.

    :param filename: The filename
    :param block_size: size of the blocks in the file. This will be the max size of returned value
    :param raise_error: If should raise errors while reading the file,
        if this is False (default) and there are errors reading the file, it will return None
    :return: The data block
    :rtype: bytes|None
    """
    content = None
    try:
        # seek from end is not available on text files, only read in binary mode
        with open(filename, "rb") as fh:
            with suppress(IOError):
                fh.seek(-1 * block_size, io.SEEK_END)
            content = fh.read()
    except OSError as err:
        if raise_error:
            raise err
    return content


def read_last_lines(
    filename, lines=1, block_size=4096, encoding="utf-8", raise_error=False
):
    """Return last lines from the specified text file, without reading the
    complete file.

    :param filename: The filename
    :param lines: max number of lines returned
    :param block_size: size of the last block to read from file to extract lines
    :param raise_error: If should raise errors while reading the file,
    :param encoding: encoding of the text file (default is utf-8)
    :param raise_error: If should raise errors while reading the file,
        if this is False (default) and there are errors reading the file, it will return None
    :return: The data block
    :rtype: bytes|None
    """
    last_block = read_last_block(filename, block_size, raise_error)
    if last_block is None:
        return None

    last_lines = last_block.decode(encoding).splitlines()
    return last_lines[-lines:]


def file_size_mb(filename):
    """Return the size of the file in mega bytes.

    :param str filename: The file name
    :return: filesize in MB
    :rtype: float
    """
    return os.stat(filename).st_size / (1024**2)


def human_readable_file_size(file: os.PathLike) -> str:
    size_in_bytes = os.stat(file).st_size
    return add_binary_prefix(size_in_bytes) + "B"


def human_readable_size_to_bytes(size: str) -> int:
    """Convert the human-readable size to bytes using numfmt (GNU coreutils).

    :param str size: The human-readable size string
    :return int: number of bytes the size string represents
    """

    # numfmt only accepts suffixes like M,G,T,P, etc.
    # So we strip unnecessary B suffix, so it converts 4GB to 4G
    size = size.upper().rstrip("B")

    numfmt_command = (
        "/opt/homebrew/bin/numfmt" if system() == "Darwin" else "/usr/bin/numfmt"
    )
    result = check_output([numfmt_command, "--from", "iec", size])

    return int(result)


def add_binary_prefix(num: float) -> str:
    if num <= 0:
        return str(round(num, 1))
    prefixes = ["", "k", "M", "G", "T", "P", "E", "Z", "Y"]
    logarithm = int(log(num, 1024))
    capped_log = min(len(prefixes) - 1, logarithm)
    return str(round(num / (1024**capped_log), 1)) + prefixes[capped_log]


@contextmanager
def capture_file_size_change(filename):
    """Capture file size changes in a context (in MB).
    The context variable a list, starting with the size of the file (MB) before
    the context, and after the context finishes, the file size
    after the context is added to the list.

    :param str filename: filename to capture size changes
    """
    size_samples = [file_size_mb(filename)]
    yield size_samples
    size_samples.append(file_size_mb(filename))


def is_risky_sxid_file(path):
    """
    This function tells if the file has the setuid or setgid bit set.
    A file with the setuid or setgid bit is marked as risky if the file is
    owned by a different user or group than the app user or group.

    Access privileges are in octal, only take the important number because the
    other access privileges are not important if the value is above 0 then the
    number is set to some kind of elevated privileges and we want to check.

    :param path: path of the file we want to know if it is a setuid or setgid file
    :return: True or False depending on if the file is a sxid file
    """
    file_stat = os.stat(path)
    mode = file_stat.st_mode
    permission_flag = int(oct(mode)[-4:-3])
    owned_by_app_uid_and_gid = (
        file_stat.st_uid == APP_USER_UID and file_stat.st_gid == APP_USER_GID
    )
    return permission_flag > 0 and not owned_by_app_uid_and_gid


def get_sha512_hash_file(path, chunk_size=4096):
    """
    Hash the file in path and return the hash
    :param path: File you want to hash
    :param chunk_size: Size of the chunks used for hashing, default on 4096
     since this is the memory page size on Hypernodes hosted by COMBELL
    :return: Hash of the file in path
    """
    hasher = hashlib.sha512()
    if os.path.isfile(path):
        with open(path, "rb") as f:
            for chunk in read_file_in_chunks(f, chunk_size):
                hasher.update(chunk)
        file_hash = hasher.hexdigest()
    else:
        file_hash = ""
    return file_hash


def get_sha512_hash_filtered_file(path, filter_pattern):
    """
    Hash the filtered contents of the file in path and return the hash
    :param path: File you want to hash
    :param filter_pattern: Filter you want to apply, must be compatible with sed extended regular expressions.
     Matching text will be left out
    :return: Hash of the file in path
    """
    if os.path.isfile(path):
        output = subprocess.check_output(
            f"/usr/bin/cat {path} | /usr/bin/sed -E 's/{filter_pattern}//g' | /usr/bin/sha512sum",
            shell=True,
        )
        file_hash = output.decode("utf-8").rstrip(" -\n")
    else:
        file_hash = ""
    return file_hash


def read_file_in_chunks(file_handle, chunk_size=4096):
    """
    Read a file in chunks
    :param file_handle: file object after open() is called on the path
    :param chunk_size: Size of the chunks used for hashing, default on 4096
     since this is the memory page size on Hypernodes hosted by COMBELL
    :return: chunks
    """
    while True:
        data = file_handle.read(chunk_size)
        if not data:
            break
        yield data


def get_file_filtered_hashes(files, pattern):
    """
    Hashes all the files from a given list and makes a dict from the
    path and hashes.

    :param files: all the paths of the files you want to have hashed
    :param pattern: the pattern to be filtered out of the file contents
    :return: dict with the file paths as keys and the hash as the value
    """
    file_hashes = {}
    for filepath in files:
        filehash = get_sha512_hash_filtered_file(filepath, pattern)
        file_hashes[filepath] = filehash
    return file_hashes


def get_file_hashes(files):
    """
    Hashes all the files from a given list and makes a dict from the
    path and hashes.

    :param files: all the paths of the files you want to have hashed
    :return: dict with the file paths as keys and the hash as the value
    """
    file_hashes = {}
    for filepath in files:
        filehash = get_sha512_hash_file(filepath)
        file_hashes[filepath] = filehash
    return file_hashes


def compare_file_hashes_to_earlier_defined_list(
    original_file_hashes, found_files, filter_pattern: str = None
):
    """
    Check if the file is legal and if legal make sure no one changed the file
    :param original_file_hashes: The list with legal files and their hashes
    :param found_files: A list of files found on the system
    :param filter_pattern: A filter to apply to the contents of the files
    :return: All the files that are marked as illegal or changed
    """
    illegal_files = []
    changed_files = []
    legal_files = original_file_hashes.keys()

    if filter_pattern:
        file_hashes = get_file_filtered_hashes(legal_files, filter_pattern)
    else:
        file_hashes = get_file_hashes(legal_files)

    for file_path in found_files:
        if file_path in legal_files:
            if isinstance(original_file_hashes[file_path], str):
                if (
                    original_file_hashes[file_path]
                    and file_hashes[file_path]
                    and file_hashes[file_path] != original_file_hashes[file_path]
                ):
                    changed_files.append(file_path)
            else:
                if file_hashes[file_path] not in original_file_hashes[file_path]:
                    changed_files.append(file_path)
        else:
            illegal_files.append(file_path)
    return illegal_files, changed_files
