import subprocess
from logging import getLogger
from os import remove
from random import randrange
from sys import stderr

from kamikaze3.common.lineinfile import lineinfile

# Dehydrated options
DEHYDRATED_DOMAINS_TXT = "/data/web/.dehydrated/domains.txt"
DEHYDRATED_ROOT_ACCOUNTS_DIR = "/etc/dehydrated/accounts"
DEHYDRATED_ACCOUNT_CMD = (
    "/usr/bin/dehydrated --register --accept-terms --account-dir {}"
)

# Cron options
DEHYDRATED_CRON_CMD = (
    "/usr/bin/dehydrated --domain {} --cron --create-dirs --account-dir {}"
)
TMP_CRONTAB_FILE = "/var/cache/managed_vhosts/tmp_crontab"

logger = getLogger(__name__)


class DehydratedError(RuntimeError):
    pass


def manage_letsencrypt_certificate(new_config):
    """
    Manage the letsencrypt certificate for the given config.
    Will create or remove a new certificate if needed.
    :rtype: None
    """
    # Add or remove servername to /data/web/.dehydrated/domains.txt based on https config
    logger.debug("Adding {} to dehydrated domains.txt".format(new_config["servername"]))
    lineinfile(
        line=new_config["servername"],
        filename=DEHYDRATED_DOMAINS_TXT,
        state="present" if new_config["https"] else "absent",
        create=True,
    )

    # Register dehydrated account and accept terms if letsencrypt is True
    dehydrated_account_cmd = DEHYDRATED_ACCOUNT_CMD.format(DEHYDRATED_ROOT_ACCOUNTS_DIR)
    dehydrated_cron_cmd = DEHYDRATED_CRON_CMD.format(
        new_config["servername"], DEHYDRATED_ROOT_ACCOUNTS_DIR
    )
    # Only do hours in the night so that dehydrated isn't triggered during officehours
    cron_cmd_prefix = "{minute} {hour} * * {weekday}".format(
        minute=randrange(60), hour=randrange(6), weekday=randrange(7)
    )
    if new_config["https"]:
        logger.info(
            "Generating letsencrypt certificate for {}".format(new_config["servername"])
        )
        try:
            logger.debug(
                "Ensuring dehydrated account; running the command '{}'".format(
                    dehydrated_account_cmd
                )
            )
            subprocess.check_call(dehydrated_account_cmd.split(" "), stdout=stderr)

            logger.debug(
                "Requesting certificate; running the command '{}'".format(
                    dehydrated_cron_cmd
                )
            )
            subprocess.check_call(dehydrated_cron_cmd.split(" "), stdout=stderr)
        except subprocess.CalledProcessError:
            logger.exception(
                "Encountered an error while running '{}; {}'"
                "".format(dehydrated_account_cmd, dehydrated_cron_cmd)
            )
            raise DehydratedError

    # Manage cronfile
    logger.debug("Ensuring dehydraded cron job for {}".format(new_config["servername"]))
    prepare_for_crontab_change()
    lineinfile(
        filename=TMP_CRONTAB_FILE,
        regex=dehydrated_cron_cmd if new_config["https"] else None,
        line="{} {}".format(cron_cmd_prefix, dehydrated_cron_cmd),
        state="present" if new_config["https"] else "absent",
    )
    flush_crontab()


def remove_letsencrypt_config(servername):
    """
    Remove all configurations for generating letsencrypt certificates
    for the given servername
    :param str servername: Servername to remove configs for
    :rtype: None
    """
    # Remove domain from domains.txt
    lineinfile(line=servername, filename=DEHYDRATED_DOMAINS_TXT, state="absent")

    dehydrated_cron_cmd = DEHYDRATED_CRON_CMD.format(
        servername, DEHYDRATED_ROOT_ACCOUNTS_DIR
    )

    # Remove line from crontab
    prepare_for_crontab_change()
    lineinfile(filename=TMP_CRONTAB_FILE, regex=dehydrated_cron_cmd, state="absent")
    flush_crontab()


def prepare_for_crontab_change():
    """
    Prepare the temporary crontab file by writing the current crontab contents to it.
    If there is no crontab, write an empty string to the file.
    :rtype: None
    """
    try:
        contents = subprocess.check_output(["/usr/bin/crontab", "-u", "app", "-l"])
        contents = contents.decode()  # Convert bytestring
    except subprocess.CalledProcessError:
        logger.debug("Crontab doesn't exist yet, placing empty tmp crontab file")
        contents = ""
    with open(TMP_CRONTAB_FILE, "w") as fp:
        fp.write(contents)


def flush_crontab():
    """
    Update crontab with the contents of the temporary file file
    :rtype: None
    """
    subprocess.check_call(["/usr/bin/crontab", "-u", "app", TMP_CRONTAB_FILE])
    remove(TMP_CRONTAB_FILE)
