import re
from pathlib import Path


def lineinfile(filename, state="present", line=None, regex=None, create=False):
    """
    Python version of the ansible lineinfile module.
    See https://docs.ansible.com/ansible/latest/modules/lineinfile_module.html
    :param str filename: Name of the file that should be managed
    :param str state: State of the line in the file. Should be either 'present'
    or 'absent'
    :param str line: Line that should be managed in the file. Doesn't go together
    with the regex parameter.
    :param str regex: Pattern to look for in the file. Doesn't go together with
    the line parameter
    :param bool create: Ensure file exists before performing operations.
    :return str: New content of the file
    """
    validate_lineinfile_input(line=line, regex=regex, state=state)

    if create:
        # Creates file if it doesn't exist, and update utime
        Path(filename).touch()

    with open(filename, "r") as fp:
        line_list = [l.strip() for l in fp.readlines()]

    if regex and state == "absent":
        newlines = remove_line(line_list, regex=regex)
    elif line and state == "absent":
        newlines = remove_line(line_list, line=line)
    elif regex and state == "present":
        newlines = replace_line(line_list, regex=regex, line=line)
    elif not regex and state == "present":
        newlines = add_line(line_list, line=line)
    else:
        raise AssertionError(
            "Undefined behaviour for lineinfile. Args: filename={}, state={}, line={}, regex={}".format(
                filename, state, line, regex
            )
        )

    new_filecontents = "\n".join(newlines)
    # Append newline to EOF
    new_filecontents = new_filecontents + "\n"

    # Write new lines to file
    with open(filename, "w") as fp:
        fp.write(new_filecontents)

    return new_filecontents


def validate_lineinfile_input(line, regex, state):
    """
    Check if the given input is valid for the lineinfile function.
    Raise AssertionError if input is invalid.
    :param str line: Line that should be managed in the file. Doesn't go together
    with the regex parameter.
    :param str regex: Pattern to look for in the file. Doesn't go together with
    the line parameter
    :param str state: State of the line in the file. Should be either 'present'
    or 'absent'
    :rtype: None|AssertionError
    """
    if state not in ["present", "absent"]:
        raise AssertionError("State should be either present or absent")

    if not (line or regex):
        raise AssertionError(
            "Need to have at least one of line or regex parameter in lineinfile"
        )

    if regex and line and state == "absent":
        raise AssertionError(
            "Unclear behaviour for state absent on both regex and line"
        )

    if not line and state == "present":
        raise AssertionError("Line is required if state is present")


def remove_line(lines, line=None, regex=None):
    """
    Remove line from file.
    :param list[str, ..] lines: Lines that are currently in the file.
    :param str line: Line that should be managed in the file. Doesn't go together
    with the regex parameter.
    :param str regex: Pattern to look for in the file. Doesn't go together with
    the line parameter
    :return: Lines that remain in the file
    """
    if line:
        delete_line = line
    else:
        delete_line = find_matching_line(lines, regex)
    if delete_line in lines:
        lines.remove(delete_line)
    return lines


def add_line(lines, line=None, index=None):
    """
    Add line to the file at the specified index. If index is not specified,
    line will be appended at the end of the list.
    :param list[str, ..] lines: Lines that are currently in the file.
    :param str line: Line that should be added.
    :param int|None index: Add line to this specific index.
    :return: Lines with the newly added line.
    """
    if line not in lines:
        if index:
            lines.insert(index, line)
        else:
            lines.append(line)
    return lines


def replace_line(lines, line, regex):
    """
    Replace line that matches the regex with the given line. Will add the line
    at the end of the file if no line matches the regex pattern.
    :param list[str, ..] lines: Lines that are currently in the file.
    :param str line: Line that should be managed in the file. Doesn't go together
    with the regex parameter.
    :param str regex: Pattern to look for in the file. Doesn't go together with
    the line parameter
    :return: Lines with the replaced line.
    """
    delete_line = find_matching_line(lines, regex)
    index = None
    if delete_line:
        index = lines.index(delete_line)
        lines = remove_line(lines, line=delete_line)
    lines = add_line(lines, line=line, index=index)
    return lines


def find_matching_line(lines, regex):
    """
    Find the first line that matches the given pattern.
    Returns None if no strings match the line
    :param list[str, ..] lines: Lines that are currently in the file.
    :param str regex: Regex file that should match a line.
    :rtype: str|None
    """
    pattern = re.compile(regex)
    for l in lines:
        if pattern.search(l):
            return l
