import os
import sys
from logging import DEBUG, INFO, Formatter, Logger, StreamHandler
from logging.handlers import SysLogHandler


def create_syslog_handler(address="/dev/log", name_override=None, level=DEBUG):
    """Create a SysLogHandler with proper formatter and decent default settings

    :param str|tuple address: syslog address, string for unix or (address, port) for UDP
    :param str name_override: override the logger name in formatter
    :param int level: logging level for this handler
    :rtype: logging.handlers.SysLogHandler
    """
    if isinstance(address, str) and not os.path.exists(address):
        # Docker doesn't have syslog daemon, use default settings instead to connect to localhost
        syslog_handler = SysLogHandler()
    else:
        syslog_handler = SysLogHandler(address=address)
    syslog_handler.setLevel(level)
    name_in_format = "%(name)s" if name_override is None else name_override
    syslog_handler.setFormatter(
        Formatter("[%(process)d] {} [%(levelname)s] %(message)s".format(name_in_format))
    )
    return syslog_handler


def create_stream_handler(
    stream=None, level=INFO, log_format="%(levelname)s: %(message)s"
):
    """Create a StreamHandler with decent default settings for stdout

    :param file stream: file like object (default is sys.stdout)
    :param int level: logging level for this handler (default is INFO)
    :param str log_format: format of the log message
    :rtype: logging.handlers.StreamHandler
    """
    if not stream:
        stream = sys.stderr
    stream_handler = StreamHandler(stream)
    stream_handler.setLevel(level)
    stream_handler.setFormatter(Formatter(log_format))
    return stream_handler


def setup_logging(logger: Logger, verbose: bool = False):
    """Default logging setup with syslog handler and stdout stream handler
    :param logging.Logger logger: logger to configure
    :param bool verbose: whether debug level messages should be written to stdout
    """
    logger.setLevel(DEBUG)
    logger.addHandler(create_syslog_handler(level=DEBUG if verbose else INFO))
    logger.addHandler(create_stream_handler(level=DEBUG if verbose else INFO))
