"""
kamikaze3.common.postfix
========================
functionality related to postfix, processing logs, etc.
"""

import re
from datetime import datetime
from logging import getLogger
from os import path
from typing import List, Set

from kamikaze3.common.api import HttpApiClient
from kamikaze3.common.syslog import log_file_earliest_entry, log_file_lines
from kamikaze3.utils import get_app_name

LOG_FILE = "/var/log/mail.log"
# our dns_check endpoint does a bunch of checks, not all of them relevant for mail, these are:
DNS_CHECK_MAIL_CHECKS = ("mail_auth", "dkim_hypernode", "spf_record")
# These domains we should not check
SPECIAL_DOMAINS = ["hypernode.io"]  # We sign with a different DKIM key for this domain

logger = getLogger(__name__)


def is_mail_failed_log(log: str) -> bool:
    """Return True if the log line is for a deferred/bounced mail message"""
    log = log.upper()
    return ("STATUS=DEFERRED" in log) or ("STATUS=BOUNCE" in log)


def get_transip_auth_issue(log: str) -> str:
    """If the log relates to TransIP mail authentication failure, return the details of the issue parsed out from
    the log line, otherwise return empty string
    """

    # sample log line:
    # Aug 17 11:28:11 5vi9lw-erikhyperdev-magweb-cmbl postfix/smtp[5532]: 5C5DC40E4F: to=<test@example.org>, \
    # relay=vps.transip.email[86.105.244.7]:587, delay=4.3, delays=0.02/0.01/1.1/3.1, dsn=4.7.0, status=deferred \
    # (host vps.transip.email[86.105.244.7] said: 421 4.7.0 No valid 'x-transip-mail-auth' DNS record found for \
    # 'erikhyperdev.hypernode.io'. Please see https://www.transip.eu/knowledgebase/entry/349 for more information. \
    # (4BVWxl3c21z7t9L) (in reply to end of DATA command))
    match = re.search(
        r"(No valid 'x-transip-mail-auth' DNS record found for.+)\s+\(.+\)\s+\(in reply to",
        log,
        re.IGNORECASE,
    )
    return match.group(1).strip() if match else ""


def get_transip_spf_issue(log: str) -> str:
    """If the log relates to TransIP mail SPF check failure, return the details of the issue parsed out from
    the log line, otherwise return empty string
    """

    # we know of 2 types of errors:
    # sample log line 1:
    # Apr 22 13:44:30 wifamk-testfarzad-magweb-cmbl postfix/smtp[5628]: 013A860BF0: to=<test@example.org>, relay=vps.transip.email[86.105.244.7]:587, \
    # delay=0.51, delays=0.03/0.03/0.18/0.26, dsn=4.7.0, status=deferred (host vps.transip.email[86.105.244.7] said: 421 4.7.0 SPF records for testfarzad.hypernode.io \
    # have not been configured correctly. Please see https://www.transip.eu/knowledgebase/entry/350 for more information. \
    # (4FQzDZ2Csbz18H4J) (in reply to end of DATA command)
    #
    # sample log line 2:
    # Apr 23 13:01:09 cj8jyi-vdloo-magweb-aws postfix/smtp[16986]: 9C20C20DF6: to=<test@example.org>, relay=vps.transip.email[86.105.244.7]:587, \
    # delay=0.48, delays=0/0/0.25/0.22, dsn=4.7.0, status=deferred (host vps.transip.email[86.105.244.7] said: 421 4.7.0 An Internal Server Error has occurred. \
    # SPF records for testfarzad.hypernode.io could not be verified. Please try again later. Please see https://www.transip.eu/question/110000767 for more information. \
    # (4FRZD471jBz2Bd3L) (in reply to end of DATA command)
    match = re.search(
        r"(SPF records for.+)\s+\(.+\)\s+\(in reply to", log, re.IGNORECASE
    )
    return match.group(1).strip() if match else ""


def get_transip_dkim_issue(log: str) -> str:
    """If the log relates to TransIP mail DKIM verification failure, return the details of the issue parsed out from
    the log line, otherwise return empty string
    """

    # sample log line:
    # May 17 08:08:16 wifamk-testfarzad-magweb-cmbl postfix/smtp[5821]: 0DBAD60AEB: to=<test@example.org>, \
    # relay=vps.transip.email[86.105.244.7]:587, delay=0.36, delays=0.03/0.01/0.18/0.14, dsn=4.7.0, status=deferred \
    # (host vps.transip.email[86.105.244.7] said: 421 4.7.0 \
    # message failed DKIM verification with a temporary error (4FkBb41y7Gz1BG2s) (in reply to end of DATA command))
    match = re.search(
        r"(message failed DKIM verification with .+)\s+\(.+\)\s+\(in reply to",
        log,
        re.IGNORECASE,
    )
    return match.group(1).strip() if match else ""


def find_failed_transip_mails_from_logs(
    after: datetime = None, limit: int = 0
) -> List[str]:
    """Find the first occurrence of an authentication failure to TransIP mail in postfix logs.
    Returns the description of the failure as logged by postfix.
    Return empty string if no such log was found.
    Search faster by specifying a datetime object for "after".
    """
    mail_issues = []  # type: List[str]
    if not path.exists(LOG_FILE):
        logger.info("no mail log file exists on '{}'".format(LOG_FILE))
        return mail_issues

    for line, line_num in log_file_lines(
        LOG_FILE, filter_=is_mail_failed_log, after=after
    ):
        mail_issue = get_transip_auth_issue(line)
        if mail_issue:
            logger.info(
                "found mail transip auth failure log in line {} of {}".format(
                    line_num, LOG_FILE
                )
            )
            mail_issues.append(mail_issue)
        else:
            mail_issue = get_transip_spf_issue(line)
            if mail_issue:
                logger.info(
                    "found mail transip SPF failure log in line {} of {}".format(
                        line_num, LOG_FILE
                    )
                )
                mail_issues.append(mail_issue)
            else:
                mail_issue = get_transip_dkim_issue(line)
                if mail_issue:
                    logger.info(
                        "found mail transip DKIM failure log in line {} of {}".format(
                            line_num, LOG_FILE
                        )
                    )
                    mail_issues.append(mail_issue)

        if limit and len(mail_issues) >= limit:
            logger.warning(
                "reached the limit on transip mail failures of {} in file {}. skipping the rest".format(
                    limit, LOG_FILE
                )
            )
            break

    if len(mail_issues) < 1:
        logger.info("didn't find any mail issue logs in {}".format(LOG_FILE))
    return mail_issues


def get_from_warn_domain(log: str) -> str:
    """Get the domain a WARN mail was sent from from the relevant log line"""
    # sample log line:
    # Aug 12 09:14:45 abbrwj-vdloo-magweb-cmbl postfix/smtp[19809]: BA38240E62: warning: header From: Komkommerbanaan.nu <komkommerbanaan.nu@komkommer-jaja.nl>
    if "<" in log and not log.endswith(">"):
        return ""
    match = re.search(r"@(.*?)(>?)$", log)
    return match.group(1).strip().split(" ")[0] if match else ""


def get_from_qmgr_domain(log: str) -> str:
    """Get the domain a qmgr mail was sent from from the relevant log line"""
    # sample log line:
    # Jun 30 13:47:48 s2ltgs-phvr-magweb-aws postfix/qmgr[82633]: 8866C20A66: from=<customerservice@phvr.fr>, size=16534, nrcpt=1 (queue active)
    match = re.search(r"from=<.+@(.+)>,\s+size=", log, re.IGNORECASE)
    return match.group(1).strip() if match else ""


def is_mail_from_relevant_warn_domain(log: str) -> bool:
    """Return True if the log line is for a WARN header check"""
    return "warning: header From:" in log


def is_mail_from_relevant_qmgr_domain(log: str) -> bool:
    """Return True if the log line is for a deferred/bounced mail message"""
    log = log.upper()
    # a couple of quick returns for performance
    if "FROM=" not in log:
        return False
    if "POSTFIX/QMGR" not in log:
        return False
    if "NOREPLY@HYPERNODE.IO" in log:
        # our default address, not relevant here
        return False
    if "@" not in log:
        # skip invalid or empty addresses, eg a mail from a system user like root or app
        return False
    return True


def add_domains_by_filter(
    mail_domains: set, domain_filter: callable, domain_getter: callable
) -> Set[str]:
    """
    Add newly found domains to the set based on the
    filter function specified.
    :param set mail_domains: The set of found domains
    :param func domain_filter: The filter to apply to find lines
    matching our domain pattern
    :param func domain_getter: The function that returns our parsed
    domain from the identified relevant line
    :return set: The updated set
    """
    for line, _ in log_file_lines(LOG_FILE, filter_=domain_filter):
        domain = domain_getter(line)
        if domain:
            mail_domains.add(domain)
    return mail_domains


def find_relevant_mail_domains_from_logs() -> Set[str]:
    """Find all domains used to mail from in the mail log."""
    mail_domains = set()
    if not path.exists(LOG_FILE):
        logger.info("no mail log file exists on '{}'".format(LOG_FILE))
        return mail_domains
    logger.info(
        "found mail logs from {} onward...".format(log_file_earliest_entry(LOG_FILE))
    )
    mail_domains = add_domains_by_filter(
        mail_domains, is_mail_from_relevant_qmgr_domain, get_from_qmgr_domain
    )
    mail_domains = add_domains_by_filter(
        mail_domains, is_mail_from_relevant_warn_domain, get_from_warn_domain
    )
    mail_domains = {
        mail_domain
        for mail_domain in mail_domains
        if mail_domain not in SPECIAL_DOMAINS
    }
    return mail_domains


def check_dns_for_domain(domain):
    """gather check results for the domain
    Return str
    """
    cli = HttpApiClient()
    app_name = get_app_name()
    response = cli.get("v2/check_domain/{}/?domain={}".format(app_name, domain))
    return {check: response[check]["level"] for check in DNS_CHECK_MAIL_CHECKS}
