import re
from glob import glob
from logging import getLogger

from OpenSSL.crypto import FILETYPE_PEM, load_certificate
from pyasn1.codec.der import decoder as der_decoder
from pyasn1_modules.rfc2459 import SubjectAltName

ROOT_SSL_PATH_PATTERN = "/etc/nginx/ssl/*.crtkeyca"
APP_SSL_PATH_PATTERN = "/data/web/nginx/ssl/*.crtkeyca"

logger = getLogger(__name__)


def get_alternative_names(crt):
    """
    Get all alternative DNS-names from a PEM-encoded certificate
    ref: http://www.pyopenssl.org/en/stable/api/crypto.html?highlight=get_extension#OpenSSL.crypto.X509
    :param OpenSSl.crypto.X509 crt: a PEM-encoded X509 certificate
    :return: a list with all alternative names found in the certificate.
        If the certificate has no 'Alternative Names' extension, return an empty list.
    :rtype: list[str]
    """
    for i in range(crt.get_extension_count()):
        ext = crt.get_extension(i)
        if ext.get_short_name() == b"subjectAltName":
            # http://pyasn1.sourceforge.net/docs/codec/der/contents.html#pyasn1.codec.der.decoder.decode
            # DER decoder returns a tuple containing a pyasn1 object it recovered and the unprocessed remains.
            # we are only interested in the pyasn1 object, which contains the SANs values:
            decoded_data = der_decoder.decode(
                ext.get_data(), asn1Spec=SubjectAltName()
            )[0]
            sans = []
            for d in decoded_data:
                san = d.getComponent()._value
                # For version compatibility, we need to check for the string type
                sans.append(san.decode() if isinstance(san, bytes) else san)
            return sans
    return []


def parse_x509(crtkeyca_path):
    """
    Parse the x509 from the crtkeyca file.
    :param str crtkeyca_path: Path to the crtkeyca file
    :return OpenSSl.crypto.X509: a PEM-encoded X509 certificate
    """
    with open(crtkeyca_path, "rb") as fp:
        return load_certificate(FILETYPE_PEM, fp.read())


def san_to_regex_pattern(san):
    """
    Transform a Subject Alternative Name to a regex pattern
    which we can use to compare to domainnames.
    :param str san: Subject Alternative Name from a certificate.
    :return str: Regex compatible string
    """
    escaped_san = re.escape(san)
    # Only transform the left-most wildcard: https://tools.ietf.org/html/rfc6125#section-6.4.3
    wildcard_san = re.sub(r"^\\\*", r"[^\.]*", escaped_san)
    # wrap san in ^$ to prevent example.com from matching example.com.nl.jp
    san_re = "^{}$".format(wildcard_san)
    return san_re


def is_servername_in_certificate(servername, crtkeyca_file):
    """
    Check if the given servername is supported by the certificate
    in the given file.
    :param str servername: Servername for which we check if an SSL
    :param str crtkeyca_path: Path to the crtkeyca file
    :rtype: bool
    """
    crt = parse_x509(crtkeyca_file)
    names = get_alternative_names(crt)
    logger.debug("Found names {} in certificate file {}".format(names, crtkeyca_file))
    for san in names:
        pattern = san_to_regex_pattern(san)
        if re.match(pattern, servername):
            return True
    return False


def get_servername_ssl_cert(servername, ssl_path_glob=ROOT_SSL_PATH_PATTERN):
    """
    Get the SSL certificate file that matches the given
    servername. This can also mean that wildcards match this
    servername.
    :param str servername: Servername for which we check if an SSL
    certificate matches.
    :param str ssl_path_glob: Path in which we look for crtkeyca files
    :return str|None: Path to the certificate which matches the
    servername, or None if no certificate matches.
    """
    for crtkeyca_file in glob(ssl_path_glob):
        if is_servername_in_certificate(servername, crtkeyca_file):
            logger.info(
                "{} matches the certificate {}".format(servername, crtkeyca_file)
            )
            return crtkeyca_file
    logger.info("Servername {} not found in root SSL certificates".format(servername))
