"""
kamikaze3.common.syslog
=======================
functionality related to system syslog, parsing logs, etc.
"""

from collections.abc import Callable, Generator
from datetime import datetime


def parse_log_datetime(log: str) -> datetime:
    """Parse a datetime from a given log line"""

    # sample:
    # Sep  1 12:19:43 3nlisx-testfarzad-magweb-cmbl crontab[6796]: (app) LIST (app)
    month, day, time_ = (v.strip() for v in log.strip().split(maxsplit=3)[0:3])
    if len(day) == 1:  # zero pad the day as strptime reads
        day = f"0{day}"
    try:
        return datetime.strptime(
            f"{datetime.today().year} {month} {day} {time_}",
            "%Y %b %d %H:%M:%S",
        )
    except ValueError:
        # try parsing as iso8601
        value = log.strip().split()[0].strip()
        return datetime.fromisoformat(value)


def log_file_lines(
    filename: str, filter_: Callable[[str], bool] = None, after: datetime = None
) -> Generator[tuple[str, int], None, None]:
    """A generator to read log file lines, with features to filter out unwanted lines quickly.
    Yields a tuple of 2 items to the caller:
    - The log line (whitespaces stripped)
    - line number in the log file

    Empty lines (with only whitespaces) are skipped.
    A filter_ function helps to skip unwanted lines faster (by returning Falsy values), keep it simple for fast
    selection, filter works before checking for timestamps.

    If a datetime is passed for "after", logs before that time will be skipped.
    """
    counter = 0
    check_time = True if after else False

    # infuse filter_ with skipping empty lines into a lambda, to avoid checking if filter_ is passed for each line.
    # an if expression was less readable, hence the if/else block
    if filter_:
        _skip_line = lambda x: x == "" or not filter_(x)
    else:
        _skip_line = lambda x: x == ""

    # @TODO: replace this "while True" and "readline()" with iterating over the file handle.
    # Due to lack of iterator interface support from "mock_open" in Python 3.5,
    # we couldn't use the iterator way, otherwise tests will fail on Python 3.5.
    # I prefer the iterator way, and didn't like to change code for this, but
    # our Xenial packaging won't work otherwise. Really Looking forward to not having
    # to support Xenial and Python 3.5.
    with open(filename) as fh:
        while True:
            line = fh.readline()
            if not line:  # EOF
                break
            line = line.strip()
            counter += 1
            if _skip_line(line):
                continue
            if check_time:
                if parse_log_datetime(line) < after:  # type: ignore
                    continue
                check_time = False  # log is later than after, from now on no need to check for log timestamps anymore
            yield line, counter


def log_file_earliest_entry(filename: str) -> datetime | None:
    """Parse a datetime from the first line of a log and return it,
    so we can tell how old the logs are"""
    with open(filename) as fh:
        line = fh.readline()
        if not line:
            return None
        return parse_log_datetime(line.strip())
