"""
kamikaze3.common.system
~~~~~~~~~~~~~~~~~~~~~~~
Common system/os related functionality
"""

import os
import platform
import socket

from kamikaze3.common.filesystem import APP_USER_UID


def walk_with_depth(path, depth=1):
    """
    Walk function that performs the same as `os.walk` with the added ability of specifying a depth. Specifying a depth
    of 1 means we will walk the directory of `path` and all subdirectories, but no further. Same can be applied for
    depth 2, accept we go one more subdirectory deeper etc.
    :param path: The path to walk
    :param depth: Depth to search for. Depth of 1 means that we will only go one directory deep for searches.
    :return: regular `os.walk` return values
    """
    # Based off of https://stackoverflow.com/a/234329/1294864
    # We need to strip any trailing `/` here, otherwise our root separator count won't be accurate.
    # I.e. if we receive `/my/path/` instead of `/my/path` it means we will check for a separator count of
    # 4 (root_separator_count + depth). This means we would accidentally search one directory too deep, because
    # os.walk will return current_path values with trailing `/` stripped. So the separator count for
    # `current_path` would be 2, and our maximum count would be 4. Meaning we will search at a depth of 2
    # instead of 1.
    path = path.rstrip(os.path.sep)
    root_separator_count = path.count(os.path.sep)
    for current_path, dirs, files in os.walk(path):
        yield current_path, dirs, files
        current_separator_count = current_path.count(os.path.sep)
        if root_separator_count + depth <= current_separator_count:
            del dirs[:]


def user_is_root():
    """
    Returns true if the users if root, returns false if the user is not root
    :return: boolean
    """
    return os.getuid() == 0


def user_is_app():
    """
    Returns true if the user is app, returns false if the user is not app
    :return: boolean
    """
    return os.getuid() == APP_USER_UID


def is_hostname_localhost(host: str):
    """
    Returns true if the host is localhost, returns false if the host is not localhost
    :param str host:
    :return: boolean
    """
    host = socket.gethostbyname(host)
    return host in ["127.0.0.1", "::1", "0.0.0.0"]


def get_os_release() -> str | None:
    if hasattr(platform, "freedesktop_os_release"):
        return platform.freedesktop_os_release()["VERSION_CODENAME"]

    with open("/etc/os-release") as f:
        for line in f:
            if line.startswith("VERSION_CODENAME"):
                return line.split("=")[1].strip().strip('"')

    return None
