from logging import getLogger
from pathlib import Path
from typing import List

import dasbus
from dasbus.typing import unwrap_variant

from kamikaze3.dbus.interfaces.systemd import SYSTEMD

logger = getLogger(__name__)


class SystemdException(BaseException):
    pass


class Unit:
    def __init__(self, unit_path: str):
        self._unit_attributes = {}
        self._unit_interface = SYSTEMD.get_proxy(
            object_path=unit_path, interface_name="org.freedesktop.systemd1.Unit"
        )
        self._unit_properties_interface = SYSTEMD.get_proxy(
            object_path=unit_path, interface_name="org.freedesktop.DBus.Properties"
        )
        self._load_unit_properties()

    def _load_unit_properties(self):
        self._unit_attributes.clear()
        properties = self._unit_properties_interface.GetAll(
            "org.freedesktop.systemd1.Unit"
        )
        for key, value in properties.items():
            self._unit_attributes[str(key)] = unwrap_variant(value)

    def __getitem__(self, item: str):
        if item in self._unit_attributes:
            return self._unit_attributes[item]
        raise KeyError(item)


class Service(Unit):
    def __init__(self, unit_path: str):
        super().__init__(unit_path)
        self._service_attributes = {}
        self._service_interface = SYSTEMD.get_proxy(
            object_path=unit_path, interface_name="org.freedesktop.systemd1.Service"
        )
        self._service_properties_interface = SYSTEMD.get_proxy(
            object_path=unit_path, interface_name="org.freedesktop.DBus.Properties"
        )
        self._load_service_properties()

    def _load_service_properties(self) -> None:
        self._service_attributes.clear()
        properties = self._service_properties_interface.GetAll(
            "org.freedesktop.systemd1.Service"
        )
        for key, value in properties.items():
            self._service_attributes[str(key)] = unwrap_variant(value)

    def __getitem__(self, item: str):
        if item in self._service_attributes:
            return self._service_attributes[item]
        elif item in self._unit_attributes:
            return self._unit_attributes[item]
        raise KeyError(item)


def get_systemd_manager():
    return SYSTEMD.get_proxy(interface_name="org.freedesktop.systemd1.Manager")


def get_timer_files():
    systemd_manager = get_systemd_manager()
    list_unit_files = systemd_manager.ListUnitFiles()

    result: List[Path] = []
    for unit_file in list_unit_files:
        unit_file_path: str = unit_file[0]
        if unit_file_path.endswith(".timer"):
            result.append(Path(unit_file_path))

    return result


def get_unit_path_by_name(name: str) -> str:
    systemd_manager = get_systemd_manager()
    try:
        return systemd_manager.GetUnit(name)
    except dasbus.error.DBusError as e:
        raise SystemdException(e)


def get_unit_by_name(name: str) -> Unit:
    return Unit(get_unit_path_by_name(name))


def get_service_by_name(name: str) -> Service:
    if not name.endswith(".service"):
        name = f"{name}.service"

    return Service(get_unit_path_by_name(name))


def get_service_active_state(name: str) -> str:
    service = get_service_by_name(name)
    return service["ActiveState"]


def check_service_running(name: str) -> bool:
    try:
        service_state = get_service_active_state(name)
        return service_state == "active"
    except (dasbus.error.DBusError, SystemdException):
        return False


def get_service_sub_state(name: str) -> str:
    service = get_service_by_name(name)
    return service["SubState"]
