from grp import getgrall
from pwd import getpwall

# UIDs for the following users: root = 0, toor = 0, app = 1000, support = 1002
UID_CHECKS = ["0", "1000", "1002"]


class SubUidInfo:
    def __init__(self, user, subuid, subuid_count):
        self.user = user
        self.subuid = subuid
        self.subuid_count = subuid_count
        self.start_subuid = self.subuid
        self.end_subuid = self.subuid + max(0, self.subuid_count - 1)

    def __contains__(self, item):
        return self.subuid_count and self.start_subuid <= item <= self.end_subuid


def get_subuid_info_for_user(user) -> SubUidInfo:
    """
    Get the subuid information for a user
    :param user: the user to get the subuid information for
    :return: SubUidInfo object with the subuid information
    """
    try:
        with open("/etc/subuid", "r") as f:
            for line in f.read().splitlines():
                if line.startswith(user + ":"):
                    _, subuid, subuid_count = line.split(":")
                    return SubUidInfo(user, int(subuid), int(subuid_count))
    except FileNotFoundError:
        pass
    return SubUidInfo(user, 0, 0)


def gather_users_and_groups_information():
    """
    Gather all important information about the users and groups on a Hypernode.
    This is then put in a usable datastructure.
    :return: dict with all necessary information about the users and groups
    """
    all_users = getpwall()
    # Get ids from users to check
    uid_important = {}
    for uid in UID_CHECKS:
        uid_int = int(uid)
        selected_users = [user.pw_name for user in all_users if user.pw_uid == uid_int]
        uid_important[uid] = selected_users
    # Get user their name, home directory and shell
    users = [
        {"name": user.pw_name, "dir": user.pw_dir, "shell": user.pw_shell}
        for user in all_users
    ]
    # Get group name and members
    groups = [
        {"name": group.gr_name, "members": sorted(group.gr_mem)} for group in getgrall()
    ]
    return {"uid_important": uid_important, "users": users, "groups": groups}
