import datetime
import getpass
import json
import logging
import sys
from argparse import ArgumentParser
from operator import xor
from os import EX_IOERR, EX_OK

from kamikaze3.common.env import (
    get_active_php_version_info,
    get_magento_2_install_dir,
    is_magento_1,
)
from kamikaze3.common.log import create_stream_handler
from kamikaze3.diagnostic.node_configuration import (
    get_cron_jobs,
    get_magento_info,
    get_magento_one_cache_info,
    get_media_folder_size,
)
from kamikaze3.diagnostic.text import result_to_human_text

logger = logging.getLogger("kamikaze3")
logger.setLevel(logging.INFO)
logger.addHandler(create_stream_handler(log_format="%(message)s"))


def parse_arguments(cmdline=None):
    parser = ArgumentParser(description="Run diagnostic for Hypernode and Magento.")
    parser.add_argument(
        "--json", help="output diagnostic in json format", action="store_true"
    )
    parser.add_argument(
        "--path",
        help="where to look for the magento installation",
        nargs="?",
        default="/data/web/",
    )
    return parser.parse_args(cmdline)


def check_if_run_by_app_user():
    effective_user = getpass.getuser()

    if effective_user != "app":
        sys.stderr.write(
            "You should run the diagnostic tool as app user, not as {}.\n".format(
                effective_user
            )
        )
        exit(1)


def main():
    args = parse_arguments()
    check_if_run_by_app_user()

    results = run_diagnostic(args.path)
    if not results:
        logger.info(
            "Inconclusive data reached. This might be due to multiple magento installations being present. "
            "You can try specifying --path to search in a specific location. Just now we tried the "
            "following path: {}".format(args.path)
        )
        return EX_IOERR
    print(json.dumps(results) if args.json else result_to_human_text(results))
    return EX_OK


def run_diagnostic(path):
    is_magento_one = bool(is_magento_1())
    is_magento_two = bool(get_magento_2_install_dir(path))

    results = {}
    if xor(is_magento_one, is_magento_two):
        magento_info = get_magento_info(is_magento_one, path)

        results = {
            "is_magento_one": is_magento_one,
            "php_version": get_active_php_version_info(),
            "media_size": get_media_folder_size(),
            "cron_jobs": get_cron_jobs(),
            "date_time": get_current_time(),
            "mag_version_string": magento_info["version"],
            "mag_edition": magento_info["edition"],
        }

        if is_magento_one:
            results["caching_fpc"] = get_magento_one_cache_info().get("fpc", "N/A")
        else:
            results["mag_application_mode"] = magento_info["application_mode"]

    return results


def get_current_time():
    now = datetime.datetime.now()
    return now.strftime("%Y-%m-%d %H:%M:%S")


if __name__ == "__main__":
    main()
