import json
import subprocess
from subprocess import CalledProcessError, check_output

from kamikaze3.common.env import get_magento_2_install_dir
from kamikaze3.utils import camel_to_snake


def get_media_folder_size():
    try:
        media_size = (
            check_output(["/usr/bin/du", "/data/web/public/media/", "-hs"])
            .decode()
            .strip()
        )
    except (CalledProcessError, OSError, ValueError) as err:
        raise RuntimeError("Failed to detect media folder size: {}".format(err))

    return media_size.split()[0]


def get_cron_jobs():
    try:
        crons = (
            check_output(["/usr/bin/crontab", "-l"], stderr=subprocess.PIPE)
            .decode()
            .strip()
        )
    except (CalledProcessError, OSError, ValueError):
        return ["N/A"]

    return [
        line
        for line in crons.splitlines()
        if line.lstrip() and not line.lstrip().startswith("#")
    ]


def get_magento_info(is_magento_one, path):
    magerun_version = (
        "/usr/local/bin/magerun" if is_magento_one else "/usr/local/bin/magerun2"
    )
    cwd = "/data/web/public" if is_magento_one else get_magento_2_install_dir(path)
    json_string = "{}"

    if cwd:
        try:
            sys_info_json_string = (
                check_output([magerun_version, "sys:info", "--format=json"], cwd=cwd)
                .decode()
                .strip()
            )
        except (CalledProcessError, OSError, ValueError) as err:
            raise RuntimeError("Failed to retrieve magerun info: {}".format(err))

        json_string = _clean_magerun_json_string(sys_info_json_string)
    return _convert_magerun_dict_to_key_value_dict(json.loads(json_string))


def _clean_magerun_json_string(string):
    # Skip anything that might be printed ahead of the actual JSON. This can be warnings for example:
    # dbModel read resource does not implement Zend_Db_Adapter_Abstract
    # {
    #     "0": {
    return "".join(string.partition("{")[1:])


def _convert_magerun_dict_to_key_value_dict(magerun_dict, key="name", value="value"):
    return {camel_to_snake(d[key]): d[value] for d in magerun_dict.values()}


def get_magento_one_cache_info():
    try:
        cache_list_json_string = (
            check_output(
                ["/usr/local/bin/magerun", "cache:list", "--format=json"],
                cwd="/data/web/public",
            )
            .decode()
            .strip()
        )
    except (CalledProcessError, OSError, ValueError) as err:
        raise RuntimeError("Failed to retrieve cache list: {}".format(err))

    return _convert_magerun_dict_to_key_value_dict(
        json.loads(cache_list_json_string), key="code", value="status"
    )
