import argparse
import asyncio
import json
import logging
import ssl
import traceback
from typing import Optional

import nats
from nats.aio.client import Client

from kamikaze3.common.env import get_hypernode_name
from kamikaze3.event_listener.settings import PROGRAM_NAME
from kamikaze3.event_listener.utils import (
    get_default_nats_server,
    get_default_nats_ssl_context,
)

logger = logging.getLogger(PROGRAM_NAME)


async def error_cb(e: Exception):
    logger.error(f"NATS message handling error: {e} at {traceback.format_exc()}")


async def reconnected_cb():
    logger.info("Got reconnected to NATS...")


async def get_nats_client(server, tls_hostname) -> Client:
    logger.debug(f"Connecting to NATS server on {server}")
    options = {
        "servers": [server],
        "error_cb": error_cb,
        "reconnected_cb": reconnected_cb,
        "drain_timeout": 3,
        "max_reconnect_attempts": 2,  # 3 tries in total
        "connect_timeout": 1,
        "reconnect_time_wait": 1,
    }

    ssl_context = get_ssl_context()
    if ssl_context:
        ssl_ctx = ssl.create_default_context(purpose=ssl.Purpose.SERVER_AUTH)
        ssl_ctx.load_verify_locations(ssl_context.get("ca"))
        ssl_ctx.load_cert_chain(
            certfile=ssl_context.get("crt"), keyfile=ssl_context.get("key")
        )
        options["tls"] = ssl_ctx
        options["tls_hostname"] = tls_hostname

    nc = await nats.connect(**options)
    return nc


def get_ssl_context(args: Optional[argparse.Namespace] = None):
    if args and args.key and args.cert and args.ca:
        return {"crt": args.cert, "key": args.key, "ca": args.ca}
    return get_default_nats_ssl_context()


async def publish_event(
    subject: str,
    data: Optional[str] = None,
    nc: Optional[Client] = None,
    nats_url: Optional[str] = None,
    tls_hostname: Optional[str] = None,
):
    nc = nc or await get_nats_client(
        nats_url or get_default_nats_server(),
        tls_hostname=tls_hostname or get_hypernode_name(),
    )
    logger.debug(f"Publishing event to {subject}")
    await nc.publish(subject, data.encode("utf-8") if data else b"")


async def request_event(
    subject: str,
    data: Optional[str] = None,
    nc: Optional[Client] = None,
    nats_url: Optional[str] = None,
    tls_hostname: Optional[str] = None,
):
    nc = nc or await get_nats_client(
        nats_url or get_default_nats_server(),
        tls_hostname=tls_hostname or get_hypernode_name(),
    )
    logger.debug(f"Requesting event {subject}")
    msg = await nc.request(subject, data.encode("utf-8") if data else b"")
    return msg


def publish_event_sync(
    subject: str, data: Optional[str] = None, nats_url: Optional[str] = None
):
    loop = asyncio.get_event_loop()
    return loop.run_until_complete(
        publish_event(subject=subject, data=data, nats_url=nats_url)
    )


def request_event_sync(
    subject: str, data: Optional[str] = None, nats_url: Optional[str] = None
):
    loop = asyncio.get_event_loop()
    return loop.run_until_complete(
        request_event(subject=subject, data=data, nats_url=nats_url)
    )


def create_response(message, status=0, data=None):
    response_data = {
        "message": message,
        "status": status,
        "data": data,
    }
    return json.dumps(response_data).encode(errors="ignore")
