import json
import logging
from pathlib import Path
from collections.abc import Awaitable, Callable
from subprocess import check_output, CalledProcessError

from nats.aio.msg import Msg

from kamikaze3 import ftp
from kamikaze3.analysis.app_locator import find_applications as _find_applications
from kamikaze3.autoscaling.magento2.main import (
    run_validations as run_magento2_validations,
)
from kamikaze3.common.args import OUTPUT_FORMAT_JSON
from kamikaze3.autoscaling.varnish.main import (
    run_validations as run_varnish_validations,
)
from kamikaze3.common.env import get_hn_common_var
from kamikaze3.dbus.interfaces.nginx_config_reloader import NGINX_CONFIG_RELOADER
from kamikaze3.event_listener.client import create_response
from kamikaze3.event_listener.settings import PROGRAM_NAME

logger = logging.getLogger(PROGRAM_NAME)


def event_handler_wrapper(
    handler: Callable[[Msg], Awaitable]
) -> Callable[[Msg], Awaitable]:
    async def _wrapper(msg: Msg):
        logger.info(f"Received event {msg.subject}")
        resp = await handler(msg)
        logger.debug(f"Handled event {msg.subject}")
        return resp

    return _wrapper


async def reload_hmv(msg: Msg):
    logger.debug("Reloading HMV")
    check_output(
        [
            "/usr/bin/sudo",
            "-uapp",
            "/usr/bin/hypernode-manage-vhosts",
            "--no",
            "--all",
            "--ssl-noclobber",
            "--skip-event",
        ]
    )
    logger.debug("Done reloading HMV")


async def list_vhosts(msg: Msg):
    vhosts = check_output(
        r"/usr/sbin/nginx -T 2>&1 | /usr/bin/grep -E '^\s*server_name ' | /usr/bin/awk '{print$2}' | /usr/bin/cut -d';' -f1 | /usr/bin/sort | /usr/bin/uniq",
        shell=True,
    )
    logger.debug(f"Found vhosts: {vhosts.decode()}")
    return await msg.respond(vhosts)


async def show_fpm_status(msg: Msg):
    status = check_output("/usr/bin/hypernode-fpm-status", shell=True)
    logger.debug(f"Found FPM status: {status.decode()}")
    return await msg.respond(status)


async def show_ftp_users(msg: Msg):
    try:
        users = ftp.list_users()
    except OSError:
        logger.error("IOError: Could not list FTP users")
        return await msg.respond(
            create_response(message="Could not list FTP users", status=1)
        )

    logger.debug(f"Found FTP users: {users}")
    return await msg.respond(create_response(message=None, status=0, data=users))


async def add_ftp_user(msg: Msg):
    try:
        data = json.loads(msg.data.decode())
    except json.JSONDecodeError as e:
        logger.error(f"FTP user could not be added due {str(e)}")
        return await msg.respond(
            create_response(message="Ftp user could not be added", status=1)
        )

    for required_field in ["user", "password", "homedir"]:
        if required_field in data:
            continue
        logger.error(f"FTP user could not be added. Missing {required_field}")
        return await msg.respond(
            create_response(
                message=f"FTP user could not be added. Missing {required_field}",
                status=1,
            )
        )

    try:
        user = ftp.add_user(data["user"], data["password"], data["homedir"])
    except SystemExit as e:
        logger.error(f"FTP user could not be added due {str(e)}")
        return await msg.respond(create_response(message=str(e), status=1))

    logger.info(f"FTP user added: {data['user']} for {data['homedir']}")
    return await msg.respond(create_response(message=None, status=0, data=user))


async def remove_ftp_user(msg: Msg):
    try:
        data = json.loads(msg.data.decode())
    except json.JSONDecodeError as e:
        logger.error(f"FTP user could not be removed due {str(e)}")
        return await msg.respond(
            create_response(message="Ftp user could not be removed", status=1)
        )

    if "user" not in data:
        logger.error("FTP user could not be removed. Missing username in msg")
        return await msg.respond(
            create_response(
                message="FTP user could not be removed, field user is required",
                status=1,
            )
        )

    try:
        user = ftp.remove_user(data["user"])
    except SystemExit as e:
        logger.error(f"FTP user could not be removed due {str(e)}")
        return await msg.respond(create_response(message=str(e), status=1))

    logger.info(f"FTP user removed: {data['user']}")
    return await msg.respond(create_response(message=None, status=0, data=user))


# Useful for debugging
async def pong(msg: Msg):
    logger.debug(f"Received ping from {msg.subject}")
    return await msg.respond(b"pong")


async def reload_nginx(msg: Msg):
    logger.debug("Triggering nginx reload")
    nginx_config_reloader_proxy = NGINX_CONFIG_RELOADER.get_proxy()
    nginx_config_reloader_proxy.Reload()
    return await msg.respond(b"reloaded")


async def find_applications(msg: Msg):
    try:
        result = _find_applications(Path("/data/web"))
        installed_apps = json.dumps(list(map(lambda x: x.serialize(), result)))
    except OSError:
        logger.error("OSError: Could not locate installed applications")
        return await msg.respond(
            create_response(message="Could not locate installed applications", status=1)
        )

    logger.debug("Found installed applications: {}".format(len(result)))
    return await msg.respond(
        create_response(message=None, status=0, data=installed_apps)
    )


async def check_varnish_configuration_for_autoscaling(msg: Msg):
    result = []
    try:
        data = run_varnish_validations(OUTPUT_FORMAT_JSON)
        result.extend(json.loads(data))
    except OSError:
        logger.error("OSError: Could not show VCL")
        return await msg.respond(
            create_response(
                message="Could not validate Varnish status and configuration", status=1
            )
        )

    logger.debug("Varnish validation Status: {}".format(data[0]))
    return await msg.respond(create_response(message=None, status=0, data=result))


async def check_magento2_configuration_for_autoscaling(msg: Msg):
    try:
        data = json.loads(msg.data.decode())
    except json.JSONDecodeError as e:
        logger.error(
            f"Magento2 configuration could not be checked for autoscaling due to {str(e)}"
        )
        return await msg.respond(
            create_response(
                message="Magento2 configuration could not be checked", status=1
            )
        )

    if "path" not in data:
        logger.error("Magento2 configuration could not be checked. Missing path in msg")
        return await msg.respond(
            create_response(
                message="Magento2 configuration could not be checked, field path is required",
                status=1,
            )
        )

    result = []
    try:
        response = run_magento2_validations(OUTPUT_FORMAT_JSON, data["path"])
        result.extend(json.loads(response))
    except OSError:
        logger.error("OSError: Could not validate Magento2 configuration")
        return await msg.respond(
            create_response(
                message="Could not validate Magento2 configuration", status=1
            )
        )

    logger.debug("Magento2 validation Status: {}".format(data[0]))
    return await msg.respond(create_response(message=None, status=0, data=result))


async def show_autoscale_status(msg: Msg):
    status = []
    try:
        check_output(["/usr/bin/systemctl", "is-active", "--quiet", "haproxy"])
    except CalledProcessError:
        status.append("Haproxy service not available and ready for use.")

    cluster_ip = get_hn_common_var("cluster_ip", None)
    if not cluster_ip:
        status.append("Cluster IP not configured")
    else:
        try:
            check_output(["/usr/bin/ping", "-c1", cluster_ip])
        except CalledProcessError:
            status.append(f"Unable to ping configured cluster_ip {cluster_ip}")

    response_data = {
        "status": "1" if not status else "0",
        "data": status,
    }

    return await msg.respond(json.dumps(response_data).encode(errors="ignore"))
