import logging
import os
from argparse import ArgumentParser
from pprint import pprint
from socket import getfqdn
from subprocess import CalledProcessError, check_output

import requests
from requests.status_codes import codes

from kamikaze3.common.api import get_hypernode_api_token
from kamikaze3.common.env import host_is_a_docker, host_is_a_vagrant
from kamikaze3.firewall.iptables import Firewall, FirewallException
from kamikaze3.firewall.log import configure_logging
from kamikaze3.settings import USR_CNF

FLAG = "-f"


logger = logging.getLogger("kamikaze3.firewall")


def _get_firewall_url():
    try:
        format_url = check_output((USR_CNF, FLAG)).decode("utf-8").strip()
    except (FileNotFoundError, CalledProcessError):
        raise RuntimeError("Unable to retrieve hn-api URL from USR_CNF")
    return format_url


def parse_args(args=None):
    parser = ArgumentParser(
        description="Script to pull the centralized Hypernode firewall "
        "to this node and apply via iptables"
    )
    parser.add_argument(
        "-v", "--debug", action="store_true", help="Enable debug logging"
    )
    parser.add_argument(
        "-d",
        "--dry-run",
        action="store_true",
        help="Write rules to STDOUT instead " "of applying to iptables",
    )
    parser.add_argument(
        "--flush-firewall",
        action="store_true",
        help="Flush the firewall chain and exit (for debugging)",
    )
    parser.add_argument(
        "--safe-mode",
        action="store_true",
        help="Enable firewall safe-mode " "(ask before making dangerous changes)",
    )

    return parser.parse_args(args=args)


def handle_errors(response: requests.Response) -> None:
    if response.status_code == codes.UNAUTHORIZED:
        logger.warning(
            "hn-api is returning a 401 (unauthorized) error. "
            "If this script is running on a test hypernode or locally "
            "created hypernode, this behavior is expected."
        )
        exit(0)
    response.raise_for_status()


def get_firewall_entries_from_hn_api(url):
    """
    Tries to get a JSON object from hn-api <URL>
    :param url: str: The URL to get
    :return: json-object / object: If json then json object, otherwise the thing that was returned by hn-api
    """
    response = requests.get(
        url,
        headers={
            "Authorization": "Token {}".format(get_hypernode_api_token()),
            "Content-Type": "application/json",
            "Accept": "application/json",
            "User-Agent": "{}:{}".format(getfqdn(), "hypernode-kamikaze3"),
        },
    )
    handle_errors(response)
    return response.json()


def add_chain_rules(requested, firewall, table="mangle"):
    """
    Process firewall changes
    First remove old entries (not found in requested, but present in current)
    Then add new entries (not found in current, but present in requested)
    :param requested: tuple: The request rules (or requested state)
    :param table: str: The IPtables table to work on
    :param firewall: iptables.Firewall object
    """
    # This method assumes that the firewall is already correctly configured and that the rules are correct!
    # Add new entries
    for rule in requested:
        logger.debug("Adding rule: {} to the {} table".format(rule, table))
        if not firewall.add_iptables_rule(rule, table):
            logger.warning("Adding firewall rule failed")


def construct_requested_rules_tuple(entries, firewall):
    """
    Make a set of requested rules based on the info retrieved from hn-api
    :param entries: list of dicts: The requested state for this action
    :param firewall: iptables.Firewall object
    :return: set: The requested state for this action
    """
    requested_state = []
    for entry in entries:
        if entry["action"] == "deny":
            rule = firewall.construct_iptables_rule(
                entry["ip"],
                cidr=entry["cidr"],
                chain=firewall.firewall_chain,
                action="DROP",
            )
        else:
            rule = firewall.construct_iptables_rule(
                entry["ip"],
                cidr=entry["cidr"],
                chain=firewall.firewall_chain,
                action="RETURN",
            )
        if rule is None:
            logger.warning(
                "Entry {} received from hn-api does not seem to be valid, skipping...".format(
                    entry["ip"]
                )
            )
            continue
        requested_state.append(rule)
    return tuple(requested_state)


def flush_chain(chain, firewall, table="mangle"):
    """
    Flush an IPtables chain
    :param chain: str: The IPtables chain to flush
    :param table: str: The IPtables table to operate on
    :param firewall: iptables.Firewall object
    :return bool: True if successful, False otherwise
    """
    logger.debug("Flushing the {} chain in table {}".format(chain, table))
    if not firewall.flush_iptables_chain(chain, table=table):
        logger.error("Flushing the chain {} failed".format(chain))
        return False
    return True


def process_chain_entries(firewall, dry_run=False):
    """
    Retrieve <type> entries from hn-api
    Flush the current <type> chain
    Fill the current <type> chain with entries received from hn-api
    :param firewall: iptables.Firewall object
    :param dry_run: bool: don't change anything
    :return: The entries received from hn-api, None if none
    """
    logger.debug("Retrieving firewall entries from hn-api")
    firewall_url = _get_firewall_url()
    allow_entries = get_firewall_entries_from_hn_api(f"{firewall_url}/?action=allow")
    deny_entries = get_firewall_entries_from_hn_api(f"{firewall_url}/?action=deny")

    if not allow_entries and not deny_entries:
        logger.warning("No entries received from hn-api")
        # Return early, because there might be something wrong with hn-api at this point so let's keep the old entries
        return
    table = "mangle"
    deny_rules = construct_requested_rules_tuple(deny_entries, firewall)
    allow_rules = construct_requested_rules_tuple(allow_entries, firewall)
    if not dry_run:
        if not flush_chain(firewall.firewall_chain, firewall, table=table):
            raise FirewallException(
                f"Flushing chain {firewall.firewall_chain} in table {table} failed"
            )
        for rule_type, rules in (("allow", allow_rules), ("deny", deny_rules)):
            logger.debug(f"Adding the {rule_type} rules received from hn-api")
            add_chain_rules(rules, firewall, table=table)
            logger.debug(f"{rule_type} entries processed")
    # Finally return the entries for later use
    return [*allow_entries, *deny_entries]


def main(args=None):
    args = parse_args(args)
    configure_logging(debug=(args.debug or args.dry_run))
    logger.debug("Logger configured")

    if args.dry_run:
        logger.debug("Dry run flag set, not applying any IPtables")

    # ROOT only!
    if os.getuid() != 0:
        raise FirewallException("This script can only be run by root!")

    # Non of these virtual imposter's
    if host_is_a_docker() or host_is_a_vagrant():
        raise FirewallException(
            "This command can only be executed on non virtual environments"
        )

    # Setup the firewall and ensure everything is as we expect it to be
    logger.debug("Ensuring IPtables firewall is in the correct state")
    firewall = Firewall(safe_mode=args.safe_mode)
    # Don't configure the firewall on a dry_run
    if not args.dry_run and (
        not firewall.ensure_firewall_chains()
        or not firewall.ensure_firewall_chain_references()
    ):
        raise FirewallException(
            "Firewall not configured correctly and unable to rectify, please investigate"
        )

    # FLUSH?
    if args.flush_firewall and not args.dry_run:
        if not flush_chain(firewall.firewall_chain, firewall, table="mangle"):
            exit(1)
        exit(0)

    logger.debug("Processing firewall entries")
    new_firewall_entries = process_chain_entries(firewall, args.dry_run)

    # Dump entries if dry_run
    if args.dry_run:
        logger.debug("Dry-running, printing the info")
        current_firewall_entries = firewall.get_chain_entries(
            firewall.firewall_chain, table="mangle"
        )
        logger.debug("Current IPtables entries")
        pprint(current_firewall_entries)
        logger.debug("Firewall entries received from hn-api")
        pprint(new_firewall_entries)
