import logging
import os
import subprocess
import sys
import time
from argparse import ArgumentParser
from contextlib import suppress
from datetime import datetime, timezone
from signal import SIGINT

import requests

from kamikaze3.analysis.fpmstatus import FpmSource, worker_as_text
from kamikaze3.common.env import get_node_roles
from kamikaze3.settings import ROLE_LOADBALANCER

logger = logging.getLogger(__name__)


FPM_TIMEOUT = 2
ANALYSERS = [
    (("/usr/bin/hypernode-fpm-status",), "fpm-status.txt"),
    (("/usr/bin/hypernode-fpm-status-json",), "fpm-status.json"),
    (("/usr/bin/mysql", "-e", "show full processlist"), "mysql-process-list.txt"),
    (("/bin/netstat", "-ntp"), "tcp-connections.txt"),
    (("/bin/ps", "uxaf"), "process-list.txt"),
]


def worker_is_gone(worker, methods):
    return (
        worker["worker_status"] == "GONE"
        and worker["method"] in methods
        and not worker["uri"].startswith("/admin/")
    )


def worker_is_gone_get(worker):
    return worker_is_gone(worker, ["GET", "HEAD"])


def worker_is_gone_post(worker):
    return worker_is_gone(worker, ["POST"])


class FPMMonitor(object):
    cooldown_iterations = 0
    logger = None

    def __init__(
        self,
        fpm_status_url="http://localhost/fpm_status",
        reports_directory="/data/web/incidents/",
    ):

        self.fpm_status_url = fpm_status_url
        self.reports_directory = reports_directory
        self.logger = logger

    @staticmethod
    def now():
        return datetime.now(tz=timezone.utc).replace(microsecond=0)

    def attempt_report(self):
        """
        Attempt to write a report about the incident. If we are
        unable to write the report, suppress and continue
        :return None:
        """
        try:
            self.create_incident_folder()  # also links current/ symlink
            self.generate_status_reports()
        # If anything goes wrong writing the report, we should not error and crash because we still
        # need to perform the gone-killing later. Perhaps that the next run will be able to write a
        # successful report.
        except IOError:
            self.logger.warning(
                "Failed to write a report about the incident", exc_info=True
            )

    def run(self):
        try:
            alive = self.check_fpm_health()
        except requests.ConnectionError:
            # Cannot check FPM status. If Nginx is unavailable, there are no active FPM workers anyway.
            self.logger.debug(
                "Could not reach Nginx, so no active FPM workers. Exiting."
            )
            return 0

        if alive:
            self.logger.debug("FPM is alive, clearing out current incident symlink")
            try:
                os.unlink(os.path.join(self.reports_directory, "current"))
                self.logger.debug(
                    "Symlink %s removed"
                    % os.path.join(self.reports_directory, "current")
                )
            except OSError:
                pass
        else:
            self.logger.debug("FPM is not responding, reporting incident")
            self.attempt_report()

        self.kill_gone()
        return 0

    def check_fpm_health(self):
        try:
            return (
                requests.get(self.fpm_status_url, timeout=FPM_TIMEOUT).status_code
                == 200
            )
        except requests.Timeout:
            return False

    def create_incident_folder(self):
        directory = os.path.join(self.reports_directory, self.now().isoformat())
        symlink = os.path.join(self.reports_directory, "current")
        if not os.path.exists(symlink):
            self.logger.debug(
                "Creating incident %s and symlink" % self.now().isoformat()
            )
            os.makedirs(directory)
            os.symlink(directory, symlink)

    def generate_status_reports(self):
        for program, logfile in ANALYSERS:
            try:
                self.logger.debug("Running analyzer '%s'" % " ".join(program))
                logfile = os.path.join(
                    self.reports_directory,
                    "current",
                    self.now().isoformat() + "-" + logfile,
                )
                with open(logfile, "at") as fh:
                    fh.write(
                        subprocess.check_output(
                            program, stderr=subprocess.STDOUT
                        ).decode()
                    )
            except (OSError, subprocess.CalledProcessError, UnicodeDecodeError) as e:
                self.logger.warning(
                    "Could not perform analysis %s" % " ".join(program), exc_info=1
                )

    @staticmethod
    def find_workers_to_kill():
        to_kill = list()

        if ROLE_LOADBALANCER in get_node_roles():
            # If the server is a loadbalancer, we don't have any local workers to kill
            return to_kill

        scoreboard = FpmSource().get_scoreboard()
        num_requests = len(scoreboard)

        if num_requests:
            to_kill = list(filter(worker_is_gone_get, scoreboard))
            gone_posts = list(filter(worker_is_gone_post, scoreboard))
            num_gone_posts = len(gone_posts)
            if num_gone_posts > 2 and (num_gone_posts * 100 / num_requests >= 30):
                to_kill.extend(gone_posts)
        return to_kill

    def kill_workers(self, workers):
        killed = list()
        for worker in workers:
            self.logger.debug(" - " + worker_as_text(worker))
            with suppress(OSError):
                os.kill(int(worker["pid"]), SIGINT)
                killed.append(worker)
        self.logger.debug(
            "Killed %d workers that are busy serving clients that have already gone away."
            % len(killed)
        )
        return killed

    def kill_gone(self):
        killed = list()
        to_kill = self.find_workers_to_kill()
        if not to_kill:
            self.logger.debug("No busy workers to kill")
            return killed

        killed.extend(self.kill_workers(to_kill))

        # There are things to kill. We assume just killing the workers once is not enough,
        # since more requests may come out of the TCP buffer queue.
        # It is very inefficient to always check the FPM queue more than once a minute,
        # but if we know it was busy just now and GONEs had to be slayed, it makes sense to assume
        # more GONEs are going to be there.
        #
        # So we go to berserk mode, checking for GONE requests to kill every 3 seconds
        # to slay upcoming GONE requests if any.
        #
        # If after the wait there was no worker to kill, there may be more
        # requests on the way. It's not efficient enough to wait for the next
        # call to this script from cron, until then lots of GONE requests
        # may pile up.
        # So we'll go to cooldown mode with longer waits (12 ADDITIONAL seconds).
        # Staying in cooldown mode for 3 iterations (45 seconds waiting over all) means
        # situation is stable and we don't need to kill any GONE request anymore
        # for now.
        cooldown_iterations = 3
        while cooldown_iterations > 0:
            time.sleep(3)
            to_kill = self.find_workers_to_kill()
            if to_kill:  # more to kill, back to berserk mode
                killed.extend(self.kill_workers(to_kill))
                cooldown_iterations = 3
            else:  # nothing to kill, let's cooldown
                cooldown_iterations -= 1
                time.sleep(12)

        return killed


class FPMMonitorApp(object):
    def __init__(self, args):
        self.args = self.parse_args(args)
        self.logger = logger

    @staticmethod
    def parse_args(args):
        parser = ArgumentParser(
            description="Check for FPM health, report failures and kill GONE requests"
        )
        parser.add_argument(
            "-v",
            "--verbose",
            action="store_true",
            required=False,
            help="run in verbose mode",
        )
        return parser.parse_args(args)

    def run(self):
        log_level = logging.DEBUG if self.args.verbose else logging.ERROR
        logging.basicConfig(level=log_level)
        self.logger.setLevel(log_level)
        self.logger.debug("Starting FPMMonitor")

        monitor = FPMMonitor()
        monitor.logger = self.logger
        return monitor.run()


def main():
    sys.exit(FPMMonitorApp(sys.argv[1:]).run())
