import argparse
import getpass
import grp
import json
import logging
import os
import pwd
import sys
from operator import itemgetter
from typing import Dict, List

from kamikaze3.common.password import get_password_hash

FTPASSWD_LOCATION = "/data/var/ftp/ftpasswd"


logging.basicConfig(level=logging.DEBUG)


def main(argv=None):
    argv = argv or sys.argv[1:]
    args = parse_args(argv)

    if args.subcommand == "remove":
        remove_user(args.username)
    elif args.subcommand == "add":
        password = args.password or get_password()
        add_user(args.username, password, args.homedir)
    elif args.subcommand == "list":
        users = list_users()

        if args.json:
            print(json.dumps(users, indent=4))
        else:
            print_users(users)


def parse_args(args):
    parser = argparse.ArgumentParser(description="Manage FTP users on your Hypernode.")
    subparsers = parser.add_subparsers(dest="subcommand")
    subparsers.required = True

    json_parser = subparsers.add_parser("list")
    json_parser.add_argument(
        "--json", "-j", action="store_true", help="Print users in JSON format."
    )

    remove_parser = subparsers.add_parser("remove")
    remove_parser.add_argument(
        "--username", "-u", required=True, help="The FTP user to remove."
    )

    add_parser = subparsers.add_parser("add")
    add_parser.add_argument(
        "--username", "-u", required=True, help="The FTP user to add."
    )
    add_parser.add_argument(
        "--password", "-p", required=False, help="The password for the user."
    )
    add_parser.add_argument(
        "--homedir",
        "-d",
        required=True,
        help="The directory the user will be restricted to.",
    )

    return parser.parse_args(args)


###
# List users
###
def list_users() -> List[Dict]:
    try:
        with open(FTPASSWD_LOCATION) as f:

            def get_user_and_homedir(line):
                entries = line.split(":")
                return {"user": entries[0], "homedir": entries[5]}

            return list(map(get_user_and_homedir, f.readlines()))
    except IOError:
        return []


def format_users(users) -> str:
    message = ""
    max_userlength = max(map(len, map(itemgetter("user"), users)))
    for user in users:
        user_str = "{1:<0}   {2}".format(max_userlength, user["user"], user["homedir"])
        message += user_str + "\n"
    return message


def print_users(users):
    if not users:
        print("No users in ftpasswd file")
    else:
        print("Users found:")
        print(format_users(users))


###
# Add user
###
def get_password():
    password = getpass.getpass("Please choose a strong password: ")

    if not password:
        print(sys.stderr, "Password cannot be empty")
        sys.exit(1)

    confirmation = getpass.getpass("Retype password: ")
    if password != confirmation:
        print(sys.stderr, "Passwords do not match, try again")
        sys.exit(1)

    return password


def add_user(user, password, homedir):
    if not homedir.startswith("/"):
        message = "Cannot add user with relative home directory. Please specify absolute path."
        print(sys.stderr, message)
        sys.exit(message)

    if not os.path.isdir(homedir):
        message = (
            f"Given home directory '{homedir}' does not exist or is not a directory."
        )
        print(sys.stderr, message)
        sys.exit(message)

    password_hash = get_password_hash(password)
    remove_user(user)
    uid, gid = get_app_uid_and_gid()
    with open(FTPASSWD_LOCATION, "a") as f:
        passwd_line = f"{user}:{password_hash}:{uid}:{gid}::{homedir}:/bin/false\n"
        f.write(passwd_line)

    print(
        "Account has been added.\n\n"
        "Please remember to add your IP to the allowlist, either via the Control Panel or the hypernode-systemctl \n"
        "command, in order to connect to the FTP server."
    )

    return {"user": user, "homedir": homedir}


def get_app_uid_and_gid():
    # will throw errors by design when app user does not exist :)
    uid = pwd.getpwnam("app").pw_uid
    gid = grp.getgrnam("app").gr_gid
    return uid, gid


###
# Remove users
###
def remove_user(user):
    response = None
    if not os.path.exists(FTPASSWD_LOCATION):
        message = "No users present in ftpasswd file."
        print(sys.stderr, message)
        sys.exit(message)
    with open(FTPASSWD_LOCATION, "r") as f:
        lines = f.readlines()
    with open(FTPASSWD_LOCATION, "w") as f:
        for line in lines:
            if not line.startswith(user + ":"):
                f.write(line)
            else:
                username = line.split(":")[0]
                response = {"user": username}
                print(f"Removed user {user}")

    return response
