from typing import List, Optional

from kamikaze3.haproxy.client import SocketClient
from kamikaze3.haproxy.input import parse_args
from kamikaze3.haproxy.settings import SERVER_STATS_DEFAULTS, USER_SOCKET_PATH
from kamikaze3.haproxy.structs import ServerStats, SystemInfo
from kamikaze3.utils import camel_to_snake


def apply_server_stats_defaults(data):
    for key, value in data.items():
        if value == "":
            data[key] = SERVER_STATS_DEFAULTS.get(key, "-")
    return data


class StatsCommand:
    def __init__(self):
        self.client = SocketClient(USER_SOCKET_PATH)

    def get_system_info(self) -> SystemInfo:
        result = self.client.send_and_receive("show info")
        lines = result.strip().splitlines()
        system_info = {}
        for line in lines:
            key, value = line.split(":", maxsplit=1)
            key = camel_to_snake(key.replace("-", "_"))
            system_info[key] = value.strip()
        return SystemInfo(**system_info)

    def get_server_stats(self) -> List[ServerStats]:
        result = self.client.send_and_receive("show stat")
        lines = result.strip().splitlines()
        header = lines.pop(0).strip("#, ").split(",")
        rows = []
        for line in lines:
            parsed_line = line.split(",")
            data = dict(zip(header, parsed_line))
            # In the new Haproxy version there is a metric called "-" with value "-".
            # We need to unset this because this is not allowed to be present in the ServerStats struct.
            data.pop("-", None)
            data = apply_server_stats_defaults(data)
            rows.append(ServerStats(**data))
        return rows

    @staticmethod
    def render_server_stats(server: ServerStats):
        downtime = f"{server.downtime}sec" if server.status == "DOWN" else "None"

        print(
            f"# {server.svname} ("
            f"status: {server.status},"
            f" algorithm: {server.algo},"
            f" mode: {server.mode},"
            f" weight: {server.weight},"
            f" check: {server.check_status},"
            f" check duration: {server.check_duration}ms"
            ")",
            f"  - Downtime: {downtime}",
            f"  - Queued requests: {server.qcur} (max: {server.qmax}, limit: {server.qlimit})",
            f"  - Session rate: {server.rate} (max: {server.rate_max}, limit: {server.rate_lim})",
            f"  - Request rate: {server.req_rate} (max: {server.req_rate_max}, total: {server.req_tot})",
            f"  - Connection rate: {server.conn_rate} (max: {server.conn_rate_max}, total: {server.conn_tot})",
            f"  - Session: {server.scur} (max: {server.smax}, limit: {server.slim}, total: {server.stot})",
            f"  - Bytes: {server.bin} in, {server.bout} out",
            f"  - Denied: {server.dreq} request, {server.dcon} connection, {server.dresp} response",
            f"  - Errors: {server.ereq} request, {server.econ} connection, {server.eresp} response",
            f"  - Warnings: {server.wretr} retries, {server.wredis} redispatches",
            sep="\n",
        )

    def render_full(self):
        system_info = self.get_system_info()
        print(
            f"{system_info.name} v{system_info.version} (released {system_info.release_date})",
            f"Process ID: {system_info.pid} (process #{system_info.nbproc}, nbproc = {system_info.nbproc}, nbthread = {system_info.nbthread})",
            f"Uptime: {system_info.uptime}",
            f"maxsock: {system_info.maxsock}, current pipes: {system_info.pipes_used}/{system_info.pipes_free}, conn rate: {system_info.conn_rate}/s",
            f"Running tasks: {system_info.tasks}, idle: {system_info.idle_pct}%",
            sep="\n",
        )

        server_stats = self.get_server_stats()
        frontend_servers = [
            server for server in server_stats if server.pxname == "http_in"
        ]
        backend_servers = [
            server for server in server_stats if server.pxname == "http_backend"
        ]

        print()
        print(f"Frontend servers: {len(frontend_servers)}")
        for server in frontend_servers:
            self.render_server_stats(server)

        print()
        print(f"Backend servers: {len(backend_servers)}")
        for server in backend_servers:
            self.render_server_stats(server)

    def render_short(self):
        system_info = self.get_system_info()
        print(
            f"{system_info.name} v{system_info.version} (released: {system_info.release_date}, uptime: {system_info.uptime})"
        )
        server_stats = self.get_server_stats()
        for server in server_stats:
            if server.status == "DOWN":
                print(
                    f"{server.svname} (status={server.status}, addr={server.addr}): down for {server.downtime}sec"
                )
            else:
                print(
                    f"{server.svname} ({server.status}): request rate({server.req_rate}), session rate({server.rate}), conn rate({server.conn_rate})"
                )


def main(args: Optional[List[str]] = None):
    full = parse_args(args)
    command = StatsCommand()
    if full:
        command.render_full()
    else:
        command.render_short()


if __name__ == "__main__":
    main()
