import logging
import io
import sys
import paramiko
from contextlib import suppress
from os import path
from multiprocessing import Lock
from multiprocessing.pool import Pool
from subprocess import Popen, PIPE, STDOUT

from kamikaze3.common.env import get_app_name
from kamikaze3.log.log import configure_logging
from kamikaze3.hypertail.parsing import parse_arguments
from kamikaze3.hypertail.apps import get_target_app_names

logger = logging.getLogger("kamikaze3.hypertail.main")
stdout_lock = Lock()

MAX_AMOUNT_OF_LINES_TO_GET = 999999
HYPERNODE_APP_USER_SSH_DIR = "/etc/hypernode/.ssh/app/"
HYPERNODE_DEFAULT_LOG_USER = "app"


def main(args=None):
    """
    Tail logs on (multiple) nodes with a given role.
    """
    args = parse_arguments(args=args)
    configure_logging("kamikaze3.hypertail", args.verbose)
    hypertail(args)


def hypertail(args=None):
    local_app_name = get_app_name()
    target_app_names = get_target_app_names(args)
    if not target_app_names:
        logger.info("No target apps found to tail logs for")
        return

    pool = Pool(processes=len(target_app_names))
    try:
        results = []
        for app_name in target_app_names:
            logger.debug("parsing log lines for {}".format(app_name))
            if app_name == local_app_name:
                parse_func = parse_local_log_lines
            else:
                parse_func = parse_remote_log_lines

            results.append(pool.apply_async(parse_func, args=(app_name, args)))

        pool.close()
        pool.join()
    except KeyboardInterrupt:
        pool.terminate()


def parse_local_log_lines(app_name: str, args):
    """
    Parse the local log lines
    :param str app_name: The name of the app
    :param obj args: Parsed arguments
    """
    cmd = ["tail", args.file, "-n", str(args.lines)]
    if args.follow:
        cmd.append("-f")

    p = Popen(cmd, stdout=PIPE, stderr=STDOUT)
    with suppress(KeyboardInterrupt):
        for raw_line in iter(p.stdout.readline, ""):
            try:
                line = raw_line.decode("utf-8")
                if not line and not args.follow:
                    # Break on EOF if `--follow` is not passed
                    break
            except KeyboardInterrupt:
                break
            except UnicodeDecodeError:
                pass
            print_log_message(line, app_name)


def get_ssh_connection(app_name: str):
    ssh = paramiko.SSHClient()
    ssh.set_missing_host_key_policy(paramiko.MissingHostKeyPolicy())

    ssh_key_file = "{}/cluster_{}_hypernode".format(
        HYPERNODE_APP_USER_SSH_DIR, app_name
    )
    if not path.exists(ssh_key_file):
        return None

    private_key_file = open(ssh_key_file, "r")
    private_key_content_raw = private_key_file.read()
    private_key_content = io.StringIO(private_key_content_raw)
    pkey = paramiko.RSAKey.from_private_key(private_key_content)

    ssh.connect(
        "{}.hypernode.io".format(app_name),
        username=HYPERNODE_DEFAULT_LOG_USER,
        timeout=3,
        pkey=pkey,
    )

    return ssh


def parse_remote_log_lines(app_name: str, args):
    """
    Parse the log lines on the remote node
    :param str app_name: Name of the target node to tail the logs
    :param args: Parsed arguments
    :return return_buff: Buffer filled with the lines to log
    """
    cmd = "tail {} -n {}".format(args.file, args.lines)
    if args.follow:
        cmd += " -f"

    ssh = get_ssh_connection(app_name)
    if not ssh:
        return

    stdin, stdout, stderr = ssh.exec_command(cmd)
    logger.debug("tailing logs for {}".format(app_name))

    return_buff = []

    line = "dummyline"
    while line:
        try:
            line = stdout.readline()
            if line:
                print_log_message(line, app_name)
        except KeyboardInterrupt:
            break
        # except to prevent encoding issues when reading the lines.
        except UnicodeDecodeError:
            pass

    ssh.close()
    return return_buff


def print_log_message(message: str, app_name: str):
    """
    Acquire the stdout lock to make sure no lines getting mixed on screen,
    write the message to stdout flush and release the lock again
    :param str message: The message to log
    :param str app_name: Appname for the log
    """
    stdout_lock.acquire()
    sys.stdout.write("[" + app_name + "] " + message + "\n")
    sys.stdout.flush()
    stdout_lock.release()


if __name__ == "__main__":
    main()
