import logging
from os.path import dirname, join
from re import search, sub
from shlex import quote
from xml.etree import ElementTree

from kamikaze3.importer.abstract import ImporterSubClass
from kamikaze3.importer.error_codes import (
    DB_NAME_SPECIAL_CHAR_RC,
    DB_NAME_TOO_LONG_RC,
    DB_TABLE_PREFIX_SPECIAL_CHAR_RC,
    DB_USERNAME_TOO_LONG_RC,
    MAG_2_NOT_SUPPORTED_RC,
    NO_CMS_RC,
    REMOTE_DB_CONF_NOT_FOUND_RC,
    REMOTE_DB_CONF_PARSE_ERR_RC,
    REMOTE_DB_M2_CONF_NOT_FOUND_RC,
    REMOTE_DB_M2_CONF_PARSE_ERR_RC,
    REMOTE_DB_SW5_CONF_NOT_FOUND_RC,
    REMOTE_DB_SW5_CONF_PARSE_ERR_RC,
    REMOTE_DB_SW6_CONF_NOT_FOUND_RC,
    REMOTE_DB_SW6_CONF_PARSE_ERR_RC,
    WRONG_PATH_RC,
)
from kamikaze3.importer.utils import has_only_safe_chars
from kamikaze3.importer.validators import (
    db_name_fits_in_hypernode,
    db_username_fits_in_hypernode,
)
from kamikaze3.privileged.manage_vhosts.input import check_conditions_before_starting
from kamikaze3.settings import HYPERNODE_NGINX_PATH

log = logging.getLogger(__name__)


class CMSNotFound(RuntimeError):
    pass


class DBIdentifierException(RuntimeError):
    pass


class SpecialCharacterException(DBIdentifierException):
    pass


class SpecialCharactersInDBName(SpecialCharacterException):
    pass


class SpecialCharactersInTablePrefix(SpecialCharacterException):
    pass


class DBUsernameTooLongException(DBIdentifierException):
    pass


class DBNameTooLongException(DBIdentifierException):
    pass


class DiscoverDriver(object):
    def check_parameters(self):
        """
        Check the constructor received an object to use as the Discovery Connection
        and get the connection details from.
        :return None:
        """
        if self.discover_connection is None:
            raise RuntimeError(
                "You must pass the DiscoverConnection object to the constructor "
                "of an DiscoverDriver object because the configured attributes are "
                "used to determine the connection details of the remote "
                "server."
            )

    def __init__(self, discover_connection):
        """
        Initialize the DiscoverDriver object and store the passed db
        connection on the self. Check if the constructor received
        the required parameters.
        :param obj discover_connection: The discover connection object with
        the details to use to perform operations on the remote machine with.
        """
        self.discover_connection = discover_connection
        self.check_parameters()

    def check_if_remote_path_is_symlink(self, path):
        """
        Check if the remote path is a symlink
        :param str path: the path to check for if it is a symlink
        :return bool is_symlink: True or false
        """

        # NOTE(vdloo): test is is coreutils, this exists on more systems than
        # what we previously used which was 'file /tmp/file | grep symbolic'.
        check_remote_path_is_symlink_command = "test -L {path}".format(path=path)
        return self.discover_connection.importer.ssh_conn.driver.exec_command_check_zero_exit(
            command=check_remote_path_is_symlink_command,
            success_msg="Looks like the provided path is a symlink",
            failure_msg="Detected that the provided path is not a symlink",
        )

    def resolve_remote_symlink(self, unsafe_path):
        """
        If the specified path is a symlink, return the real path
        :param str unsafe_path: the unsafe path to resolve the symlink for
        :return str unsafe_resolved_path: Either the path of the resolved symlink or the original path.
        """
        safe_path = quote(unsafe_path)
        resolve_path_command = "readlink {}".format(safe_path)
        (
            return_code,
            raw_resolved_path,
            _,
        ) = self.discover_connection.importer.ssh_conn.driver.exec_command(
            resolve_path_command
        )
        unsafe_resolved_path = raw_resolved_path.decode("utf-8").strip()
        if return_code == 0:
            log.debug("Resolved remote path to: {}".format(unsafe_resolved_path))
            return unsafe_resolved_path or unsafe_path
        else:
            log.debug("Path does not look like a symlink: {}. OK!".format(unsafe_path))
            return unsafe_path

    def find_remote_cms_directory(self, path, cms_type):
        """
        Find a directory that looks like a cms installation, Shopware {5, 6} & Magento {1, 2} in a path on the
        remote host in the provided path. If there are more than one the first one
        returned by find will be used, if there is none an empty string will be
        returned by run_remote_command. If the directory does not exist on the
        remote machine, the exec_command will return nonzero and make this
        function return None.
        :param str path: the directory to look for a CMS installation in
        :param str cms_type: CMS and versions, Magento {1, 2}, and Shopware version {5, 6}
        :return str | NoneType: path to the cms installation or None if not found
        """

        find_cms_commands = {
            "Magento 1": "find {} -name local.xml 2>&1 | grep 'app/etc/local.xml' | sed 's/app\\/etc\\/local.xml//g' |"
            " grep -v 'Permission denied\\|No such file or directory' -i | sort | head -n 1".format(
                path
            ),
            "Magento 2": "find {} -name static.php 2>&1 | grep -v 'Permission denied\\|No such file or directory' -i "
            "| grep -v 'vendor/' | sort | head -n 1 | sed 's/static.php//'".format(
                path
            ),
            "Shopware 5": "find {} -name Kernel.php 2>&1 | grep 'Shopware/Kernel.php'| sed 's/Shopware\\/Kernel.php//g' "
            "| grep -v 'Permission denied\\|No such file or directory' -i | sort | head -n 1".format(
                path
            ),
            "Shopware 6": "find {} -name Kernel.php 2>&1 | grep 'src/Kernel.php'| sed 's/src\\/Kernel.php//g' "
            "| grep -v 'Permission denied\\|No such file or directory' -i | sort | head -n 1".format(
                path
            ),
        }
        find_cms_command = find_cms_commands[cms_type]
        (
            return_code,
            raw_cms_directory,
            _,
        ) = self.discover_connection.importer.ssh_conn.driver.exec_command(
            find_cms_command
        )
        cms_directory = raw_cms_directory.decode("utf-8").strip()
        if return_code == 0 and cms_directory:
            log.info("Found {} installation in {}!".format(cms_type, cms_directory))
            return cms_directory
        else:
            log.debug("No {} installation was found in {}".format(cms_type, path))

    def find_remote_cms_installation(self, path):
        """
        Look for a Magento (1, 2) or Shopware (5, 6) installation on the remote host in the provided path
        If no CMS installation is found raise a RuntimeError
        If a Magento 2 installation is found raise a NotImplementedError
        :param str path: the directory to look for a Magento installation in
        :return str: path to the CMS installation
        """
        log.debug("Connecting to the remote machine to look for Magento")
        magento_1_path = self.find_remote_cms_directory(path, cms_type="Magento 1")
        if magento_1_path:
            self.discover_connection.importer.set_site_type("magento1")
            return magento_1_path
        magento_2_pub_path = self.find_remote_cms_directory(path, cms_type="Magento 2")
        if magento_2_pub_path:
            self.discover_connection.importer.set_site_type("magento2")
            magento_2_path = (
                dirname(self.resolve_remote_symlink(magento_2_pub_path).rstrip("/"))
                + "/"
            )
            return magento_2_path
        shopware_5_eng_path = self.find_remote_cms_directory(
            path, cms_type="Shopware 5"
        )
        if shopware_5_eng_path:
            self.discover_connection.importer.set_site_type("shopware5")
            shopware_5_path = (
                dirname(self.resolve_remote_symlink(shopware_5_eng_path).rstrip("/"))
                + "/"
            )
            return shopware_5_path
        shopware_6_src_path = self.find_remote_cms_directory(
            path, cms_type="Shopware 6"
        )
        if shopware_6_src_path:
            self.discover_connection.importer.set_site_type("shopware6")
            shopware_6_path = (
                self.resolve_remote_symlink(shopware_6_src_path).rstrip("/")
            ) + "/"
            return shopware_6_path
        raise CMSNotFound("No CMS installation found in {}".format(path))

    def find_cms_in_expected_locations(self, path=None, autodetect=True):
        """
        If no path is provided, the importer will look on the remote host in the default directories
        $HOME, /var/www, /var/html and /www. Once the CMS installation is found the function will
        return and look no further.
        Raises a RuntimeError when the CMS installation can not be found in either the path or the
        default directories if no path was specified. The error tells the user to be more specific
        with --path
        :param str path: the directory to look for a CMS installation in
        :param bool autodetect: Find CMS in default locations if CMS was not found in the provided path
        :return str: path to the CMS installation
        """
        if path:
            # If a path is specified, only look in that path
            cms_locations = [path]
        elif not autodetect:
            # If no path is specified and autodetect is disabled, only
            # look in the first default location (the user HOME dir)
            cms_locations = [self.discover_connection.importer.cms_default_locations[0]]
        else:
            # If no path is specified and autodetect is enabled, look in all default locations.
            cms_locations = self.discover_connection.importer.cms_default_locations

        for search_location in cms_locations:
            log.debug(
                "Looking for CMS installation in {}".format(
                    path or "the usual locations"
                )
            )
            try:
                try:
                    cms_installation_path = self.find_remote_cms_installation(
                        search_location
                    )
                    return cms_installation_path
                except NotImplementedError:
                    self.discover_connection.importer.return_code = (
                        MAG_2_NOT_SUPPORTED_RC
                    )
                    raise
            except CMSNotFound:
                log.debug(
                    "Couldn't find a CMS installation in {}!".format(search_location)
                )
                if path and autodetect:
                    log.info(
                        "Will try to autodetect the CMS installation in known installation directories now.."
                    )
                    found_cms = (
                        self.find_cms_in_expected_locations()
                    )  # without the path
                    self.discover_connection.importer.return_code = WRONG_PATH_RC
                    raise CMSNotFound(
                        "Could not find CMS installation in the provided path, "
                        "however an installation was found in a different path. \n"
                        "Perhaps you meant to run the importer with --path {}".format(
                            found_cms
                        )
                    )
        self.discover_connection.importer.return_code = NO_CMS_RC
        raise CMSNotFound(
            "Could not find CMS installation on the remote host! "
            "Specify the path to the CMS installation with --path"
        )

    def find_cms(self):
        """
        Find CMS based on the importer configured options.
        Will look in the specified path if one is specified.
        Will attempt to autodetect the location if the installation
        can not be found or if not path is specified, but only if
        skip_autodetect is not passed as a cli flag to the importer.
        :return str: path to the CMS installation
        """
        autodetect = not self.discover_connection.importer.skip_autodetect
        unsafe_path = self.discover_connection.importer.path
        safe_path = quote(unsafe_path) if unsafe_path else None

        if unsafe_path and self.check_if_remote_path_is_symlink(safe_path):
            unsafe_path = self.resolve_remote_symlink(unsafe_path)
            safe_path = quote(unsafe_path)
        return self.find_cms_in_expected_locations(
            path=safe_path, autodetect=autodetect
        )

    @staticmethod
    def compose_database_information(raw_php_config):
        """
        Gather the required db info values from the php config file
        Simple state machine that just removes the default delimiters
        from the value. To get it 100% right we'd need to parse PHP.
        We can't just run a PHP interpreter because then we'd make it
        very easy to do RCE. For now this will suffice. If this doesn't
        work in most cases, we'll have to think of something more
        sophisticated.
        :param iter raw_php_config: The raw config file lines
        :return dict db_info: The db_info dict
        """

        def try_parse_from_line(key, inner_line, inner_db_info):
            prefix = "'{}' => ".format(key)
            if prefix in line and key not in db_info:
                item_content = inner_line.replace(prefix, "").replace(",", "").strip()
                inner_db_info[key] = item_content.lstrip("'").rstrip("'")

        db_keys_to_parse = {
            "host",
            "port",
            "username",
            "password",
            "dbname",
            "table_prefix",
        }
        db_info = {}

        db_context = False
        backend_context = False
        context_delimiter = None
        for line in raw_php_config:
            if context_delimiter and line.startswith(context_delimiter):
                db_context = False
                backend_context = False
            if "'db' =>" in line:
                db_context = True
                context_delimiter = line.split("'db' =>")[0] + "'"
            if "'backend' =>" in line:
                backend_context = True
                context_delimiter = line.split("'backend' =>")[0] + "'"

            if db_context:
                for item in db_keys_to_parse:
                    try_parse_from_line(item, line, db_info)
            if backend_context:
                try_parse_from_line("frontName", line, db_info)
        return db_info

    def compose_magento1_db_info(self, remote_xml_path, tree):
        """
        Gather the required db info values from the local.xml
        :param str remote_xml_path: The remote xml path
        :param obj tree: The XML tree object
        :return dict db_info: The db_info dict
        """
        # This function is tested as part of the parse_magento_1_database_information unittest

        # for reference, http://eli.thegreenplace.net/2012/03/15/processing-xml-in-python-with-elementtree
        # cheatsheet: https://gist.github.com/LeCoupa/8c305ec8c713aad07b14
        dbconn = tree.find(".//global/resources/default_setup/connection")
        if not dbconn:
            self.discover_connection.importer.return_code = REMOTE_DB_CONF_NOT_FOUND_RC
            raise RuntimeError(
                "Could not find database config in {} on "
                "the remote server".format(remote_xml_path)
            )

        try:
            host, _, port = dbconn.find("./host").text.partition(":")
            username = dbconn.find("./username").text
            password = dbconn.find("./password").text
            dbname = dbconn.find("./dbname").text
        except AttributeError:
            self.discover_connection.importer.return_code = REMOTE_DB_CONF_PARSE_ERR_RC
            raise RuntimeError(
                "Could not parse database config " "from {}".format(remote_xml_path)
            )

        try:
            table_prefix = tree.find(".//global/resources/db/table_prefix").text or ""
        except AttributeError:
            table_prefix = ""

        try:
            port = int(port)
        except (ValueError, TypeError):
            port = 3306

        try:
            admin_path = (
                tree.find("./admin/routers/adminhtml/args/frontName").text or "admin"
            )
        except AttributeError:
            admin_path = ""

        return {
            "host": host,
            "port": port,
            "username": username,
            "password": password,
            "db_name": dbname,
            "table_prefix": table_prefix,
            "admin_path": admin_path,
        }

    def parse_magento_1_database_information(self):
        """
        Find the magento 1 local.xml and parse the database information
        :return dict parsed_database_information: Dict with db connection details
        """
        remote_xml_path = join(
            self.discover_connection.importer.remote_path, "app/etc/local.xml"
        )
        raw_local_xml = "\n".join(
            self.discover_connection.importer.ssh_conn.driver.read_file(remote_xml_path)
        )
        tree = ElementTree.ElementTree(ElementTree.fromstring(raw_local_xml))
        db_info = self.compose_magento1_db_info(remote_xml_path, tree)
        return db_info

    def compose_magento2_db_info(self, raw_env_php):
        """
        This function assembles and returns remote database credentials
        :param raw_env_php:
        :return: parsed_db_info: dict with parsed database credentials
        """
        # This function is tested as part of the parse_magento_2_database_information unittest

        db_info = self.compose_database_information(raw_env_php)
        try:
            port = int(db_info.get("port", 3306))
        except ValueError:
            port = 3306

        parsed_db_info = {
            "host": db_info.get("host", "127.0.0.1"),
            "port": port,
            "username": db_info["username"],
            "password": db_info["password"],
            "db_name": db_info["dbname"],
            "table_prefix": db_info.get("table_prefix", ""),
            "admin_path": db_info["frontName"],
        }
        return parsed_db_info

    def parse_magento_2_database_information(self):
        """
        Find the magento 2 env.php and parse the database information
        NOTE(vdloo): Does not support multi DB
        :return dict parsed_database_information: Dict with db connection details
        """
        remote_env_path = join(
            self.discover_connection.importer.remote_path, "app/etc/env.php"
        )

        try:
            raw_env_php = self.discover_connection.importer.ssh_conn.driver.read_file(
                remote_env_path
            )
        except IOError:
            self.discover_connection.importer.return_code = (
                REMOTE_DB_M2_CONF_NOT_FOUND_RC
            )
            raise RuntimeError(
                "Could not find remote php env file at location: {}".format(
                    remote_env_path
                )
            )

        try:
            db_info = self.compose_magento2_db_info(raw_env_php)
        except KeyError:
            self.discover_connection.importer.return_code = (
                REMOTE_DB_M2_CONF_PARSE_ERR_RC
            )
            raise RuntimeError(
                "Could not parse database " "config from {}".format(remote_env_path)
            )

        return db_info

    def compose_shopware_5_database_information(self, raw_config_php):
        """
        Gather and compose shopware 5 database credentials from config.php
        :param iter raw_config_php: The raw config file lines
        :return dict db_info: The db_info dict
        """
        # This function is tested as part of the parse_shopware_5_database_information unittest

        db_info = self.compose_database_information(raw_config_php)

        try:
            port = int(db_info.get("port", 3306))
        except ValueError:
            port = 3306

        parsed_db_info = {
            "host": db_info.get("host", "127.0.0.1"),
            "port": port,
            "username": db_info["username"],
            "password": db_info["password"],
            "db_name": db_info["dbname"],
            "table_prefix": "s_",
        }
        return parsed_db_info

    def parse_shopware_5_database_information(self):
        """
        find database config.php and parse shopware 5 db information
        :return: dict parsed_database_information: Dict with db connection details
        """
        remote_config_path = join(
            self.discover_connection.importer.remote_path, "config.php"
        )

        try:
            remote_config_path = (
                self.discover_connection.importer.ssh_conn.driver.read_file(
                    remote_config_path
                )
            )
        except IOError:
            self.discover_connection.importer.return_code = (
                REMOTE_DB_SW5_CONF_NOT_FOUND_RC
            )
            raise RuntimeError(
                "Could not find remote php config file at location: {}".format(
                    remote_config_path
                )
            )

        try:
            db_info = self.compose_shopware_5_database_information(remote_config_path)
        except KeyError:
            self.discover_connection.importer.return_code = (
                REMOTE_DB_SW5_CONF_PARSE_ERR_RC
            )
            raise RuntimeError(
                "Could not parse database " "config from {}".format(remote_config_path)
            )

        return db_info

    @staticmethod
    def check_db_info(parsed_db_info):
        """
        returns a list of empty key values in parsed_db_info
        Excluding table_prefix since this key is optional.
        :param parsed_db_info: populated credentials dictionary
        :return: list of empty keys (except table_prefix) in parsed_db_info
        """
        empty_keys = []
        for keys in parsed_db_info:
            if not parsed_db_info[keys] and keys != "table_prefix":
                empty_keys.append(keys)
        return empty_keys

    def compose_shopware_6_database_information(self, raw_env):
        """
        Gather the required db info values from the .env
        :param iter raw_env: The raw .env file lines
        :return dict db_info: The db_info dict
        """
        # This function is tested as part of the parse_shopware_6_database_information unittest

        env_credentials = None
        app_url = None

        regex_patterns = {
            "username": r"(?<=\/\/)(.*?)(?=\:)",
            "password": r"(?<={}:)(.*?)(?=\@)",
            "host": r"(?<=@)(.*?)(?=:)",
            "port": r"(?<={}:)(.*?)(?=\/)",
            "database": r"(?<={}\/)(.*?)(?=$)",
        }
        for line in raw_env:
            if line.startswith("DATABASE_URL"):
                env_credentials = sub(r"=\"(.+)\"$", r"=\1", line)
            if line.startswith("APP_URL"):
                app_url = line
        self.discover_connection.importer.app_url = app_url
        db_info = {}

        username_match = search(regex_patterns["username"], env_credentials)
        if username_match:
            db_info["username"] = username_match.group(1)
        password_match = search(
            regex_patterns["password"].format(db_info["username"]), env_credentials
        )
        if password_match:
            db_info["password"] = password_match.group(1)
        host_match = search(regex_patterns["host"], env_credentials)
        if host_match:
            db_info["host"] = host_match.group(1)
        port_match = search(
            regex_patterns["port"].format(db_info["host"]), env_credentials
        )
        if port_match:
            db_info["port"] = port_match.group(1)
        database_match = search(
            regex_patterns["database"].format(db_info["port"]), env_credentials
        )
        if database_match:
            db_info["database"] = database_match.group(1)

        for key in db_info:
            if db_info[key] == "" and key != "table_prefix":
                raise RuntimeError(
                    "Could not parse {} " "from database config".format(key)
                )
        try:
            port = int(db_info.get("port", 3306))
        except ValueError:
            port = 3306

        parsed_db_info = {
            "host": db_info.get("host", "127.0.0.1"),
            "port": port,
            "username": db_info["username"],
            "password": db_info["password"],
            "db_name": db_info["database"],
            "table_prefix": "",
        }
        empty_keys = self.check_db_info(parsed_db_info)
        if len(empty_keys) > 0:
            missing_string = ""
            for key in empty_keys:
                missing_string = missing_string + key + " "
            if missing_string:
                self.discover_connection.importer.return_code = (
                    REMOTE_DB_SW5_CONF_PARSE_ERR_RC
                )
                raise KeyError(
                    "Could not parse database credentials, unable to acquire {}".format(
                        parsed_db_info
                    )
                )

        return parsed_db_info

    def parse_shopware_6_database_information(self):
        """
        attempt to find database ".env" and parse Shopware 6 database information
        :return: dict parsed_database_information: Dict with db connection details
        """
        remote_env_path = join(self.discover_connection.importer.remote_path, ".env")
        try:
            if self.discover_connection.importer.ssh_conn.driver.read_file(
                remote_env_path
            ):
                raw_env = self.discover_connection.importer.ssh_conn.driver.read_file(
                    remote_env_path
                )
                log.debug(".env config exists!")
        except IOError:
            self.discover_connection.importer.return_code = (
                REMOTE_DB_SW6_CONF_NOT_FOUND_RC
            )
            raise RuntimeError(
                "Could not find remote env file at location: {}".format(remote_env_path)
            )
        try:
            db_info = self.compose_shopware_6_database_information(raw_env)
        except RuntimeError:
            self.discover_connection.importer.return_code = (
                REMOTE_DB_SW6_CONF_PARSE_ERR_RC
            )
            raise RuntimeError(
                "Could not parse database " "config from {}".format(remote_env_path)
            )

        return db_info

    def find_database_information(self):
        """
        Find the remote database connection details so we can later connect with these.
        :return dict remote_db_info: The remote database connection details
        """
        log.debug("Retrieving remote database information")
        if self.discover_connection.importer.site_type == "magento1":
            return self.parse_magento_1_database_information()
        elif self.discover_connection.importer.site_type == "magento2":
            return self.parse_magento_2_database_information()
        elif self.discover_connection.importer.site_type == "shopware5":
            return self.parse_shopware_5_database_information()
        elif self.discover_connection.importer.site_type == "shopware6":
            return self.parse_shopware_6_database_information()
        raise NotImplementedError(
            "Parsing {} database information has not been implemented "
            "yet".format(self.discover_connection.importer.site_type)
        )

    def check_remote_rsync(self):
        """
        Check if there is a remote rsync that we can use to transfer the shop
        :return bool available: True if rsync is available, False if not
        """

        return self.discover_connection.importer.ssh_conn.driver.exec_command_check_zero_exit(
            command="rsync --help",
            success_msg="Detected an rsync that we can use on the remote server",
            failure_msg="Did not find rsync on the remote server",
        )

    def check_remote_nginx_path(self):
        """
        Check if there is a remote nginx path that we can sync the configs from.
        This only works for imports from other Hypernodes because we check for
        our own hardcoded nginx path.
        :return str remote_nginx_path | None: The remote Nginx path or None
        """
        if self.discover_connection.importer.ssh_conn.driver.exec_command_check_zero_exit(
            command="test -d {}".format(HYPERNODE_NGINX_PATH),
            success_msg="Found an NGINX config dir that we can sync on the remote server",
            failure_msg="Did not find an NGINX config dir that we can sync on the remote host",
        ):
            return HYPERNODE_NGINX_PATH

    def check_hmv_active_remote(self):
        """
        Check if the remote server has hypernode-manage-vhosts available and enabled.
        :rtype: bool
        """
        if not self.discover_connection.importer.ssh_conn.driver.exec_command_check_zero_exit(
            command="which hypernode-manage-vhosts".format(HYPERNODE_NGINX_PATH),
            success_msg="Found the hypernode-manage-vhosts command",
            failure_msg="Did not find the hypernode-manage-vhosts command",
        ):
            return False
        if not self.discover_connection.importer.ssh_conn.driver.exec_command_check_zero_exit(
            command="/usr/bin/hypernode-manage-vhosts --list".format(
                HYPERNODE_NGINX_PATH
            ),
            success_msg="Hypernode-manage-vhosts works correctly",
            failure_msg="Hypernode-manage-vhosts is not yet activated",
        ):
            return False
        return True


class DiscoverConnection(ImporterSubClass):
    driver = None
    host = None
    default_file = None

    def __init__(self, importer):
        """
        Initialize the DiscoverConnection.
        :param obj importer: The Importer object where we will put the
        keep the global state and will put the discover connection object
        if we can determine information about the remote host successfully.
        :return None:
        """
        super(DiscoverConnection, self).__init__(importer=importer)

    def connect(self):
        """
        Connect to the local mysql instance. Returns the DiscoveryConnection
        object.

        :return obj DiscoveryConnection: The DiscoveryConnection object
        """
        log.info("Gathering information about the remote machine")
        self.driver = DiscoverDriver(discover_connection=self)
        self.importer.remote_path = self.driver.find_cms()
        self.importer.remote_db_info = self.driver.find_database_information()
        if not has_only_safe_chars(self.importer.remote_db_info["db_name"]):
            self.importer.return_code = DB_NAME_SPECIAL_CHAR_RC
            raise SpecialCharactersInDBName(
                "There are special characters in the database name. "
                "To prevent problems, we recommend that you rename the "
                "database {} to something without special "
                "characters.".format(self.importer.remote_db_info["db_name"])
            )
        if not has_only_safe_chars(self.importer.remote_db_info["table_prefix"]):
            self.importer.return_code = DB_TABLE_PREFIX_SPECIAL_CHAR_RC
            raise SpecialCharactersInTablePrefix(
                "There are special characters in the table prefix. "
                "To prevent problems, we recommend that change the "
                "table prefix {} to something without special "
                "characters.".format(self.importer.remote_db_info["db_name"])
            )
        if not db_username_fits_in_hypernode(self.importer.remote_db_info["username"]):
            self.importer.return_code = DB_USERNAME_TOO_LONG_RC
            raise DBUsernameTooLongException(
                "The remote database username seems to be too long. "
                "To prevent problems, we recommend to change the "
                'remote database user "{}" to have less than 16 '
                "characters.".format(self.importer.remote_db_info["username"])
            )
        if not db_name_fits_in_hypernode(self.importer.remote_db_info["db_name"]):
            self.importer.return_code = DB_NAME_TOO_LONG_RC
            raise DBNameTooLongException(
                "The remote database name seems to be too long. "
                "To prevent problems, we recommend to rename the "
                'remote database "{}" to have less than 64 '
                "characters.".format(self.importer.remote_db_info["db_name"])
            )
        self.importer.remote_rsync_available = self.driver.check_remote_rsync()
        self.importer.remote_nginx_path = self.driver.check_remote_nginx_path()
        self.importer.local_hmv_active = check_conditions_before_starting()
        self.importer.remote_hmv_active = self.driver.check_hmv_active_remote()
        return self
