import logging
import os
import re
import time
from datetime import datetime
from pathlib import Path

from kamikaze3.log.main import block_until_last_flow_done
from kamikaze3.systemctl.settings import set_attribute

logger = logging.getLogger(__name__)

DEFAULT_PHP_VERSION = "7.4"
JOB_POST_WAIT = 10


def is_hipex_node(importer):
    """
    Check whether the node resembles signs of a Hipex node
    :param importer: Importer object to connect to remote node
    to the node
    :rtype: bool
    """
    return_code, stdout, _ = importer.ssh_conn.driver.exec_command("hostname")
    if return_code != 0:
        logger.warn("Can't parse hostname. Assuming not a Hipex node")
        return False
    is_hipex = stdout.strip().decode("utf-8").endswith(".hipex.io")
    if is_hipex:
        logger.info("Found remote host to be a Hipex node")
    return is_hipex


def get_hipex_php_version(importer):
    """
    Get the PHP version of the domain on the Hipex node.
    If not found, we'll set the default PHP version
    :param importer: Importer object to connect to remote node
    :return str php_version | noneType None: Like '5.6' or '7.4'.
    If no PHP version can be detected None is returned.
    """
    return_code, stdout, _ = importer.ssh_conn.driver.exec_command(
        "ps x --sort -start_time"
    )
    if return_code != 0:
        logger.warn("Can't parse PHP version. PHP detection failed, returning None")
        return None

    for line in stdout.strip().decode("utf-8").split("\n"):
        if "php-fpm" in line:
            match = re.match(r".*pool php(\d{2})-.*", line)
            if not match:
                continue

            php_version = match.group(1)

            return "{}.{}".format(php_version[0], php_version[1])
    logger.warn("Could not detect PHP version, returning None")


def get_remote_mysql_version(importer):
    """
    Return the remote MySQL major version
    :param importer: Importer object to connect to remote node
    :return str mysql_version: Like '5.7' or '5.6'
    """
    mysql_version_command = 'SHOW VARIABLES WHERE Variable_name = "version";'
    results = importer.remote_db_conn.driver.exec_query(mysql_version_command)

    mysql_version = results[0][
        "Value"
    ]  # Returns a tuple of dicts containing 'Variable_name' and 'Value'
    mysql_major_version = ".".join(mysql_version.split(".")[:2])
    return mysql_major_version


def _create_hipex_state_dir() -> Path:
    date_of_import = datetime.now()
    backup_dir = Path(f"/data/web/importer_hipex_state/{date_of_import.isoformat()}")
    logger.info(f"Preserving Hipex state in {backup_dir}")
    backup_dir.mkdir(parents=True, exist_ok=True)
    return backup_dir


def _backup_hipex_crontab(importer, backup_dir):
    logger.info(f"Backing up crontab")
    return_code, stdout, stderr = importer.ssh_conn.driver.exec_command("crontab -l")
    if return_code != 0:
        logger.warn(f"Can't parse crontab, skipping crontab backup. Error: {stderr}")
    else:
        with open(backup_dir / "crontab", "w") as fp:
            fp.write(stdout.decode("utf-8", errors="ignore"))


def _backup_hipex_mysql_config(importer, backup_dir: Path):
    logger.info(f"Backing up MySQL")
    os.makedirs(backup_dir / "mysql")
    importer.file_transfer_conn.driver.download(
        "/etc/my.cnf", backup_dir / "mysql/my.cnf"
    )
    importer.file_transfer_conn.driver.download("/etc/my.cnf.d", backup_dir / "mysql")


def _backup_hipex_nginx_config(importer, backup_dir: Path):
    # Customer configs are stored in ~/domains/<domain>/var/etc/. This also contains
    # the Varnish VCL if it exists
    return_code, stdout, _ = importer.ssh_conn.driver.exec_command("ls -1 ~/domains/")
    if return_code != 0:
        logger.warn("Can't parse nginx domains. Skipping nginx backup")
    else:
        os.makedirs(backup_dir / "nginx")
        domains = stdout.decode("utf-8", errors="ignore").strip().split("\n")
        for domain_name in domains:
            logger.info(f"Backing up nginx domain {domain_name}")
            nginx_dir = f"domains/{domain_name}/var/etc"
            importer.file_transfer_conn.driver.download(
                nginx_dir, backup_dir / f"nginx/{domain_name}/"
            )


def _backup_hipex_redis_config(importer, backup_dir: Path):
    logger.info(
        f"Trying to backup any Redis files, it's OK if these fail. Those files might not be readable by the user."
    )
    os.makedirs(backup_dir / "redis")
    importer.file_transfer_conn.driver.download(
        "/etc/redis.conf", backup_dir / "redis/redis.conf"
    )
    importer.file_transfer_conn.driver.download(
        "/etc/redis-sentinel.conf", backup_dir / "redis/redis-sentinel.conf"
    )
    importer.file_transfer_conn.driver.download("/etc/redis.d", backup_dir / "redis")


def _backup_hipex_php_config(importer, backup_dir: Path):
    php_version = get_hipex_php_version(importer=importer)
    if php_version:
        logger.info(f"Backing up php '{php_version}'")
        php_short_version = php_version[0:3:2]
        os.makedirs(backup_dir / "php")
        importer.file_transfer_conn.driver.download(
            f"/etc/opt/remi/php{php_short_version}", backup_dir / f"php"
        )
    else:
        logger.info(
            "Because the PHP version was not detected we're not backing "
            "up any PHP settings for reference"
        )


def _backup_hipex_supervisor_config(importer, backup_dir: Path):
    logger.info(f"Backing up supervisor")
    os.makedirs(backup_dir / "supervisor")
    importer.file_transfer_conn.driver.download(
        f"supervisor/supervisor.d", backup_dir / f"supervisor"
    )


def backup_hipex_state(importer):
    """
    Create a copy of all existing state on the Hipex server for future debugging. This can
    be helpful when comparing the previous state on the Hipex server to the new state on
    the Hypernode.
    """
    backup_dir = _create_hipex_state_dir()
    _backup_hipex_crontab(importer=importer, backup_dir=backup_dir)
    _backup_hipex_mysql_config(importer=importer, backup_dir=backup_dir)
    _backup_hipex_nginx_config(importer=importer, backup_dir=backup_dir)
    _backup_hipex_redis_config(importer=importer, backup_dir=backup_dir)
    _backup_hipex_php_config(importer=importer, backup_dir=backup_dir)
    _backup_hipex_supervisor_config(importer=importer, backup_dir=backup_dir)


def setup_hipex_transfer(importer):
    """
    Detect all relevant settings at the Hipex server and set them at the Hypernode server.
    :param importer: Importer object to connect to remote node
    """
    # Detect PHP version
    remote_php_version = get_hipex_php_version(importer=importer)

    # Detect mysql version
    remote_mysql_version = get_remote_mysql_version(importer=importer)

    # Set mysql version
    if importer.dry_run:
        logger.info("Would have set MySQL version to {}".format(remote_mysql_version))
    else:
        logger.info("Setting MySQL version to {}".format(remote_mysql_version))
        set_attribute("mysql_version", remote_mysql_version)

    # Set php version
    if importer.dry_run:
        logger.info("Would have set PHP version to {}".format(remote_php_version))
    else:
        if remote_php_version:
            logger.info("Setting PHP version to {}".format(remote_php_version))
            set_attribute("php_version", remote_php_version)
        else:
            logger.info(
                "Not setting PHP version because we failed to detect "
                "which version on the Hipex server"
            )

    # Wait a little bit to let the job be posted on the jobboard
    logger.info(
        "Waiting {} seconds until update jobs are posted..".format(JOB_POST_WAIT)
    )
    time.sleep(JOB_POST_WAIT)

    # Wait for update
    block_until_last_flow_done()

    # Copy cronjobs
    # TODO: Use dry-run
    # TODO: Parse `crontab -l` and place it on the node
