import logging
import socket
from abc import ABC, abstractmethod
from contextlib import closing, contextmanager, suppress
from errno import EADDRNOTAVAIL, EHOSTUNREACH, ETIMEDOUT
from getpass import getpass
from shlex import quote
from subprocess import Popen
from threading import Timer
from time import sleep

from paramiko import AutoAddPolicy, SSHClient, SSHException
from paramiko.config import SSHConfig
from paramiko.ssh_exception import NoValidConnectionsError

from kamikaze3.common.env import get_hypernode_name, host_is_a_docker, host_is_a_vagrant
from kamikaze3.importer.abstract import ImporterSubClass
from kamikaze3.importer.error_codes import (
    CIRCULAR_IMPORT_RC,
    IPV6_RC,
    NO_SUCH_DNS_RC,
    SSH_AUTH_FAIL_RC,
    SSH_INFO_INVALID_RC,
    SSH_NO_PASS_RC,
    SSH_NO_ROUTE_RC,
    SSH_PORT_NOT_AVAILABLE_RC,
    SSH_SPECIAL_CHAR_RC,
    SSH_TIMEOUT_RC,
    SSH_WRONG_HN_USER_PROVIDED_RC,
)
from kamikaze3.importer.proxy import ProxyCommand
from kamikaze3.importer.utils import catch_and_raise_known_errors, compose_sshpass_env

log = logging.getLogger(__name__)


class SpecialCharacterException(RuntimeError):
    pass


def source_is_destination(host):
    """
    Check if the importer is used to import the shop from a
    Hypernode to the same Hypernode. That wouldn't work for
    obvious reasons so if that is the case we should error out.
    In case if it is a Vagrant this does not apply so we can
    return early.
    :param str host: The host that is being imported from.
    Example: vdloo.hypernode.io
    :return bool is_destination: True if the source is the
    destination, False if the source is another host than
    the local machine.
    """
    return (
        not host_is_a_vagrant()
        and not host_is_a_docker()
        and host == get_hypernode_name()
    )


def get_ssh_config_for_host(host, attribute):
    """
    Parse the ssh config (if it exists) for an identityfile for the host
    :param str host: host that should be looked up in the ssh config
    :param str attribute: the item to find. e.g. identityfile or proxycommand
    Note(vdloo): The attribute is normalized to lowercase. I.e. specify
    proxycommand, not ProxyCommand
    :return str identityfile | None:
    """
    config = SSHConfig()
    with suppress(IOError):
        config.parse(open("/data/web/.ssh/config"))
        return config.lookup(host).get(attribute)


class SSHDriver(ABC):
    # Note(vdloo): The TestCase in tests/importer/ssh/ssh_driver/test_ssh_driver.py
    # tests this abstract class entirely at once. Because we need to test the abstract
    # method functionality it makes no sense to test these methods individually
    # as they contain very little business logic. We just need to check if
    # Python actually throws TypeErrors if all abstract methods we require are
    # not implemented.
    def check_parameters(self):
        """
        Check the constructor received an object to use as the SSH Connection
        and get the connection details from.
        :return None:
        """
        if self.ssh_connection is None:
            raise RuntimeError(
                "You must pass the SSHConnection object to the constructor "
                "of an SSHDriver object because the configured attributes are "
                "used to determine the connection details of the remote "
                "server."
            )

    def __init__(self, ssh_connection):
        """
        Initialize the SSHDriver object and store the passed ssh
        connection on the self. Check if the constructor received
        the required parameters.
        :param obj ssh_connection: The SSH connection object with
        the details to use to connect to the remote SSH server with.
        """
        self.ssh_connection = ssh_connection
        self.check_parameters()

    @abstractmethod
    def exec_command(self, command):
        """
        All SSH Drivers must implement exec_command which takes a
        command as a string and executes it on the remote host.
        :param str command: The command to execute on the remote host
        :return tuple (int return_code, str standard_out, str standard_error): A
        tuple of the return code and the output streams read to strings.
        """

    def exec_command_check_zero_exit(self, command, success_msg, failure_msg):
        """
        Execute a command and check if the return value was zero,
        if so return True, if not return False. Debug logs the
        success_msg on zero, debug logs the failure_msg on nonzero
        :param str command: The command to execute on the remote server
        :param str success_msg: The success message to log
        :param str failure_msg: The failure message to log
        :return bool zero_exit: True if zero exit, False if nonzero
        """
        return_code, _, _ = self.exec_command(command)
        if return_code == 0:
            log.debug(success_msg)
            return True
        else:
            log.debug(failure_msg)
            return False

    def tcp_forward(self, remote_host, remote_port, local_port):
        """
        Open a TCP forward tunnel. Forward a port of a remote
        host to a port on the local machine.
        :param str remote_host: The remote host to forward
        :param int remote_port: The remote port to forward
        :param int local_port: The local port to forward
        :return obj proc: The process that holds the tunnel
        """
        forward_arg = "{}:{}:{}".format(local_port, quote(remote_host), remote_port)
        proxy_option = () if self.ssh_connection.proxy else ("-oProxyCommand=none",)
        ssh_pass = (
            ("/usr/bin/sshpass", "-e", "--") if self.ssh_connection.password else ()
        )
        tunnel_command = (
            *ssh_pass,
            "ssh",
            "-o",
            "StrictHostKeyChecking=no",
            "-o",
            "UserKnownHostsFile=/dev/null",
            "-4",
            "-p",
            str(self.ssh_connection.importer.port),
            "-l",
            self.ssh_connection.importer.user,
            self.ssh_connection.importer.host,
            "-L",
            forward_arg,
            "-N",
            *proxy_option,
        )
        tunnel_environment = compose_sshpass_env(password=self.ssh_connection.password)
        proc = Popen(tunnel_command, env=tunnel_environment)
        sleep(5)  # give the subprocess some time to establish the tunnel
        return proc


class SSHParamikoDriver(SSHDriver):
    """
    Use Paramiko to connect to the remote host and allow for the execution of
    shell commands on the remote server.
    """

    @property
    def connect_timeout(self):
        return 10

    paramiko_ssh_client = None
    paramiko_proxy = None
    connection = None

    def configure_ssh_client(self):
        """
        Configure the Paramiko SSH client
        :return None:
        """
        self.paramiko_ssh_client = SSHClient()
        self.paramiko_ssh_client.set_missing_host_key_policy(AutoAddPolicy())

    def configure_proxy(self):
        """
        Configure the Paramiko proxy if we need one. Rewrites the standard port of the
        ProxyCommand if a nonstandard port needs to be proxied to.
        :return None:
        """
        self.paramiko_proxy = get_ssh_config_for_host(
            self.ssh_connection.importer.host, "proxycommand"
        )
        if self.paramiko_proxy:
            # Replace the nonstandard port in a SSH builtin secure channel forward proxycommand
            # Looks like: ssh -W vdloo.hypernode.io:22 app@apenvliegenniet.hypernode.io
            self.paramiko_proxy = self.paramiko_proxy.replace(
                ":22 ", ":{} ".format(self.ssh_connection.importer.port)
            )
            # Replace the nonstandard port in a netcat proxycommand
            # Looks like: ssh vdloo.hypernode.io nc apenvliegenniet.hypernode.io 22 2> /dev/null
            self.paramiko_proxy = self.paramiko_proxy.replace(
                " 22 ", " {} ".format(self.ssh_connection.importer.port)
            )
            self.ssh_connection.proxy = True

    @contextmanager
    def close_conn_on_timeout(self):
        """
        Work around a bug in the python-paramiko packaged in Xenial
        https://github.com/paramiko/paramiko/issues/520

        Without this hack any SSH exception hangs on a deadlock
        in an internal thread of the transport layer of Paramiko
        if an exception occurs when authenticating at the remote sshd.

        Closes connection after 10 seconds if connecting takes longer
        than allowed amount of time, which is what happens when the
        Paramiko internal threads get deadlocked because they are
        waiting for a .join that never happens because of the exception.

        :yield NoneType None: Yields nothing.
        """
        timer = Timer(self.connect_timeout, self.paramiko_ssh_client.close)
        timer.start()
        yield
        timer.cancel()

    def get_connection(self, force_no_proxy=False):
        """
        Get a Paramiko SSH connection
        :param bool force_no_proxy: Force not using a jumphost even
        though one might be configured.
        :return None:
        """
        key_filename = get_ssh_config_for_host(
            self.ssh_connection.importer.host, "identityfile"
        )

        # Note(vdloo): this is imported from the proxy.py that we include in this project
        # instead of the one from the system installed paramiko, which at the time of writing
        # is 1.16.0-1. The system ProxyCommand implementation in Paramiko is broken in Python 3.
        # See:
        # https://github.com/paramiko/paramiko/issues/673
        # https://github.com/paramiko/paramiko/pull/681
        # The proxy.py in this module was imported from the current Paramiko master (as it
        # contains even more fixes and is still compatible with 1.16.0-1):
        # https://github.com/paramiko/paramiko/blob/ecb297518267487853813109da946d7240d717e1/paramiko/proxy.py
        proxy = (
            ProxyCommand(self.paramiko_proxy)
            if self.paramiko_proxy and not force_no_proxy
            else None
        )

        # Note(vdloo): we could consider adding a fallback strategy here that tries to connect
        # to the remote server through the proxy using the ssh config configured identityfile.
        # Normally if you have a block like this:
        #
        # Host
        #     henkslaaf.hypernode.io
        #     ProxyCommand ssh -W %h:%p app@apenvliegenniet.hypernode.io
        #     IdentityFile /tmp/some_key.id.rsa
        #
        # And you run `ssh app@henkslaaf.hypernode.io`, SSH will try to connect
        # to henkslaaf.hypernode.io by first connecting to
        # apenvliegenniet.hypernode.io and ssh-ing from there. But to connect
        # to apenvliegenniet.hypernode.io the configured IdentityFile for henkslaaf
        # will not be used. This is expected SSH behavior but I can imagine that this
        # is not the most known SSH feature, so automatically retrying to connect
        # to the proxy using the configured IdentityFile for the host we are connecting
        # to through the proxy might save some headaches and make things just 'work'.
        # But because we don't really know if this ever happens at all and if it does
        # how often, we'll leave that out for now. If this becomes a known problem in
        # the future we can attempt this strategy to mitigate that.

        with self.close_conn_on_timeout():
            self.paramiko_ssh_client.connect(
                self.ssh_connection.importer.host,
                port=self.ssh_connection.importer.port,
                username=self.ssh_connection.importer.user,
                password=self.ssh_connection.password,
                # The paramiko ssh client connect method takes key_filename as string
                # or a list of strings. The get_ssh_config_for_host always returns a
                # list of str filenames even if only one is found. If no is configured,
                # None will be returned.
                key_filename=key_filename,
                sock=proxy,
            )
        self.connection = self.paramiko_ssh_client

    def try_get_connection(self):
        """
        Try establishing an SSH connection with the remote host using Paramiko.
        Tries to fall back to no proxy if it failed and a proxy was specified.
        :return None:
        """
        try:
            self.get_connection()
        except SSHException:
            # If we can't connect using the proxy, try again without the proxy
            if self.paramiko_proxy:
                # Note(vdloo): if the Hyperform needs to resort to this strategy and it succeeds,
                # it could be that the site is not importable from the eventual trial (with a
                # different IP) if there is a firewall in place. In that case the operator
                # could manually set up a tunnel to the remote server through the hyperform server
                # like: ssh -L 1222:the_remote_server.nl:22 hyperform.hypernode.io and then import
                # from the 2222 port bound on the localhost of the trial you want to import the shop
                # to.
                log.debug(
                    "Paramiko could not connect to the remote host using the specified jumphost\n"
                    "Trying again without the proxy to see if we can connect directly instead. "
                )
                self.get_connection(force_no_proxy=True)
            else:
                # If there was no proxy to begin with, don't catch the exception
                raise

    def __init__(self, ssh_connection):
        """
        Initialize the SSHParamikoDriver. Will throw an SSHException if the
        connection will not be able to execute shell commands remotely.
        :param obj ssh_connection: The SSH connection object with
        the details to use to connect to the remote SSH server with.
        """
        log.debug("Attempting to connect to the remote server using Paramiko")
        super(SSHParamikoDriver, self).__init__(ssh_connection=ssh_connection)
        self.configure_ssh_client()
        self.configure_proxy()
        self.try_get_connection()
        log.debug(
            "Checking if we can execute remote commands using the Paramiko connection"
        )
        self.exec_command("/bin/true")

    def exec_command(self, command):
        """
        Execute a command as string on the remote host
        :param str command: The shell command as a string
        :return None:
        """
        _, stdout, stderr = self.connection.exec_command(command)
        return stdout.channel.recv_exit_status(), stdout.read(), stderr.read()

    @contextmanager
    def open_file_on_node(self, file_name, flags="r", mode=0o600):
        """
        Get a paramiko sftp file handle on the remote server
        :param str file_name: file to get a handle on
        :param str flags: flags contains the requested mode for opening (read-only, write-append, etc)
        See http://docs.paramiko.org/en/2.0/api/sftp.html. As string, not as bitset
        :param octal mode: The permissions to open the file with if O_CREAT is used.
        :yield obj remote_file: The handle
        """
        sftp_client = self.connection.open_sftp()
        with closing(sftp_client.open(file_name, flags, mode)) as remote_file:
            yield remote_file

    def remove_file_from_node(self, path):
        sftp_client = self.connection.open_sftp()
        sftp_client.remove(path)
        sftp_client.close()

    def read_file(self, file_name):
        """
        Read a file on a node and return the contents as a string
        :param str file_name: The file to read
        :return iter read_lines: Returns the result of readlines, which is an iterable
        """
        with self.open_file_on_node(file_name) as remote_file:
            lines = remote_file.readlines()
            return lines


class SSHConnection(ImporterSubClass):
    password = None
    driver = None
    proxy = None

    def check_parameters(self):
        """
        Check the constructor received an object to use as the Importer
        and get the connection details from.
        :return None:
        """
        if source_is_destination(self.importer.host):
            self.importer.return_code = CIRCULAR_IMPORT_RC
            raise RuntimeError(
                "Looks like you tried importing a production Hypernode "
                "from the same production Hypernode!\n Perhaps you meant "
                "to import the production Hypernode into a local development "
                "environment like hypernode-docker?\nIf so, enter the command "
                "in the shell of that environment, not this production shell."
            )

    def __init__(self, importer):
        """
        Instantiate an SSHConnection object
        :param obj importer: The Importer object containing all
        configurations required to connect to the remote SSH server.
        :return None:
        """
        self.proxy = False
        super(SSHConnection, self).__init__(importer=importer)
        self.check_parameters()

    def attempt_connect(self):
        """
        Connect to the remote SSH server using any of the configured
        SSH connection drivers. These are objects that implement
        'exec_command' which runs a command over SSH on the remote
        host. In some cases Paramiko doesn't work, if so we can try
        to fall back to using the 'ssh' tool over the commandline.
        This will give degraded performance because instantiating
        connections is slower, but it's better than bailing out.
        :return None:
        """
        log.debug(
            "Connecting to {}@{} {} password".format(
                self.importer.user,
                self.importer.host,
                "using" if self.password is not None else "without",
            )
        )
        self.driver = SSHParamikoDriver(ssh_connection=self)

    def prompt_for_password(self):
        """
        Prompt the user for an SSH password. If we can't login using the
        provided key from the default locations on disk or from the agent
        that doesn't always mean the user wants to use password auth but
        we prompt anyway because that's one less decision to make up front
        when specifying the importer options in case you do want to use
        password authentication. Getpass will read from STDIN when the
        importer is running in a non-interactive shell. So you can pipe
        a string to hypernode-importer and it will be used as the password
        in case the key authentication method fails. When running interactively
        the user can type in their password.
        :return None:
        """
        self.password = getpass(
            "Can't login to {}@{} using SSH key or agent, please provide "
            "a password: ".format(self.importer.user, self.importer.host)
        )

    def attempt_password_auth_connect(self):
        """
        Prompt for a password from the user and use that to attempt
        to login to the remote SSH server. Catches known exceptions
        and sets specific return codes for them on the importer object
        in order to help translate the exception to the user.
        :return None:
        """
        try:
            self.prompt_for_password()
            self.attempt_connect()
        except SSHException:
            self.importer.return_code = SSH_AUTH_FAIL_RC
            raise
        except SpecialCharacterException:
            self.importer.return_code = SSH_SPECIAL_CHAR_RC
            raise
        except EOFError:
            self.importer.return_code = SSH_NO_PASS_RC
            raise RuntimeError(
                "Key authentication failed and no password was provided. "
                "Can not log in to remote SSH server."
            )

    def catch_known_ssh_errors(self, e):
        """
        This method checks a caught SSHException for known socket errors
        that we can translate to something more user-friendly for the user
        of the program. Sets specific exit codes and raises alternative
        (not SSHException) errors if the error is known.
        :param obj SSHException e: A caught SSHException as e.
        :return None:
        """
        if isinstance(e, UnicodeError):
            self.importer.return_code = SSH_INFO_INVALID_RC
            raise RuntimeError(
                "The SSH information you entered does not seem valid. Is it correct? "
                "Did you perhaps make a typo?"
            )
        if e.errno == ETIMEDOUT:
            self.importer.return_code = SSH_TIMEOUT_RC
            raise RuntimeError(
                "Could not connect to the provided SSH address (timed out), "
                "is it correct? Maybe there is a firewall?"
            )
        if e.errno == EHOSTUNREACH:
            self.importer.return_code = SSH_NO_ROUTE_RC
            raise RuntimeError(
                "Could not connect to the provided SSH address (no route to host), "
                "is it correct? Maybe there is a firewall?"
            )
        if e.errno == socket.EAI_NONAME or e.errno == socket.EAI_AGAIN:
            self.importer.return_code = NO_SUCH_DNS_RC
            raise RuntimeError(
                "Could not resolve the provided SSH address. Is it correct? Does "
                "the DNS exist? Name or service not known. If you just updated "
                "your DNS it is possible that the TTL hasn't expired yet."
            )
        if e.errno == EADDRNOTAVAIL:
            self.importer.return_code = IPV6_RC
            raise RuntimeError(
                "Could not resolve remote address, because it was most likely an IPv6 address. Is it possible to "
                "change this address to an IPv4 address?"
            )
        # Can't be tested since NoValidConnectionsError breaks in python3 once you try to instantiate it correctly
        if isinstance(e, NoValidConnectionsError):
            self.importer.return_code = SSH_PORT_NOT_AVAILABLE_RC
            raise RuntimeError(
                "Something went wrong. Most likely this is because we couldn't SSH through the provided SSH "
                "port. Is it possible to enable access to this port or supply a different one? Usually "
                "port 22 is used by default for SSH."
            )

    def connect(self):
        """
        Connect to a remote SSH server. Returns the SSHConnection
        object and a password if one was prompted for and allowed
        the program to connect successfully to the remote server.

        Use the specified connection details to connect to the remote
        server and instantiate the first SSH driver object that works.
        We attempt to use Paramiko first but in case that does not work
        for any reason we try to fall back to connecting using the
        commandline SSH utility. If we can't connect with any of the SSH
        keys available to the program, we prompt for a password and try
        again once more.

        :return tuple (ssh_connection, password): The SSHConnection
        object and the password if any.
        """
        log.info("Connecting to the remote SSH server")
        with catch_and_raise_known_errors(
            exceptions=(OSError, UnicodeError),
            catch_function=self.catch_known_ssh_errors,
        ):
            try:
                self.attempt_connect()
            except SSHException:
                self.attempt_password_auth_connect()

        if self.importer.return_code == SSH_NO_PASS_RC:
            if self.importer.user != "app" and self.importer.host.endswith(
                "hypernode.io"
            ):
                self.importer.return_code = SSH_WRONG_HN_USER_PROVIDED_RC

        return self, self.password
