import logging
import os
from abc import abstractmethod
from contextlib import contextmanager, suppress
from shlex import quote

from kamikaze3.common.env import get_hypernode_name
from kamikaze3.common.shell import replace_in_file, run_local_command
from kamikaze3.importer.abstract import ImporterSubClass
from kamikaze3.importer.error_codes import (
    DISK_TOO_SMALL_RC,
    EXEC_QUERY_CANT_SET_REMOTE_MAINTENANCE,
    REDIS_FLUSH_ERR_RC,
    SSH_CANT_SET_MAINTENANCE_RC,
    SSH_CANT_UNSET_MAINTENANCE_RC,
)
from kamikaze3.importer.hipex import (
    backup_hipex_state,
    is_hipex_node,
    setup_hipex_transfer,
)
from kamikaze3.importer.nginx import copy_app_nginx_files, copy_app_nginx_hmv_files
from kamikaze3.importer.validators import check_mysql_major, check_php_major
from kamikaze3.utils import async_pool, request_confirmation

log = logging.getLogger(__name__)
DB_SKIP_TABLES = {
    "magento1": [
        "log_url",
        "log_url_info",
        "log_customer",
        "log_quote",
        "report_event",
        "dataflow_batch_export",
        "dataflow_batch_import",
        "core_file_storage",
        "log_visitor_info",
        "log_visitor",
        "ecomdev_stock_import_tmp",
        "innodb_monitor",
    ],
    "magento2": ["report_event", "customer_visitor", "innodb_monitor"],
    "shopware5": ["campaigns_logs", "core_log"],
    "shopware6": ["log_entry"],
}
# Time out the database transfer after 3 hours. If it takes longer than that it is
# probably stuck or the database is truly huge or the connection is very slow.
DATABASE_SYNC_TIMEOUT = 10800


def dump_remote_database_into_local(
    default_file, remote_db_info, structure_only=False, extra_arguments=None
):
    """
    Perform the mysqldump of the remote server into the local server
    :param str default_file: The path to the .cnf file to get the
    credentials from for the remote database
    :param dict remote_db_info: The dict with the connection
    details for the remote database
    :param bool structure_only: Only copy table definitions, no data
    :param str extra_arguments: A string of extra arguments
    :return None:
    """
    no_data = "--no-data" if structure_only else ""
    extra_arguments = extra_arguments if extra_arguments else ""
    mysqldump_command = (
        "mysqldump --defaults-file={} --opt --databases {} --single-transaction "
        "--force --insert-ignore --quote-names {} {} | "
        'pv -ab --name "Syncing database" 1> /dev/stdout | '
        "mysql --defaults-file={}"
        "".format(
            default_file,
            quote(remote_db_info["db_name"]),
            no_data,
            extra_arguments,
            "/data/web/.my.cnf",
        )
    )
    run_local_command(
        ["bash", "-c", "set -o pipefail; " + mysqldump_command],
        shell=False,
        timeout=DATABASE_SYNC_TIMEOUT,
    )


def import_database(default_file, remote_db_info, extra_arguments=None):
    """
    Import the database of the shop
    # NOTE(vdloo): This is a function instead of an object method on
    ImportCMS because in order to run in it in a thread pool
    python needs to 'pickle' it, but pickling objects or object
    methods is not trivial in python. We could encode the whole
    thing with messagepack or something but for now this is fine.
    :param str default_file: The path to the .cnf file to get the
    credentials from for the remote database
    :param dict remote_db_info: The dict with the connection
    details for the remote database
    :param str extra_arguments: A string of extra arguments. Is only
    added on the real sync, not the initial structure only sync
    :return None:
    """
    log.info("Starting database synchronization")
    # The dump runs twice, one to copy all table definitions, one
    # to sync data from select tables only
    dump_remote_database_into_local(default_file, remote_db_info, structure_only=True)
    dump_remote_database_into_local(
        default_file,
        remote_db_info,
        structure_only=False,
        extra_arguments=extra_arguments,
    )


class ImportCMS(ImporterSubClass):
    local_database_config = None
    local_env_config = None
    base_url = None
    admin_url = None

    @property
    def source_maintenance_path(self):
        raise NotImplementedError()

    @property
    def destination_maintenance_path(self):
        raise NotImplementedError()

    def __init__(self, importer):
        """
        Initialize the ImportCMS object.
        :param obj importer: The Importer object where we will put the
        keep the global state. and which will call the execute method.
        :return None:
        """
        super(ImportCMS, self).__init__(importer=importer)
        self.skip_db_import = self.importer.skip_db_import
        self.skip_file_import = self.importer.skip_file_import

    def prepare_database(self):
        """
        Prepare the local database for the imported database
        :return None:
        """
        if not self.skip_db_import:
            log.debug("Preparing local database for import")
            self.importer.local_db_conn.driver.drop_database()
            self.importer.local_db_conn.driver.create_database()
            # Keep the connection alive by pinging the MySQL server.
            # The import can take a while and by the time it is finished
            # the connection to the local DB might have timed out.
            # Causing MySQL has gone away errors like 2003 and 2006
            # See https://dev.mysql.com/doc/refman/5.7/en/gone-away.html
            # Note that pinging is computationally expensive though,
            # that is why we only do it in case we know we will actually
            # be doing a database import and we only do it on the local db.
            # Also see https://stackoverflow.com/a/29331237/4158804
            self.importer.local_db_conn.driver.connection.ping(True)

        self.importer.local_db_conn.driver.update_my_cnf_if_needed()
        self.importer.local_db_conn.driver.connection.select_db(
            self.importer.remote_db_info["db_name"]
        )

    @abstractmethod
    def import_filesystem(self):
        """
        Import the filesystem of the shop
        :return None:
        """
        self.download_nginx_configs()

    def purge_redis_cache(self):
        """
        Flush the redis database.
        :return None:
        """
        log.info("Purging Redis cache")
        return_code = run_local_command(["redis-cli", "flushall"])
        if return_code != 0:
            self.importer.return_code = REDIS_FLUSH_ERR_RC
            raise RuntimeError("Failed to flush the Redis cache.")

    def fix_up_local_database_config(self):
        """
        Replace the MySQL DB hostname in the Magento
        or shopware config to the hostname of the database on the
        Hypernode the shop was imported to. 'mysqlmaster'
        is defined as an alias in /etc/hosts
        :return None:
        """
        if not self.local_database_config:
            raise RuntimeError(
                "Parent object of ImportCMS needs to "
                "define a self.local_database_config property"
            )
        log.debug("Reconfiguring the local database config")
        replace_in_file(
            self.local_database_config,
            self.importer.remote_db_info["host"].encode("utf-8"),
            b"mysqlmaster",
        )

    def compose_skip_tables_args(self, tables_to_skip):
        """
        Compose the skip table arguments for a mysqldump
        :param iter[str, ..] tables_to_skip: An iterable of table names to skip
        :return str skip_tables_arguments: String of arguments for
        mysqldump to skip tables
        """

        def ignore_table_flag(table_name):
            composed_table_name = "{}.{}{}".format(
                self.importer.remote_db_info["db_name"],
                self.importer.remote_db_info["table_prefix"],
                table_name,
            )
            return "--ignore-table={}".format(quote(composed_table_name))

        return " ".join(map(ignore_table_flag, tables_to_skip))

    def sync_data_serial(self, import_db_args, import_db_kwds):
        """
        Import the database and the filesystem serially, does not run
        the two imports at the same time
        :param iter import_db_args: The import_database function args
        :param dict import_db_kwds: The import database function kwargs
        :return None:
        """
        if not self.skip_db_import:
            import_database(*import_db_args, **import_db_kwds)
        if not self.skip_file_import:
            self.import_filesystem()

    def sync_data_parallel(self, import_db_args, import_db_kwds):
        """
        Import the database and the filesystem at the same time. Saves
        time but increases the load on both the source and the destination.
        :param iter import_db_args: The import_database function args
        :param dict import_db_kwds: The import database function kwargs
        :return None:
        """
        # NOTE(vdloo): Any exceptions will be thrown as part of the .join() call
        # after the context exits. So if import filesystem excepts we'll never see
        # the database error if it also excepts. That should be OK because people
        # can fix the first error and then resync and discover the second error if
        # it still exists.
        with async_pool(max_procs=1) as pool:
            if not self.skip_db_import:
                pool.apply_async(
                    import_database, args=import_db_args, kwds=import_db_kwds
                )
            # Need to run the rsync in the main thread because
            # we can't pickle the importer object. That's OK,
            # because we can run the mysqldump in the background
            # and join only after the rsync is complete, effectively
            # making it two threads anyway.
            if not self.skip_file_import:
                self.import_filesystem()

    def skip_tables_args(self, cms_type):
        """
        Compose the skip table arguments for the Magento 1 mysqldump
        :return str skip_tables_arguments: String of arguments for
        mysqldump to skip tables
        """
        tables_to_skip = DB_SKIP_TABLES[cms_type]

        return self.compose_skip_tables_args(tables_to_skip)

    def sync_data(self, extra_mysqldump_arguments=None, cms_type=None):
        """
        Sync the shop data. The filesystem and the database in parallel.
        :param cms_type: the type of content management system for skip arguments
        :param str extra_mysqldump_arguments: A string of extra arguments.
        Is only added on the real sync, not the initial structure only sync
        :return None:
        """
        extra_mysqldump_arguments = (
            extra_mysqldump_arguments if extra_mysqldump_arguments else ""
        )
        extra_mysqldump_arguments += self.skip_tables_args(cms_type)
        if self.skip_db_import:
            import_db_args = []
            import_db_kwds = {}
        else:
            import_db_args = (
                self.importer.remote_db_conn.driver.default_file,
                self.importer.remote_db_info,
            )
            import_db_kwds = {"extra_arguments": extra_mysqldump_arguments}
        if self.importer.single_threaded:
            self.sync_data_serial(import_db_args, import_db_kwds)
        else:
            self.sync_data_parallel(import_db_args, import_db_kwds)

    def print_success(self):
        """
        Print the success message with information about the shop
        :return None:
        """
        log.info("Success! Import completed.")
        if self.base_url and self.admin_url:
            log.info("You can test your Hypernode here:\n")
            log.info(" - {}".format(self.base_url))
            log.info(" - {}".format(self.admin_url))
        # log warning messages here if any, to make sure the warning messages are still visible after the import.
        if self.importer.import_warnings:
            log.warning("")
            for warning in self.importer.import_warnings:
                log.warning(warning)
        log.info("\nFor next steps, see http://support.hypernode.com/going-live.html\n")

    def download_nginx_configs(self):
        """
        If there are Nginx configs on the remote host in the expected
        location for a Hypernode, then copy those over to the local machine
        :return None:
        """
        if self.importer.remote_nginx_path:
            if self.importer.skip_nginx_import:
                log.info(
                    "Found NGINX configs on the remote in {}, not syncing them because "
                    "--skip-nginx-import was specified".format(
                        self.importer.remote_nginx_path
                    )
                )
            else:
                log.info("Starting NGINX config synchronization")
                if self.importer.import_nginx_strategy == "standard":
                    copy_app_nginx_files(self.importer)
                elif self.importer.import_nginx_strategy == "hmv-copy-vhosts":
                    copy_app_nginx_hmv_files(self.importer)
                elif self.importer.import_nginx_strategy == "hmv-on-remote":
                    raise RuntimeError(
                        "Importing nginx configuration files from a remote server with hypernode-manage-vhosts to a "
                        "host without this feature enabled is currently not supported. Please use the "
                        "--skip-nginx-import flag to continue. Check out our support documentation on "
                        "hypernode-manage-vhosts to generate new VHosts: "
                        "https://support.hypernode.com/knowledgebase/hypernode-managed-vhosts/"
                    )
                else:
                    raise RuntimeError(
                        "Unknown Nginx import strategy {}".format(
                            self.importer.import_nginx_strategy
                        )
                    )

        else:
            log.debug("No NGINX configs found to sync, skipping")

    def check_disk_usage(self):
        """
        Check with du what the disk usage is at the remote server and compare it with the disk size on the local server.
        If the disk size is less the disk usage we throw an error.
        :return None:
        """
        if self.get_remote_disk_usage() > self.get_disk_size():
            self.importer.return_code = DISK_TOO_SMALL_RC
            raise RuntimeError(
                "The disk size on the hypernode is lower than the size of the magento installation."
            )

    def get_remote_disk_usage(self):
        """
        Get with du what the disk usage is at the remote server. If we are not able to check the disk space return 0.
        :return Int:
        """
        return_code, du_output, _ = self.importer.ssh_conn.driver.exec_command(
            "du --apparent-size --max-depth=0 {}".format(
                quote(self.importer.remote_path)
            )
        )
        if return_code == 0:
            return int(du_output.split()[0])
        else:
            return 0

    def get_disk_size(self):
        """
        Get the disk size on the local server.
        :return Int:
        """
        statvfs = os.statvfs("/data/web")
        return statvfs.f_frsize * statvfs.f_blocks

    @contextmanager
    def manage_maintenance(self):
        """
        Make sure maintenance flags are correctly placed on the source and destination nodes. Available as context
        manager. This allows, for example, setting the maintenance flag on the source, syncing data, and removing the
        flag from the destination.

        Please note the comment below why this context manager does not catch errors.
        :return: None
        """
        if self.importer.set_maintenance_on_source:
            if self.importer.dry_run:
                log.info("Skipping settings maintenance mode, doing a dry run!")
            else:
                log.info("Setting maintenance mode, source shop will go offline now!")
                self.set_maintenance_flag()

            # Note: this contextmanager should NOT catch the exception and not run the exit functions before returning.
            # Finally is not used here, because when something goes wrong after setting the maintenance flag on the
            # source host we wouldn't want the destination host (the Hypernode) to become available, because the sync
            # is probably incomplete. If that happens, fix whatever is wrong that caused the importer to fail and then
            # run it again (managing maintenance mode is idempotent this way).
            yield

            if self.importer.dry_run:
                log.info(
                    "Skip ensuring that maintenance mode is unset on the new copy, doing a dry run!"
                )
            else:
                log.info(
                    "Ensuring that the maintenance flag is removed from the copy of the site."
                )
                self.unset_maintenance_flag_on_destination()
        else:
            log.debug("Not managing maintenance mode.")
            yield

    def set_maintenance_flag(self):
        """
        Set the maintenance flag on the (old) Magento site.
        :return: None:
        """
        try:
            with self.importer.ssh_conn.driver.open_file_on_node(
                self.source_maintenance_path, flags="w"
            ) as f:
                f.write("")
        except IOError:
            self.importer.return_code = SSH_CANT_SET_MAINTENANCE_RC
            raise

    def unset_maintenance_flag_on_destination(self):
        """
        Unset the maintenance flag on the new Magento site.
        :return: None:
        """
        with suppress(FileNotFoundError):
            try:
                os.remove(self.destination_maintenance_path)
            except IOError:
                self.importer.return_code = SSH_CANT_UNSET_MAINTENANCE_RC
                raise

    def execute(self):
        """
        Import a content management system using the shop information
        stored on the passed importer object
        :return: None
        """
        self.check_disk_usage()
        if is_hipex_node(importer=self.importer):
            backup_hipex_state(importer=self.importer)
            setup_hipex_transfer(importer=self.importer)
        self.prepare_database()
        with self.manage_maintenance():
            self.sync_data()


class ImportMagento(ImportCMS):
    @property
    def magerun(self):
        raise NotImplementedError()

    def save_last_known_base_url(self):
        """
        Store the last known base url from the remote shop on disk
        so we can later use this to find the shop to benchmark against
        in hypernode-control even if the customer supplied an SSH host
        that is not also the domain of the shop that is being imported.
        :return None:
        """
        log.info(
            "Attempting to save the last known base url in /tmp/last_known_base_url"
        )

        save_base_url = (
            "cd /data/web/public && {} sys:url:list | head -n 1 "
            "> /tmp/last_known_base_url".format(self.magerun)
        )
        return_code = run_local_command(save_base_url, shell=True)
        if return_code != 0:
            log.info("Failed to save last known base url, that's OK!")

    def set_base_url(self, base_url):
        """
        Set the store base url
        :param str base_url: The base url to set
        :return None:
        """
        log.debug("Setting the web/secure/base_url and web/unsecure/base_url")
        core_config_data_table = (
            self.importer.remote_db_info["table_prefix"] + "core_config_data"
        )
        update_base_url_query = (
            "UPDATE `{}` SET value = %s WHERE path "
            "IN ('web/secure/base_url', '"
            "web/unsecure/base_url')"
            "".format(core_config_data_table)
        )
        update_base_url_params = (base_url,)
        self.importer.local_db_conn.driver.exec_query(
            update_base_url_query, update_base_url_params
        )

    def set_default_url(self):
        """
        Set the store default url. Both for the base url and the admin url.
        :return None:
        """
        log.info("Setting the default URL")
        hostname = get_hypernode_name()

        self.base_url = "http://{}/".format(hostname)
        self.admin_url = "{}{}/".format(
            self.base_url, self.importer.remote_db_info["admin_path"]
        )
        self.set_base_url(self.base_url)


class ImportShopware(ImportCMS):
    def verify_remote_db(
        self, expected_table=None, expected_row=None, expected_value=None
    ):
        """
        Verify the existence of all parts of the remote database, and maintenance entry.
        return: True if all parts return True.
        return: False one or part returns False.
        """
        return (
            self.verify_remote_database_exists()
            and self.verify_remote_shops_table(expected_table=expected_table)
            and self.verify_remote_maintenance_row(
                expected_table=expected_table,
                expected_row=expected_row,
                expected_value=expected_value,
            )
        )

    def verify_remote_database_exists(self):
        """
        Run a query to check if the remote database has a database entry as expected.
        return: True if the result returns a filled dictionary with results.
        return: False if return empty tuple
        """
        query_check_db_exists = "SHOW DATABASES LIKE '{}';".format(
            self.importer.remote_db_info["db_name"]
        )
        if (
            len(self.importer.remote_db_conn.driver.exec_query(query_check_db_exists))
            > 0
        ):
            log.debug("Remote database exists")
            return True
        return False

    def verify_remote_shops_table(self, expected_table):
        """
        Run a query to check if the remote database has a table entry as expected.
        return: True if the result returns a filled dictionary with results.
        return: False if return empty tuple
        """
        self.importer.remote_db_conn.driver.exec_query(
            "use {};".format(self.importer.remote_db_info["db_name"])
        )
        query_check_table_exists = "SHOW TABLES LIKE '{}'".format(expected_table)
        if (
            len(
                self.importer.remote_db_conn.driver.exec_query(query_check_table_exists)
            )
            > 0
        ):
            log.debug("Remote table {} exists".format(expected_table))
            return True
        return False

    def verify_remote_maintenance_row(
        self, expected_table, expected_row, expected_value
    ):
        """
        Run a query to check if the remote database has a maintenance row .
        return: True if the result returns a filled dictionary with results.
        return: False if return empty tuple
        """
        query_check_row_exists = "SELECT * FROM {} WHERE {} = {} LIMIT 1;".format(
            expected_table, expected_row, expected_value
        )
        if (
            len(self.importer.remote_db_conn.driver.exec_query(query_check_row_exists))
            > 0
        ):
            log.debug("Remote row {} exists".format(expected_row))
            return True
        return False

    def set_maintenance_flag(self):
        if self.verify_set_maintenance():
            request_confirmation(
                message="Part of setting maintenance is to clear shop caches, do you want to proceed?"
            )
            try:
                self.importer.remote_db_conn.driver.exec_query(
                    "use {};".format(self.importer.remote_db_info["db_name"])
                )
                self.importer.remote_db_conn.driver.exec_query(
                    self.compose_query_set_maintenance
                )
                self.clear_caches_on_source()
            except Exception as e:
                self.importer.return_code = EXEC_QUERY_CANT_SET_REMOTE_MAINTENANCE
                log.exception(
                    "Caught unexpected exception trying to set maintenance on source, aborting."
                    " Exception was {}".format(e)
                )
                raise

    @abstractmethod
    def clear_caches_on_source(self):
        pass

    @abstractmethod
    def verify_set_maintenance(self):
        pass

    @property
    @abstractmethod
    def compose_query_set_maintenance(self):
        pass

    @property
    @abstractmethod
    def compose_query_unset_maintenance(self):
        pass

    def unset_maintenance_flag_on_destination(self):
        self.importer.local_db_conn.driver.exec_query(
            self.compose_query_unset_maintenance
        )

    @abstractmethod
    def update_configuration(self):
        pass

    @property
    @abstractmethod
    def download_path(self):
        pass

    def update_settings(self):
        self.update_configuration()
        php_output = check_php_major()
        if php_output:
            log.info(php_output)
        self.print_success()

    def import_filesystem(self):
        """
        Import the filesystem of the shop
        :return None:
        """
        log.info("Starting filesystem synchronization")
        self.importer.file_transfer_conn.driver.download(
            self.importer.remote_path, self.download_path
        )
        super(ImportShopware, self).import_filesystem()

    @staticmethod
    def has_mysql_57():
        """
        Check if the local environment has a mysql version above 5.6, and RuntimeError if it is not.
        """
        mysql_output = check_mysql_major()
        if mysql_output:
            raise RuntimeError("{}".format(mysql_output))
