import logging

from kamikaze3.importer.abstract import ImporterSubClass
from kamikaze3.importer.strategy.magento1 import ImportMagento1
from kamikaze3.importer.strategy.magento2 import ImportMagento2
from kamikaze3.importer.strategy.shopware5 import ImportShopware5
from kamikaze3.importer.strategy.shopware6 import ImportShopware6

log = logging.getLogger(__name__)


class ImportStrategy(ImporterSubClass):
    execute = None

    def determine_strategy(self):
        """
        Determine the import strategy to use for the shop that
        was encountered on the other side.
        :return:
        """
        log.info("Determining import strategy")
        if self.importer.site_type == "magento1":
            log.debug("Setting magento 1 import strategy")
            self.execute = ImportMagento1(importer=self.importer).execute
        elif self.importer.site_type == "magento2":
            log.debug("Setting magento 2 import strategy")
            self.execute = ImportMagento2(importer=self.importer).execute
        elif self.importer.site_type == "shopware5":
            log.debug("Setting shopware 5 import strategy")
            self.execute = ImportShopware5(importer=self.importer).execute
        elif self.importer.site_type == "shopware6":
            log.debug("Setting shopware 6 import strategy")
            self.execute = ImportShopware6(importer=self.importer).execute
        if not callable(self.execute):
            log.debug(
                "Failed to determine an import strategy "
                "for site_type: {}".format(self.importer.site_type)
            )
            raise NotImplementedError(
                "The hypernode-importer has no import strategy for "
                "site type: {}".format(self.importer.site_type)
            )

    def determine_nginx_strategy(self):
        """
        Determine how to import nginx configs when dealing with hypernode-manage-vhosts
        :rtype: str
        """
        if not self.importer.remote_hmv_active:
            self.importer.import_nginx_strategy = "standard"
        elif self.importer.local_hmv_active:
            self.importer.import_nginx_strategy = "hmv-copy-vhosts"
        else:
            self.importer.import_nginx_strategy = "hmv-on-remote"

    def __init__(self, importer):
        """
        Initialize the ImportStrategy.
        :param obj importer: The Importer object where we will put the
        keep the global state and which will call the execute method.
        :return None:
        """
        super(ImportStrategy, self).__init__(importer=importer)
        self.determine_strategy()
        self.determine_nginx_strategy()
