import logging
from contextlib import suppress
from os import path, remove
from xml.etree import ElementTree

from kamikaze3.common.shell import run_local_command
from kamikaze3.importer.nginx import MAGENTO_2_NGINX_FLAG
from kamikaze3.importer.strategy.cms import ImportMagento

log = logging.getLogger(__name__)

default_redis_cache_backend = """
    <cache>
        <backend>Cm_Cache_Backend_Redis</backend>
        <backend_options>
            <server>redismaster</server>
            <port>6379</port>
            <persistent></persistent>
            <database>{}</database>
            <password></password>
            <force_standalone>0</force_standalone>
            <connect_retries>1</connect_retries>
            <read_timeout>10</read_timeout>
            <automatic_cleaning_factor>0</automatic_cleaning_factor>
            <compress_data>1</compress_data>
            <compress_tags>1</compress_tags>
            <compress_threshold>20480</compress_threshold>
            <compression_lib>gzip</compression_lib>
        </backend_options>
    </cache>
"""


class ImportMagento1(ImportMagento):
    local_database_config = "/data/web/public/app/etc/local.xml"
    local_fpc_config = "/data/web/public/app/etc/fpc.xml"
    magerun = "magerun"

    @property
    def source_maintenance_path(self):
        return path.join(self.importer.remote_path, "maintenance.flag")

    @property
    def destination_maintenance_path(self):
        return "/data/web/public/maintenance.flag"

    def import_filesystem(self):
        """
        Import the filesystem of the shop
        :return None:
        """
        log.info("Starting filesystem synchronization")
        self.importer.file_transfer_conn.driver.download(
            self.importer.remote_path, "/data/web/public"
        )
        super(ImportMagento1, self).import_filesystem()

    def sync_data(self, extra_mysqldump_arguments=None, cms_type=None):
        """
        Sync the shop data. The filesystem and the database in parallel.
        :param cms_type: the cms type and version
        :param str extra_mysqldump_arguments: A string of extra arguments.
        Is only added on the real sync, not the initial structure only sync
        :return None:
        """
        super(ImportMagento1, self).sync_data(cms_type="magento1")

    def enable_custom_admin_url(self):
        """
        Enable custom admin url
        :return None:
        """
        log.debug("Setting custom admin url to enabled")
        core_config_data_table = (
            self.importer.remote_db_info["table_prefix"] + "core_config_data"
        )
        update_base_url_query = (
            "UPDATE `{}` SET value = 0 WHERE path = 'admin/url/use_custom'"
            "".format(core_config_data_table)
        )
        self.importer.local_db_conn.driver.exec_query(update_base_url_query)

    def set_default_url(self):
        """
        Set the store default url. Both for the base url and the admin url.
        :return None:
        """
        super(ImportMagento1, self).set_default_url()
        self.enable_custom_admin_url()

    def purge_redis_cache(self):
        """
        Use magerun to flush the magento cache
        :return None:
        """
        log.info("Flushing the Magento 1 cache")
        magento_flush_command = [
            "n98-magerun",
            "cache:flush",
            "--root-dir=/data/web/public",
        ]
        run_local_command(magento_flush_command)
        super(ImportMagento1, self).purge_redis_cache()

    @staticmethod
    def fix_cache_config(cache_path, tag="cache", database=0):
        """
        Fix the Redis cache config
        :return None:
        """
        log.info("Fixing {} configuration in {}".format(tag, cache_path))
        tree = ElementTree.parse(cache_path)

        globalconfig = tree.find("./global")

        for cache in globalconfig.findall(tag):
            globalconfig.remove(cache)

        redis_backend = ElementTree.fromstring(
            default_redis_cache_backend.format(database)
        )
        redis_backend.tag = tag  # set the cache element to the requested tag type

        globalconfig.append(redis_backend)

        tree.write(cache_path, encoding="utf-8", xml_declaration=True)

    def update_settings(self):
        """
        Update the imported shop settings to fit the local environment
        :return None:
        """
        self.fix_cache_config(self.local_database_config)
        if path.exists(self.local_fpc_config):
            self.fix_cache_config(self.local_fpc_config, tag="fpc", database=1)

        self.fix_up_local_database_config()
        self.save_last_known_base_url()
        if self.importer.set_default_url:
            self.set_default_url()
        self.purge_redis_cache()

    @staticmethod
    def remove_magento_2_nginx_flag():
        """
        Remove any existing Magento 2 nginx flag
        See; https://support.hypernode.com/knowledgebase/installing-magento-2-on-hypernode/#Activate_Magento_2_mode
        :return None:
        """
        with suppress(FileNotFoundError):
            remove(MAGENTO_2_NGINX_FLAG)

    def execute(self):
        """
        Import a Magento 1 shop using the shop information
        stored on the passed importer object
        :return: None
        """
        log.info("Executing Magento 1 import strategy")
        super(ImportMagento1, self).execute()
        self.update_settings()
        self.remove_magento_2_nginx_flag()
        self.print_success()
