import logging
from urllib.parse import urlparse

from kamikaze3.common.env import get_hypernode_name
from kamikaze3.common.shell import run_local_command
from kamikaze3.importer.error_codes import (
    EXEC_COMMAND_CANT_CLEAR_CACHE_REMOTE_MAINTENANCE,
    REMOTE_APP_NAME_NOT_FOUND_SW6_RC,
)
from kamikaze3.importer.strategy.cms import ImportShopware

log = logging.getLogger(__name__)


class ImportShopware6(ImportShopware):
    local_database_config = "/data/web/.my.cnf"

    @property
    def download_path(self):
        return "/data/web/shopware6/"

    def sync_data(self, extra_mysqldump_arguments=None, cms_type="shopware6"):
        """
        Sync the shop data. The filesystem and the database in parallel.
        :param cms_type: cms type and version for skip table arguments
        :param str extra_mysqldump_arguments: A string of extra arguments.
        Is only added on the real sync, not the initial structure only sync
        :return None:
        """
        super(ImportShopware6, self).sync_data(cms_type="shopware6")
        log.debug("Symlink /data/web/shopware6/public to /data/web/public")
        run_local_command("ln -sf /data/web/shopware6/public /data/web/", shell=True)

    def get_remote_app_name(self):
        """
        Find remote app name from .env APP_URL line
        :return: remote_app_name: The remote app name
        """
        error_message = "Unable to acquire remote app name from .env (APP_URL)"
        try:
            if self.importer.app_url:
                remote_app_name = self.importer.app_url.split("//")
                remote_app_name = remote_app_name[1].rstrip()
            else:
                self.importer.return_code = REMOTE_APP_NAME_NOT_FOUND_SW6_RC
                raise RuntimeError(error_message)
        except IndexError:
            self.importer.return_code = REMOTE_APP_NAME_NOT_FOUND_SW6_RC
            raise RuntimeError(error_message)
        return remote_app_name

    def set_base_url(self):
        """
        Set the store default url. for the base url Shopware 6 type database.
        :return None:
        """
        log.info("Setting the default URL for Shopware 6")

        sales_channel_table = (
            self.importer.remote_db_info["table_prefix"] + "sales_channel_domain"
        )
        remote_url_query = "select url from {} where url" ' like "%{}%"'.format(
            sales_channel_table, self.get_remote_app_name()
        )
        remote_url_result = self.importer.local_db_conn.driver.exec_query(
            remote_url_query
        )

        for entry in remote_url_result:
            url = entry["url"]
            parsed_url = urlparse(url)
            updated_url = (
                parsed_url.scheme + "://" + get_hypernode_name() + parsed_url.path
            )
            update_url_query = (
                'UPDATE sales_channel_domain SET url = "{}" where url like "{}"'.format(
                    updated_url, url
                )
            )
            self.importer.local_db_conn.driver.exec_query(update_url_query)

    def set_default_url(self):
        """
        Set the store default url. base_url
        :return None:
        """
        self.set_base_url()

    def remove_caches(self):
        """
        Removes Shopware caches in /var/cache and /web/cache
        :return: None:
        """
        self.set_environment()
        run_local_command(["php", "/data/web/shopware6/bin/console", "cache:clear"])

    def read_lines(self, file_name):
        try:
            with open(file_name, "r") as file:
                lines = file.readlines()
                return lines
        except FileNotFoundError:
            log.debug("Local database configuration file was not found")

    def parse_env(self):
        """
        Gather the local .env credentials
        :return:
        """
        env_info = {}
        try:
            port = int(env_info.get("port", 3306))
        except ValueError:
            port = 3306

        local_env_info = {
            "user": "",
            "password": "",
            "host": "",
            "db_name": self.importer.remote_db_info["db_name"],
            "port": port,
        }

        local_config_file = self.importer.local_db_cnf_file
        cnf_lines = self.read_lines(local_config_file)
        for line in cnf_lines:
            if line.startswith("user"):
                line_parts = line.split("= ")
                unstripped_user = line_parts[1]
                user_line = unstripped_user.rstrip()
                local_env_info["user"] = user_line
            elif line.startswith("password"):
                line_parts = line.split('= "')
                unstripped_user = line_parts[1]
                quoted_password_line = unstripped_user.rstrip()
                password_line = quoted_password_line.rstrip('"')
                local_env_info["password"] = password_line
            elif line.startswith("host"):
                line_parts = line.split("= ")
                unstripped_host = line_parts[1]
                host_line = unstripped_host.rstrip()
                local_env_info["host"] = host_line
        return local_env_info

    def compose_env(self, database_url, app_url):
        """
        compose all the local database credentials that where gathered from .env
        :param database_url: The replacement database_url for the new .env
        :param app_url: The replacement app_url for the new .env
        :return:
        """
        env_lines = []
        lines = self.read_lines(self.local_env_config)
        for line in lines:
            if line.startswith("DATABASE_URL"):
                env_lines.append(database_url)
            elif line.startswith("APP_URL"):
                env_lines.append(app_url)
            else:
                env_lines.append(line)
        return env_lines

    def set_environment(self):
        """
        Update the .env configuration to suite the new environment in Shopware 6
        Completely rewrites .env by first reading the contents then writing everything back
        with the correct configurations for the new host.
        :return: None:
        """
        log.info("Updating database configuration")
        local_env_info = self.parse_env()
        database_url = "DATABASE_URL=mysql://{}:{}@{}:{}/{}\n".format(
            local_env_info["user"],
            local_env_info["password"],
            local_env_info["host"],
            local_env_info["port"],
            local_env_info["db_name"],
        )
        app_url = "APP_URL=http://{}\n".format(get_hypernode_name())

        env_lines = self.compose_env(database_url=database_url, app_url=app_url)
        with open(self.local_env_config, "w") as env:
            env.writelines(env_lines)

    def update_configuration(self):
        """
        Update the imported shop settings to fit the local environment
        :return None:
        """
        log.debug("Updating Shopware 6 host")
        self.fix_up_local_database_config()
        self.purge_redis_cache()
        self.local_env_config = "/data/web/shopware6/.env"
        self.remove_caches()
        if self.importer.set_default_url:
            self.set_base_url()

    def verify_set_maintenance(self):
        """
        Verify Shopware 6 type maintenance entry in remote database before database edit
        @return bool: True if all the rows and entries exist, False if not.
        """
        return super(ImportShopware6, self).verify_remote_db(
            expected_table="sales_channel", expected_row="maintenance", expected_value=0
        )

    @property
    def compose_query_set_maintenance(self):
        """
        Compose a query to set maintenance on Shopware 6 type shops.
        :string update_query_set_maintenance: set maintenance query for Shopware 6
        """
        values_table = self.importer.remote_db_info["table_prefix"] + "sales_channel"
        update_query_set_maintenance = (
            "UPDATE {} SET maintenance = 1 WHERE maintenance = 0"
            "".format(values_table)
        )
        return update_query_set_maintenance

    @property
    def compose_query_unset_maintenance(self):
        """
        Compose an update query to unset maintenance / service mode on Shopware 6 type shops.
        :string update_query_unset_maintenance: unset maintenance query for Shopware 6
        """
        values_table = self.importer.remote_db_info["table_prefix"] + "sales_channel"
        update_query_unset_maintenance = (
            "UPDATE {} SET maintenance = 0 WHERE maintenance = 1"
            "".format(values_table)
        )
        return update_query_unset_maintenance

    def clear_caches_on_source(self):
        """
        Runs a bin/console clear:cache on the remote server to make sure maintenance mode is activated.
        """
        try:
            self.importer.ssh_conn.driver.exec_command(
                "cd {} && bin/console c:c".format(self.importer.remote_path)
            )
        except Exception as e:
            self.importer.return_code = EXEC_COMMAND_CANT_CLEAR_CACHE_REMOTE_MAINTENANCE
            log.exception(
                "Caught unexpected exception trying to clear cache on source, aborting."
                " Exception was {}".format(e)
            )
            raise

    def execute(self):
        """
        Import a Shopware 6 shop using the shop information
        stored on the passed importer object
        :return: None
        """
        log.info("Executing Shopware 6 import strategy")
        super(ImportShopware6, self).has_mysql_57()
        super(ImportShopware6, self).execute()
        super(ImportShopware6, self).update_settings()
