import json
import logging
import socket
import time
from argparse import ArgumentParser

import requests

from kamikaze3.common.api import HYPERNODE_API_URL, get_hypernode_api_token
from kamikaze3.common.args import (
    OUTPUT_FORMAT_JSON,
    OUTPUT_FORMAT_TEXT,
    add_default_args,
)
from kamikaze3.common.env import get_app_name
from kamikaze3.log.log import configure_logging

log = logging.getLogger(__name__)


HYPERNODE_API_FLOW_LIST_URL = HYPERNODE_API_URL + "/logbook/v1/logbooks/{}/flows/"
HYPERNODE_API_FLOW_DETAIL_URL = (
    HYPERNODE_API_URL + "/logbook/v1/logbooks/{}/flows/?tracker_uuid={}"
)
# If people want to see this they can still retrieve it from the API themselves.
# But for most users this information is kinda spammy in the log and not as
# relevant as more important flows like update_app and xgrade_app.
FLOWS_TO_SKIP = [
    "check_sudoers_files_integrity",
    "check_users_groups",
    "check_crons",
]
BLOCK_JOB_TIMEOUT = 600
BLOCK_JOB_INTERVAL = 1
GET_FLOW_RETRY_ATTEMPTS = 5


def parse_arguments():
    """
    Parse the hypernode-log arguments
    :return obj args: The parsed arguments
    """
    parser = ArgumentParser(
        description="List pending and previous system " "actions for this Hypernode"
    )
    parser.add_argument(
        "--lines",
        default=50,
        required=False,
        type=int,
        help="Amount of tasks to display. Uses 50 lines by default",
    )
    add_default_args(parser)
    parsed_args = parser.parse_args()
    return parsed_args


def _try_get(url):
    response = requests.get(
        url,
        headers={
            "Authorization": "Token {}".format(get_hypernode_api_token()),
            "Content-Type": "application/json",
            "Accept": "application/json",
            "User-Agent": "{}:{}".format(socket.getfqdn(), "hypernode-kamikaze3"),
        },
    )
    try:
        response.raise_for_status()
        return response.json()
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 404:
            print("No recent activity")
            exit(0)
        else:
            raise


def _print_flow_status(flow: dict):
    job_id = flow["tracker"]["uuid"]
    job_name = flow.get("name")
    state = flow.get("state") or "waiting"
    if state == "waiting":
        # Job is not picked up yet, probably some other job is still running
        state_msg = f"Waiting for job {job_name} ({job_id}). State: {state}"
    else:
        current_task = (
            flow["progress"]["running"][0] if flow["progress"]["running"] else ""
        )
        nr_completed = flow["progress"]["completed"]
        nr_total = flow["progress"]["total"]
        state_msg = f"Waiting for job {job_name} ({job_id}). State: {state}. Progress: {nr_completed}/{nr_total}. Current task: {current_task}"
    log.info(state_msg)


def block_until_job_is_done(job_tracker_uuid: str):
    """
    Wait until the job is done.
    """
    time_elapsed = 0
    previous_task = "dummy text"

    while time_elapsed < BLOCK_JOB_TIMEOUT:
        flow = get_flow_details(job_tracker_uuid)
        if flow.get("state") in ("reverted", "success"):
            break
        current_task = (
            flow["progress"]["running"][0] if flow["progress"]["running"] else ""
        )
        # Only show status if it changed
        if current_task != previous_task:
            _print_flow_status(flow)
            previous_task = current_task
        time.sleep(BLOCK_JOB_INTERVAL)
        time_elapsed += BLOCK_JOB_INTERVAL
    log.info(f"Job {job_tracker_uuid} completed after {time_elapsed} seconds")


def block_until_last_flow_done(interval=BLOCK_JOB_INTERVAL, timeout=BLOCK_JOB_TIMEOUT):
    """
    Wait until there are no more flows running.
    """
    time_elapsed = 0

    print("Waiting until all flows finished running")
    while time_elapsed < timeout:
        flows = get_latest_flows(amount=1)
        # Check if the number of total tasks is the same as the number of completed tasks
        if all([f["progress"]["total"] == f["progress"]["completed"] for f in flows]):
            break

        time.sleep(interval)
        time_elapsed += interval
    print("All jobs are completed")


def get_latest_flows(amount=50):
    flows = list()
    resp_handle = {"next": HYPERNODE_API_FLOW_LIST_URL.format(get_app_name())}
    while len(flows) < amount and resp_handle.get("next"):
        next_url = resp_handle["next"]
        resp_handle = _try_get(next_url)
        flows.extend(resp_handle["results"])
    return flows[:amount]


def get_flow_details(job_tracker_uuid: str) -> dict:
    retry_attempts = GET_FLOW_RETRY_ATTEMPTS
    while retry_attempts > 0:
        job = _try_get(
            HYPERNODE_API_FLOW_DETAIL_URL.format(get_app_name(), job_tracker_uuid)
        )
        if not job or not job.get("results"):
            # Job takes some time to be picked up
            log.warning(f"Couldn't find job, retrying {retry_attempts} more times")
            retry_attempts -= 1
            time.sleep(BLOCK_JOB_INTERVAL)
            continue
        return job["results"][0]
    raise RuntimeError(f"Couldn't find job {job_tracker_uuid}")


def print_list_line(
    flow_name="ACTION",
    state="STATE",
    start="START",
    end="END",
    progress="TASKS",
    running="RUNNING",
    finished=False,
):

    print(
        "{:<30}\t{:<20}\t{:<20}\t{:<8}\t{:<4}\t{}".format(
            flow_name[:30],
            start or "",
            end or "",
            state or "WAITING",
            progress or "",
            running or ("finished" if finished else "waiting"),
        )
    )


def last_jobs(lines, output_format: str = OUTPUT_FORMAT_TEXT):
    """
    List the last flows run for this app. This is the same as ./bin/last_jobs in
    hn-control but then only for this specific app.
    :param int lines: The amount of flows to display
    :param str output_format: The output format to use
    :return None:
    """
    latest_flows = get_latest_flows(amount=lines)

    prepared_flows = []
    for f in latest_flows:
        failed = f["state"] in ("reverted",)
        all_tasks_complete = f["progress"]["completed"] == f["progress"]["total"]
        finished = f["progress"]["total"] > 0 and all_tasks_complete
        flow_name = f["name"]
        state = f["state"]
        if flow_name in FLOWS_TO_SKIP:
            continue
        if flow_name == "create_backup" and state == "reverted":
            state = "background"
        progress = "{}/{}".format(f["progress"]["completed"], f["progress"]["total"])
        prepared_flows.append(
            {
                "flow_name": flow_name,
                "state": state or "WAITING",
                "start": f["created_at"],
                "end": f["updated_at"],
                "progress": progress,
                "running": " ".join(f["progress"]["running"])
                or ("finished" if finished else "waiting"),
                "finished": finished or failed,
            }
        )

    if output_format == OUTPUT_FORMAT_JSON:
        print(json.dumps(prepared_flows))
    else:
        print_list_line()
        for f in prepared_flows:
            print_list_line(**f)


def main():
    args = parse_arguments()
    configure_logging("kamikaze3.log", args.verbose)
    last_jobs(lines=args.lines, output_format=args.output_format)
