import logging
import subprocess

from kamikaze3.manage_floating_ip.metadata import (
    ANCHOR_GATEWAY,
    FLOATING_IP_ADDRESS,
    GATEWAY_L3_MAC,
    INSTANCE_ADDRESS,
    INSTANCE_GATEWAY,
    attempt_get_metadata,
)

logger = logging.getLogger(__name__)


def determine_stderr_output(silence_output):
    """
    Return the correct destination to silence the output of check_call based on the given boolean.
    :param bool silence_output: Whether to silence the output of the error logs from check_call or not.
    :rtype: subprocess.DEVNULL|None
    """
    return subprocess.DEVNULL if silence_output else None


def manage_arp_entry(enable, ip=None, mac=GATEWAY_L3_MAC, silence_output=False):
    """
    Adds or removes a static ARP entry.
    To ensure we don't get clashes we flush the ARP entry in the same command.
    :param bool enable: Whether to add or remove
    :param str ip: The IP to add a static entry for
    :param str mac: The MAC address to point to
    :param bool silence_output: Whether to silence the output of the error logs from check_call or not.
    """
    ip = ip if ip else attempt_get_metadata(ANCHOR_GATEWAY)
    logger.info(
        "{} static ARP entry for {} to {}".format(
            "Adding" if enable else "Removing", ip, mac
        )
    )
    flush_command = "ip neigh flush {}".format(ip)
    # Replace the entry instead of adding, becasue 'replace' will overwrite an existing or FAILED entry, or add if it does not exist.
    neigh_command = "ip neigh {} {} lladdr {} dev eth0".format(
        "replace" if enable else "del", ip, mac
    )
    subprocess.check_call(
        flush_command + " ; " + neigh_command,
        stderr=determine_stderr_output(silence_output),
        shell=True,
    )


def manage_ip_rule_for_mail_route(
    enable, table_name="mail-route", silence_output=False
):
    """
    Sets a rule that adds a mark to the given table
    :param bool enable: Whether to add or remove the mark from the given table
    :param str table_name: The name of the mail table
    :param bool silence_output: Whether to silence the output of the error logs from check_call or not.
    :rtype: None
    """
    logger.info(
        "Managing IP rule to table {} by {} mark".format(
            table_name, "adding" if enable else "removing"
        )
    )
    ip_rule_tables = subprocess.check_output(["/usr/sbin/ip", "rule", "ls"]).decode(
        "utf-8"
    )
    if not enable and table_name not in ip_rule_tables:
        logger.info(
            "Rules for table {} not set. Skipping manage IP rule for mail route..".format(
                table_name
            )
        )
        return
    command = [
        "/usr/sbin/ip",
        "rule",
        ("add" if enable else "del"),
        "fwmark",
        "0x1",
        "table",
        table_name,
    ]
    subprocess.check_call(command, stderr=determine_stderr_output(silence_output))


def manage_mail_route_via_instance_gateway(
    enable, table_name="mail-route", instance_gateway=None, silence_output=False
):
    """
    Manages the route for the mail table via the instance gateway
    :param bool enable: Whether to add or remove the route
    :param str table_name: The name of the mail table
    :param str instance_gateway: The instance gateway. Defaults to retrieving the gateway via the API or disk
    :param bool silence_output: Whether to silence the output of the error logs from check_call or not.
    :rtype: None
    """
    refreshed_instance_gateway = instance_gateway or attempt_get_metadata(
        INSTANCE_GATEWAY
    )
    route = [
        "default",
        "via",
        refreshed_instance_gateway,
        "dev",
        "eth0",
        "table",
        table_name,
    ]

    logger.info(
        "{} route for mail route via instance gateway {}".format(
            "Adding" if enable else "Removing", refreshed_instance_gateway
        )
    )
    ip_rule_tables = subprocess.check_output(["/usr/sbin/ip", "rule", "ls"]).decode(
        "utf-8"
    )
    if not enable and table_name not in ip_rule_tables:
        logger.info(
            "Rules for table {} not set. Skipping manage mail route via instance gateway..".format(
                table_name
            )
        )
        return
    command = ["ip", "route", ("add" if enable else "del")] + route
    subprocess.check_call(command, stderr=determine_stderr_output(silence_output))


def manage_mail_route_table(
    enable,
    mr_name="mail-route",
    mr_id="200",
    table_file_location="/etc/iproute2/rt_tables",
    silence_output=False,
):
    """
    Manages the table in /etc/iproute2/rt_tables
    :param bool enable: Whether to add or remove the line in the config file
    :param str mr_name: The table name
    :param int mr_id: The table ID
    :param str table_file_location: the table file
    This is NOT USED but it is to make the signature of this function match the other ones in the outer loop.
    :param bool silence_output: Whether to silence the output of the error logs from check_call or not.
    :rtype: None
    """
    logger.info(
        "Managing mail route table {} on id {} by writing to file {}".format(
            mr_name, mr_id, table_file_location
        )
    )
    table_line = "{}\t{}\n".format(mr_id, mr_name)
    table_line_exists = False
    try:
        with open(table_file_location, "r") as table_file:
            table_contents = [l.strip() + "\n" for l in table_file.readlines()]
        with open(table_file_location, "w") as table_file:
            # Check if the line already exists
            if [l for l in table_contents if table_line in l]:
                logger.info(
                    "Table {} already has ID {} assigned".format(mr_name, mr_id)
                )
                table_line_exists = True
            # If the line does not already exist, add it
            if enable and not table_line_exists:
                logger.info("Writing ID {} to table {}".format(mr_id, mr_name))
                table_contents.append(table_line)
            # If the line already exists, do nothing
            elif enable and table_line_exists:
                logger.info(
                    "Trying to add table {} with ID {} but entry exists, skipping...".format(
                        mr_name, mr_id
                    )
                )
            # If the line should not exist and it does, remove it
            elif not enable and table_line_exists:
                logger.info("Removing table {} with ID {}".format(mr_name, mr_id))
                table_contents = [l for l in table_contents if l != table_line]
            # If the line should not exist and it does not exist, do nothing
            else:
                logger.info(
                    "Trying to remove table {} with ID {} but entry doesnt exist, skipping...".format(
                        mr_name, mr_id
                    )
                )
            # Overwrite the table file with the new content. Also writes if nothing changed.
            logger.info("Writing new content to file {}".format(table_file_location))
            table_file.writelines(table_contents)
    except (OSError, IOError):
        logger.error("Could not open {}!".format(table_file_location))


def manage_mail_mangle_output_mark(
    enable, interface="eth0", ports=None, silence_output=False
):
    """
    Manages iptables config that mangles on the given interface and mark for a list of ports
    :param bool enable: Whether to add or remove the iptables config
    :param str interface: The interface that should be configured
    :param list ports: List of ports that should be configured
    :param bool silence_output: Whether to silence the output of the error logs from check_call or not.
    :rtype: None
    """
    ports = ports or [25, 587]
    logger.info("Managing mail mangle output marking")
    command = [
        "iptables",
        ("-A" if enable else "-D"),
        "OUTPUT",
        "-t",
        "mangle",
        "-o",
        interface,
        "-p",
        "tcp",
        "-j",
        "MARK",
        "--set-mark",
        "1",
        "--dport",
    ]
    for port in ports:
        logger.info(
            "{} mangle for port {}".format("Setting" if enable else "Removing", port)
        )
        subprocess.check_call(
            command + [str(port)], stderr=determine_stderr_output(silence_output)
        )


def manage_mail_snat_to_instance_ip(
    enable, interface="eth0", ports=None, instance_ip=None, silence_output=False
):
    """
    Manages iptables config for SNATing the traffic on the given interface and ports to the instance IP
    :param bool enable: Whether to add or remove the iptables config
    :param str interface: The interface that should be configured
    :param list ports: List of ports that the configuration is for
    :param str instance_ip: The instance IP that traffic should be targeted to.
    :param bool silence_output: Whether to silence the output of the error logs from check_call or not.
    :rtype: None
    """
    refreshed_instance_ip = instance_ip or attempt_get_metadata(INSTANCE_ADDRESS)
    ports = ports or [25, 587]
    logger.info(
        "{} mail SNAT to public IP {}".format(
            "Adding" if enable else "Removing", refreshed_instance_ip
        )
    )
    command = [
        "iptables",
        ("-A" if enable else "-D"),
        "POSTROUTING",
        "-t",
        "nat",
        "-o",
        interface,
        "-p",
        "tcp",
        "-j",
        "SNAT",
        "--to",
        refreshed_instance_ip,
        "--dport",
    ]
    for port in ports:
        logger.info(
            "Setting SNAT to public IP {} for port {}".format(
                refreshed_instance_ip, port
            )
        )
        subprocess.check_call(
            command + [str(port)], stderr=determine_stderr_output(silence_output)
        )


def manage_floating_ip_loopback(enable, ip_address=None, silence_output=False):
    """
    Manages the routing for adding loopback to floating IP
    :param bool enable: Whether to add or remove the loopback rule
    :param str ip_address: The IP address that the loopback is for
    :param bool silence_output: Whether to silence the output of the error logs from check_call or not.
    :rtype: None
    """
    refreshed_ip_address = ip_address or attempt_get_metadata(FLOATING_IP_ADDRESS)
    route = [refreshed_ip_address + "/32", "dev", "lo"]

    logger.info(
        "{} floating IP loopback on IP {}".format(
            "Adding" if enable else "Removing", refreshed_ip_address
        )
    )
    command = ["ip", "addr", ("add" if enable else "del")] + route
    subprocess.check_call(command, stderr=determine_stderr_output(silence_output))


def manage_route_to_digitalocean_api(
    enable, instance_gateway=None, silence_output=False
):
    """
    Manages the routing to the DigitalOcean API
    :param bool enable: Whether to add or remove the route
    :param str instance_gateway: The gateway that should target the DigitalOcean API
    :param bool silence_output: Whether to silence the output of the error logs from check_call or not.
    :rtype: None
    """
    refreshed_instance_gateway = instance_gateway or attempt_get_metadata(
        INSTANCE_GATEWAY
    )
    route = ["169.254.0.0/16", "via", refreshed_instance_gateway]

    logger.info(
        "Managing route to DigitalOcean API by {} it".format(
            "adding" if enable else "removing"
        )
    )
    command = ["ip", "route", ("add" if enable else "del")] + route
    subprocess.check_call(command, stderr=determine_stderr_output(silence_output))


def ensure_idempotency(manage_function, enable, **kwargs):
    """
    Ensure that the given function can be run idempotently. This means that if the function is being
    called to enable a feature, then this function will try to disable it first to ensure the manage_function
    can do it's job correctly.
    :param func manage_function: The function that will be run later. This function MUST accept arg 'enable'.
    :param bool enable: Whether the manage_function will be enabled later on.
    :param kwargs: Any kwargs that have to be passed to the manage_function.
    :rtype: None
    """
    try:
        logger.info(
            "Trying to {} function {} to ensure idempotency".format(
                "undo" if enable else "run", manage_function.__name__
            )
        )
        manage_function(enable=not enable, silence_output=True, **kwargs)
    except subprocess.CalledProcessError:
        logger.info("Environment already set correctly!")
    logger.info(
        "Idempotency is ensured for function {}. Running it now..".format(
            manage_function.__name__
        )
    )
    manage_function(enable=enable, **kwargs)
