"""
kamikaze3.metrics.long_processes
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Provide metrics about long-running processes
"""

import sys
from argparse import ArgumentParser
from logging import DEBUG, INFO, getLogger
from os import EX_OK, EX_SOFTWARE, EX_TEMPFAIL

from kamikaze3.common.csv import print_dicts_as_csv
from kamikaze3.common.log import create_syslog_handler
from kamikaze3.metrics.imageoptimizer_cron import gather_imageoptimizer_metrics
from kamikaze3.metrics.long_processes import gather_long_processes_metrics
from kamikaze3.metrics.node_info import gather_node_info_metrics
from kamikaze3.metrics.web_requests import gather_web_requests_metrics

program_name = "hypernode-gather-metrics"
logger = getLogger(__name__)


def parse_args(args):
    parser = ArgumentParser(
        prog=program_name,
        description="Provide metrics as CSV for consumption by telegraf",
    )
    parser.add_argument(
        "--verbose", action="store_true", help="print verbose messages to stderr"
    )
    parser.add_argument(
        "--prometheus", action="store_true", help="print metrics in Prometheus format"
    )
    return parser.parse_args(args)


def main(args=None):
    """Program entry point, create a report of longest running processes:
    prints cvs of the duration of the longest-running MySQL query, shell PHP process, and fpm process, in seconds.

    :return: the exit code
    :rtype: int
    """

    opts = parse_args(args)

    logger.setLevel(DEBUG if opts.verbose else INFO)
    logger.addHandler(
        create_syslog_handler(
            name_override=program_name if __name__ == "__main__" else None
        )
    )

    try:
        web_metrics = gather_web_requests_metrics()
        long_processes_metrics = gather_long_processes_metrics()
        imageoptimizer_metrics = gather_imageoptimizer_metrics()
        node_info_metrics = gather_node_info_metrics()
        if opts.prometheus:
            for key in web_metrics:
                if key == "requests":
                    # requests is a special case, it's the total number of requests.
                    # We don't want to expose this as a metric, as it's not very useful.
                    # This metric can be calculated by summing all the other metrics.
                    continue
                prometheus_key = key.replace("requests_", "").replace("_requests", "")
                print('web_request{type="%s"} %s' % (prometheus_key, web_metrics[key]))
            for key in long_processes_metrics:
                print('long_process{type="%s"} %s' % (key, long_processes_metrics[key]))
            for key in node_info_metrics:
                print("%s %s" % (key, node_info_metrics[key]))
            print(
                "hypernode_image_optimizer {}".format(
                    imageoptimizer_metrics["hypernode-image-optimizer"]
                )
            )
        else:
            result = {
                **web_metrics,
                **long_processes_metrics,
                **imageoptimizer_metrics,
            }
            print_dicts_as_csv([result])
        return EX_OK
    except KeyboardInterrupt:
        logger.warning("process interrupted by user")
        return EX_TEMPFAIL
    except Exception as err:
        logger.exception(err)
        return EX_SOFTWARE


if __name__ == "__main__":
    sys.exit(main())
