import argparse
import re
import subprocess
import sys
from os import getuid
from typing import Iterator, List, Optional, Tuple

EXPORTER_PATH = "/var/lib/prometheus/node-exporter/redis.prom"
ALLOWED_METRICS = (
    "used_memory",
    "maxmemory",
    "total_",
    "instantaneous_",
    "sync_",
    "expired_",
    "evicted_",
    "keyspace_",
    "active_defrag_",
    "connected_slaves_",
    "master_repl_",
    "repl_backlog_",
    "connected_clients_",
    "maxclients",
    "blocked_clients",
    "tracking_clients",
    "clients_in_timeout_table",
    "db[0-9]+",
)
ALLOWED_METRICS_PATTERN = re.compile("^({})".format("|".join(ALLOWED_METRICS)))
BLOCKED_METRICS_PATTERN = re.compile("(_human|_perc|_policy)$")
LSOF_TIMEOUT = 5


class RedisDiscoveryError(Exception):
    pass


def get_redis_instances() -> List[int]:
    output = subprocess.check_output(
        ["/usr/bin/pgrep", "-f", "redis-server"], encoding="utf-8"
    )
    output = output.strip()
    if not output:
        return []
    return [int(pid) for pid in output.split("\n")]


def get_redis_command(pid: int) -> Tuple[List[str], str]:
    try:
        lsof_output = subprocess.check_output(
            ["/usr/bin/lsof", "-P", "-a", "-p", str(pid)],
            encoding="utf-8",
            timeout=LSOF_TIMEOUT,
        )
        lsof_output = lsof_output.strip()
    except subprocess.TimeoutExpired:
        raise RedisDiscoveryError(
            f"Timed out trying to find tcp/socket information for pid: {pid}"
        )

    if not lsof_output:
        raise RedisDiscoveryError(
            f"Failed to find tcp/socket information for pid: {pid}"
        )

    port: Optional[int] = None
    socket: Optional[str] = None
    for line in lsof_output.strip().split("\n")[1:]:
        _, _, _, _, fd_type, _, _, *values = line.split()
        if fd_type == "IPv4" and values[-1].endswith("(LISTEN)"):
            port = int(values[-2].split(":")[-1])
        elif fd_type == "unix":
            socket = values[1]

    if port:
        return ["/usr/bin/redis-cli", "-p", str(port), "info"], str(port)
    elif socket:
        if not socket.startswith("/"):
            raise RedisDiscoveryError("Found redis with non-absolute socket")
        return ["/usr/bin/redis-cli", "-s", socket, "info"], socket
    else:
        raise RedisDiscoveryError(f"Failed to find port/socket for pid: {pid}")


def parse_redis_info_line(line: str, target: str) -> Iterator[str]:
    if not line or line.startswith("#") or not ALLOWED_METRICS_PATTERN.match(line):
        return []

    stat, value = [c.strip() for c in line.split(":")]
    if re.match(r"^db[0-9]+$", stat):
        # keys=258,expires=0,avg_ttl=0
        for record in value.split(","):
            db_stat, db_value = record.split("=")
            yield 'redis_db_{}{{target="{}",db="{}"}} {}'.format(
                db_stat, target, stat, db_value
            )
    elif not BLOCKED_METRICS_PATTERN.search(stat):
        yield 'redis_{}{{target="{}"}} {}'.format(stat, target, value)

    return []


def get_redis_info(pid) -> List[str]:
    result = []
    try:
        command, target = get_redis_command(pid)
        output = subprocess.check_output(command, encoding="utf-8")
        for line in output.strip().split("\n"):
            for parsed in parse_redis_info_line(line, target):
                result.append(parsed)
    except RedisDiscoveryError as e:
        print(f"ERROR: {e}", file=sys.stderr)

    return result


def parse_args(args: Optional[List[str]] = None) -> bool:
    parser = argparse.ArgumentParser(
        description="Collect and write redis information for prometheus node exporter"
    )
    parser.add_argument(
        "--verbose",
        "-v",
        action="store_true",
        default=False,
        help="Enable verbose logging",
    )

    parsed_args = parser.parse_args(args)
    return parsed_args.verbose


def main():
    verbose = parse_args()

    if getuid() != 0:
        print("ERROR: This command must be run as root.", file=sys.stderr)
        sys.exit(1)

    pids = get_redis_instances()
    result = []

    for pid in pids:
        result.extend(get_redis_info(pid))

    with open(EXPORTER_PATH, "w", encoding="utf-8") as f:
        for metric in result:
            if verbose:
                print(metric)
            f.write(metric + "\n")


if __name__ == "__main__":
    main()
