"""
kamikaze3.metrics.top_resource_procs
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Provide metrics about processes consuming most of the system resources.
"""

import sys
from argparse import ArgumentParser
from logging import DEBUG, INFO, getLogger
from operator import methodcaller
from os import EX_OK, EX_SOFTWARE, EX_TEMPFAIL

from psutil import process_iter

from kamikaze3.common.csv import print_dicts_as_csv
from kamikaze3.common.log import create_stream_handler, create_syslog_handler

program_name = "hypernode-top-resource-procs"
RESOURCE_TYPES = ("cpu", "memory")
logger = getLogger(__name__)


def top_resource_consuming_procs(type_, count=5):
    """Returns processes that are consuming most of the specified resource.

    :param str type_: the resource type (see RESOURCE_TYPES)
    :param int count: the number of returned processes
    :return list[psutil.Process]: list of processes, sorted decendingly by resource consumption
    """
    if type_ not in RESOURCE_TYPES:
        raise ValueError(
            'Invalid resource type "{}". Should be any of: {}'.format(
                type_, ", ".join(RESOURCE_TYPES)
            )
        )
    logger.debug("detecting top {} consuming processes ...".format(type_))
    return sorted(
        list(process_iter()), key=methodcaller("{}_percent".format(type_)), reverse=True
    )[:count]


def metrics_from_procs(procs, **kwargs):
    """Returns metrics from the list of processes as a list of dictionaries.
    An 'order' key is provided, whose meaning depends on if the processes
    are sorted based on any attribute.
    Extra meta data can be attached to each metric dict as kwargs.

    :param list[psutil.Process]: list of processes
    :return list[dict]: list of process information
    """
    results = []
    for idx, proc in enumerate(procs):
        metric = kwargs.copy()
        metric.update(
            dict(
                proc_name=proc.name(),
                pid=proc.pid,
                user=proc.username(),
                status=proc.status(),
                cpu_pct=round(proc.cpu_percent(), 2),
                mem_pct=round(proc.memory_percent(), 2),
                order=(idx + 1),  # enumerate indices start at 0
            )
        )
        results.append(metric)
    return results


def parse_args(args=None):
    """Parse command line arguments and return parsed options"""
    parser = ArgumentParser(
        prog=program_name,
        description="Display top processes consuming system resources. Logs actions to syslog",
    )
    parser.add_argument(
        "--number",
        "-n",
        type=int,
        default=5,
        help="number of processes for each resource",
    )
    parser.add_argument(
        "--debug", action="store_true", help="log debug information (to stderr)"
    )
    return parser.parse_args(args)


def main(args=None):
    """Program entry point, create a report of top resource consuming processes

    :param list args: list of arguements
    :return: the exit code
    :rtype: int
    """
    opts = parse_args(args)

    logger.setLevel(DEBUG if opts.debug else INFO)
    logger.addHandler(
        create_syslog_handler(
            name_override=program_name if __name__ == "__main__" else None
        )
    )
    if opts.debug:
        # log to stderr so stdout is only for generated reports
        logger.addHandler(create_stream_handler(sys.stderr, level=DEBUG))

    try:
        proc_metrics = []
        for type_ in RESOURCE_TYPES:
            procs = top_resource_consuming_procs(type_, opts.number)
            logger.debug(
                "top {} processes consuming {}: {}".format(
                    len(procs),
                    type_,
                    ", ".join(['"{}[{}]"'.format(p.name(), p.pid) for p in procs]),
                )
            )
            proc_metrics.extend(metrics_from_procs(procs, type="top_{}".format(type_)))

        print_dicts_as_csv(proc_metrics)
        return EX_OK
    except KeyboardInterrupt:
        logger.warning("process interrupted by user")
        return EX_TEMPFAIL
    except Exception as err:
        logger.exception(err)
        return EX_SOFTWARE


if __name__ == "__main__":
    sys.exit(main())
