"""
kamikaze3.metrics.web_requests
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Provide web request metrics
"""

import json
import os
import re
import subprocess
import sys
from argparse import ArgumentParser
from logging import DEBUG, INFO, getLogger
from os import EX_OK, EX_SOFTWARE, EX_TEMPFAIL

from kamikaze3.common.csv import print_dicts_as_csv
from kamikaze3.common.log import create_stream_handler, create_syslog_handler

program_name = "hypernode-web-requests"
log_filename = "/var/log/nginx/access.log"
rotated_log_filename = "/var/log/nginx/access.log.1"
cache_filename = "/var/cache/telegraf/web_requests.cache"
logger = getLogger(__name__)

statuses_to_report = (
    200,
    301,
    302,
    304,
    400,
    401,
    403,
    404,
    405,
    410,
    499,
    500,
    502,
    503,
    "3xx",
    "4xx",
    "5xx",
)
status_code_key = "requests_{}"


def parse_args(args):
    parser = ArgumentParser(
        prog=program_name, description="Provide web request metrics"
    )
    parser.add_argument(
        "--verbose", action="store_true", help="print verbose messages to stderr"
    )
    return parser.parse_args(args)


def analyze(lines):
    """
    :param iterator lines: takes an iterator containing parsed loglines
    :return: dict[str, float]: a dictionary containing statistics:

     - "requests": number of requests
     - "php_requests": any request handled by handlers "phpfpm"
     - "php_response_time": average PHP response time of all handled php_requests
    """

    # Cannot use a defaultdict here, because response.keys() would not give
    # the three keys when no php requests are there. That would result in missing
    # keys at JSON serialization.
    response = {
        "requests": 0,
        "php_requests": 0,
        "php_response_time": 0,
        "php_bot_requests": 0,
    }

    # Always report these, even if there aren't any in the logs.
    # This makes ServerDensity have them available for every timeslot.
    for status in statuses_to_report:
        response.setdefault(status_code_key.format(status), 0)

    # Iterate over all the lines
    for line in lines:
        if line["request"] == "GET /status.php HTTP/1.1":
            continue

        response["requests"] += 1

        # gather e.g. requests_503, status is all strings here
        if int(line["status"]) in statuses_to_report:
            response[status_code_key.format(line["status"])] += 1

        # Summarize all 3xx 4xx and 5xx for graphing in MageReport --WdG
        if int(line["status"][0]) in [3, 4, 5]:
            http_error_class = line["status"][0] + "xx"
            response[status_code_key.format(http_error_class)] += 1

        if line["handler"] in ["phpfpm"]:
            response["php_requests"] += 1
            response["php_response_time"] += float(line["request_time"])

            if re.search(
                "http|bot|crawl|spider|heartbeat", line["user_agent"], re.IGNORECASE
            ):
                response["php_bot_requests"] += 1

    # average out the PHP response time
    if response["php_requests"]:
        response["php_response_time"] /= response["php_requests"]

    return response


def get_new_lines(cache):
    """
    Gets all new loglines from the logfile.

    Returns a list of lines from the logfile. If the logfile
    has rotated since the last invokation of main(), the logfile will
    be read entirely and the data from the rotated logfile will be added to it.

    So in effect, the contents of both files is returned as a list.

    :param dict cache: a dict with the last read linenr and timestamp from the previous invocation
    :return: a list containing raw loglines
    """

    if cache:
        atcachednr = subprocess.check_output(
            ["/usr/bin/sed", "-n", str(cache["linenr"]) + "p", log_filename]
        )
        nextline = str(cache["linenr"] + 1)
        if (
            atcachednr
            and parse_line(atcachednr.decode("utf-8", errors="replace"))["time"]
            == cache["time"]
        ):
            # logfile matches cache, so it hasn't been rotated
            return (
                subprocess.check_output(
                    ["/usr/bin/tail", "-n+" + nextline, log_filename]
                )
                .decode("utf-8", errors="replace")
                .split("\n")
            )
        else:
            # logfile has been rotated: return the new lines of the rotated file plus the entire new file
            if os.path.isfile(rotated_log_filename):
                lines = (
                    subprocess.check_output(
                        ["/usr/bin/tail", "-n+" + nextline, rotated_log_filename]
                    )
                    .decode("utf-8", errors="replace")
                    .split("\n")
                )
            else:
                lines = []
            lines += [l for l in open(log_filename, encoding="utf-8", errors="replace")]
            return lines
    else:
        # first run or cache got lost, so return nothing
        return [""]


def parse_line(line):
    """
    Parse Hypernode Nginx loglines

    :param line: a raw logline in either JSON-like or less-JSON-like format
    :return: dict[str, str]: a dictionary containing all the logged keys, all values are strings

    notably (with examples)
     - "request": "GET / status.php HTTP/1.1"
     - "remote_addr": "1.2.3.4"
     - "remote_user": "-"
     - "user_agent": "Mozilla"
     - "request_time": "0.847" (in seconds, as a string)
     - "status_code": "200" (as a string)
    """

    def format_unicode_number(groupmatch):
        # cast match from hex, then format that hex as unicode code point
        hexcode = int("0x{}".format(groupmatch.group(1)), 0)
        return "\\u{:04x}".format(hexcode)

    proper_line = re.sub(r"\\x([A-Z0-9]+)", format_unicode_number, line)
    return json.loads(proper_line)


def parse_lines(lines):
    """
    Takes a list of strings and calls parse_line on each.
    Will skip any lines that we cannot parse if the parser throws a ValueError

    :param T <= Iterator lines: lines to parse
    :return: T <= Iterator: the dicts (parsed lines)
    """

    def safe_parse(line):
        try:
            return parse_line(line)
        except ValueError:
            return None

    return filter(bool, map(safe_parse, lines))


def get_cache():
    os.makedirs(os.path.dirname(cache_filename), exist_ok=True)
    try:
        with open(cache_filename, "r") as f:
            cache = json.load(f)
    except FileNotFoundError:
        # FileNotFoundError is an IOError, if not caught here, os.remove() would fail when file not found
        cache = None
    except (IOError, json.JSONDecodeError):
        # cache got corrupted, remove it
        os.remove(cache_filename)
        cache = None
    return cache


def set_cache():
    try:
        cache = {
            "linenr": int(
                subprocess.check_output(["/usr/bin/wc", "-l", log_filename]).split()[0]
            ),
            "time": parse_line(
                subprocess.check_output(["/usr/bin/tail", "-n1", log_filename]).decode(
                    "utf-8", errors="replace"
                )
            )["time"],
        }
    except (IOError, UnicodeDecodeError):
        # problem with log, remove old cache and bail out
        os.remove(cache_filename)
        return
    # If the logfile was just rotated it is empty, but reporting "0" breaks stuff
    if cache["linenr"] == 0:
        cache["linenr"] = 1
    with open(cache_filename, "w") as f:
        json.dump(cache, f)


def gather_web_requests_metrics():
    if not os.path.isfile(log_filename):
        logger.warning("logfile {} not found".format(log_filename))
        return EX_SOFTWARE
    result = analyze(parse_lines(get_new_lines(get_cache())))
    set_cache()
    return result


def main(args=None):
    """Program entry point, print analysis of newly appeared requests from nginx log

    Newly appeared means between two consecutive runs of run().
    The first time main() is ran, it will skip all content in the logfile.
    The second time and later, it will process all new content.

    If the logfile is rotated, the WebRequests will follow the new file.

    :return: the exit code
    :rtype: int
    """
    opts = parse_args(args)

    logger.setLevel(DEBUG if opts.verbose else INFO)
    logger.addHandler(create_syslog_handler(name_override=program_name))

    if opts.verbose:
        logger.addHandler(create_stream_handler(stream=sys.stderr, level=DEBUG))

    try:
        print_dicts_as_csv([gather_web_requests_metrics()])
        return EX_OK

    except KeyboardInterrupt:
        logger.warning("process interrupted by user")
        return EX_TEMPFAIL
    except Exception as err:
        logger.exception(err)
        return EX_SOFTWARE


if __name__ == "__main__":
    sys.exit(main())
