#!/usr/bin/env python
import datetime
import json
import sys
from argparse import ArgumentParser
from os import EX_OK, EX_UNAVAILABLE, path
from os.path import exists
from subprocess import check_output
from time import time

import MySQLdb

from kamikaze3.settings import (
    BACKUP_DATE_FORMAT,
    ROLE_MYSQL_MASTER,
    NODE_CREATED_AT_PATH,
)

from ..common.env import get_node_roles
from .utils import fetch_created_at

OFFSITE_DATABASE_BACKUP_STATEFILE = (
    "/data/var/statefiles/offsite_database_backup_status.json"
)
WARNING_AGE_LIMIT = 36  # hours
ONE_DAY_IN_SECONDS = 86400
BACKUP_MONITOR_FILE = "/data/var/mysql_backup_status.json"
BELOW_UPTIME = 48  # hours


def parse_arguments(args=None):
    """
    Parse the arguments for the hypernode-monitoring-offsite-backup-database command
    :return obj args: The parsed arguments
    """
    parser = ArgumentParser(
        description="Check if we have a valid offsite database backup for last night. "
        "Saves the state in {}".format(OFFSITE_DATABASE_BACKUP_STATEFILE)
    )

    parser.add_argument(
        "--write-json",
        help="Saves state to output file at {}".format(
            OFFSITE_DATABASE_BACKUP_STATEFILE
        ),
        action="store_true",
        default=False,
    )

    return parser.parse_args(args)


def check_offsite_backup_database():
    """
    Check if the latest backup is older than a certain amount
    of hours, if so then that indicates that something is wrong
    with the offsite database backups
    :return int: The exitcode
    """
    if check_if_backup_not_necessary():
        return EX_OK

    if check_if_node_is_less_old_than_one_day():
        return EX_OK

    if check_if_state_file_is_missing():
        return EX_UNAVAILABLE

    if check_most_recent_backup_is_too_old():
        return EX_UNAVAILABLE

    return EX_OK


def check_if_backup_not_necessary():
    return ROLE_MYSQL_MASTER not in get_node_roles()


def check_if_state_file_is_missing():
    """
    Check if state file exists.
    :return bool:
        True:  if state file is missing
        False: if state file exists
    """
    return not exists(BACKUP_MONITOR_FILE)


def check_most_recent_backup_is_too_old():
    """
    Check if most recent backup older than one day
    :return bool older_than_one_day: True if older than one day,
    False if not older than one day
    """
    backup_state = _get_backup_state()

    # Check for backups that are too old
    old_age_count = count_backups_older_than_one_day(backup_state)

    if old_age_count > 0:
        return True

    # Check for DBs without backups
    missed_backups = count_missing_backups(backup_state)

    return missed_backups > 0


def main(args=None):
    """
    Perform the offsite database backup check and write a file
    with the current status.
    :return None:
    """
    arguments = parse_arguments(args)
    state = check_offsite_backup_database()
    state_ok = state == EX_OK

    if arguments.write_json:
        write_state_file(state=state_ok)
    else:
        print("offsite_database_backup_healthy_state", int(state_ok))


def write_state_file(state):
    """
    Writes state file
    """
    message = ["offsite_database_backup_healthy_state=", str(state)]

    with open(OFFSITE_DATABASE_BACKUP_STATEFILE, "w") as f:
        f.write(json.dumps(message))


def count_backups_older_than_one_day(backup_state, now=None):
    """
    Counts the amount of backups that are older than given :hours:
    returns:
        integer - the amount of backups older than :hours:
    """
    if now is None:
        now = datetime.datetime.utcnow()

    count = 0
    for db in backup_state:
        try:
            dt = datetime.datetime.strptime(
                backup_state[db]["last_backup"], BACKUP_DATE_FORMAT
            )
            if (now - dt) > datetime.timedelta(hours=WARNING_AGE_LIMIT):
                count += 1

        except KeyError:
            pass

    return count


def count_missing_backups(backup_state, now=datetime.datetime.utcnow()):
    """
    Counts the amount of databases without backups (yet),
    where databases have been created before :max_hours_without_backups:
    returns:
        no_backup_counter    amount of missing backups older than one day
    """
    dbs = set(_fetch_list_of_databases())
    databases_in_backup_state = set(backup_state.keys())
    dbs_not_in_state_file = dbs - databases_in_backup_state

    no_backup_counter = 0
    for db in dbs_not_in_state_file:
        # create_time gives creation times of tables within DB
        try:
            query = (
                "SELECT create_time FROM INFORMATION_SCHEMA.TABLES "
                "WHERE table_schema = '%s' AND create_time IS NOT NULL "
                "ORDER BY create_time DESC LIMIT 1;"
            ) % db
            cmd = ["/usr/bin/mysql", "--defaults-file=/root/.my.cnf", "-B", "-e", query]
            output = check_output(cmd, encoding="utf-8")

            # Output is empty when database has no tables, so creation time cannot be determined.
            # Ignore this entry.
            if not output:
                continue

            creation_datetime_string = output.split("\n")[1]  # Strip off header
            creation_datetime = datetime.datetime.strptime(
                creation_datetime_string, "%Y-%m-%d %H:%M:%S"
            )

            if (now - creation_datetime) >= datetime.timedelta(hours=WARNING_AGE_LIMIT):
                no_backup_counter += 1

        except Exception as e:
            print(e)
            exit(EX_UNAVAILABLE)

    return no_backup_counter


def check_if_node_is_less_old_than_one_day():
    """
    Check if node is less old than one day
    :return bool older_than_one_day: True if not older than one day,
    False if older than one day
    """
    # Check first if /etc/hypernode/node_created_at actually exists before we check the age of the node.
    # It can occur that the node is currently being installed and the statfile isn't present yet.
    if not path.isfile(NODE_CREATED_AT_PATH):
        print(
            "OK: node_created_at cannot be determined, because the statfile is not present yet. Backups don't have to be present yet!",
        )
        return True

    current_timestamp = int(time())  # converted to int to truncate to seconds
    if (current_timestamp - fetch_created_at()) <= ONE_DAY_IN_SECONDS:
        print(
            "OK: Node has been created in the last 24 hours. Backups don't have to be present yet!",
            file=sys.stderr,
        )
        return True
    return False


def _fetch_list_of_databases():
    """
    Fetches a list of all databases on the server,
    excluding Database|information_schema|performance_schema
    returns:
        List of all databases as strings
    """
    excluded = ["information_schema", "performance_schema"]

    db = MySQLdb.connect(read_default_file="/root/.my.cnf")
    cur = db.cursor()
    cur.execute("show databases")
    databases = cur.fetchall()
    databases = [row[0] for row in databases if row[0] not in excluded]

    return databases


def _get_backup_state():
    with open(BACKUP_MONITOR_FILE, "r") as f:
        backup_state = json.loads(f.read())

    return backup_state
