#!/usr/bin/env python
import json
import re
import sys
from argparse import ArgumentParser
from datetime import datetime, timedelta
from operator import itemgetter
from os import EX_OK, EX_UNAVAILABLE, path
from subprocess import check_output
from time import time

from kamikaze3.settings import NODE_CREATED_AT_PATH

from .utils import fetch_created_at

OFFSITE_FILESYSTEM_BACKUP_STATEFILE = (
    "/data/var/statefiles/offsite_filesystem_backup_status.json"
)
WARNING_AGE_LIMIT = 25  # hours
ONE_DAY_IN_SECONDS = 86400
BACKUP_REGEX = (
    r"^(Incremental|Full)\s+((Mon|Tue|Wed|Thu|Fri|Sat|Sun)\s+(Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|"
    r"Dec)\s+\d+\s+\d+:\d+:\d+\s+\d+)"
)


def _check_that_most_recent_backup_is_older_than_one_day(parsed_backups):
    """
    :param list(dict): parsed_backups: all parsed backup types (full/incr) and dates in a list of dicts
    :return bool younger_than_one_day: True if not older than one day,
    False if older than one day
    """
    sorted_backups = sorted(parsed_backups, key=itemgetter("backup_date"), reverse=True)
    most_recent_backup = sorted_backups[:1]
    if not len(most_recent_backup):
        return True

    most_recent_backup = most_recent_backup[0]

    limit = datetime.now() - timedelta(hours=WARNING_AGE_LIMIT)
    return most_recent_backup["backup_date"] < limit


def _parse_all_backup_datetimes(duply_output):
    """
    :param str duply_output: the output of `dyply hyp status`
    :return list(dict): all parsed backup types (full/incr) and dates in a list of dicts
    """
    result_dates = []

    for line in duply_output.splitlines():
        m = re.match(BACKUP_REGEX, line.decode().strip())
        if m:
            result_dates.append(
                {
                    "backup_type": m.group(1),
                    "backup_date": datetime.strptime(
                        str(m.group(2)).strip(), "%a %b %d %H:%M:%S %Y"
                    ),
                }
            )

    return result_dates


def _filter_backups_by_age(parsed_backups, age=7):
    """
    Return all backups younger than :age: days.
    :param list (str, datetime) parsed_backups: all parsed backup dates in a list of string and tuples
    :return: list  all backups younger than :age: days
    :rtype list (str, datetime):
    """
    max_age = datetime.utcnow() - timedelta(days=age)
    return [i for i in parsed_backups if i["backup_date"] > max_age]


def parse_arguments(args=None):
    """
    Parse the arguments for the hypernode-monitoring-offsite-backup-filesystem command
    :return obj args: The parsed arguments
    """
    parser = ArgumentParser(
        description="Check if we have a valid offsite filesystem backup for last night. "
        "Saves the state in {}".format(OFFSITE_FILESYSTEM_BACKUP_STATEFILE)
    )

    parser.add_argument(
        "--write-json",
        help="Saves state to output file at {}".format(
            OFFSITE_FILESYSTEM_BACKUP_STATEFILE
        ),
        action="store_true",
        default=False,
    )

    return parser.parse_args(args)


def check_if_node_is_less_old_than_one_day():
    """
    Check if node is less old than one day
    :return bool older_than_one_day: True if not older than one day,
    False if older than one day
    """
    # Check first if /etc/hypernode/node_created_at actually exists before we check the age of the node.
    # It can occur that the node is currently being installed and the statfile isn't present yet.
    if not path.isfile(NODE_CREATED_AT_PATH):
        print(
            "OK: node_created_at cannot be determined, because the statfile is not present yet. Backups don't have to be present yet!",
        )
        return True

    current_timestamp = int(time())  # converted to int to truncate to seconds
    if (current_timestamp - fetch_created_at()) <= ONE_DAY_IN_SECONDS:
        print(
            "OK: Node has been created in the last 24 hours. Backups don't have to be present yet!",
            file=sys.stderr,
        )
        return True
    return False


def check_most_recent_backup_is_too_old():
    """
    Check if most recent backup older than one day
    :return bool older_than_one_day: True if older than one day,
    False if not older than one day
    """
    duply_output = check_output(["/usr/bin/duply", "hyp", "status"])
    parsed_backups = _parse_all_backup_datetimes(duply_output)
    return _check_that_most_recent_backup_is_older_than_one_day(parsed_backups)


def check_offsite_backup_filesystem():
    """
    Check if the latest backup is older than a certain amount
    of hours, if so then that indicates that someting is wrong
    with the offsite filesystem backups
    :return None:
    """
    if check_if_node_is_less_old_than_one_day():
        return EX_OK

    if check_most_recent_backup_is_too_old():
        return EX_UNAVAILABLE

    return EX_OK


def write_state_file(state: bool):
    """
    Writes state file
    """
    message = ["offsite_filesystem_backup_healthy_state=", str(state)]

    with open(OFFSITE_FILESYSTEM_BACKUP_STATEFILE, "w") as f:
        f.write(json.dumps(message))


def read_state_file() -> int:
    try:
        with open(OFFSITE_FILESYSTEM_BACKUP_STATEFILE) as f:
            message = json.load(f)
            if len(message) == 2:
                _, value = message
                return int(value)
            return 0
    except OSError as e:
        print(e, file=sys.stderr)
        return 0


def main(args=None):
    """
    Perform the offsite filesystem backup check and write a file
    with the current status.
    :return None:
    """
    arguments = parse_arguments(args)

    if arguments.write_json:
        state = check_offsite_backup_filesystem()
        state_ok = state == EX_OK
        write_state_file(state=state_ok)
    else:
        print("offsite_filesystem_backup_healthy_state", read_state_file())
