import argparse
import sys
from re import match

from kamikaze3.common.env import get_node_roles
from kamikaze3.log.log import configure_logging
from kamikaze3.mysqlbackup.backupstate import BackupState
from kamikaze3.mysqlbackup.exceptions import (
    BrokenViewException,
    InvalidDatabaseNameException,
)
from kamikaze3.mysqlbackup.mysqldumper import stream_new_backup_and_check_exit_codes
from kamikaze3.mysqlbackup.settings import DB_VALIDATION_REGEX
from kamikaze3.settings import ROLE_MYSQL_MASTER

from kamikaze3.mysqlbackup.logger import get_logger
from kamikaze3.mysqlbackup.mysql import fetch_list_of_databases

logger = get_logger("hypernode.mysql.backup")

backupstate = BackupState()


def main(args=None):
    args = parse_arguments(args=args)
    configure_logging("hypernode.mysql.backup", args.verbose)

    logger.info("Starting database backup run")
    logger.debug("Arguments used: {}".format(args))

    mysqlbackup(args.database)


def mysqlbackup(database):
    exitcode = 0
    try:
        if ROLE_MYSQL_MASTER not in get_node_roles():
            logger.info("Node doesn't have mysql role, nothing to do here...")
            return

        # Listify database name when --database flag is used
        if database:
            databases = [database]
        else:
            databases = fetch_list_of_databases()

        for db in databases:
            try:
                if match(DB_VALIDATION_REGEX, db):
                    stream_new_backup_and_check_exit_codes(database=db)
                    backupstate.backup_succeeded(db)
                else:
                    raise InvalidDatabaseNameException(
                        "Invalid character in database name found: {}".format(db)
                    )
            except (InvalidDatabaseNameException, BrokenViewException, Exception):
                exitcode = 1
                logger.exception("Backup of database {} failed!".format(db))
        logger.info("Finished database backup run")
    except Exception:
        logger.exception("An error occurred during database backup")
        exitcode = 1

    backupstate.remove_obsolete_entries()
    sys.exit(exitcode)


def parse_arguments(args=None):
    """
    Parses script arguments from commandline into variables
    """

    parser = argparse.ArgumentParser()

    parser.add_argument("-v", "--verbose", action="store_true")

    exclusive_group = parser.add_mutually_exclusive_group(required=True)
    exclusive_group.add_argument(
        "-a",
        "--all",
        required=False,
        action="store_true",
        help="Create a backup of all databases on this host",
    )
    exclusive_group.add_argument(
        "-d",
        "--database",
        required=False,
        type=str,
        help="Provide a specific database to back up.",
    )

    args = parser.parse_args(args)

    return args


if __name__ == "__main__":
    main()
