import json
import shlex
from datetime import datetime
from os import environ
from subprocess import PIPE, Popen

from kamikaze3.common.command import retry_on_exception
from kamikaze3.common.env import get_hn_mysqlmaster_var

from .logger import get_logger
from .settings import (
    ARCHIVER_CMD,
    AWS_CLI_VIRTUAL_ENV,
    FOLDER_NAME,
    MYSQLDUMP_CMD,
    VARS_FILE,
)

log = get_logger(__name__)


@retry_on_exception(max_retries=1, wait_time=300)
def stream_new_backup_and_check_exit_codes(database):
    """
    This is the MAIN entrypoint for creating a database backup and handling everything else.

    It checks for exit codes and raises a RuntimeError when any exit code is non-zero.
    It logs a processes its stderr output to the CRITICAL level if any exit code is non-zero.

    With one exception: When mysqldump exits with code 2 and reports error code 1356 in it's stderr output,
    it encountered a corrupt view. This is no problem since we run mysqldump with --force. The backup is still
    successful. We log whenever this happens (to info level) and continue.

    Arguments:
        database : database name

    Returns:
        None

    Raises:
        RuntimeError if any exit code != 0
    """
    log.info(f"[{database}] Creating backup of database")
    failed_on_corrupt_view = False

    results = stream_compressed_mysqldump_to_s3(database=database)

    # Check if any process failed
    (
        archiver_exit_code,
        mysqldump_exit_code,
        s3_exit_code,
        archiver_stderr,
        mysqldump_stderr,
        s3_stderr,
    ) = results

    # Check for corrupt view errors. We can safely ignore those.
    # Sample line: Couldn't execute 'show create table `qsel_av_printen`': View 'db036441_4dms_portal_e.qsel_av_printen'
    # references invalid table(s) or column(s) or function(s) or definer/invoker of view lack rights to use them (1356)
    if mysqldump_exit_code == 2 and all(
        ["(1356)" in l for l in mysqldump_stderr.decode().strip().split("\n")]
    ):
        log.info(
            "Mysqldump reported a failure during backup caused by a failed view (error code 1356). "
            "Logging full stderr output:"
        )
        log.info(mysqldump_stderr)
        failed_on_corrupt_view = True

    if mysqldump_exit_code != 0 and not failed_on_corrupt_view:
        log.critical(f"[{database}] Mysqldump exit code: {mysqldump_exit_code}")
        log.critical(f"[{database}] Mysqldump stderr: {mysqldump_stderr}")

    if archiver_exit_code != 0:
        log.critical(f"[{database}] Archiver exit code: {archiver_exit_code}")
        log.critical(f"[{database}] Archiver stderr: {archiver_stderr}")

    if s3_exit_code != 0:
        log.critical(f"[{database}] S3 exit code: {s3_exit_code}")
        log.critical(f"[{database}] S3 stderr: {s3_stderr}")

    if any(
        (
            (mysqldump_exit_code != 0 and not failed_on_corrupt_view),
            archiver_exit_code,
            s3_exit_code,
        )
    ):
        raise RuntimeError(
            "[{}] At least one process exited with a non-zero exit code. "
            "stderr output has been logged.".format(database)
        )

    # Finish backup when successful
    log.info(f"[{database}] Backup successful")


def stream_compressed_mysqldump_to_s3(database):
    """
    Streams a compressed mysqldump of given :database: to
    our remote backup host via SSH.

    Inputs:
        database: local database name to backup

    Returns a tuple of 6 items:
        [0] - archiver exit code
        [1] - mysqldump exit code
        [2] - s3 exit code
        [3] - archiver stderr output (empty string if nonexistent)
        [4] - mysqldump stderr output (empty string if nonexistent)
        [5] - s3 stderr output (empty string if nonexistent)
    """
    log.info(f"[{database}] Streaming backup of database '{database}' to s3")

    conf_data = _read_json_from_file(VARS_FILE)
    ignored_tables = get_hn_mysqlmaster_var("mysql_backup_ignored_tables")

    aws_credentials = {
        "AWS_ACCESS_KEY_ID": conf_data["s3_user"],
        "AWS_SECRET_ACCESS_KEY": conf_data["s3_pass"],
    }

    # Set up processes
    log.debug(f"[{database}] Starting backup subprocesses...")
    mysqldump_process = start_dumper(database=database, ignored_tables=ignored_tables)
    archiver_process = start_archiver(input_pipe=mysqldump_process.stdout)
    s3_stream_process = start_s3_backup_streamer(
        database,
        conf_data["appname"],
        aws_credentials,
        conf_data["backup_bucket_name"],
        input_pipe=archiver_process.stdout,
    )

    # Connect them & wait for them to finish
    log.info(f"[{database}] Creating backup...")
    s3_stderr = s3_stream_process.communicate()[1]
    archiver_stderr = archiver_process.communicate()[1]
    mysqldump_stderr = mysqldump_process.communicate()[1]
    archiver_exit_code = archiver_process.poll()
    mysqldump_exit_code = mysqldump_process.poll()
    s3_exit_code = s3_stream_process.poll()

    # Close pipes
    log.debug(f"[{database}] Closing pipes")
    archiver_process.stdout.close()
    mysqldump_process.stdout.close()

    log.debug(f"[{database}] Backup process finished")
    return (
        archiver_exit_code,
        mysqldump_exit_code,
        s3_exit_code,
        archiver_stderr,
        mysqldump_stderr,
        s3_stderr,
    )


######################
#  HELPER FUNCTIONS  #
######################


def start_archiver(input_pipe):
    """
    Helper function for stream_compressed_mysqldump_to_remote_host.
    Starts archiver process, connects stdin with :input_pipe: and stdout with :output_pipe:.
    Returns Popen object.
    """
    log.debug(f"Starting archiver process: '{ARCHIVER_CMD}'")
    archiver_process = Popen(
        shlex.split(ARCHIVER_CMD),
        shell=False,
        stdin=input_pipe,
        stdout=PIPE,
        stderr=PIPE,
        bufsize=4096,
    )
    return archiver_process


def start_dumper(database, ignored_tables: list[str] | None = None):
    """
    Helper function for stream_compressed_mysqldump_to_remote_host.
    Starts mysqldump process with params :database: and :ignored_tables:
    if given.
    Returns Popen object.
    """

    ignored_tables_options = " ".join(
        map(
            lambda table: f"--ignore-table='{table}'",
            ignored_tables or [],
        )
    )
    formatted_mysqldump_cmd = MYSQLDUMP_CMD % {
        "ignored_tables": ignored_tables_options,
        "database": database,
    }
    log.debug(f"[{database}] Starting dumper process: '{formatted_mysqldump_cmd}'")
    mysqldump_process = Popen(
        shlex.split(formatted_mysqldump_cmd),
        shell=False,
        stdout=PIPE,
        stderr=PIPE,
        bufsize=4096,
    )
    return mysqldump_process


def start_s3_backup_streamer(dbname, appname, aws_credentials, bucket_name, input_pipe):
    """
    Helper function for stream_compressed_mysqldump_to_remote_host.
    Starts ssh process to fileserver and gives the remote script the right
    "receive" command. Connects stdin pipe to given :input_pipe:.
    Returns the Popen object.

    arguments:
            dbname:           Database name
            appname:          The name of the hypernode
            aws_credentials:  A dict containing the AWS credentials, namely:
                    -  AWS_ACCESS_KEY_ID
                    -  AWS_SECRET_ACCESS_KEY
            bucket_name:      The name of the s3 bucket to be used
            input_pipe:       STDIN pipe
    """
    s3_command = [
        "env",
        "-i",
        "VIRTUAL_ENV={}".format(
            AWS_CLI_VIRTUAL_ENV
        ),  # TODO: Can be removed when precise is gone
        "AWS_DEFAULT_REGION=eu-west-1",  # this is aws-ireland
        "AWS_ACCESS_KEY_ID={}".format(aws_credentials["AWS_ACCESS_KEY_ID"]),
        "AWS_SECRET_ACCESS_KEY={}".format(aws_credentials["AWS_SECRET_ACCESS_KEY"]),
        "PATH={path}:{virtual_env}/bin/".format(
            path=environ["PATH"], virtual_env=AWS_CLI_VIRTUAL_ENV
        ),
        "aws",
        "s3",
        "cp",
        "--storage-class",
        "REDUCED_REDUNDANCY",
        "-",
        generate_s3_pathname(dbname, appname, bucket_name),
    ]

    log.debug(f"[{dbname}] Starting s3 pusher process")
    s3_process = Popen(
        s3_command,
        shell=False,
        stdin=input_pipe,
        stdout=PIPE,
        stderr=PIPE,
        bufsize=4096,
    )
    return s3_process


def generate_s3_pathname(dbname, appname, bucket_name):
    """
    Generates a path and filename for a new backup based on appname, database name
    and current date&time. Extension is '.sql.gz'.
    Returns directory/filename as a string.

    :type dbname: string
    :param dbname: The database to be backed up
    :type appname: string
    :param appname: The app name from which we're backup up from
    :type bucket_name: string
    :param bucket_name: The s3 bucket we're backing up towards
    """
    datetimestring = datetime.utcnow().strftime("%Y-%m-%d_%H-%M-%S_UTC")
    extension = "sql.gz"
    return "s3://{}/{}/mysqlbackups/{}/{}_{}.{}".format(
        bucket_name,
        FOLDER_NAME.format(appname=appname),
        dbname,
        dbname,
        datetimestring,
        extension,
    )


def _read_json_from_file(file):
    with open(file) as f:
        return json.loads(f.read())
