import argparse
import sys
from datetime import datetime
from subprocess import check_output

from pytz import timezone, utc

from kamikaze3.mysqlbackuprestore.list_backup import list_database_backups
from kamikaze3.mysqlbackuprestore.restore_backup import (
    download_or_import_databases,
    list_databases_in_latest_backup,
)
from kamikaze3.mysqlbackuprestore.settings import RESTORE_PATH
from kamikaze3.mysqlbackuprotator.logger import get_logger
from kamikaze3.mysqlbackuprotator.s3_commands import get_bucket_and_config

CET = timezone("Europe/Amsterdam")
log = get_logger("hypernode.mysql.backuprestore")


def main():
    assert_called_user_is_root()
    bucket, config = get_bucket_and_config()

    if not list_databases_in_latest_backup():
        log.warning(
            "No database backups available because there are no databases on this host"
        )
        sys.exit(0)
    else:
        args = parse_arguments()
        try:
            if args.list:
                # list available backups
                list_database_backups(args.list, bucket, config)
            else:
                # restore/download specific or all databases
                download_or_import_databases(args)

        except Exception:
            log.exception("An error occurred during database restore")
            sys.exit(1)


def parse_arguments():
    """
    Parses script arguments from commandline into variables
    """

    parser = argparse.ArgumentParser()

    exclusive_group = parser.add_mutually_exclusive_group(required=True)
    exclusive_group.add_argument(
        "--restore-all-databases",
        action="store_true",
        help="Restore all database backups directly into mysql!",
    )
    exclusive_group.add_argument(
        "--download-all-databases",
        action="store_true",
        help="Download all database backups directly onto disk!",
    )
    exclusive_group.add_argument(
        "--database",
        choices=list_databases_in_latest_backup(),
        help="Provide specific databases to restore",
    )
    exclusive_group.add_argument(
        "--list",
        type=str,
        choices=list_databases_in_latest_backup(),
        help="List all available backups for the given database name.",
    )

    download_or_import = parser.add_mutually_exclusive_group(required=False)
    download_or_import.add_argument(
        "--download",
        action="store_true",
        help="Download database backup to {}".format(RESTORE_PATH),
    )
    download_or_import.add_argument(
        "--import-as",
        metavar="DATABASE_NAME",
        type=str,
        help="Import database backup towards given database name",
    )

    date_or_latest = parser.add_mutually_exclusive_group(required=False)
    date_or_latest.add_argument(
        "--date",
        metavar="DATE",
        type=validate_date,
        help="Retrieve backups closest to given date and time (timezone used is CET), expected "
        "format is 2016-02-23_15:59",
    )
    date_or_latest.add_argument(
        "--latest", action="store_true", help="Use the most recent backup available"
    )

    parser.add_argument(
        "--force",
        action="store_true",
        help="Overwrite existing databases without"
        " warning! Be very careful with this!",
    )

    parser.add_argument(
        "--max-backup-delta",
        type=int,
        default=30,
        help="Given a date (or now in case of --latest), how "
        "far back/forward should we look for backups? In days.",
    )

    parser.add_argument(
        "--exclude-database",
        nargs="+",
        help="Exclude specified databases when used in combination with --restore-all-databases or "
        "--download-all-databases",
    )

    args = parser.parse_args()

    if args.restore_all_databases is True:
        if args.force is not True:
            parser.print_help()
            print(
                "\nError: To restore all available database --force must be provided!\n"
                "!!! Warning, all databases will be overwritten !!!"
            )
            sys.exit(1)

    if args.database:
        # check that either --download or --import-as is provided
        if not any([args.download, args.import_as]):
            parser.print_help()
            print(
                "\nError: Please provide a restore target via --download or --import-as"
            )
            sys.exit(1)

    # check that either --date or --latest is provided
    if not any([args.latest, args.date]) and not args.list:
        parser.print_help()
        print(
            "\nError: Please provide --date or --latest to select which backup to restore"
        )
        sys.exit(1)

    if args.exclude_database and not (
        args.restore_all_databases or args.download_all_databases
    ):
        parser.print_help()
        print(
            "\nError: --exlude-database can only be used in combination with --download-all-databases or "
            "--restore-all-databases."
        )
        sys.exit(1)

    return args


def validate_date(string):
    try:
        # convert from CET to UTC
        res = CET.localize(datetime.strptime(string, "%Y-%m-%d_%H:%M")).astimezone(utc)
        # convert to non timezone aware object, to allow datetime comparisons
        return res.replace(tzinfo=None)
    except ValueError:
        msg = "Not a valid date: '{}'. Expected date like: 2016-02-23_15:59".format(
            string
        )
        raise argparse.ArgumentTypeError(msg)


def assert_called_user_is_root():
    assert (
        check_output("/usr/bin/whoami").decode() == "root\n"
    ), "This tool is meant for administrator (root) usage!"


if __name__ == "__main__":
    main()
