import shlex
from datetime import datetime
from subprocess import PIPE, Popen

from kamikaze3.mysqlbackuprestore.logger import get_logger
from kamikaze3.mysqlbackuprestore.settings import UNARCHIVER_CMD
from kamikaze3.mysqlbackuprotator.parser import parse_backups_for_database
from kamikaze3.mysqlbackuprotator.retention_calculator import find_closest
from kamikaze3.mysqlbackuprotator.settings import S3_MYSQL_BACKUP_FOLDER

log = get_logger("hypernode.mysql.backuprestore")


def stream_backup_from_s3(pipe, bucket, backup, appname):
    """
    Streams backup data from s3 and writes it to :pipe:

    :param pipe: the backup data will be written to this pipe
    :param bucket: Boto bucket object
    :param MysqlBackup backup: the backup which will be provided
    :param str appname: name of the app for which we're fetching a backup
    :return:
    """
    key = bucket.get_key(
        "{}{}/{}".format(
            S3_MYSQL_BACKUP_FOLDER.format(appname=appname),
            backup.database,
            backup.filename,
        )
    )
    key.get_contents_to_file(pipe)


def start_decompressor():
    """
    Starts a uncompress process through which data will flow

    :return: Popen object of started process
    """
    archiver_process = Popen(
        shlex.split(UNARCHIVER_CMD),
        shell=False,
        stdin=PIPE,
        stdout=PIPE,
        stderr=PIPE,
        bufsize=4096,
    )
    return archiver_process


def start_mysql_importer(input_pipe, database):
    """
    Starts the mysql cli client for import. The SQL data will be writen to the :input_pipe:

    :param subprocess.PIPE input_pipe: pipe object to which SQL data will be written to
    :param database: database name under which data will be imported
    :return: Popen object of started process
    """
    mysql_process = Popen(
        ["/usr/bin/mysql", database],
        shell=False,
        stdin=input_pipe,
        stdout=PIPE,
        stderr=PIPE,
        bufsize=4096,
    )

    return mysql_process


def ensure_database(database):
    """
    Create :database: if it doesn't exist yet

    :param database: the database name which needs to be created
    """
    mysql_process = Popen(
        ["/usr/bin/mysql", "-e", "CREATE DATABASE IF NOT EXISTS `{}`".format(database)],
        shell=False,
        stdout=PIPE,
        stderr=PIPE,
        bufsize=4096,
    )
    mysqlimport_stderr = mysql_process.communicate()[1]
    mysqlimport_exit_code = mysql_process.poll()
    if mysqlimport_exit_code != 0:
        raise RuntimeError(
            "Creating database {} if not exists failed! "
            "Mysql exited with a non-zero exit code {}".format(
                database, mysqlimport_stderr
            )
        )


def find_closest_backup(bucket, database, appname, target_date=None, max_delta=None):
    """
    For a given database in the backup, fetch the closest available backup to :target_date:

    :param bucket: Boto bucket object
    :param str database: database name from which we're looking for a close backup
    :param str appname: name of the hypernode app
    :param int max_delta: how many days before/after :target_date: should we look for backups
    :param datetime target_date:  naive datetime object in UTC

    :returns a tuple of (datetime datetime, str filename) which represents the closest available
             backup to :target_date:
    """
    target_date = datetime.utcnow() if target_date is None else target_date
    backups = parse_backups_for_database(bucket, database, appname=appname)
    return find_closest(target_date, backups, max_delta=max_delta)


def filter_databases(databases, backups, exclude=False):
    """
    Given a list of :databases: and :backups:, return only :backups: which are asked for in :databases:
    If :exclude: is True then do the reverse: return only :backups: which are not excluded in :databases:

    :param list(str) databases: a list of databases
    :param list(MysqlBackup) backups: a list of MysqlBackup objects
    :param bool exclude: if True return all :backups: which are not present in :databases:
    :return: return a list of all tuples which are asked for in :databases:
    """
    databases_to_be_restored = [b.database for b in backups]
    for database in databases:
        if database not in databases_to_be_restored and not exclude:
            raise RuntimeError(
                "Specified invalid database, database {} is not in the latest backup".format(
                    database
                )
            )

    if exclude:
        return list(filter(lambda backup: backup.database not in databases, backups))
    else:
        return list(filter(lambda backup: backup.database in databases, backups))


def assert_database_does_not_exist(database):
    """
    Assert that the given :database: name does not exist yet

    :param str database: the database name
    :raises AssertionError: if database does already exist
    """

    mysql_process = Popen(
        ["/usr/bin/mysql", "-B", "-e", "SHOW DATABASES"],
        shell=False,
        stdout=PIPE,
        stderr=PIPE,
        bufsize=4096,
    )
    stdout, stderr = mysql_process.communicate()
    mysql_exit_code = mysql_process.poll()
    if mysql_exit_code != 0:
        raise RuntimeError("Checking if database exists failed. Current state unknown!")

    database_list = stdout.split(b"\n")
    database_list.remove(b"Database")  # header of output

    assert database not in database_list, (
        "The database {} already exists in MySQL! Aborting!\n You can specify "
        "--force to overwrite the existing database.".format(database)
    )


def find_closest_backups(bucket, databases, appname, target_date=None, max_delta=None):
    """
    For a given list of databases we find the closest backup to :target_date:

    :param bucket: Boto bucket object
    :param list(str) database: a list of databases for which to find a close backup
    :param str appname: name of the hypernode app
    :param int max_delta: how many days before/after :target_date: should we look for backups
    :param datetime target_date:  naive datetime object in UTC
    :returns a list of MysqlBackup objects containing the
             closest available backups to :target_date: for each database
    :rtype list(MysqlBackup):
    """

    target_date = datetime.utcnow() if target_date is None else target_date

    backups_to_be_restored = []
    for db in databases:
        log.debug("Finding latest backup for database {}".format(db))
        backup = find_closest_backup(
            bucket, db, appname, target_date=target_date, max_delta=max_delta
        )
        if backup is not None:
            backups_to_be_restored.append(backup)
            log.debug("Found appropriate backup for database {}".format(db))
        else:
            log.debug("Backup for database {} too old, skipping".format(db))
    return backups_to_be_restored
