from datetime import datetime

from kamikaze3.mysqlbackuprotator.helpers import parse_json_file
from kamikaze3.mysqlbackuprotator.logger import get_logger
from kamikaze3.mysqlbackuprotator.parser import parse_backups_for_database
from kamikaze3.mysqlbackuprotator.retention_calculator import (
    backups_to_remove,
    generate_keep_dates,
)
from kamikaze3.mysqlbackuprotator.s3_commands import (
    create_s3_connection,
    delete_file,
    list_database_directories,
)
from kamikaze3.mysqlbackuprotator.settings import BACKUP_RETENTION, VARS_FILE

NOW = datetime.utcnow()
log = get_logger(__name__)


def rotate_backups(from_datetime=NOW):
    """
    This function lists all backed up databases at s3 for this app and removes leftover backupes

    parameters:
        from_datetime:  datetime from which the retentions get calculated from. used to test the function
    """
    kwargs = parse_json_file(VARS_FILE)
    bucket = create_s3_connection(
        kwargs["s3_user"], kwargs["s3_pass"], kwargs["backup_bucket_name"]
    )

    databases = list_database_directories(bucket, **kwargs)
    for db in databases:
        remove_old_backups(
            bucket, db, BACKUP_RETENTION, from_datetime=from_datetime, **kwargs
        )


def remove_old_backups(bucket, database, retention, from_datetime=NOW, **kwargs):
    """
    Deletes all superfluous backups for :database:.

    Any backup that remain after matching all backups in :database: according to :retention:
    will be deleted.
    """
    keep_dates = generate_keep_dates(retention, from_datetime)

    backups = parse_backups_for_database(bucket, database, **kwargs)
    bk_to_delete = backups_to_remove(keep_dates, backups)

    log.info(
        "Found {}, keeping {}, deleting {}. Database '{}'".format(
            len(backups), retention.total_backup_count(), len(bk_to_delete), database
        )
    )

    log.debug("Logging from_datetime")
    log.debug(str(from_datetime))

    log.debug("Logging our backups:")
    log.debug(str(backups))

    log.debug("Logging our bk_to_delete:")
    log.debug(str(bk_to_delete))

    for backup in bk_to_delete:
        log.info(
            "Removing old backup: {} datetime: {}".format(
                backup.filename, backup.backup_date
            )
        )
        delete_file(bucket, backup, **kwargs)
