from argparse import ArgumentParser

from MySQLdb import OperationalError, connect
from MySQLdb.cursors import Cursor

from kamikaze3.common.config_parsers import MysqlConfigParser
from kamikaze3.importer.validators import get_mysql_major_version
from kamikaze3.mysqlfixer.logger import get_logger
from kamikaze3.mysqlfixer.settings import (
    ACCESS_USERS,
    EXPECTED_SUPERUSER_PERMISSIONS,
    GET_ACTUAL_USER_PERMISSIONS,
    SET_CORRECT_SUPERUSER_PERMISSIONS,
    MySQLUser,
)

log = get_logger()


def main():
    process_args()
    fix_mysql()


def process_args():
    parser = ArgumentParser(
        description="Fix the Mysql user permissions and ensure that "
        "the configured passwords on disk are correct."
    )
    parser.parse_args()


def fix_mysql():
    cursor = get_root_cursor()

    for user in ACCESS_USERS:

        ensure_user_present(user, cursor)

        if not are_user_permissions_correct(user=user.name, cursor=cursor):
            fix_mysql_user_permissions(user=user.name, cursor=cursor)

        if not user_can_login(user):
            fix_mysql_user_password(user, cursor)

    log.info("Fixing is finished!")


def ensure_user_present(user: MySQLUser, cursor: Cursor) -> None:
    password = get_password_from_cnf_file(user.cnf_file_path)
    for host in user.hosts:
        log.info("Checking existence of user {} at host {}".format(user.name, host))
        cursor.execute(
            "SELECT COUNT(*) FROM user WHERE User='{}' AND Host='{}'".format(
                user.name, host
            )
        )
        user_is_present = cursor.fetchone()[0]
        if not user_is_present:
            # Sometimes the root user disappears from the mysql.user table but creating a new one gives a MYSQL 1396 error.
            # Dropping the user again fixes this, see:
            # https://teamdotblue.slack.com/archives/C0RFNERTQ/p1641309359251300?thread_ts=1641308721.246100&cid=C0RFNERTQ
            if user.name == "root":
                try:
                    cursor.execute(
                        "DROP USER '{}'@'{}'; FLUSH PRIVILEGES;".format(user.name, host)
                    )
                    log.info("Dropped user {} at host {}".format(user.name, host))
                except OperationalError:
                    pass

            log.info("Creating user {} at host {}".format(user.name, host))
            cursor.execute(
                "CREATE USER '{}'@'{}' IDENTIFIED BY '{}'".format(
                    user.name, host, password
                )
            )


def change_mysql_password(cursor, user: str, password: str, host: str) -> int:
    """
    Change mysql password. Behaviour changes based on mysql version
    """
    mysql_version = get_mysql_major_version()
    log.warn(
        "Setting password {} for {} on MySQL {}".format(password, user, mysql_version)
    )
    if mysql_version == "5.6":
        rows_updated = cursor.execute(
            "UPDATE `mysql`.`user` SET `password` = PASSWORD(%s) WHERE `user` = %s",
            (password, user),
        )
    else:
        rows_updated = cursor.execute(
            "ALTER USER '{user}'@'{host}' IDENTIFIED BY '{passw}'".format(
                user=user, passw=password, host=host
            )
        )

    cursor.execute("FLUSH PRIVILEGES")
    return rows_updated


def fix_mysql_user_password(user: MySQLUser, cursor: Cursor) -> int:

    assert user.name != "root", "safeguard: am not changing password for root"

    password = get_password_from_cnf_file(user.cnf_file_path)

    assert len(password) >= 8, "password too short"
    rows_updated = 0
    for host in user.hosts:
        rows_updated += change_mysql_password(
            cursor=cursor, user=user.name, password=password, host=host
        )

    return rows_updated


def get_password_from_cnf_file(cnf: str) -> str:
    config = MysqlConfigParser()
    config.read(cnf)
    return config.get("client", "password")


def user_can_login(user: MySQLUser) -> bool:
    try:
        get_mysql_conn(user.cnf_file_path)
    except OperationalError as e:
        # (1045, "Access denied for user 'debian-sys-maint'@'localhost' (using password: YES)")
        code, msg = e.args
        if code == 1045:
            log.warn("User {} cannot login".format(user.name))
            return False
        else:
            raise

    return True


def get_root_cursor() -> Cursor:
    """
    Tries the various root accounts and returns a root cursor if found
    :return: MySQLdb.cursor with root privileges
    """
    log.info("Testing for root account")
    for user in ACCESS_USERS:
        try:
            cursor = get_mysql_conn(user.cnf_file_path)
            cursor.execute(
                "SELECT * FROM user"
            )  # simple test to check if we have superuser privileges
            log.info("Found root account: {}".format(user.name))
            return cursor
        except OperationalError:
            pass
    raise RuntimeError("No working root mysql credentials were found!")


def get_mysql_conn(my_cnf_location=None):
    """
    Login to mysql and return a cursor with credentials from :my_cnf_location:

    :param  str my_cnf_location: the path to the my.cnf file with login credentials
    :return: MySQLdb.cursor
    """
    db = connect(db="mysql", host="localhost", read_default_file=my_cnf_location)
    return db.cursor()


def fix_mysql_user_permissions(user, cursor):
    """
    Set's correct superuser equivalent privileges for :user:
    :param str user: the mysql username
    :param cursor: a MysqlDB cursor with root permissions
    """

    log.info("Fixing mysql permissions for user {}".format(user))

    if user == "app":
        cursor.execute(
            "DELETE FROM user WHERE user = '{}' AND host != '%'".format(user)
        )
        if cursor.rowcount != 0:
            log.info(
                "Deleted %d excess entries for user {}".format(cursor.rowcount, user)
            )
    cursor.execute(SET_CORRECT_SUPERUSER_PERMISSIONS, (user,))
    cursor.execute("flush privileges")


def are_user_permissions_correct(user: str, cursor: Cursor) -> bool:
    """
    Checks if user's permissions are correctly configured

    :returns bool: True if permissions are correct, false if permissions are false
    """
    log.info("checking MySQL permissions of user: {}".format(user))

    cursor.execute(GET_ACTUAL_USER_PERMISSIONS, (user,))
    result = cursor.fetchall()

    for permissions_tuple in result:
        if permissions_tuple != EXPECTED_SUPERUSER_PERMISSIONS:
            log.warning(
                "MySQL permissions of user {} not what we expected!".format(user)
            )
            return False

    log.info("MySQL permissions for user {} look good!".format(user))
    return True


if __name__ == "__main__":
    # execute only if run as a script
    main()
