from argparse import ArgumentParser
from multiprocessing import Pool
from os.path import isfile
from re import match
from subprocess import CalledProcessError, check_output
from time import sleep

REQUESTS_PER_DOMAIN = 15
DEHYDRATED_DOMAINS_FILES = [
    "/etc/dehydrated/domains.txt",
    "/data/web/.dehydrated/domains.txt",
]
ACCEPTED_DOMAIN_REGEX = r"^(\w+|\.|-)+$"  # Regex to validate domains


def main(args=None):
    parser = ArgumentParser(
        description="This tool warms the nginx OCSP cache in a very simple way. We make {} requests to the HTTPS"
        "address to force nginx to update it's OCSP cache for that domain.".format(
            REQUESTS_PER_DOMAIN
        )
    )

    parser.parse_args(args=args)

    existing_files = [f for f in DEHYDRATED_DOMAINS_FILES if isfile(f)]
    if existing_files == []:
        print(
            "No existing domains file found among {}, not doing anything.".format(
                DEHYDRATED_DOMAINS_FILES
            )
        )
        return

    # We use a set to exclude duplicate domains
    domain_names = set()
    for filename in existing_files:
        domain_names = domain_names.union(get_domain_names(filename))
    print("Going to warm OCSP cache for domains: {}".format(" ".join(domain_names)))
    process_pool = Pool(processes=10)

    # warm cache
    for domain in domain_names:
        process_pool.apply_async(
            warm_ocsp_cache_for_domain, (domain,)
        )  # must pass tuple

    # wait 5 seconds, then kill all processes regardless
    sleep(5)
    process_pool.terminate()


def get_domain_names(filename):
    """
    Read domains defined to dehydraded from :filename: and sanitize through :ACCEPTED_DOMAIN_REGEX:.
    All non-ascii characters are stripped.

    :param str filename: path of file to check for domain names
    :return: set of domains
    :rtype set[str]:
    """
    with open(filename) as f:
        lines = f.read()

    result = []
    for item in lines.split():
        domain = item.strip()
        if match(ACCEPTED_DOMAIN_REGEX, domain):
            result.append(domain)

    return set(result)


def warm_ocsp_cache_for_domain(domain_name):
    """
    Make requests to the nginx webserver to force it to warm up the OCSP cache
    :param str domain_name:
    :return None:
    """
    for i in range(REQUESTS_PER_DOMAIN):
        url = "https://{}/.well-known/not_existing_file_for_fast_ocsp_cache_warming.txt".format(
            domain_name
        )
        print(
            "{:<35}: Filling OCSP cache, request {} of {}".format(
                domain_name, i + 1, REQUESTS_PER_DOMAIN
            )
        )

        try:
            # We're using perl 'GET' here because it has OCSP support which python requests does not have.
            # Doing the exact same request with requests.get() will not prime the OCSP cache. :sadpanda:

            # We're hitting the `.well-known` directory because it's defined in mag1/2 situations and does not
            # hit the php-fpm handler.

            check_output(["/usr/bin/GET", "-t", "1", url])

        except CalledProcessError:
            pass  # we always get this error, since the file we're calling does not exist (intentionally)

        except Exception:
            print(
                "{:<35}: Unexpected exception received. Ignoring.".format(domain_name)
            )
