import json
import os.path
import re
import subprocess

import dateutil.parser
import psutil
from psutil import NoSuchProcess

from kamikaze3.onboarding.check import Check, ReCheck

SPHINX_PROCESS_NAME = "searchd"
AUTH_LOG_PATH = "/var/log/auth.log"
NGINX_HANDLER_CONFIG_PATH = "/data/web/nginx/handler.conf"
MAVO_SETTINGS = "/bin/su app -c '/usr/lib/byte-mavo/plugins/check_settings.php --magentopath /data/web/public/'"
MAGENTO2_FLAG = "/data/web/nginx/magento2.flag"


class SphinxCheck(Check):
    name = "sphinx"

    def run(self):
        def _get_process_name(proc):
            # work around psutil's backwards incompatibility
            return proc.name() if hasattr(proc.name, "__call__") else proc.name

        for process in psutil.process_iter():
            try:
                if _get_process_name(process) == SPHINX_PROCESS_NAME:
                    return True
            except NoSuchProcess:
                # Process might have disappeared between fetching the process list and
                # determining its name.
                pass

        return False


class RedisCheck(Check):
    name = "redis"

    def run(self):
        exit_code = subprocess.call('redis-cli info | grep -s ":keys="', shell=True)
        return exit_code == 0


class Magento1Active(Check):
    name = "magento1-active"

    def run(self):
        if not os.path.exists(MAGENTO2_FLAG):
            Magento2Active().unmark_check()
            return True
        return False


class Magento2Active(Check):
    name = "magento2-active"

    def run(self):
        if os.path.exists(MAGENTO2_FLAG):
            Magento1Active().unmark_check()
            return True
        return False


class AdminLoginCheck(ReCheck):
    name = "admin-login"
    logfile = "/var/log/nginx/access.log"

    """
    Check if the customer accessed his admin backend in the last three days. We search
    the current access log in reverse order for the last occurrence of the admin URL
    (which is determined using MAVO). Since the access logs are typically rotated every
    day, we need a way to keep track of the last time the customer accessed the admin.

    The mtime of the state file of this check is used as a simple storage for the last
    time the customer accessed the admin.

    * success: admin access found, and more recent than previous access (or no access)
    * fail: no admin access found
    * fail: admin access found, but previous access is more recent
    """

    def run(self):
        admin_url = self.get_admin_url()

        if admin_url:
            new_atime = self.last_admin_access_time(admin_url)

            if self.verify(new_atime):
                self.new_success_time = new_atime
                return True
            return False

    def verify(self, new_atime):
        return new_atime and (
            self.last_success is None or new_atime > self.last_success
        )

    def get_admin_url(self):
        try:
            settings = self.get_mavo_settings()
            admin_url = settings["global_settings"]["admin_url"]
        except (subprocess.CalledProcessError, ValueError, KeyError):
            return None

        url_match = re.search(
            r"^(https?://[-_.a-z0-9]+/)?(/?index\.php)?/?(.*)$", admin_url
        )
        if url_match:
            return url_match.group(3).rstrip("/")

    def get_mavo_settings(self):
        return json.loads(self._call_mavo())

    def _call_mavo(self):
        return subprocess.check_output(
            MAVO_SETTINGS.split(), stderr=subprocess.STDOUT, shell=False
        )

    def last_admin_access_time(self, admin_url):
        return self.last_url_occurrence(admin_url)

    def last_url_occurrence(self, url):
        try:
            line = self._search_log(url)
            if isinstance(line, (bytes, bytearray)):
                line = line.decode("utf-8")
            tokens = json.loads(line)
            return dateutil.parser.parse(tokens["time"])
            # Should we force UTC here? nginx timestamps contain utc offset,
            # which dateutil.parser translates into tzutc()
        except (TypeError, KeyError, ValueError):
            return None

    def _search_log(self, token):
        if re.search(r"[^-./_:a-zA-Z0-9]", token):
            raise RuntimeError("Unsupported characters in search object")
        if re.search(r"[^-./_a-zA-Z0-9]", self.logfile):
            raise RuntimeError("Unsupported characters in search subject")

        grep_regex = '"POST (/index.php)?/%s.*, "status":"302"' % token
        try:
            return subprocess.check_output(
                "/usr/bin/tac {} | /usr/bin/grep -m 1 -E '{}'".format(
                    self.logfile, grep_regex
                ),
                shell=True,
            )
        except subprocess.CalledProcessError as e:
            if e.returncode == 1:
                return None
            raise RuntimeError("Command failed: %s" % e.output)


service_checks = [
    SphinxCheck,
    AdminLoginCheck,
    Magento1Active,
    Magento2Active,
    RedisCheck,
]
