import logging
import os
from ctypes import cdll

from kamikaze3.oom_monitor.strategy import WaitStrategy

logger = logging.getLogger("kamikaze3.oom_monitor")


# We need eventfd (http://man7.org/linux/man-pages/man2/eventfd.2.html) to use
# the cgroups notification API, we could either 1. create a c extension (or use
# one like 'butter') or 2. use the one in the glibc shared library which is
# available on all our systems anyway. For more info see:
# http://sgros-students.blogspot.nl/2013/05/calling-eventfd-from-python.html
def eventfd(init_val, flags):
    libc = cdll.LoadLibrary("libc.so.6")
    return libc.eventfd(init_val, flags)


class CGroupV1WaitStrategy(WaitStrategy):
    def __init__(self, cgroup_slice: str = ""):
        self.event_control_path = (
            f"/sys/fs/cgroup/memory/{cgroup_slice}/cgroup.event_control"
        )
        self.oom_control_path = (
            f"/sys/fs/cgroup/memory/{cgroup_slice}/memory.oom_control"
        )
        self._event_socket = self.start_event_listener()

    def start_event_listener(self):
        # We will use the cgroup notification API. 1. start eventfd subsystem. 2. pass 2 control file descriptors by
        # writing them as integers to the special event_fd socket. See:
        # https://access.redhat.com/documentation/en-US/Red_Hat_Enterprise_Linux/6/html/Resource_Management_Guide/sec-Using_the_Notification_API.html
        event_fd = eventfd(0, 0)
        cfd = os.open(self.event_control_path, os.O_RDWR)
        ofd = os.open(self.oom_control_path, os.O_RDONLY)
        # Note that there's no .format for bytestrings, there was a PEP for that but
        # that was rejected: https://www.python.org/dev/peps/pep-0461/ printf style
        # formatting it is then.
        os.write(cfd, b"%d %d" % (event_fd, ofd))
        return event_fd

    def wait(self):
        os.read(self._event_socket, 8)  # events are 8 bytes

    def close(self):
        os.close(self._event_socket)
