"""
Used for making openvpn client configurations, please note that when this template changes ALL client configurations
need to be regenerated if we want it to be changed everywhere.
"""

import logging
from os import remove
from subprocess import PIPE, Popen

from kamikaze3.openvpn_config_generator.inventory import get_client_certificate

logger = logging.getLogger(__name__)

OPENVPN_CLIENT_CONFIG_DIR = "/data/openvpn/openvpn_client_configs"
OPENVPN_CLIENT_CONFIG_TEMPLATE = """
ifconfig-push {}
"""
DEFAULT_PATH = "/data/openvpn/client-configs"
TEMPLATE = """
client
dev tun
proto udp
remote {hn_name} 11194
nobind
comp-lzo
verb 3
key-direction 1
cipher AES-256-CBC
data-ciphers AES-256-CBC
<cert>
{cert}
</cert>
<key>
{key}
</key>
<ca>
{ca}
</ca>
<tls-auth>
{tls}
</tls-auth>
"""


def generate_openvpn_client_cert(name, hypernode_name):
    """
    Generates a openvpn client certificate using the easy-rsa pkitool
    :param name: The name of the client certificate to generate (eg. db-at-office)
    :param hypernode_name: The name of the hypernode (eg. erikhyperdev.hypernode.io)
    :return: int: The pkitool command exit code
    """
    easy_rsa_dir = "/etc/openvpn/easy-rsa"
    command = [
        "/bin/bash",
        "-c",
        ". {e_rsa_dir}/vars && {e_rsa_dir}/pkitool {name}.{hn_name}".format(
            e_rsa_dir=easy_rsa_dir, name=name, hn_name=hypernode_name
        ),
    ]
    process = Popen(command, stdout=PIPE)
    output, error = process.communicate(timeout=90)
    if process.returncode is not None:
        logger.info(f"Openvpn easy-rsa output: {output}")
        return process.returncode
    else:
        logger.error(f"OpenVPN easy-rsa output: \nstdout: {output} \nstderr: {error}")
        raise RuntimeError("Easy-rsa command timeout during certificate generation!")


def generate_client_config(name, hypernode_name, path=DEFAULT_PATH):
    """
    Get all needed certificates from disk and paste them into the template, then write template to "path"
    :param name: The name of the certificate to get (eg. db-at-office)
    :param hypernode_name: the name of the hypernode (eg. erikhyperdev.hypernode.io)
    :param path: the path to write the client config to
    :return: bool: True if success, False if failure
    """
    try:
        ta, ca, cert, key = get_client_certificate(".".join((name, hypernode_name)))
    except OSError as e:
        logger.exception(
            "FAIL: Unable to get client certificate for {}. Exited with error: {}".format(
                hypernode_name, e
            )
        )
        return False
    try:
        with open(f"{path}/{name}.{hypernode_name}.ovpn", "w") as client_config_file:
            client_config_file.write(
                TEMPLATE.format(
                    hn_name=hypernode_name, cert=cert, key=key, ca=ca, tls=ta
                )
            )
    except OSError as e:
        logger.exception(
            "FAIL: Unable to write client certificate for {}. Exited with error: {}.".format(
                hypernode_name, e
            )
        )
        return False
    return True


def _remove_file(path):
    """
    Attempt to remove a file and report if it was successful.
    :param str path: The path to remove
    :return bool success: True if successful,
    False if an IOError was encountered.
    """
    try:
        remove(path)
        return True
    except OSError:
        return False


def remove_client_config(name, hypernode_name, path=DEFAULT_PATH):
    """
    Remove the client config
    :param name: str: The name of the custom client config to remove
    :param str hypernode_name: The hypernode name. Like 'banaan.hypernode.io'
    :param str path: The path to prepend to the config file.
    :return bool success: True if successful,
    False if an IOError was encountered.
    """
    client_config_path = f"{path}/{name}.{hypernode_name}.ovpn"
    return _remove_file(client_config_path)


def remove_openvpn_client_config(name, hypernode_name, path=OPENVPN_CLIENT_CONFIG_DIR):
    """
    Remove the openVPN client config
    :param name: str: The name of the custom client config to remove
    :param str hypernode_name: The hypernode name. Like 'banaan.hypernode.io'
    :param str path: The path to prepend to the config file.
    :return bool success: True if successful,
    False if an IOError was encountered.
    """
    openvpn_config_path = f"{path}/{name}.{hypernode_name}"
    return _remove_file(openvpn_config_path)


def place_openvpn_client_config(
    name, hypernode_name, ip, path=OPENVPN_CLIENT_CONFIG_DIR
):
    try:
        with open(f"{path}/{name}.{hypernode_name}", "w") as client_config:
            client_config.write(OPENVPN_CLIENT_CONFIG_TEMPLATE.format(ip))
            return True
    except OSError:
        return False
