import re
from argparse import ArgumentParser
from subprocess import CalledProcessError, check_call, check_output


def get_php_services():
    """
    Get a list of all the installed php services on the Hypernode.
    :return list: list of the php versions e.g. ['7.3', '7.1']
    """
    try:
        output = (
            check_output(
                "/usr/bin/find /lib/systemd/system/php*-fpm.service | grep -o '[0-9][\\.][0-9]\\+'",
                shell=True,
            )
            .decode()
            .strip()
        )

        return output.split("\n") if output else []
    except CalledProcessError:
        return []


def get_php_alternatives():
    """
    Get a list of all the installed php alternatives on the Hypernode.
    :return list: list of the php versions e.g. ['7.3', '7.1']
    """
    try:
        output = (
            check_output(
                "/usr/bin/update-alternatives --list php",
                shell=True,
            )
            .decode()
            .strip()
        )

        if not output:
            return []

        pattern = r"([0-9]+.[0-9]+)$"
        return re.findall(pattern, output, re.MULTILINE)
    except CalledProcessError:
        return []


def process_args():
    """
    Process the arguments given to the php switch command
    :return ArgumentParser: parsed arguments for the switch php command
    """
    parser = ArgumentParser(description="Set the PHP version")
    parser.add_argument(
        "version", help="The PHP version", choices=get_php_alternatives()
    )
    parser.add_argument(
        "--skip-fpm",
        action="store_true",
        help="Skip enabling and starting the php-fpm service",
    )
    return parser.parse_args()


def disable_fpm_service(old_version):
    """
    Disable php fpm for a php version
    :param str old_version: the php version to disable fpm for
    """
    check_call(["systemctl", "disable", "php{}-fpm".format(old_version)])


def stop_fpm_service(old_version):
    """
    Stop php fpm for a php version
    :param str old_version: the php version to stop fpm for
    """
    check_call(["systemctl", "stop", "php{}-fpm".format(old_version)])


def _get_path_for_php_version(version):
    """
    Get the path for php executable of :version
    :param str version: the php version to fetch the path of
    :return str: the php executable path of given php version
    """
    php_location = check_output(["/usr/bin/which", "php" + version]).strip()

    return php_location


def update_alternatives(new_version):
    """
    Set the node to a new php version
    :param str new_version: the new php version to set for the node
    :return bool: True if alternative was updated, False if not
    """
    new_php_location = _get_path_for_php_version(new_version)

    output = check_output(
        [
            "/usr/bin/update-alternatives",
            "--set",
            "php",
            new_php_location,
        ]
    )

    return bool(output)


def enable_fpm_service(new_version):
    """
    Enable php fpm for a php version
    :param str new_version: the php version to enable fpm for
    """
    check_call(["/usr/bin/systemctl", "enable", "php{}-fpm".format(new_version)])


def start_fpm_service(new_version):
    """
    Start php fpm for a php version
    :param str new_version: the php version to start fpm for
    """
    check_call(["/usr/bin/systemctl", "start", "php{}-fpm".format(new_version)])


def clean_wsdl_cache():
    """
    Remove WSDL cache files from /tmp directory.
    This is important to do, because WSDL cache files may be incompatible between PHP versions.
    Incompatible WSDL cache files result in fatal errors on the application side.
    """
    check_call(
        [
            "/usr/bin/find",
            "/tmp",
            "-mindepth",
            "1",
            "-maxdepth",
            "1",
            "-type",
            "f",
            "-name",
            "wsdl-*",
            "-delete",
        ]
    )


def switch_php(new_version: str, skip_fpm: bool = False):
    """
    Switch the PHP version for this node
    :param str new_version: The PHP version to switch this node to
    :param bool skip_fpm: Skip managing the php-fpm service
    """
    php_services = get_php_services()
    php_services_to_disable = [
        service for service in php_services if service != new_version
    ]

    if not skip_fpm:
        for old_version in php_services_to_disable:
            disable_fpm_service(old_version)
            stop_fpm_service(old_version)

    if update_alternatives(new_version):
        print("Detected PHP version change, clearing WSDL cache")
        clean_wsdl_cache()
    else:
        print("PHP version did not change")

    if new_version in php_services and not skip_fpm:
        enable_fpm_service(new_version)
        start_fpm_service(new_version)


def main():
    """
    Main function for the php switch command
    """
    args = process_args()
    switch_php(new_version=args.version, skip_fpm=args.skip_fpm)


if __name__ == "__main__":
    main()
