import os
from difflib import Differ
from hashlib import md5
from logging import getLogger
from shutil import copy
from typing import Iterator, List, NewType, Optional, cast

from colorama import Fore

from kamikaze3.privileged.manage_vhosts import program_name, settings
from kamikaze3.privileged.manage_vhosts.render import render_template

logger = getLogger(program_name)
Checksum = NewType("Checksum", str)


def get_savefile_path(filepath: str) -> str:
    """
    Get the path of the savefile relative to the given path
    :param str filepath: Path to the nginx file owned by app
    :rtype: str
    """
    return settings.LAST_TEMPLATE_MATCH_PATH.format(
        filepath.replace(settings.APP_NGINX_DIR, "")
    )


def color_diff(diff: List[str]) -> Iterator[str]:
    """
    Color lines based on the difflib prefix
    :param [str, ..] diff: List of lines generated by difflib
    :rtype: [str, ..]
    """
    for line in diff:
        if line.startswith("+"):
            yield Fore.GREEN + line + Fore.RESET
        elif line.startswith("-"):
            yield Fore.RED + line + Fore.RESET
        elif line.startswith("^"):
            yield Fore.BLUE + line + Fore.RESET
        else:
            yield line


def get_mismatch_diff(filepath: str) -> str:
    """
    Get a printable diff between the last template match and the current file
    :param str filepath: Path to the nginx file owned by app
    :rtype: str
    """
    try:
        with open(filepath, "r") as fp:
            file_lines = [l for l in fp.read().split("\n") if should_be_checksummed(l)]
    except FileNotFoundError:
        logger.debug("{} does not exist, continuing diff with no lines")
        file_lines = []

    try:
        with open(get_savefile_path(filepath), "r") as fp:
            last_template_match_lines = [
                l for l in fp.read().split("\n") if should_be_checksummed(l)
            ]
    except FileNotFoundError:
        logger.debug("Savefile {} does not exist, continuing diff with no lines")
        last_template_match_lines = []
    diff = Differ().compare(last_template_match_lines, file_lines)
    return "\n".join(color_diff(diff))


def get_checksum_line(filepath: str) -> Optional[Checksum]:
    """
    Check if there is a checksum and return it if present
    :param str filepath: Path to the nginx file owned by app
    :rtype: str|None
    """
    with open(filepath, "r") as fp:
        firstline = fp.readline()
        match = settings.CHECKSUM_COMPILED_PATTERN.match(firstline)
        if match:
            checksum = cast(Checksum, match.group(1))
            logger.debug("Found inline checksum for {}: {}".format(filepath, checksum))
            return checksum
    logger.debug("No inline checksum found for {}".format(filepath))


def place_checksum_line(filepath: str, checksum: Checksum):
    """
    Add an inline checksum on the first line of the given file
    :param str filepath: Path to the nginx file owned by app
    :param str checksum: Checksum to be placed in the file
    :rtype: None
    """
    logger.debug("Placing checksum {} in {}".format(checksum, filepath))
    with open(filepath, "r+") as fp:
        contents = fp.read()
        fp.seek(0, 0)
        checksum_line = settings.CHECKSUM_LINE_FORMAT.format(checksum=checksum)
        fp.write(checksum_line + contents)


def replace_checksum_line(filepath: str, checksum: Checksum):
    """
    Replace the existing checksum line in the given file
    :param str filepath: Path to the nginx file owned by app
    :param str checksum: Checksum to be placed in the file
    :rtype: None
    """
    logger.debug("Updating checksum in {} to {}".format(filepath, checksum))
    with open(filepath, "r") as fp:
        lines = fp.readlines()
    if not settings.CHECKSUM_COMPILED_PATTERN.match(lines[0]):
        # TODO: Isn't raising an exception a bit too rigorous?
        raise RuntimeError(
            "No checksum found while replacing checksum in {}".format(filepath)
        )

    new_checksum_line = settings.CHECKSUM_LINE_FORMAT.format(checksum=checksum)
    lines[0] = new_checksum_line
    with open(filepath, "w") as fp:
        fp.writelines(lines)


def ensure_savefile(filepath: str, template_path: str, config: dict):
    """
    Ensure that a savefile exists for the given file.
    If there is no savefile present, place a rendered template in the savefile
    This is important to make sure that there is a savefile present before doing
    a diff, otherwise there is no comparison to be made to the original file if
    there is a mismatch.
    :param str filepath: Path to the nginx file owned by app
    :param str template_path: Path to the template for generating the nginx file
    :param dict config: VHost configuration that corresponds to the filepath
    :rtype: None
    """
    save_filepath = get_savefile_path(filepath)
    if os.path.isfile(save_filepath):
        return
    logger.debug(
        "No savefile found for {}, creating savefile based on template".format(filepath)
    )
    os.makedirs(os.path.dirname(save_filepath), exist_ok=True)
    rendered_template = render_template(template_path, config)
    with open(save_filepath, "w") as fp:
        fp.write(rendered_template)


def copy_file_to_savefile(filepath: str):
    """
    Save file that matches the template for later diff comparison
    :param str filepath: Path to the nginx file owned by app
    :rtype: None
    """
    save_filepath = get_savefile_path(filepath)
    os.makedirs(os.path.dirname(save_filepath), exist_ok=True)
    # TODO: Prevent the user from changing this file, but don't place a wildcard in the sudoers file
    # like in this commit: c292a49
    copy(filepath, save_filepath)


def ensure_checksum_line(filepath: str, checksum: Checksum):
    """
    Ensure that a checksum line is present with the given checksum
    :param str filepath: Path to the nginx file owned by app
    :param str checksum: Checksum to be placed in the file
    :rtype: None
    """
    if get_checksum_line(filepath):
        replace_checksum_line(filepath, checksum)
    else:
        place_checksum_line(filepath, checksum)
    copy_file_to_savefile(filepath)


def get_rendered_template_checksum(template_path: str, config: dict) -> Checksum:
    rendered = render_template(template_path, config)
    rendered_lines = rendered.split("\n")
    filtered_lines = [l.strip() for l in rendered_lines if should_be_checksummed(l)]
    checksum = checksumlines(filtered_lines)
    logger.debug(
        "Calculated checksum for rendered template {}: {}".format(
            template_path, checksum
        )
    )
    return checksum


def get_file_checksum(filepath: str) -> Checksum:
    checksummable_lines = []
    with open(filepath, "r") as fp:
        for l in fp.readlines():
            stripped_line = l.strip()
            if should_be_checksummed(stripped_line):
                checksummable_lines.append(stripped_line)
    checksum = checksumlines(checksummable_lines)
    logger.debug("Calculated checksum for {}: {}".format(filepath, checksum))
    return checksum


def should_be_checksummed(line: str) -> bool:
    """
    Filter out whether or not the line should be part of the contents that is checksummed
    :param str line: Line from an nginx config file
    :rtype: bool
    """
    # The checksum line matches the comment pattern so it does not need to be filtered here
    if settings.COMMENT_PATTERN.match(line):
        return False
    if settings.EMPTY_LINE_PATTERN.match(line):
        return False
    return True


def checksumlines(lines: List[str]) -> Checksum:
    """
    Convert a list of filelines to a checksum
    :param [str, ..] lines: List of filelines that are stripped of contents that are not relevant
    to the checksum
    :rtype: str
    """
    # MD5 requires bytestrings, so encoding is necessary
    combined_lines = "".join(lines).encode()
    return md5(combined_lines).hexdigest()


def does_inline_checksum_match(
    filepath: str, file_checksum: Checksum
) -> Optional[bool]:
    """
    If checksum line is present, check if it matches the current file
    If checksum matches, *automatically change varnish*
    If checksum doesn't match, *cannot automatically change varnish*
    :param str filepath: Path to the nginx file owned by app
    :param str file_checksum: Checksum to be compared to the in-line checksum
    :return bool|None: Bool if in-line checksum is found, otherwise None
    """
    line_checksum = get_checksum_line(filepath)
    if line_checksum:
        if file_checksum == line_checksum:
            logger.debug(
                "{} checksum matches in-line checksum: {}".format(
                    filepath, line_checksum
                )
            )
            return True
        logger.debug(
            "Checksum {} for file {} does not match the in-line checksum {}".format(
                file_checksum, filepath, line_checksum
            )
        )
        return False


def does_rendered_template_checksum_match(
    template_path: str, filepath: str, config: dict, file_checksum: Checksum
) -> bool:
    """
    If checksum line not present, check if file matches template
    If file matches template, place an inline checksum that matches the pattern
    and *automatically change varnish*
    If file does not match, *cannot automatically change varnish*
    :param str template_path: Path to the template for generating the nginx file
    :param str filepath: Path to the nginx file owned by app
    :param dict config: VHost configuration that corresponds to the
    :param str file_checksum: Checksum to be compared to the template checksum
    :rtype: bool
    """
    rendered_template_checksum = get_rendered_template_checksum(
        template_path=template_path, config=config
    )
    if rendered_template_checksum == file_checksum:
        ensure_checksum_line(filepath, file_checksum)
        return True
    logger.debug(
        "Cannot automatically change config for {} because template checksum {} does not match file checksum "
        "{}".format(filepath, rendered_template_checksum, file_checksum)
    )
    return False


def is_valid_checksum(template_path: str, filepath: str, config: dict) -> bool:
    """
    Check if the file related to the template still uses the original template by comparing the checksum from either
    the template itself or a checksum line if present.
    :param str template_path: Path to the template for generating the nginx file
    :param str filepath: Path to the nginx file owned by app
    :param dict config: VHost configuration that corresponds to the
    :rtype: bool
    """
    try:
        file_checksum = get_file_checksum(filepath)
    except FileNotFoundError:
        logger.debug("No file to checksum at {}".format(filepath))
        return False
    inline_cs_match = does_inline_checksum_match(
        filepath=filepath, file_checksum=file_checksum
    )
    if inline_cs_match is not None:
        return inline_cs_match
    return does_rendered_template_checksum_match(
        template_path=template_path,
        filepath=filepath,
        config=config,
        file_checksum=file_checksum,
    )
