from logging import getLogger

from kamikaze3.privileged.manage_vhosts import program_name
from kamikaze3.privileged.manage_vhosts.state import (
    get_all_configs,
    get_current_config,
    update_statefile,
)

logger = getLogger(program_name)


def get_default_server(config):
    """
    Retrieve the servername that is the current default_server.
    :param dict config: Config of all managed VHosts
    :rtype: str|None
    """
    for servername, servername_config in config.items():
        if servername_config.get("default_server"):
            return servername


def manage_default_server(servername):
    """
    Manage the switch from default_server to the given servername by managing
    the configs of the relevant servernames in the statefile.
    :param str servername: Servername which is going to be the new default_server
    :return [str, ..]: List of updated servernames for which the configs should
    be regenerated.
    """
    all_configs = get_all_configs()
    old_default_servername = get_default_server(all_configs)
    new_default_server_config = all_configs.get(servername) or get_current_config(
        servername
    )
    new_default_server_config["default_server"] = True
    all_configs[servername] = new_default_server_config
    updated_servernames = [servername]

    if old_default_servername != servername:
        logger.info(
            "Updating default_server from {} to {}".format(
                old_default_servername, servername
            )
        )
        if old_default_servername:
            all_configs[old_default_servername]["default_server"] = False
            updated_servernames.append(old_default_servername)

        for sn in updated_servernames:
            all_configs[sn][
                "servername"
            ] = sn  # To make sure the config is saved under the correct servername
            update_statefile(all_configs[sn])
    return updated_servernames
