import json
from logging import getLogger
from os import EX_OK, EX_TEMPFAIL
from typing import List

from prettytable import PrettyTable

from kamikaze3.privileged.manage_vhosts import program_name, settings
from kamikaze3.privileged.manage_vhosts.state import get_all_configs

logger = getLogger(program_name)


def _print_table(vhost_configs, headers=None):
    """
    Print config in table format
    :param dict vhost_configs: Vhost configs that need to be printed
    :param [str, ..] headers: List of headers that should be printed
    :rtype: None
    """
    headers = headers or settings.PRINT_HEADERS
    vhost_info_table = PrettyTable(headers)
    sorted_servernames = sorted(vhost_configs.keys())
    for servername in sorted_servernames:
        vhost_info_table.add_row([str(vhost_configs[servername][k]) for k in headers])
    print(vhost_info_table)


def _print_json(vhost_configs, headers=None):
    """
    Print config in json format
    :param dict vhost_configs: Vhost configs that need to be printed
    :param [str, ..] headers: List of headers that should be printed
    :rtype: None
    """
    headers = headers or settings.PRINT_HEADERS
    json_data = {}
    sorted_servernames = sorted(vhost_configs.keys())
    for servername in sorted_servernames:
        json_data[servername] = {
            # If the statefile is somehow missing a key (perhaps we added more keys and
            # this is an old configuration) print 'Unconfigured'. For example, if 'default_server'
            # is missing we can assume "False" but for other options it might not be so clear.
            # Unconfigured is a good safe catchall and it will indicate to the operator that
            # something fishy might be going on instead of throwing a big fat keyerror.
            k: vhost_configs[servername].get(k, "Unconfigured")
            for k in headers
            if k not in ["servername"]
        }
    print(json.dumps(json_data, indent=2, sort_keys=True))


def _print_raw(vhost_configs, headers=None, delim=";", show_headers=True):
    """
    Print config in raw format
    :param dict vhost_configs: Vhost configs that need to be printed
    :param [str, ..] headers: List of headers that should be printed
    :param str delim: String that should function as delimiter between fields
    :param bool show_headers: Whether or not to print the headers
    :rtype: None
    """
    headers = headers or settings.PRINT_HEADERS
    if show_headers:
        print(delim.join(headers))
    sorted_servernames = sorted(vhost_configs.keys())
    for servername in sorted_servernames:
        print(delim.join([str(vhost_configs[servername][k]) for k in headers]))


def get_active_headers(vhost_configs: dict) -> List[str]:
    """
    Get a list of all headers that are used by the current config.
    Only check for relevant headers defined in PRINT_HEADERS
    :param dict vhost_configs: Vhost configs that need to be printed
    :rtype: [str, ..]
    """
    active_headers = []
    for header in settings.PRINT_HEADERS:
        if all([header in c for c in vhost_configs.values()]):
            active_headers.append(header)
    return active_headers


def list_vhosts(servernames=None, list_format="table"):
    """
    Show a pretty list of all active VHosts with info
    :param str list_format: What format should be applied to print the VHost configs
    :param [str, ...] servernames: List of servernames to pretty print
    :return int: Exitcode
    """
    all_configs = get_all_configs()
    if servernames:
        vhost_configs = {}
        for servername in servernames:
            if servername not in all_configs.keys():
                logger.warning(
                    "Could not find {} as an existing managed VHost".format(servername)
                )
                return EX_TEMPFAIL
            vhost_configs[servername] = all_configs[servername]
    else:
        vhost_configs = all_configs
    active_headers = get_active_headers(vhost_configs)

    if list_format == "table":
        _print_table(vhost_configs, headers=active_headers)
    elif list_format == "json":
        _print_json(vhost_configs, headers=active_headers)
    elif list_format == "raw":
        _print_raw(vhost_configs, headers=active_headers)
    else:
        logger.warning("Unknown format option {}".format(list_format))
        return EX_TEMPFAIL
    return EX_OK
