import copy
import os
import shutil
from contextlib import suppress
from logging import getLogger
from os import listdir, makedirs, mkdir, path, remove, unlink
from os.path import basename, dirname, isfile
from typing import Optional

from kamikaze3.common.letsencrypt import (
    DehydratedError,
    manage_letsencrypt_certificate,
    remove_letsencrypt_config,
)
from kamikaze3.common.ssl import (
    APP_SSL_PATH_PATTERN,
    ROOT_SSL_PATH_PATTERN,
    get_servername_ssl_cert,
)
from kamikaze3.privileged.manage_vhosts import program_name, settings
from kamikaze3.privileged.manage_vhosts.checksum import (
    ensure_checksum_line,
    ensure_savefile,
    get_file_checksum,
    get_mismatch_diff,
    get_rendered_template_checksum,
    get_savefile_path,
    is_valid_checksum,
)
from kamikaze3.privileged.manage_vhosts.render import render_template
from kamikaze3.privileged.manage_vhosts.state import (
    _is_legit_to_quit,
    get_all_configs,
    get_current_config,
    safe_write_root_file,
)
from kamikaze3.privileged.manage_vhosts.system_config import SystemConfig

# Nginx config locations
from kamikaze3.utils import noop, request_confirmation

logger = getLogger(program_name)


def backup_appconfig_file(configfile: str, timestamp: str):
    """
    Save backup of the given appconfig file so that the original file can be overwritten
    :param str configfile: Path to app nginx config file
    :param str timestamp: Timestamp at which the config is updated
    :rtype: None
    """
    if not isfile(configfile):
        logger.info("Couldn't find {}, skipping backup".format(configfile))
        return
    configname = basename(dirname(configfile))
    backup_file = settings.APP_HMV_BACKUP_DIR.format(
        configname=configname, timestamp=timestamp, configfile=basename(configfile)
    )
    makedirs(path.dirname(backup_file), exist_ok=True)
    logger.info("Backing up {} to {}".format(configfile, backup_file))
    shutil.copy(configfile, backup_file)


def does_appconfig_change(configfile: str, templatefile: str, config: dict) -> bool:
    """
    Check if the appconfig would change based on the last saved template file
    :param str configfile: Path to app nginx config file
    :param str templatefile: Jinja template filepath
    :param dict config: Configuration used to fill in the template variables
    :rtype: bool
    """
    last_saved_template = get_savefile_path(configfile)
    if not isfile(last_saved_template):
        logger.debug("Could not verify if {} would change".format(configfile))
        return True
    rendered_template_checksum = get_rendered_template_checksum(
        template_path=templatefile, config=config
    )
    file_checksum = get_file_checksum(last_saved_template)
    return file_checksum != rendered_template_checksum


def _should_appconfig_be_written(
    app_servername_dir_exists: bool,
    configfile_path: str,
    abs_template_filename: str,
    new_config: dict,
    automatic_confirmation: Optional[str] = None,
) -> bool:
    if not app_servername_dir_exists:
        logger.debug("Placing {} for new VHost".format(configfile_path))
        return True

    # Don't overwrite config if the template wouldn't change
    if not does_appconfig_change(configfile_path, abs_template_filename, new_config):
        logger.debug("No update required for {}".format(configfile_path))
        return False

    if is_valid_checksum(
        template_path=abs_template_filename, filepath=configfile_path, config=new_config
    ):
        config_should_be_written = True
        logger.debug("Updating {} for existing VHost".format(configfile_path))
    else:
        ensure_savefile(
            filepath=configfile_path,
            template_path=abs_template_filename,
            config=new_config,
        )
        logger.info(
            "Cannot automatically update config for {} because user changes would be overwritten".format(
                configfile_path
            )
        )
        logger.info(
            "Found the following changes: \n\n{}\n\n".format(
                get_mismatch_diff(configfile_path)
            )
        )
        config_should_be_written = False
        if not automatic_confirmation:
            config_should_be_written = request_confirmation(
                prompt="Do you want to backup your changes and overwrite the file? (yes/no) ",
                func=noop,
            )
        if not config_should_be_written and automatic_confirmation in [None, "no"]:
            logger.info("Skipping update of {}".format(configfile_path))
        else:
            config_should_be_written = True
            backup_appconfig_file(
                configfile=configfile_path, timestamp=new_config["updated_at"]
            )
            logger.info("Forcing update of {}".format(configfile_path))

    return config_should_be_written


def does_type_change(new_config: dict) -> bool:
    """
    Check if the type changes compared to the current type
    :param dict new_config: Config with new changes compared to the current situation
    :rtype: bool
    """
    all_configs = get_all_configs()
    if new_config["servername"] in all_configs:
        old_type = all_configs[new_config["servername"]]["type"]
        new_type = new_config["type"]
        type_changes = new_type != old_type
        if type_changes:
            logger.info(
                "Found type change from {} to {} for {}".format(
                    old_type, new_type, new_config["servername"]
                )
            )
        return type_changes
    # Servername doesn't exist yet so type definitely doesn't change
    return False


def overhaul_type(new_config: dict, automatic_confirmation=None) -> dict:
    """
    Type is changed so ask for confirmation to overwrite the current type.
    On confirmation, remove the app config directory for the given servername.
    If no confirmation is given, revert the type change.
    :param dict new_config: Config containing the new type
    :param str|None automatic_confirmation: Possible confirmation given at parse_args. Can
    be either 'yes', 'no' or None
    :return dict: Updated dict with the correct type
    """
    overwrite_type = automatic_confirmation == "yes"
    config_dir = settings.APP_SERVERNAME_DIR.format(new_config["config_name"])
    # Ask for confirmation if not automatic 'yes' or 'no'
    if not automatic_confirmation:
        overwrite_type = request_confirmation(
            prompt="Do you want to change the type for {} to {} and overwrite all existing files in {}?. (yes/no) ".format(
                new_config["servername"],
                new_config["type"],
                config_dir,
            ),
            func=noop,
        )

    if not overwrite_type:
        logger.info("Not changing type for {}".format(new_config["servername"]))
        # Update config so that the old type will be saved in the statefile
        # If we don't, the new type will be reflected in the statefile after the
        # manage-vhost part is done in the main loop.
        current_config = get_current_config(new_config["servername"])
        new_config["type"] = current_config["type"]
    else:
        trimmed_config_dir = config_dir.rstrip("/")
        if os.path.islink(trimmed_config_dir):
            logger.warning(
                "VHost {} is a symlink and links to {}. Removing symlink and creating VHost directory.".format(
                    new_config["config_name"], os.path.realpath(trimmed_config_dir)
                )
            )
            unlink(trimmed_config_dir)
        else:
            logger.info(
                "Removing all files in {} to update type to {} for {}".format(
                    config_dir, new_config["type"], new_config["servername"]
                )
            )
            shutil.rmtree(config_dir, ignore_errors=True)
    return new_config


def manage_http_config(new_config, automatic_confirmation=None) -> dict:
    """
    Manage the VHost http config on the node based on the given setup. Returns (potentially mutated) config.
    :param dict new_config: VHost parameters
    :param str|None automatic_confirmation: Whether or not we can assume an answer to confirmation requests
    :return dict: Updated config
    """
    if does_type_change(new_config):
        new_config = overhaul_type(
            new_config, automatic_confirmation=automatic_confirmation
        )

    # Place /data/web/nginx/{}
    config_dir = settings.APP_SERVERNAME_DIR.format(new_config["config_name"])
    try:
        mkdir(config_dir)
    except FileExistsError:
        app_servername_dir_exists = True
    else:
        app_servername_dir_exists = False

    system_config = SystemConfig()
    render_config = copy.deepcopy(new_config)
    render_config["varnish"] = render_config.get(
        "varnish", False
    ) and system_config.get_bool("varnish", "enabled", True)
    render_config["https"] = render_config.get(
        "https", False
    ) and system_config.get_bool("https", "enabled", True)
    render_config["force_https"] = (
        render_config.get("force_https", False) and render_config["https"]
    )
    # Try to place/update /data/web/nginx/{}/ defaults and CMS defaults if dir is empty
    template_dirs = [
        settings.SERVERNAME_CONFIG_TEMPLATE_DIR,
        settings.TYPE_CONFIG_TEMPLATE_DIR.format(render_config["type"]),
    ]
    for template_dir in template_dirs:
        templates = listdir(template_dir)
        for template_filename in templates:
            abs_template_filename = path.join(template_dir, template_filename)
            configfile_path = path.join(config_dir, template_filename)
            if configfile_path.endswith(".j2"):
                configfile_path = configfile_path[: -len(".j2")]
            if not isfile(abs_template_filename):
                # Skip directories
                continue

            if _should_appconfig_be_written(
                app_servername_dir_exists=app_servername_dir_exists,
                configfile_path=configfile_path,
                abs_template_filename=abs_template_filename,
                new_config=render_config,
                automatic_confirmation=automatic_confirmation,
            ):
                write_conf(
                    render_config,
                    abs_template_filename,
                    configfile_path,
                    update_checksum=True,
                )

    # Place /etc/nginx/sites/http.{}.conf from http_force_https.conf if force_https is True
    # Place /etc/nginx/sites/http.{}.conf from http.conf if force_https is NOT defined
    write_conf(
        render_config,
        settings.SITE_TEMPLATES
        + (
            "http_force_https.conf.j2"
            if render_config["force_https"]
            else "http.conf.j2"
        ),
        settings.HTTP_SITE_CONFIG.format(render_config["config_name"]),
        as_root=True,
    )
    logger.info("Writing HTTP config for {}".format(render_config["servername"]))

    return new_config


def manage_varnish_config(new_config):
    """
    Manage the VHost varnish config on the node based on the given setup.
    :param dict new_config: VHost parameters
    :rtype: None
    """
    # Place /etc/nginx/sites/varnish.{}.conf from varnish.conf if varnish is True
    system_config = SystemConfig()
    if new_config["varnish"] and system_config.get_bool("varnish", "enabled", True):
        write_conf(
            new_config,
            settings.SITE_TEMPLATES + "varnish.conf.j2",
            settings.VARNISH_SITE_CONFIG.format(new_config["config_name"]),
            as_root=True,
        )
        logger.info("Writing varnish config for {}".format(new_config["servername"]))
    else:
        with suppress(FileNotFoundError):
            remove(settings.VARNISH_SITE_CONFIG.format(new_config["config_name"]))


def get_letsencrypt_certificate_file(new_config: dict) -> Optional[str]:
    """
    Retrieve a letsencrypt certificate file if possible
    :param dict new_config: VHost parameters
    :rtype: str|None
    """
    servername = new_config["servername"]
    existing_app_cert = get_servername_ssl_cert(servername, APP_SSL_PATH_PATTERN)
    if existing_app_cert and new_config.get("ssl_noclobber"):
        # Don't touch the certificate
        return existing_app_cert
    elif new_config["is_wildcard"]:
        logger.info(
            "Skipping Lets Encrypt certificate management for {} because we don't yet support wildcard "
            "certificates via Lets Encrypt".format(servername)
        )
        # A cert provided by the customer, or None
        return existing_app_cert
    else:
        try:
            manage_letsencrypt_certificate(new_config)
            # If existing_app_cert is None and our letsencrypt client did not encounter any errors,
            # we expect it will have placed a valid certificate at APP_SSL_PATH
            return existing_app_cert or settings.APP_SSL_CERT_PATH.format(servername)
        except DehydratedError:
            return None


def manage_https_config(new_config):
    """
    Manage the VHost https config on the node based on the given setup. Returns (potentially mutated) config.
    :param dict new_config: VHost parameters
    :return dict: Updated config
    """
    system_config = SystemConfig()
    configname = new_config["config_name"]
    if new_config["https"] and system_config.get_bool("https", "enabled", True):
        servername = new_config["servername"]
        certificate_file = get_servername_ssl_cert(servername, ROOT_SSL_PATH_PATTERN)
        if not certificate_file:
            certificate_file = get_letsencrypt_certificate_file(new_config)
        new_config["ssl_cert_path"] = certificate_file

        if not certificate_file:
            logger.warning(
                "Cannot place HTTPS config for {} as there is no valid SSL certificate present! "
                "Removing HTTPS config and any forced redirect to HTTPS!".format(
                    servername
                )
            )
            delete_https_config(configname)
            if new_config["force_https"]:
                # we have to undo the 'force_https' config placed by manage_http_config here
                write_conf(
                    new_config,
                    settings.SITE_TEMPLATES + "http.conf.j2",
                    settings.HTTP_SITE_CONFIG.format(configname),
                    as_root=True,
                )
        else:
            write_conf(
                new_config,
                settings.SITE_TEMPLATES + "https.conf.j2",
                settings.HTTPS_SITE_CONFIG.format(new_config["config_name"]),
                as_root=True,
            )
            logger.info("Writing HTTPS config for {}".format(servername))
    else:
        delete_https_config(configname)
    return new_config


def delete_https_config(configname):
    """
    Remove the HTTPS config for servername
    :param str servername: configname to remove configs for
    :rtype: None
    """
    with suppress(FileNotFoundError):
        remove(settings.HTTPS_SITE_CONFIG.format(configname))


def delete_vhost(servername):
    """
    Remove all configurations for the given servername
    :param str servername: Servername to remove configs for
    :rtype: None
    """
    if not _is_legit_to_quit(servername):
        return

    config = get_current_config(servername)
    deprecated_configs = [
        settings.HTTP_SITE_CONFIG,
        settings.HTTPS_SITE_CONFIG,
        settings.VARNISH_SITE_CONFIG,
    ]
    logger.info("Removing configs for {}".format(servername))
    # TODO: Create backup of vhost with backup_appconfig_file
    for deprecated_config in deprecated_configs:
        with suppress(FileNotFoundError):
            remove(deprecated_config.format(config["config_name"]))
    servername_dir = settings.APP_SERVERNAME_DIR.format(config["config_name"]).rstrip(
        "/"
    )
    if path.islink(servername_dir):
        logger.info(
            "Servername dir {} is a symlink, not removing target "
            "only removing the symlink".format(servername_dir)
        )
        unlink(servername_dir)
    else:
        with suppress(FileNotFoundError):
            shutil.rmtree(servername_dir)
    logger.debug("Removing stored statefiles for {}".format(servername))
    with suppress(FileNotFoundError):
        shutil.rmtree(settings.LAST_TEMPLATE_MATCH_PATH.format(config["config_name"]))

    if not config["is_wildcard"]:
        remove_letsencrypt_config(servername)


def write_conf(config, src_template, dst_file, as_root=False, update_checksum=False):
    """
    Render the template to the destination using the given config
    :param dict config: Configuration used to fill in the template variables
    :param str src_template: Jinja template filepath
    :param str dst_file: Target path where the rendered file will be placed
    :param bool as_root: Write the file to the given path as root
    :param bool update_checksum: Calculate the checksum of the dstfile and write it
    to the dstfile
    :rtype: None
    """
    # Remove a possible .j2 extension from the dst file
    if dst_file.endswith(".j2"):
        dst_file = dst_file[:-3]

    rendered = render_template(src_template, config)
    if as_root:
        safe_write_root_file(dst_file, rendered)
    else:
        with open(dst_file, "w") as fp:
            fp.write(rendered)

    if update_checksum:
        # When writing over the app user's nginx config, update the checksum to reflect the changes.
        # This happens when the file did not exist yet or a file exists and the template was updated
        checksum = get_file_checksum(dst_file)
        ensure_checksum_line(dst_file, checksum=checksum)
