#!/usr/bin/env python3
"""
kamikaze3.privileged.oom_protect
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This wrapper to oom_protect commands
"""
import sys
import traceback
from contextlib import suppress
from logging import getLogger
from os import EX_SOFTWARE, EX_TEMPFAIL, chmod, getpid, remove, stat, getcwd
from stat import S_IXUSR
from subprocess import Popen, check_output

from kamikaze3.utils import exclusive_lock

PID = getpid()
SUDO = "/usr/bin/sudo"
APP_COMMAND_SCRIPT_FILE = "/tmp/hypernode_oom_protect{}".format(PID)
# This is placed in the debian/install because it needs to not
# be editable by the app user. This is a shell script because
# implementing the same wrapper in python would result in more
# (and less clear) code where we just spawn a sub-shell anyway.
WRAPPER_COMMAND_SCRIPT_FILE = "/usr/sbin/hypernode_oom_protect_wrapper"
logger = getLogger(__name__)


def write_executable_script(file_name, content):
    """
    Write an executable script to a file that everyone can execute
    :param str file_name: The file to write
    :param str content: The content of the script
    :return None:
    """
    with open(file_name, "w") as f:
        f.write(content)
    st = stat(file_name)
    # Make the temporary script executable by the user (app).
    chmod(file_name, st.st_mode | S_IXUSR)


def write_app_command_script_file(args):
    """
    Write the specified command to a script so we don't have to deal
    with any shell escaping from the su that the wrapper uses to run
    the command as the app user.
    :param Iterable args: list of CLI args
    :return None:
    """
    current_working_directory = getcwd()
    write_executable_script(
        APP_COMMAND_SCRIPT_FILE,
        "#!/bin/bash\n"
        "(cd {current_working_directory}; {args})\n".format(
            current_working_directory=current_working_directory, args=" ".join(args)
        ),
    )


def execute_protected_command():
    """
    Execute the protected command by setting the oom_score_adj to -1000
    :return obj proc: The running process
    """
    command = [SUDO, WRAPPER_COMMAND_SCRIPT_FILE, "{}".format(PID)]
    proc = Popen(command, stdout=sys.stdout, stderr=sys.stderr, stdin=sys.stdin)
    return proc


def wait_for_protected_command(proc):
    """
    Wait for the protected command to complete
    :param obj proc: the running protected command
    :return int exit_code: The command return code
    """
    proc.communicate()
    return proc.returncode


def clean_up_app_command_script_file():
    """
    Try to clean up the app command script file.
    If that fails that's OK too, we can suppress
    any error here.
    :return None:
    """
    with suppress(Exception):
        remove(APP_COMMAND_SCRIPT_FILE)


def run(args):
    """
    Run program with specified arguments. Returns a proper code as exit code.

    :param iter args: list of CLI args
    :return int: exit code
    """
    try:
        with exclusive_lock(APP_COMMAND_SCRIPT_FILE):
            write_app_command_script_file(args)
            proc = execute_protected_command()
        exit_code = wait_for_protected_command(proc)
    except KeyboardInterrupt:
        logger.info("Process interrupted")
        exit_code = EX_TEMPFAIL
    except Exception as error:
        logger.error("Uncaught error: {}".format(error))
        logger.debug(traceback.format_exc())
        exit_code = EX_SOFTWARE
    finally:
        clean_up_app_command_script_file()
    return exit_code


def main(args):
    """
    Execute the oom protect command wrapper
    :param iter args: list of CLI args
    :return int: exit code
    """
    help_menu_text = """Usage:
      hypernode-oom-protect [command] [arguments]

    Options:
      -h, --help                 Display this help message
      Still got questions? Go to: https://support.hypernode.com/changelog/release-6202-mark-processes-as-unkillable-when-out-of-memory/
    """
    if not args or args[0] in ("--help", "-h"):
        print(help_menu_text)
        sys.exit(0)
    # Explicit else specified because we can't set a side effect on sys.exit
    # when we patch it. It is probably special because it is a low level python
    # function (real signature unknown). And we do want to test for that run is
    # not called when --help is specified.
    else:
        return run(args)


if __name__ == "__main__":
    arguments = sys.argv[1:]
    sys.exit(main(arguments))
