import json
import logging
import os
from argparse import ArgumentParser
from pathlib import Path
from typing import Dict, List

from kamikaze3.common import systemd
from kamikaze3.common.env import get_hypernode_name
from kamikaze3.common.filesystem import (
    compare_file_hashes_to_earlier_defined_list,
    get_file_filtered_hashes,
)
from kamikaze3.common.log import create_stream_handler
from kamikaze3.common.system import user_is_root
from kamikaze3.security.slack_security import try_send_slack_security_message

logger = logging.getLogger(__name__)
# Filter pattern to remove minutes and/or sleep seconds from crontabs
CRONTAB_FILTER = "( sleep [0-9]+;|^[# ]*[\\*0-9/]+ [\\*0-9/]+)"


def parse_arguments():
    parser = ArgumentParser(description="Check the crontab files for integrity")
    mutex_group = parser.add_mutually_exclusive_group(required=True)
    mutex_group.add_argument(
        "--json-string",
        help="json of the legal crontab file names and their hashes",
        type=str,
    )
    mutex_group.add_argument(
        "--json-file",
        help="The hash list can also be passed as a path to a json file with this argument",
        type=str,
    )
    mutex_group.add_argument(
        "--generate",
        help="Generate a list of legal crontab files and their hashes",
        action="store_true",
    )
    parser.add_argument(
        "--storepath",
        help="Where to store the generated legal crontab file. Default: $(pwd)/crontab_hashes.json",
        default=os.getcwd() + "/crontab_hashes.json",
        type=str,
    )
    parser.add_argument(
        "--slack",
        help="Post the results in a slack channel (only works as root)",
        type=str,
    )
    return parser.parse_args()


def get_crontab_file_names() -> List[Path]:
    crontab_files = [Path("/etc/crontab")]
    crontab_dirs_to_check = [
        "/etc/cron.d",
        "/etc/cron.daily",
        "/etc/cron.hourly",
        "/etc/cron.monthly",
        "/etc/cron.weekly",
    ]

    for crontab_dir in crontab_dirs_to_check:
        walker = os.walk(crontab_dir)
        for root, dirs, files in walker:
            names = dirs + files
            for name in names:
                path = Path(os.path.join(root, name))
                crontab_files.append(path)

    return crontab_files


def generate_crontab_hashes_file(storepath) -> None:
    file_names = get_crontab_file_names() + systemd.get_timer_files()
    file_names = list(map(str, file_names))
    files_with_hashes = get_file_filtered_hashes(file_names, CRONTAB_FILTER)
    files_with_hashes_json = json.dumps(files_with_hashes)
    with open(storepath, "wt") as f:
        f.write(files_with_hashes_json)


def check_crontab_files(crontab_dict: Dict[str, str]):
    found_crontab_files = get_crontab_file_names() + systemd.get_timer_files()
    found_crontab_files = list(map(str, found_crontab_files))
    illegal_files, changed_files = compare_file_hashes_to_earlier_defined_list(
        crontab_dict, found_crontab_files, CRONTAB_FILTER
    )
    return illegal_files, changed_files


def _craft_slack_message_crontabs_warning():
    """
    This function formats the message for slack
    :return str: the formatted message
    """
    message = (
        "Illegal or dirty crontab files have been found on hypernode '"
        + get_hypernode_name()
        + "'. Check up required."
    )
    return message


def _craft_print_message_crontabs_warning(illegal_files, edited_files):
    """
    This function formats the message for the print function
    :return str: the formatted message
    """
    message = (
        "On hypernode '" + get_hypernode_name() + "' the following has been found: \n\n"
    )
    if len(illegal_files) > 0:
        message += "The following crontab files seem to be illegal:\n"
        for file_path in illegal_files:
            message += file_path + "\n"
    if len(edited_files) > 0:
        message += "The following crontab files seem to have un expected changes:\n"
        for file_path in edited_files:
            message += file_path + "\n"
    return message


def main() -> int:
    args = parse_arguments()
    logger.setLevel(logging.INFO)
    logger.addHandler(create_stream_handler())

    # if the user wants to generate a list
    if args.generate:
        generate_crontab_hashes_file(args.storepath)
        return os.EX_OK

    json_contents: str
    if args.json_string:
        json_contents = args.json_string
    else:
        with open(args.json_file, "rt") as f:
            json_contents = f.read()

    try:
        crontab_dict = json.loads(json_contents)
    except json.decoder.JSONDecodeError as e:
        logger.exception(e)
        return os.EX_DATAERR

    illegal_files, dirty_files = check_crontab_files(crontab_dict)
    if len(illegal_files) or len(dirty_files):
        if args.slack and user_is_root():
            try_send_slack_security_message(
                _craft_slack_message_crontabs_warning(), channel=args.slack
            )
        print_message = _craft_print_message_crontabs_warning(
            illegal_files, dirty_files
        )
    else:
        print_message = "Scan successful, all matches up :)"

    logger.info(print_message)

    return os.EX_OK


if __name__ == "__main__":
    main()
