import json
import logging
import os
from argparse import ArgumentParser
from os import EX_DATAERR, EX_OK

from kamikaze3.common.env import get_hypernode_name
from kamikaze3.common.filesystem import (
    compare_file_hashes_to_earlier_defined_list,
    get_file_hashes,
)
from kamikaze3.common.log import create_stream_handler
from kamikaze3.common.system import user_is_root
from kamikaze3.security.slack_security import try_send_slack_security_message

logger = logging.getLogger(__name__)


def parse_arguments():
    parser = ArgumentParser(description="Check the sudoers file integrity")
    mutex_group = parser.add_mutually_exclusive_group(required=True)
    mutex_group.add_argument(
        "--json-string",
        help="json of the legal sudoers file names and their hashes",
        type=str,
    )
    mutex_group.add_argument(
        "--json-file",
        help="The hash list can also be passed as a path to a json file with this argument",
        type=str,
    )
    mutex_group.add_argument(
        "--generate",
        help="Generate a list of legal sudoers files and their hashes",
        action="store_true",
    )
    parser.add_argument(
        "--storepath",
        help="Where to store the generated legal sudeors file. Default: $(pwd)/sudoers_hashes.json",
        default=os.getcwd() + "/sudoers_hashes.json",
        type=str,
    )
    parser.add_argument(
        "--slack",
        help="Post the results in a slack channel (only works as root)",
        type=str,
    )
    return parser.parse_args()


def main():
    args = parse_arguments()
    logger.setLevel(logging.INFO)
    logger.addHandler(create_stream_handler())

    # if the user wants to generate a list
    if args.generate:
        generate_sudoers_hash_list(args.storepath)
        return EX_OK

    # code will only reach here if the user wants to check the sudoers files
    if args.json_string:
        try:
            sudoers_dict = json.loads(args.json_string)
        except json.decoder.JSONDecodeError as e:
            logger.exception(e)
            return EX_DATAERR
    else:
        with open(args.json_file, "rt") as f:
            file_contents = f.read()
        sudoers_dict = json.loads(file_contents)

    illegal_files, dirty_files = check_sudoers_files(sudoers_dict)

    if len(illegal_files) > 0 or len(dirty_files) > 0:
        if args.slack and user_is_root():
            try_send_slack_security_message(
                _craft_slack_message_sudoers_warning(), channel=args.slack
            )
        print_message = _craft_print_message_sudoers_warning(illegal_files, dirty_files)
    else:
        print_message = "Scan successful, all matches up :)"
    logger.info(print_message)
    return EX_OK


def check_sudoers_files(sudoers_dict):
    found_sudoers_files = get_sudoers_file_names()
    illegal_files, changed_files = compare_file_hashes_to_earlier_defined_list(
        sudoers_dict, found_sudoers_files
    )
    return illegal_files, changed_files


def get_sudoers_file_names():
    sudoers_files = ["/etc/sudoers"]
    walker = os.walk("/etc/sudoers.d")
    for root, dirs, files in walker:
        names = dirs + files
        for name in names:
            path = os.path.join(root, name)
            sudoers_files.append(path)
    return sudoers_files


def generate_sudoers_hash_list(storepath=os.getcwd() + "/sudoers_hashes.json"):
    file_names = get_sudoers_file_names()
    files_with_hashes = get_file_hashes(file_names)
    files_with_hashes_json = json.dumps(files_with_hashes)
    with open(storepath, "wt") as f:
        f.write(files_with_hashes_json)


def _craft_slack_message_sudoers_warning():
    """
    This function formats the message for slack
    :return str: the formatted message
    """
    message = (
        "Illegal or dirty sudoers files have been found on hypernode '"
        + get_hypernode_name()
        + "'. Check up required."
    )
    return message


def _craft_print_message_sudoers_warning(illegal_files, edited_files):
    """
    This function formats the message for the print function
    :return str: the formatted message
    """
    message = (
        "On hypernode '" + get_hypernode_name() + "' the following has been found: \n\n"
    )
    if len(illegal_files) > 0:
        message += "The following sudoers files seem to be illegal:\n"
        for file_path in illegal_files:
            message += file_path + "\n"
    if len(edited_files) > 0:
        message += "The following sudoers files seem to have un expected changes:\n"
        for file_path in edited_files:
            message += file_path + "\n"
    return message
