import json
import logging
from argparse import ArgumentParser
from os import EX_DATAERR, EX_OK, getcwd

from kamikaze3.common.env import get_hypernode_name
from kamikaze3.common.log import create_stream_handler
from kamikaze3.common.system import user_is_root
from kamikaze3.common.user_group import gather_users_and_groups_information
from kamikaze3.security.slack_security import try_send_slack_security_message

logger = logging.getLogger(__name__)


def parse_arguments():
    parser = ArgumentParser(description="Check users and groups on this hypernode")
    mutex_group = parser.add_mutually_exclusive_group(required=True)
    mutex_group.add_argument(
        "--json-string",
        help="json of the legal users and groups (found in Hypernode Control)",
        type=str,
    )
    mutex_group.add_argument(
        "--json-file",
        help="The legal users and groups list can also be passed as a path to a json file with this argument",
        type=str,
    )
    mutex_group.add_argument(
        "--generate",
        help="Generate a list of the state of the users and groups",
        action="store_true",
    )
    parser.add_argument(
        "--storepath",
        help="Where to store the generated state of the users and groups. Default: $(pwd)/users_groups.json",
        type=str,
        default=getcwd() + "/users_groups.json",
    )
    parser.add_argument(
        "--slack",
        help="Post the results in a slack channel (only works as root)",
        type=str,
    )
    return parser.parse_args()


def main():
    """
    Main function to check the users accounts and groups for changes
    """
    args = parse_arguments()
    logger.setLevel(logging.INFO)
    logger.addHandler(create_stream_handler())

    # if the user wants to generate a list
    if args.generate:
        generate_users_groups_file(args.storepath)
        return EX_OK

    # code will only reach here if the user wants to check the users and groups
    if args.json_string:
        try:
            legal_users_groups = json.loads(args.json_string)
        except json.decoder.JSONDecodeError as e:
            logger.exception(e)
            return EX_DATAERR
    else:
        with open(args.json_file, "rt") as f:
            file_contents = f.read()
        legal_users_groups = json.loads(file_contents)

    found_users_groups = gather_users_and_groups_information()
    mismatch_uid_important, mismatch_users, mismatch_groups = compare_users_groups(
        legal_users_groups, found_users_groups
    )

    if (
        len(mismatch_uid_important) > 0
        or len(mismatch_users) > 0
        or len(mismatch_groups) > 0
    ):
        if args.slack and user_is_root():
            try_send_slack_security_message(
                _craft_slack_message_users_groups_warning(), channel=args.slack
            )
        print_message = _craft_print_message_users_groups_warning(
            mismatch_uid_important, mismatch_users, mismatch_groups
        )
    else:
        print_message = "Check successful, all users and groups match up :)"
    logger.info(print_message)
    return EX_OK


def generate_users_groups_file(storepath=getcwd() + "/users_groups.json"):
    """
    Generate a list for the current Hypernode and store this in a file.
    This file will be a json and can be uploaded in Hypernode Control
    :param storepath: the path to store the json. Default: $(pwd)/users_groups.json
    """
    users_groups_json = json.dumps(gather_users_and_groups_information())
    with open(storepath, "wt") as f:
        f.write(users_groups_json)


def compare_users_groups(legal_users_groups, found_users_groups):
    """
    Function to compare the current users and groups to the list of values how the list should be.
    Returning any users or groups that changed.
    :param legal_users_groups: the list of users and groups how the list should be
    :param found_users_groups: the list of users and groups how it is found currently on the system
    :return: the mismatches, sorted in categories
    """
    mismatch_uid_important = []
    mismatch_users = []
    mismatch_groups = []
    for uid, users in found_users_groups["uid_important"].items():
        for username in users:
            if username not in legal_users_groups["uid_important"][uid]:
                mismatch_uid_important.append({"uid": uid, "name": username})
    for user in found_users_groups["users"]:
        if user not in legal_users_groups["users"]:
            mismatch_users.append(user)
    for group in found_users_groups["groups"]:
        if group not in legal_users_groups["groups"]:
            mismatch_groups.append(group)

    return mismatch_uid_important, mismatch_users, mismatch_groups


def _craft_slack_message_users_groups_warning():
    """
    This function formats the message for slack.
    This is a short warning message, for the detailed message check the logs or terminal output.
    :return str: the formatted message
    """
    message = (
        "Unexpected changes found in the users/groups configuration on hypernode '{}'."
        " Check up required.".format(get_hypernode_name())
    )
    return message


def _craft_print_message_users_groups_warning(
    mismatch_uid_important, mismatch_users, mismatch_groups
):
    """
    This function formats the message for the terminal and logs.
    This message is more detailed than the slack version and tells which users and groups are actually different.
    :param mismatch_uid_important: The users with wrong UIDs
    :param mismatch_users: users where the name, home dir or shell don't match
    :param mismatch_groups: groups where the name or members don't match
    :return str: formatted message
    """
    message = "On hypernode '{}' the following has been found: \n\n".format(
        get_hypernode_name()
    )
    if len(mismatch_uid_important) > 0:
        message += "The following user(s) have the wrong uid:\n"
        for uid_important in mismatch_uid_important:
            message += "User '{}' with uid '{}'.\n".format(
                uid_important["name"], uid_important["uid"]
            )
    if len(mismatch_users) > 0:
        message += (
            "The following user(s) have settings that do not match our records:\n"
        )
        for user in mismatch_users:
            message += "User '{}' with home dir '{}' and shell '{}'.\n".format(
                user["name"], user["dir"], user["shell"]
            )
    if len(mismatch_groups) > 0:
        message += (
            "The following group(s) have settings that do not match our records: \n"
        )
        for group in mismatch_groups:
            message += "Group '{}' with members '{}'.\n".format(
                group["name"], group["members"]
            )
    return message
