import logging
from argparse import ArgumentParser
from functools import partial
from os import listdir, makedirs, path
from os.path import exists, isdir, isfile, join

from jinja2 import Environment, FileSystemLoader

from kamikaze3.common.ssl import get_alternative_names, parse_x509

TEMPLATE_ENVIRONMENT = Environment(
    autoescape=False, loader=FileSystemLoader("/etc/hypernode/"), trim_blocks=False
)

CHAIN_TEMPLATE = "ssl-cert-key-chain-meta-template.j2"
CONF_TEMPLATE = "outside_meta_template.j2"
CERTS_DIR = "/data/web/certs"
APP_SSL_DIR = "/data/web/nginx/ssl"
OVERRIDE_FILE = "/data/web/nginx/disable_hypernode_ssl_config_generator.flag"

logger = logging.getLogger(__name__)


class FileDoesNotExistError(RuntimeError):
    pass


def set_logging_level():
    logging.basicConfig(level=logging.INFO)
    logger.setLevel(level=logging.INFO)


def get_domain_dirs():
    return (
        [d for d in listdir(CERTS_DIR) if isdir(join(CERTS_DIR, d))]
        if isdir(CERTS_DIR)
        else list()
    )


def render_cert_key_chain(domainname, private_key, certificate, ca=None):
    context = {
        "domainname": domainname,
        "private_key": private_key,
        "certificate": certificate,
        "ca": ca or "",
    }
    return TEMPLATE_ENVIRONMENT.get_template(CHAIN_TEMPLATE).render(context)


def render_config(crtkeyca):
    cert = parse_x509(crtkeyca)
    server_names = " ".join(get_alternative_names(cert))
    context = {"server_name": server_names, "crtkeyca": crtkeyca}
    return TEMPLATE_ENVIRONMENT.get_template(CONF_TEMPLATE).render(context)


def get_contents(filename):
    if not path.exists(filename):
        raise FileDoesNotExistError()

    with open(filename, "r") as f:
        return f.read()


def get_certs_dir_for_domain(domain):
    return join(CERTS_DIR, domain)


def get_file_path_for_domain(domain, filename):
    certs_dir_for_domain = get_certs_dir_for_domain(domain)
    return join(certs_dir_for_domain, filename)


def get_file_content_for_domain(domain, filename):
    return get_contents(get_file_path_for_domain(domain, filename))


def get_crtkeyca_path_for_domain(domain):
    return join(APP_SSL_DIR, domain + ".crtkeyca")


def get_conf_path_for_domain(domain):
    return join(APP_SSL_DIR, domain + ".conf")


def write_crtkeyca(domain):
    rendered = render_cert_key_chain(
        domain,
        *map(
            partial(get_file_content_for_domain, domain),
            ("privkey.pem", "fullchain.pem"),
        )
    )
    with open(get_crtkeyca_path_for_domain(domain), "w") as f:
        f.write(rendered)


def write_conf(domain):
    rendered = render_config(get_crtkeyca_path_for_domain(domain))
    with open(get_conf_path_for_domain(domain), "w") as f:
        f.write(rendered)


def ensure_directory(directory):
    if not exists(directory):
        makedirs(directory)


def generate_ssl_config(generate_configs=True):
    ensure_directory(APP_SSL_DIR)
    if isfile(OVERRIDE_FILE):
        logger.info(
            "The hypernode-ssl-config-generator tool is "
            "disabled because {} exists".format(OVERRIDE_FILE)
        )
    else:
        for domain in get_domain_dirs():
            logger.info("Generating ssl config for {}".format(domain))
            try:
                write_crtkeyca(domain)
                if generate_configs:
                    write_conf(domain)
            except FileDoesNotExistError:
                logger.info(
                    "Could not read privkey.pem or fullchain.pem for domain: {}, skipping!".format(
                        domain
                    )
                )


def parse_arguments():
    parser = ArgumentParser(
        description="Generate ssl configs for ssl certs in {}".format(CERTS_DIR)
    )
    parser.add_argument(
        "--skip-configs",
        action="store_true",
        help="Skip the generation of nginx configs and only " "generate certificates",
    )
    return parser.parse_args()


def main():
    args = parse_arguments()
    set_logging_level()
    generate_ssl_config(generate_configs=not args.skip_configs)
