import functools
import json
import os
import re
import shutil
import socket
import subprocess
from collections.abc import Callable
from logging import getLogger

import MySQLdb
import nats.errors
import requests
from MySQLdb import OperationalError

from kamikaze3.cluster.health import (
    get_nfs_mountpoints,
    get_parent_info_for_relation_type,
    has_parent_with_relation_type,
)
from kamikaze3.common.env import get_hn_app_var, get_node_roles, is_docker_enabled
from kamikaze3.common.log import setup_logging
from kamikaze3.common.system import is_hostname_localhost
from kamikaze3.common.systemd import check_service_running
from kamikaze3.event_listener.client import request_event_sync
from kamikaze3.event_listener.settings import EVENT_PING
from kamikaze3.haproxy.main import StatsCommand
from kamikaze3.settings import (
    ROLE_ELASTICSEARCH_MASTER,
    ROLE_LOADBALANCER,
    ROLE_MAGWEB,
    ROLE_MYSQL_MASTER,
    ROLE_NFS,
    ROLE_NFS_CHILD,
    ROLE_RABBITMQ_MASTER,
    ROLE_REDIS_MASTER,
)
from kamikaze3.status import cache
from kamikaze3.status.cache import (
    is_report_cache_available,
    try_get_report_cache,
    try_save_report_cache,
)
from kamikaze3.status.fs import check_disk_full, check_disk_inodes_full
from kamikaze3.status.input import StatusConfig, parse_args
from kamikaze3.status.settings import (
    DEFAULT_SOCKET_TIMEOUT,
    ELASTICSEARCH_TIMEOUT,
    PHP_TIMEOUT,
    REDIS_PERSISTENCE_PORT,
    REDIS_PORT,
)

HAPROXY_STATUS_URI = "haproxy_stats;csv"
INITIAL_APP_CONF = "initial_appconfiguration"
logger = getLogger(__name__)


def check_if_disk_full(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs) -> bool:
        if check_disk_full() or check_disk_inodes_full():
            # We stop checking MySQL if the data disk is full. This falls outside of our
            # responsibility, so we don't find it unhealthy at this point.
            return True
        return func(*args, **kwargs)

    return wrapper


def check_if_service_local(hostname_arg: str):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs) -> bool:
            config: StatusConfig = args[0]  # type: ignore
            hostname = getattr(config, hostname_arg)
            if config.local_only and not is_hostname_localhost(hostname):
                return True
            return func(*args, **kwargs)

        return wrapper

    return decorator


@check_if_disk_full
def check_mysql(config: StatusConfig) -> bool:
    try:
        MySQLdb.connect(config.mysql_host)
    except OperationalError as e:
        error_code, _ = e.args
        # 1045: Permission denied, mysql is running but permission is denied because we don't authenticate
        # 2054: Unknown authentication method, mysql is running but authentication method doesn't work
        if error_code not in (1045, 2054):
            logger.debug("Got MySQL exception: %s", e)
            return False
    return True


@check_if_disk_full
def check_php(config: StatusConfig) -> bool:
    try:
        headers = {"Host": "magweb"}
        if config.report_cache:
            headers["HN-ALLOW-CACHED"] = "1"

        resp = requests.get(
            "http://localhost/status.php",
            headers=headers,
            timeout=PHP_TIMEOUT,
        )
        return resp.text == "OK"
    except Exception as e:
        logger.debug("Got php-fpm connect error: %s", e)
        return False


@check_if_disk_full
@check_if_service_local("redis_host")
def check_redis(config: StatusConfig) -> bool:
    try:
        subprocess.check_call(
            ["redis-cli", "-h", config.redis_host, "-p", str(REDIS_PORT), "ping"],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        return True
    except subprocess.CalledProcessError as e:
        logger.debug("Got redis-cli error: %s", e)
        return False


@check_if_disk_full
@check_if_service_local("redis_host")
def check_redis_persistent(config: StatusConfig) -> bool:
    try:
        subprocess.check_call(
            [
                "redis-cli",
                "-h",
                config.redis_host,
                "-p",
                str(REDIS_PERSISTENCE_PORT),
                "ping",
            ],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        return True
    except subprocess.CalledProcessError as e:
        logger.debug("Got redis-cli error: %s", e)
        return False


@check_if_disk_full
def check_varnish(config: StatusConfig) -> bool:
    try:
        with socket.create_connection(
            (config.varnish_host, config.varnish_port), timeout=DEFAULT_SOCKET_TIMEOUT
        ) as sock:
            _ = sock.recv(4096)  # Read banner
            sock.send(b"ping\r\n")
            pong = sock.recv(4096)

            if not pong:
                return False

            return bool(re.match(r"^200 19\s+PONG ", pong.decode("utf-8")))
    except Exception as e:
        logger.debug("Got varnish connect error: %s", e)
        return False


@check_if_disk_full
@check_if_service_local("rabbitmq_host")
def check_rabbitmq(config: StatusConfig) -> bool:
    try:
        sock = socket.create_connection(
            (config.rabbitmq_host, config.rabbitmq_port), timeout=DEFAULT_SOCKET_TIMEOUT
        )
        sock.close()
        return True
    except OSError as e:
        logger.debug("Got rabbitmq connect error: %s", e)
        return False


def check_nfs(config: StatusConfig) -> bool:
    if config.local_only:
        return True

    parent_info = get_parent_info_for_relation_type(ROLE_NFS)
    if parent_info["parent"]["private_network_ip"] is not None:
        cluster_ip = parent_info["parent"]["private_network_ip"]
    else:
        cluster_ip = parent_info["parent"]["cluster_ip"]

    result = True
    output = subprocess.check_output(["/usr/bin/netstat", "-atun"], encoding="utf8")
    if f"{cluster_ip}:{config.nfs_port}" not in output:
        logger.debug(
            "Nfs service not running on %s port %s.",
            str(cluster_ip),
            str(config.nfs_port),
        )
        result = False

    mount_points = get_nfs_mountpoints()
    for mount_point in mount_points:
        if not os.path.ismount(mount_point):
            logger.debug("Nfs mount not mounted. mountpoint %s", str(mount_point))
            result = False

    return result


def check_haproxy(config: StatusConfig) -> bool:
    if config.local_only:
        return True

    stats_command = StatsCommand()
    keys_to_check = ["http_backend", "https_backend"]

    try:
        for server in stats_command.get_server_stats():
            if server.pxname in keys_to_check and server.status not in (
                "UP",
                "no check",
            ):
                logger.debug(
                    f"{server.pxname} of instance {server.svname} not UP, got status {server.status}"
                )
                return False
    except Exception as e:
        logger.debug("Got haproxy error: %s", e)
        return False
    return True


def _check_event_listener_ping() -> bool:
    try:
        request_event_sync(EVENT_PING)
    except nats.errors.Error as e:
        logger.debug(f"Could not connect to hypernode-event-listener: {e}")
        return False
    return True


def _temporary_event_listener_wrapper(*args, **kwargs):
    # We don't want to fully enable the event listener in the status check, so we
    # run it but catch all exceptions and return True always
    # TODO: remove this wrapper when we're confident the event listener is stable
    try:
        check_event_listener(*args, **kwargs)
    except Exception as e:
        logger.debug(f"Got exception in event listener check: {e}")

    return True


def check_event_listener(config: StatusConfig) -> bool:
    if not check_service_running("nats-server"):
        logger.debug(f"nats-server is not running")
        return False
    if not check_service_running("hypernode-event-listener"):
        logger.debug(f"hypernode-event-listener is not running")
        return False
    return _check_event_listener_ping()


@check_if_disk_full
@check_if_service_local("elasticsearch_host")
def check_elasticsearch(config: StatusConfig) -> bool:
    try:
        resp = requests.get(
            f"http://{config.elasticsearch_host}:{config.elasticsearch_port}/_cluster/health",
            timeout=ELASTICSEARCH_TIMEOUT,
        )
        if resp.status_code != 200:
            logger.debug(
                "Failed to get elasticsearch status, got http status code {}. Response body: {}".format(
                    resp.status_code, resp.text
                )
            )
            return False
    except requests.exceptions.ConnectionError:
        logger.debug("Failed to reach elasticsearch, is it running?")
        return False

    return True


def check_opendkim(config: StatusConfig) -> bool:
    socket_path = "/var/spool/postfix/opendkim/opendkim.sock"
    try:
        subprocess.check_call(["test", "-S", socket_path])
    except subprocess.CalledProcessError:
        logger.debug("OpenDKIM socket is not available at %s", socket_path)
        return False
    return True


def check_readonly_fs(config: StatusConfig) -> bool:
    result = True
    try:
        with open("/proc/mounts", encoding="utf-8") as f:
            for line in f.readlines():
                items = line.split()
                if len(items) < 2:
                    continue
                _, mount, fstype, options = items[:4]
                if mount not in ("/", "/data"):
                    continue
                if "ro" in options.split(","):
                    logger.debug("Mount point %s has ro option", mount)
                    result = False
    except OSError as e:
        logger.debug("Something went wrong reading /proc/mounts. Exception: %s", e)
        return False
    return result


def check_root_free_space(config: StatusConfig) -> bool:
    _, _, free = shutil.disk_usage("/")
    threshold = 250 * 1024 * 1024  # 250 megabytes
    return free >= threshold


def check_root_free_inodes(config: StatusConfig) -> bool:
    max_used_inodes_factor = 0.98
    stat = os.statvfs("/")
    usage_factor = 1.0 - (float(stat.f_ffree) / float(stat.f_files))
    return usage_factor < max_used_inodes_factor


def get_checks() -> list[Callable[[StatusConfig], bool]]:
    checks = [
        check_readonly_fs,
        check_root_free_space,
        check_root_free_inodes,
    ]
    roles = get_node_roles()

    if ROLE_MYSQL_MASTER in roles:
        checks.append(check_mysql)

    if ROLE_REDIS_MASTER in roles:
        checks.append(check_redis)
        if get_hn_app_var("redis_persistent_instance", default_val=False):
            checks.append(check_redis_persistent)

    if ROLE_RABBITMQ_MASTER in roles and get_hn_app_var("rabbitmq_enabled"):
        checks.append(check_rabbitmq)

    if ROLE_MAGWEB in roles:
        checks.append(check_php)

    if ROLE_LOADBALANCER in roles:
        checks.append(check_haproxy)

    if ROLE_ELASTICSEARCH_MASTER in roles and (
        get_hn_app_var("elasticsearch_enabled") or get_hn_app_var("opensearch_enabled")
    ):
        checks.append(check_elasticsearch)

    if (
        get_hn_app_var("transip_mail", "enabled")
        and get_hn_app_var("node_status") == "production"
    ):
        checks.append(check_opendkim)

    if (
        ROLE_MAGWEB in roles
        and not has_parent_with_relation_type(ROLE_LOADBALANCER)
        and get_hn_app_var("varnish_enabled")
    ) or (ROLE_LOADBALANCER in roles and get_hn_app_var("varnish_enabled")):
        checks.append(check_varnish)

    if ROLE_NFS_CHILD in roles and has_parent_with_relation_type(ROLE_NFS):
        checks.append(check_nfs)

    if not is_docker_enabled():
        checks.append(_temporary_event_listener_wrapper)

    return checks


def main(args: list[str] | None = None) -> None:
    config = parse_args(args)
    setup_logging(logger, config.verbose)
    setup_logging(cache.logger, config.verbose)

    if config.report_cache and is_report_cache_available():
        result = try_get_report_cache()
    else:
        result = {}

    if not result:
        for check_fn in get_checks():
            result[check_fn.__name__] = check_fn(config)

        try_save_report_cache(result)

    log_message = f"HYPERNODE_STATUS: {json.dumps(result)}"
    if all(result.values()):
        logger.debug(log_message)
    else:
        logger.error(log_message)

    print(json.dumps(result, indent=4))


if __name__ == "__main__":
    main()
