import logging
import os

from kamikaze3.settings import REDIS_PERSISTENT_TEMPLATE, REDIS_TEMPLATE
from kamikaze3.supervisorctl.manage_configs import (
    ensure_destination_dir_exists,
    write_config,
)

logger = logging.getLogger(__name__)

REDIS_CONF_DIR = "/data/web/redis/"
REDIS_LOG_DIR = "/data/var/log/redis/"


def ensure_redis_config(application_name, persistent, ram):
    """
    Place redis supervisor configuration for requested app name and type
    :param string application_name: name of the app
    :param bool persistent: True if redis persistent requested, otherwise False
    :param int ram: amount of memory to assign to the service
    :return string destination_path: Destination path of the redis configuration
    """
    ensure_destination_dir_exists(REDIS_LOG_DIR)

    path = get_config_path(application_name, persistent)
    ensure_destination_dir_exists(path)
    destination_path = get_destination_path(application_name, persistent)

    if not os.path.exists(destination_path):
        logger.info("Place redis configurations in {}".format(destination_path))
        config = get_my_configs(application_name, ram)
        template_path = get_template_path(persistent)
        write_config(config, template_path, destination_path)

    return destination_path


def get_config_path(application_name, persistent):
    """
    Get redis destination / config path
    :param str application_name: Name of the application
    :param bool persistent: True if persistent, otherwise false
    :return str path:
    """
    if persistent:
        return "".join([REDIS_CONF_DIR, application_name, "_persistent"])
    return "".join([REDIS_CONF_DIR, application_name])


def get_template_path(persistent):
    """
    Get template path for redis type
    :param bool persistent: True if redis persistent requested, otherwise False
    :return string template_path: destination path of the redis config
    """
    redis_version = get_installed_redis_version()
    if redis_version == 5:
        if persistent:
            return REDIS_PERSISTENT_TEMPLATE

        return REDIS_TEMPLATE


def get_installed_redis_version():
    """
    Get installed redis version, so we can load the correct config templates. For now we always return 5, if we support
    multiple versions in the future we can check the installed version here and return the correct one.
    :return int redis_version: Installed redis version Number
    """
    return 5


def get_destination_path(application_name, persistent):
    """
    Get destination path for application_name
    :param string application_name: name of the app
    :param bool persistent: True if redis persistent requested, otherwise False
    :return string destination_path: destination path of the redis config
    """
    if persistent:
        return "".join(
            [REDIS_CONF_DIR, application_name, "_persistent/redis.persistent.conf"]
        )

    return "".join([REDIS_CONF_DIR, application_name, "/redis.conf"])


def get_process_name(application_name, persistent):
    """
    Get the process_name of the redis process
    :param string application_name: Name of the app
    :param bool persistent: True if persistent instance, otherwise False
    :return string process_name: Name of the process
    """
    if persistent:
        return "".join([application_name, "_redis_persistent"])

    return "".join([application_name, "_redis"])


def get_my_configs(application_name, ram):
    """
    Get template configs for application_name
    :param string application_name: Name of the app
    :param int ram: amount of memory to assign to the service
    :return Dict: dict with the template configs
    """
    return {"app_name": application_name, "memory": ram}
