import logging

from kamikaze3.supervisorctl.redis.manage_configs import (
    ensure_redis_config,
    get_process_name,
)
from kamikaze3.supervisorctl.service import (
    ensure_process_running,
    ensure_supervisor_running,
    manage_supervisor_process_config,
)

logger = logging.getLogger(__name__)


def handle_redis_setup(application_name, instance_type, ram):
    """
    Handle redis setup for appname.
    :param string application_name: name of the app redis instance is requested.
    :param string instance_type: requested instance type
    :param int ram: amount of memory to assign to the service
    :return None:
    """
    ensure_supervisor_running()

    persistent = is_redis_persistent_setup(instance_type)
    configure_redis(application_name, persistent, ram)


def configure_redis(application_name, persistent, ram):
    """
    Configure requested redis instance for app and make sure process is running after setup
    :param string application_name: name of the app redis instance is requested
    :param bool persistent: True if persistent instance, otherwise False
    :param int ram: amount of memory to assign to the service
    :return None:
    """
    logger.info(
        "Configuring redis for application with name {}".format(application_name)
    )
    redis_config_path = ensure_redis_config(application_name, persistent, ram)
    process_name = get_process_name(application_name, persistent)
    manage_supervisor_process_config(
        command="".join(["/usr/bin/redis-server ", redis_config_path]),
        process_name=process_name,
    )
    ensure_process_running_and_log_output(application_name, process_name)


def ensure_process_running_and_log_output(application_name, process_name):
    """
    Ensure process running and log the process output
    :param string application_name: Name of the application
    :param process_name: Name of the process
    """
    process_running = ensure_process_running(process_name)
    if not process_running:
        logger.warning(
            "Supervisor process {} not running. Try `supervisorctl start {}` to start process manually. \n"
            "You can also use `$ tail -f /data/var/log/redis/{}.log` for more log output.".format(
                process_name, process_name, process_name
            )
        )
    else:
        logger.info(
            "Supervisor process {} has been started and entered running state".format(
                process_name
            )
        )
        logger.info(
            "You can use the following settings in your configuration files: \n Host: {} \n Port: 0".format(
                "/data/var/run/app/redis/".join([application_name, ".sock"])
            )
        )


def is_redis_persistent_setup(instance_type):
    """
    Check if setup is redis-persistent setup.
    :param string instance_type: requested instance type
    :return bool: True if is redis-persistent setup, False otherwise
    """
    return "redis-persistent" in instance_type
