import logging
import os
import random
import string
from subprocess import check_output

from kamikaze3.settings import (
    VARNISH_BASE_DIRECTORY,
    VARNISH_RUN_DIR,
    VARNISH_VCL_TEMPLATE,
)
from kamikaze3.supervisorctl.manage_configs import (
    ensure_destination_dir_exists,
    write_config,
)

logger = logging.getLogger(__name__)

START_PORT_RANGE = 16001
END_PORT_RANGE = 17000


def ensure_varnish_configs(application_name):
    """
    Place varnish supervisor configuration for requested application name
    :param string application_name: name of the app
    :param int ram: amount of memory to assign to the service
    :return dict paths: Dict with paths to vcl and secret
    """
    destination_path = "".join([VARNISH_BASE_DIRECTORY, application_name, "/"])
    ensure_destination_dir_exists(destination_path)
    varnish_secret_dest = ensure_varnish_secret(destination_path)
    varnish_vcl_dest = ensure_varnish_vcl(destination_path)

    varnish_run_dir = "".join([VARNISH_RUN_DIR, application_name])
    ensure_destination_dir_exists(varnish_run_dir)

    return {
        "secret": varnish_secret_dest,
        "vcl": varnish_vcl_dest,
        "run_dir": varnish_run_dir,
    }


def ensure_varnish_vcl(destination_path):
    """
    Get destination of the varnish VCL
    :param string destination_path: Destination path the file should exist
    :return string destination_path: Destination path of the varnish VCL
    """
    varnish_vcl_dest = "".join([destination_path, "varnish.vcl"])
    if not os.path.exists(varnish_vcl_dest):
        write_config({}, VARNISH_VCL_TEMPLATE, varnish_vcl_dest)

    return varnish_vcl_dest


def ensure_varnish_secret(destination_path):
    """
    Ensure varnish secret exists
    :param string destination_path: Destination path the file should exist
    :return string destination_path: Destination path of the varnish secret file
    """
    varnish_secret_dest = "".join([destination_path, "varnish.secret"])
    if not os.path.exists(varnish_secret_dest):
        with open(varnish_secret_dest, "w+") as fp:
            logger.info("Writing varnish secret {}".format(varnish_secret_dest))
            fp.write(get_random_string(32))

    return varnish_secret_dest


def get_process_name(application_name):
    """
    Get the process_name of the varnish process
    :param string application_name: Name of the app
    :return string process_name: Name of the process
    """
    return "".join([application_name, "_varnish"])


def get_random_port():
    """
    Get a random port and check if the port is available
    :return string port_number: Available port number to use
    """
    counter = 0
    retry_times = 5

    while counter < retry_times:
        port = random.randint(START_PORT_RANGE, END_PORT_RANGE)
        if is_port_available(port):
            logger.info("Found available port: {}".format(port))
            return port

        counter += 1

    logger.warning(
        "Unable to find available port, please try to run `hypernode-manage-supervisor --service varnish` again."
    )
    return None


def is_port_available(port):
    """
    Check if port number is available, and no other process currently running on this port
    :param int port: Port number to check
    :return bool: True if port available, otherwise False.
    """
    command = ["/usr/bin/netstat", "-atun"]
    output = check_output(command, encoding="utf-8")
    return str(port) not in output


def get_random_string(length):
    """
    Get random string
    :param int length: Number of characters / length of the requested random string
    :return string random_string: The random string
    """
    chars = string.ascii_lowercase
    return "".join(random.choice(chars) for i in range(length))


def get_command(
    frontend_port,
    backend_port,
    ram,
    varnish_vcl_path,
    varnish_secret_path,
    varnish_run_dir,
):
    """
    Get varnish command
    :param int frontend_port: Frontend port
    :param int backend_port: Backend port
    :param int ram: Amount of memory
    :param string varnish_vcl_path: Varnish vcl destination path
    :param string varnish_secret_path: Varnish secret destination path
    :param string varnish_run_dir: Varnish run dir
    :return string command: The command
    """
    return "".join(
        [
            "/usr/sbin/varnishd ",
            "-p feature=+esi_ignore_other_elements ",
            "-p vcc_allow_inline_c=on ",
            "-a :{} ".format(frontend_port),
            "-T 127.0.0.1:{} ".format(backend_port),
            "-f {} ".format(varnish_vcl_path),
            "-S {} ".format(varnish_secret_path),
            "-s malloc,{}m ".format(ram),
            "-F ",
            "-n {} ".format(varnish_run_dir),
        ]
    )
