import logging
import os

from kamikaze3.supervisorctl.service import (
    ensure_process_running,
    ensure_supervisor_running,
    manage_supervisor_process_config,
)
from kamikaze3.supervisorctl.system_service import ensure_system_service_running
from kamikaze3.supervisorctl.varnish.manage_configs import (
    ensure_varnish_configs,
    get_command,
    get_process_name,
    get_random_port,
)

VARNISH_BIN = "/usr/sbin/varnishd"
logger = logging.getLogger(__name__)


def handle_varnish_setup(application_name, ram):
    """
    Handle varnish setup for appname.
    :param string application_name: name of the app the varnish instance is requested.
    :param int ram: amount of memory to assign to the service
    :return None:
    """
    ensure_supervisor_running()
    ensure_system_service_running(
        service_name="varnish", hypernode_systemctl_setting_name="varnish_enabled"
    )

    if not varnish_binary_exists():
        logger.warning(
            "Varnish cannot be enabled for this application. "
            "Please upgrade to a plan that supports Varnish if you want to make use of this feature."
        )
        return

    configure_varnish(application_name, ram)


def varnish_binary_exists():
    """
    Check Varnish binary exists
    :return bool binary_exists: Returns true if binary exists, otherwise false
    """
    return os.path.exists(VARNISH_BIN)


def configure_varnish(application_name, ram):
    """
    Configure requested varnish instance for application and make sure process is running after setup
    :param string application_name: name of the app the varnish instance is requested
    :param int ram: amount of memory to assign to the service
    :return None:
    """
    logger.info(
        "Configuring varnish for application with name {}".format(application_name)
    )

    # get frontend and backend port
    frontend_port = get_random_port()
    if frontend_port is None:
        return

    backend_port = get_random_port()
    if backend_port is None:
        return

    configs = ensure_varnish_configs(application_name)
    process_name = get_process_name(application_name)
    manage_supervisor_process_config(
        command=get_command(
            frontend_port=frontend_port,
            backend_port=backend_port,
            ram=ram,
            varnish_vcl_path=configs["vcl"],
            varnish_secret_path=configs["secret"],
            varnish_run_dir=configs["run_dir"],
        ),
        process_name=process_name,
    )
    ensure_process_running_and_log_output(
        process_name, frontend_port, backend_port, configs["vcl"], configs["secret"]
    )


def ensure_process_running_and_log_output(
    process_name, frontend_port, backend_port, vcl_dest, secret_dest
):
    """
    Ensure process running and log the process output
    :param process_name: Name of the process
    :param frontend_port: Varnish Frontend port
    :param backend_port: Varnish backend port
    :param vcl_dest: Destination varnish VCL file
    :param secret_dest: Destination varnish secret file
    """
    process_running = ensure_process_running(process_name)
    if not process_running:
        logger.warning(
            "Supervisor process {} not running. Try `supervisorctl start {}` to start process manually.".format(
                process_name, process_name
            )
        )
    else:
        logger.info(
            "Supervisor process {} has been started and entered running state".format(
                process_name
            )
        )
        logger.info(
            "You can use the following settings in your configuration files: "
            "\n Host: 127.0.0.1 \n Frontend-port: {} \n Backend-port: {} "
            "\n Location vcl: {} \n Location secret: {}".format(
                frontend_port, backend_port, vcl_dest, secret_dest
            )
        )
