"""
kamikaze3.system_services.fail2ban
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Functionality regarding fail2ban service
"""

import sqlite3
from contextlib import closing


def purge_fail2ban_database(db_path, vacuum=True):
    """Purge fail2ban database tables.
    Note: Fail2ban tool up until version 0.11 can't purge DB properly,
    so we purge the DB directly.

    Note: Current versions of Fail2ban DB uses auto_vacuum=0 pragma, so
    after purging records, it's essential to VACUUM the DB otherwise no disk
    space is freed from the DB file. Vacuuming requires free disk space of double
    the size of the DB itself to work.

    :param str db_path: the SQLite database path
    :param bool vacuum: vacuum the DB
    """
    try:
        db_con = sqlite3.connect(
            db_path, timeout=10
        )  # deleting larger tables may take more time
        with closing(db_con):
            db_con.execute("DELETE FROM bans")
            # @TODO: newer versions of fail2ban will have the table "bips" which we can purge too
            db_con.commit()  # SQLite implicitly starts a transaction for DML queries
            if vacuum:
                db_con.execute("VACUUM")
    except sqlite3.DatabaseError as err:
        raise RuntimeError("Failed to purge fail2ban DB: {}".format(err))
