import logging
import requests

from kamikaze3.common.env import get_app_name, get_hn_app_var
from kamikaze3.common.api import get_hypernode_api_token, SYSTEMCTL_USER_AGENT
from kamikaze3.common.api import (
    HYPERNODE_API_URL,
    handle_api_response,
)

logger = logging.getLogger(__name__)

HYPERNODE_API_APP_DETAIL_URL = HYPERNODE_API_URL + "/v2/app/{}/"
ALTERNATIVE_PHP_VERSIONS_EAV = "alternative_php_versions"


def disable_php_version(php_version):
    """
    Disable php version by removing it from the enabled alternative_php_versions
    :param str php_version: The PHP version to disable
    """
    current_enabled_php_versions = get_current_enabled_php_versions()
    if php_version not in current_enabled_php_versions:
        logger.info("PHP version {} is already disabled.".format(php_version))
        return

    versions = []
    for enabled_php_version in current_enabled_php_versions:
        if enabled_php_version == php_version:
            continue

        versions.append(enabled_php_version)

    update_php_version_eav(versions)


def enable_php_version(php_version):
    """
    Enable php version by adding it to the enabled alternative_php_versions
    :param str php_version: The PHP version to enable
    """
    current_enabled_php_versions = get_current_enabled_php_versions()
    if php_version in current_enabled_php_versions:
        logger.info("PHP version {} is already enabled.".format(php_version))
        return

    current_enabled_php_versions.append(php_version)
    update_php_version_eav(current_enabled_php_versions)


def list_enabled_php_versions():
    """
    List the current enabled alternative php_versions
    """
    logger.info(
        "Currently enabled PHP versions are {}.".format(
            get_current_enabled_php_versions()
        )
    )


def update_php_version_eav(php_versions):
    """
    Update the alternative_php_versions eav by calling Hypernode API.
    Hypernode API will post an update_app job
    """
    response = requests.patch(
        HYPERNODE_API_APP_DETAIL_URL.format(get_app_name()),
        headers={
            "Authorization": "Token {}".format(get_hypernode_api_token()),
            "User-Agent": SYSTEMCTL_USER_AGENT,
        },
        data={ALTERNATIVE_PHP_VERSIONS_EAV: ";".join(php_versions)},
    )

    success_msg = (
        "Operation was successful and is being processed. Please allow a few minutes for "
        "the settings to be applied. Run 'livelog' to see the progress."
    )

    handle_api_response(response, success_message=success_msg)


def get_current_enabled_php_versions():
    try:
        return get_hn_app_var(ALTERNATIVE_PHP_VERSIONS_EAV, default_val=[])
    except RuntimeError:
        return []
