import json
import logging
from copy import deepcopy
from typing import List

import requests
from prettytable import PrettyTable

from kamikaze3.common.api import (
    HYPERNODE_API_URL,
    SYSTEMCTL_USER_AGENT,
    get_hypernode_api_token,
    handle_api_response,
)
from kamikaze3.common.env import get_app_name

logger = logging.getLogger(__name__)

LIST_BRANCHER_ENDPOINT = HYPERNODE_API_URL + "/v2/app/?type=brancher"
CREATE_BRANCHER_ENDPOINT = HYPERNODE_API_URL + "/v2/brancher/app/{}/"
DELETE_BRANCHER_ENDPOINT = HYPERNODE_API_URL + "/v2/brancher/{}/"
LIST_STOPWATCHES_ENDPOINT = HYPERNODE_API_URL + "/v2/app/{}/stopwatches/"


def _labels_list_to_labels_dict(labels):
    labels_dict = {}
    for label in labels:
        if label.count("=") == 1:
            label_key, label_value = label.split("=")
            labels_dict[label_key] = label_value
    return labels_dict


def print_list_brancher_table(app_infos, show_labels=False, machine_readable=False):
    """
    Format and print the list of brancher apps
    :param list[dict, ..] app_infos: A list of (brancher) app infos
    :param bool show_labels: Whether to show labels or not
    :param bool machine_readable: Print as JSON instead of humanly readable
    :return None:
    """
    headers = ["Name", "IP", "Host", "Minutes", "Unbilled minutes"]
    if show_labels:
        headers.append("Labels")

    output = list() if machine_readable else PrettyTable(headers)
    for app_info in app_infos:
        row_to_add = [
            app_info["name"],
            app_info["ip"] or ("Terminated" if app_info["destroyed"] else "Creating"),
            app_info["domainname"],
            app_info["stopwatch"]["total_minutes"],
            app_info["stopwatch"]["unbilled_minutes"],
        ]
        if show_labels:
            row_to_add.append(", ".join(app_info["labels"]))
            if machine_readable:
                headers.append("Labels_data")
                labels_as_dict = _labels_list_to_labels_dict(app_info["labels"])
                row_to_add.append(labels_as_dict)
        (
            output.append(
                dict(zip([h.lower().replace(" ", "_") for h in headers], row_to_add))
            )
            if machine_readable
            else output.add_row(row_to_add)
        )
    print(json.dumps(output) if machine_readable else output.get_string())


def merge_stopwatches_into_brancher_app_infos(brancher_app_infos):
    """
    This merges the stopwatch info into the brancher app info. This is done
    because the stopwatch info is not returned by the brancher list endpoint
    :param list[dict, ..] brancher_app_infos: A list of brancher app infos
    :return list[dict, ..]: A list of brancher app infos with stopwatch info merged in
    """
    app_name = get_app_name()
    request_url = LIST_STOPWATCHES_ENDPOINT.format(app_name)
    headers = {
        "Authorization": "Token {}".format(get_hypernode_api_token()),
        "User-Agent": SYSTEMCTL_USER_AGENT,
    }
    response = requests.get(request_url, headers=headers)
    app_infos_with_stopwatch = []
    for app_info in brancher_app_infos:
        for stopwatch in response.json():
            if stopwatch["brancher_app"] == app_info["name"]:
                filled_app_info = deepcopy(app_info)
                filled_app_info["stopwatch"] = stopwatch
                app_infos_with_stopwatch.append(filled_app_info)
                break
    return app_infos_with_stopwatch


def list_brancher(show_terminated=False, show_labels=False, machine_readable=False):
    """
    List any Brancher Apps associated with the App on which this command is running
    Prints output to the terminal
    This is similar to: curl -s -X GET https://api.hypernode.com/v2/app/?type=brancher -H "Authorization: Token $TOKEN"
    Note that this currently also prints the parent app (type=all), this is useful information for people to understand
    what Brancher apps really are (a 'branched' temporary copy of a normal Hypernode)
    :param bool show_terminated: Whether to show terminated brancher apps
    :param bool show_labels: Whether to show labels
    :param bool machine_readable: print as json so people can parse the output
    :return None:
    """
    next_url = LIST_BRANCHER_ENDPOINT
    if not show_terminated:
        next_url += "&destroyed=false"

    headers = {
        "Authorization": "Token {}".format(get_hypernode_api_token()),
        "User-Agent": SYSTEMCTL_USER_AGENT,
    }

    app_infos = []
    while next_url:
        response = requests.get(next_url, headers=headers)
        if not handle_api_response(response):
            return
        data = response.json()
        app_infos.extend(data["results"])
        next_url = data["next"]

    app_infos = merge_stopwatches_into_brancher_app_infos(app_infos)
    print_list_brancher_table(
        app_infos, show_labels=show_labels, machine_readable=machine_readable
    )


def create_brancher(labels: List[str] = None, clear_services: List[str] = None):
    """
    Creates a new Brancher App based on the app this command is running on
    Once it's 'created' it still needs to finish booting and all that, so in
    the meantime the user could look at hypernode-systemctl brancher --list
    This is similar to curl -s -X POST https://api.hypernode.com/v2/app/<APPNAME>/<APPNAME>/ -H "Authorization: Token $TOKEN"
    :return None:
    """
    app_name = get_app_name()
    request_url = CREATE_BRANCHER_ENDPOINT.format(app_name)
    headers = {
        "Authorization": "Token {}".format(get_hypernode_api_token()),
        "User-Agent": SYSTEMCTL_USER_AGENT,
    }

    data = {}
    if labels:
        data["labels"] = labels
    if clear_services:
        data["clear_services"] = clear_services

    response = requests.post(request_url, headers=headers, json=data or None)

    message = (
        "Brancher App created for app '{}'. See hypernode-systemctl brancher "
        "--list for the progress".format(app_name)
    )
    if not handle_api_response(response, success_message=message):
        return

    brancher_app_info = response.json()
    print(
        "app_name: {}\n"
        "parent: {}\n"
        "Host: {}\n"
        "Labels: {}\n"
        "Services with data to be cleared: {}\n"
        "IP: will become available in a couple of minutes\n".format(
            brancher_app_info["name"],
            brancher_app_info["parent"],
            brancher_app_info["domainname"],
            (
                ", ".join(brancher_app_info["labels"])
                if brancher_app_info["labels"]
                else "None"
            ),
            ", ".join(clear_services) if clear_services else "None",
        )
    )


def delete_brancher(brancher_app_name):
    """
    This deletes an Brancher App. It is marked as 'Destroyed' which means
    the garbage collector is going clean it up.
    This is similar to curl -s -X POST https://api.hypernode.com/v2/app/<APPNAME>/cancel/ -H "Authorization: Token $TOKEN"
    :param str brancher_app_name: The name of the app to destroy
    :return None:
    """
    request_url = DELETE_BRANCHER_ENDPOINT.format(brancher_app_name)
    headers = {
        "Authorization": "Token {}".format(get_hypernode_api_token()),
        "User-Agent": SYSTEMCTL_USER_AGENT,
    }
    response = requests.delete(request_url, headers=headers)

    message = (
        "Brancher App '{}' deleted. See hypernode-systemctl brancher "
        "--list for the list of remaining brancher apps.".format(brancher_app_name)
    )
    handle_api_response(response, success_message=message)
