import logging
from datetime import datetime

import requests

from kamikaze3.common.api import (
    HYPERNODE_API_URL,
    SYSTEMCTL_USER_AGENT,
    get_hypernode_api_token,
    handle_api_response,
)
from kamikaze3.common.env import get_app_name

logger = logging.getLogger(__name__)
COLUMN_SPACING = 2


def _make_datetime_string_py35_compatible(datetime_string):
    """
    Python 3.5 doesn't allow a timezone like "+01:00". It should
    be formatted like "+0100" instead
    :param str datetime_string:
    :rtype: str
    """
    t, tz = datetime_string.split("+")
    tz_py35_compatible = tz.replace(":", "")
    return t + "+" + tz_py35_compatible


def strip_microseconds(datetime_string):
    """
    Take an iso datetime string and remove the microseconds
    :param str datetime_string: Isoformatted datetime string
    :rtype: datetime
    """
    dt = datetime.strptime(datetime_string, "%Y-%m-%dT%H:%M:%S.%f%z")
    new_dt = dt.replace(microsecond=0)
    return new_dt


def ensure_datetime_microseconds_stripped(datetime_string):
    datetime_string = _make_datetime_string_py35_compatible(datetime_string)
    try:
        # Check if datetime is already stripped of microseconds
        dt = datetime.strptime(datetime_string, "%Y-%m-%dT%H:%M:%S%z")
    except ValueError:
        dt = strip_microseconds(datetime_string)
    return dt.isoformat()


def pprint_backup_records(backup_records):
    """
    Print the logs nicely in a terminal
    :param [dict, ...] backup_records:
    :rtype: None
    """
    columns = [
        # Title, key, min_length
        ("Backup ID", "backup_id", 36),  # Length of uuid4
        ("Type", "type", max(len("instant"), len("periodic"))),
        ("Created at", "backup_created_at", 25),  # Length of isoformat time
        ("Expires at", "expired_at", 25),  # Length of time stripped of microseconds
    ]
    # Headers
    header_string = ""
    for title, _, length in columns:
        header_string += ("{:" + str(length + COLUMN_SPACING) + "s}").format(title)
    print(header_string)
    # Backup records
    for backup_record in backup_records:
        record_str = ""
        for _, key, length in columns:
            if key in ["backup_created_at", "expired_at"]:
                backup_record[key] = ensure_datetime_microseconds_stripped(
                    backup_record[key]
                )
            record_str += ("{:" + str(length + COLUMN_SPACING) + "s}").format(
                backup_record[key]
            )
        print(record_str)


def list_backups():
    """
    Show the list of available backups to attach
    :return dict response_json: The json response as a dict
    """
    response = requests.get(
        HYPERNODE_API_URL + "/v2/app/{}/backup/".format(get_app_name()),
        headers={
            "Authorization": "Token {}".format(get_hypernode_api_token()),
            "User-Agent": SYSTEMCTL_USER_AGENT,
        },
    )
    if not handle_api_response(response):
        return
    # Results are paginated
    backup_records = response.json()["results"]
    pprint_backup_records(backup_records)
