from argparse import ArgumentParser

from kamikaze3.systemctl.alternative_php_versions.parsing import (
    setup_alternative_php_parsing,
    _check_valid_value,
)
from kamikaze3.systemctl.attach_backup.parsing import setup_attach_backup_parsing
from kamikaze3.systemctl.autoscaling.parsing import (
    setup_autoscaling_parsing,
    _handle_prompt_input,
)
from kamikaze3.systemctl.block_attack.parsing import setup_block_attack_parsing
from kamikaze3.systemctl.brancher.parsing import setup_brancher_parsing
from kamikaze3.systemctl.create_backup.parsing import setup_create_backup_parsing
from kamikaze3.systemctl.list_backups.parsing import setup_list_backups_parsing
from kamikaze3.systemctl.list_xgrades.parsing import setup_list_xgrades_parsing
from kamikaze3.systemctl.preinstall.parsing import setup_preinstall_parsing
from kamikaze3.systemctl.settings.parsing import (
    _check_valid_attribute_values,
    get_opensearch_warning,
    setup_settings_parsing,
)
from kamikaze3.systemctl.show.parsing import setup_show_parsing
from kamikaze3.systemctl.whitelist.parsing import setup_whitelist_parsing
from kamikaze3.systemctl.xgrade.parsing import setup_xgrade_parsing
from kamikaze3.utils import request_confirmation

REQUIRES_VALIDATION = {
    "settings": {
        "mysql_version": {
            "5.7": "WARNING: Upgrading to MySQL 5.7 is a dangerous process. The MySQL upgrade script could "
            "encounter a problem if there are problems with the content of the database. Before you "
            "perform this action make sure you have a recent backup at hand. You can for example create "
            "a backup using 'hypernode-systemctl create_backup --help' or test out your change first on "
            "an identical dev-Hypernode",
            "8.0": "WARNING: Upgrading to MySQL 8.0 will cause irreversible changes! This might cause "
            "incompatibilities with some applications. For example, only from Magento 2.4 and later it has "
            "been officially compatible. Are you sure you want to upgrade to MySQL 8.0? You can't go back "
            "without contacting support. Also upgrading to MySQL 5.7 is a dangerous process. The MySQL "
            "upgrade script could encounter a problem if there are problems with the content of the database. "
            "Before you perform this action make sure you have a recent backup at hand. You can for example "
            "create a backup using 'hypernode-systemctl create_backup --help' or test out your change first on "
            "an identical dev-Hypernode",
        },
        "varnish_version": {
            "4.0": "WARNING: When you change Varnish version the default passthrough VCL will be set. "
            "Make sure you copy your /data/var/varnish/default.vcl somewhere safe first!",
            "6.0": "WARNING: When you change Varnish version the default passthrough VCL will be set. "
            "Make sure you copy your /data/var/varnish/default.vcl somewhere safe first!",
        },
        "redis_version": {
            "5.0": "WARNING: When you change Redis version the data in /data/redis will be deleted. If you "
            "use persistent Redis for sessions for example, this means those sessions will be lost.",
            "6.x": "WARNING: When you change Redis version the data in /data/redis will be deleted. If you "
            "use persistent Redis for sessions for example, this means those sessions will be lost.",
            "7.x": "WARNING: When you change Redis version the data in /data/redis will be deleted. If you "
            "use persistent Redis for sessions for example, this means those sessions will be lost.",
        },
    }
}

VALIDATION_CALLBACK = {
    "settings": {
        "opensearch_enabled": get_opensearch_warning,
        "elasticsearch_enabled": get_opensearch_warning,
    }
}


def requires_validation(parsed_args):
    """
    Check if the parsed arguments require validation
    :param Namespace parsed_args: Parsed arguments from argparser
    :rtype: str|None
    """
    result = None
    if parsed_args.command in REQUIRES_VALIDATION:
        requires_validation_command = REQUIRES_VALIDATION[parsed_args.command]
        if (
            parsed_args.attribute
            and parsed_args.attribute in requires_validation_command
        ):
            requires_validation_attribute = requires_validation_command[
                parsed_args.attribute
            ]
            new_value = parsed_args.positional_value or parsed_args.deprecated_value
            if new_value and new_value in requires_validation_attribute:
                result = requires_validation_attribute[new_value]

    if not result and parsed_args.command in VALIDATION_CALLBACK:
        validation_callback_command = VALIDATION_CALLBACK[parsed_args.command]
        if (
            parsed_args.attribute
            and parsed_args.attribute in validation_callback_command
        ):
            validation_callback_attribute = validation_callback_command[
                parsed_args.attribute
            ]
            new_value = parsed_args.positional_value or parsed_args.deprecated_value
            result = validation_callback_attribute(parsed_args.attribute, new_value)

    return result


def parse_arguments(args=None):
    """
    Parse the hypernode-configure arguments
    :param list args: Optional, defaults to sys.argv
    :return obj args: The parsed arguments
    """
    parser = ArgumentParser()
    parser.add_argument(
        "--verbose", "-v", help="Display debug output", action="store_true"
    )

    sub_parsers = parser.add_subparsers(dest="command")
    setup_settings_parsing(sub_parsers)
    setup_alternative_php_parsing(sub_parsers)
    setup_show_parsing(sub_parsers)
    setup_whitelist_parsing(sub_parsers)
    setup_block_attack_parsing(sub_parsers)
    setup_xgrade_parsing(sub_parsers)
    setup_autoscaling_parsing(sub_parsers)
    setup_list_xgrades_parsing(sub_parsers)
    setup_attach_backup_parsing(sub_parsers)
    setup_list_backups_parsing(sub_parsers)
    setup_create_backup_parsing(sub_parsers)
    setup_preinstall_parsing(sub_parsers)
    setup_brancher_parsing(sub_parsers)

    parsed_args = parser.parse_args(args)

    if parsed_args.command == "settings" and (
        ("positional_value" in parsed_args and parsed_args.positional_value)
        or ("deprecated_value" in parsed_args and parsed_args.deprecated_value)
    ):
        _check_valid_attribute_values(parsed_args)
        parsed_args.value = parsed_args.positional_value or parsed_args.deprecated_value

    if parsed_args.command == "alternative_php_versions" and (
        ("positional_value" in parsed_args and parsed_args.positional_value)
    ):
        _check_valid_value(parsed_args)
        parsed_args.value = parsed_args.positional_value

    if parsed_args.command == "autoscaling" and parsed_args.disable:
        parsed_args = _handle_prompt_input(parsed_args)

    if not parsed_args.command:
        parser.print_help()
        parser.exit()

    validation_string = requires_validation(parsed_args)
    if validation_string:
        request_confirmation(validation_string)

    return parsed_args
