import logging

import requests

from kamikaze3.common.api import (
    HYPERNODE_API_URL,
    get_hypernode_api_token,
    handle_api_response,
    SYSTEMCTL_USER_AGENT,
)
from kamikaze3.common.env import get_app_name, eav_to_bool
from kamikaze3.systemctl.settings import HYPERNODE_API_APP_DETAIL_URL

log = logging.getLogger(__name__)

EXCLUDED_EAVS = ["alternative_php_versions"]


def setup_settings_parsing(sub_parsers):
    eav_descriptions = get_eav_descriptions()

    settings_usage = (
        "hypernode-systemctl settings php_version 7.1\n\n"
        "The possible values are: \n\n"
        + "\n".join(
            [
                "{: <25}\t{}".format(key, value)
                for key, value in eav_descriptions.items()
            ]
        )
    )
    settings_parser = sub_parsers.add_parser("settings", usage=settings_usage)

    settings_parser.add_argument("attribute", choices=eav_descriptions.keys())
    values_group = settings_parser.add_mutually_exclusive_group(required=False)
    values_group.add_argument("positional_value", nargs="?")
    values_group.add_argument(
        "--value",
        dest="deprecated_value",
        help="This option is deprecated. Use the positional value instead",
    )
    settings_parser.add_argument(
        "--block", help="Block until the job is completed", action="store_true"
    )


def get_eav_descriptions(ignore_excluded_eavs=True):
    response = requests.get(
        HYPERNODE_API_URL + "/v2/eav_descriptions/",
        headers={
            "Authorization": "Token {}".format(get_hypernode_api_token()),
            "User-Agent": SYSTEMCTL_USER_AGENT,
        },
    )
    handle_api_response(response)

    # Argparser passes arguments as strings, so we have to make sure we cast our allowed values to strings. This is
    # a lot easier than trying to convert the passed in arguments to bool or string (or god knows what) depending on
    # what value was given.
    descriptions = {}
    for key, value in response.json().items():
        if key in EXCLUDED_EAVS and ignore_excluded_eavs:
            continue

        if type(value) is str:
            descriptions.update({key: value})
        else:
            descriptions.update({key: [str(entry) for entry in value]})
    return descriptions


def _check_valid_attribute_values(parsed_args):
    if parsed_args.deprecated_value:
        log.error(
            "The --value option has been deprecated. Please use hypernode-systemctl settings <attribute> <value> "
            "instead."
        )
    new_value = parsed_args.positional_value or parsed_args.deprecated_value
    allowed_values = get_eav_descriptions()[parsed_args.attribute]
    if isinstance(allowed_values, list):
        allowed_lowercase_values = [
            v.lower() for v in allowed_values if v.lower() in ["true", "false"]
        ]
        if not (
            new_value in allowed_values or new_value.lower() in allowed_lowercase_values
        ):
            log.error(
                "{} is not an allowed value for attribute {}. Choices are {}.".format(
                    new_value, parsed_args.attribute, allowed_values
                )
            )
            exit(1)


def get_opensearch_warning(attribute, value):
    """
    Return a warning when OpenSearch/ElasticSearch are toggled.
    :param str attribute: Attribute name
    :param str value: Attribute value
    :rtype: None
    """
    if attribute not in (
        "elasticsearch_enabled",
        "opensearch_enabled",
    ) or not eav_to_bool(value):
        return None

    response = requests.get(
        HYPERNODE_API_APP_DETAIL_URL.format(get_app_name()),
        headers={
            "Authorization": "Token {}".format(get_hypernode_api_token()),
            "User-Agent": SYSTEMCTL_USER_AGENT,
        },
    )
    data = response.json()
    if attribute == "opensearch_enabled" and data["elasticsearch_enabled"]:
        return "WARNING: Enabling OpenSearch will disable ElasticSearch, which you are currently running."
    if attribute == "elasticsearch_enabled" and data["opensearch_enabled"]:
        return "WARNING: Enabling ElasticSearch will disable OpenSearch, which you are currently running."

    return None
