from contextlib import suppress
from os import scandir
from shutil import copytree, ignore_patterns, rmtree
from subprocess import check_call, check_output
from time import sleep

from invoke import task

from kamikaze3.common.shell import run_local_command
from kamikaze3.utils import async_pool

available_fixtures = [f.name for f in scandir("tasks/fixtures/") if f.is_dir()]
conf_cache = dict()


def run_shell_command_on_node(command, node_ssh, user="app", failure_message="Aborted"):
    return_code = run_local_command(
        "ssh -oStrictHostKeyChecking=no -o UserKnownHostsFile=/dev/null "
        "-i {} -p {} {}@{} {}".format(
            node_ssh["IdentityFile"],
            node_ssh["Port"],
            user,
            node_ssh["HostName"],
            command,
        ),
        shell=True,
    )
    if return_code != 0:
        raise RuntimeError(failure_message)


def run_vagrant(fixture=None):
    # NOTE(vdloo): This only works with the lxc vagrants because they have 'public' IPs
    # instead of the pre-routed 127.0.0.1 that VirtualBox uses if you're using private
    # networking.
    cmd = "cd 'tasks/fixtures/{}'; ./build.sh -1".format(fixture)
    if run_local_command(cmd, shell=True):
        raise RuntimeError(
            "Failed to run vagrant for fixture {}, "
            "retrieved nonzero exit status".format(fixture)
        )


def destroy_vagrant(fixture=None):
    cmd = "cd 'tasks/fixtures/{}/hypernode-vagrant'; vagrant destroy -f".format(fixture)
    if run_local_command(cmd, shell=True):
        raise RuntimeError(
            "Failed to destroy vagrant for fixture {}, "
            "retrieved nonzero exit status".format(fixture)
        )


def run_vagrants(fixture=None, proxy=False):
    print("Starting Vagrants for fixture {}".format(fixture))
    with async_pool() as pool:
        # Start the source vagrant
        pool.apply_async(run_vagrant, kwds={"fixture": fixture})
        # Wait a bit before we start the second Vagrant
        sleep(1)
        # Start the destination vagrant
        pool.apply_async(run_vagrant, kwds={"fixture": "clean"})
        if proxy:
            # Wait a bit before we start the third Vagrant
            sleep(1)
            # Start the bounce host vagrant
            pool.apply_async(run_vagrant, kwds={"fixture": "bouncer"})


def destroy_vagrants(fixture=None, proxy=False):
    print("Cleaning up Vagrants for fixture {}".format(fixture))
    with async_pool() as pool:
        # Destroy the source vagrant
        pool.apply_async(destroy_vagrant, kwds={"fixture": fixture})
        # Wait a bit before we destroy the second Vagrant
        sleep(1)
        # Destroy the destination vagrant
        pool.apply_async(destroy_vagrant, kwds={"fixture": "clean"})
        if proxy:
            # Wait a bit before we destroy the second Vagrant
            sleep(1)
            pool.apply_async(destroy_vagrant, kwds={"fixture": "bouncer"})


def get_vagrant_ssh_config(fixture):
    global conf_cache
    ssh_config_output = (
        check_output(
            "vagrant ssh-config",
            shell=True,
            cwd="tasks/fixtures/{}/hypernode-vagrant".format(fixture),
        )
        .decode("utf-8")
        .strip()
        .split("\n")
    )
    ssh_info = dict(map(lambda l: filter(None, l.split(" ")), ssh_config_output))
    ssh_info["Fixture"] = fixture
    conf_cache[fixture] = ssh_info
    return conf_cache[fixture]


def create_destination_key():
    destination_ssh = get_vagrant_ssh_config(fixture="clean")
    gen_key_command = (
        'bash -c \'echo -e "y\\n" | '
        "ssh-keygen -b 521 -t ecdsa "
        '-f /data/web/.ssh/id_ecdsa -q -N "" '
        "|| /bin/true'"
    )
    run_shell_command_on_node(
        gen_key_command,
        node_ssh=destination_ssh,
        failure_message="Failed to create SSH key on destination node",
    )


def create_destination_proxy_key():
    destination_ssh = get_vagrant_ssh_config(fixture="clean")
    # The hyperform connects to the bouncer using a different type of key (ecdsa)
    # than the one we generate for the user (rsa). This is a hack so that we can
    # get paramiko to autodetect the keys from the user .ssh directory and not
    # have to specify explicitly what key to use. This behavior is tested here
    # (but in reverse).
    # See: https://github.com/ByteInternet/hyperform/blob/master/provisioning/hyperform.yml#L125
    gen_key_command = (
        'bash -c \'echo -e "y\\n" | '
        "ssh-keygen -b 4096 -t rsa "
        '-f /data/web/.ssh/id_rsa -q -N "" '
        "|| /bin/true'"
    )
    run_shell_command_on_node(
        gen_key_command,
        node_ssh=destination_ssh,
        failure_message="Failed to create SSH proxy key on destination node",
    )


def get_destination_public_key():
    destination_ssh = get_vagrant_ssh_config(fixture="clean")
    return (
        check_output(
            "ssh -q -oStrictHostKeyChecking=no -o UserKnownHostsFile=/dev/null "
            "-i {} -p {} app@{} cat /data/web/.ssh/id_ecdsa.pub".format(
                destination_ssh["IdentityFile"],
                destination_ssh["Port"],
                destination_ssh["HostName"],
            ),
            shell=True,
        )
        .decode("utf-8")
        .strip("\n")
    )


def get_destination_proxy_public_key():
    destination_ssh = get_vagrant_ssh_config(fixture="clean")
    return (
        check_output(
            "ssh -q -oStrictHostKeyChecking=no -o UserKnownHostsFile=/dev/null "
            "-i {} -p {} app@{} cat /data/web/.ssh/id_rsa.pub".format(
                destination_ssh["IdentityFile"],
                destination_ssh["Port"],
                destination_ssh["HostName"],
            ),
            shell=True,
        )
        .decode("utf-8")
        .strip("\n")
    )


def authorize_key_on_source(public_key, fixture=None):
    source_ssh = get_vagrant_ssh_config(fixture=fixture)
    if run_local_command(
        "echo {} | ssh -oStrictHostKeyChecking=no -o UserKnownHostsFile=/dev/null "
        "-i {} -p {} app@{} 'cat >> /data/web/.ssh/authorized_keys'".format(
            public_key,
            source_ssh["IdentityFile"],
            source_ssh["Port"],
            source_ssh["HostName"],
        ),
        shell=True,
    ):
        raise RuntimeError("Failed to authorize key on source node")


def add_key_to_known_hosts(fixture=None):
    source_ssh = get_vagrant_ssh_config(fixture=fixture)
    destination_ssh = get_vagrant_ssh_config(fixture="clean")
    add_key_command = "ssh -p {} app@{} echo testing if we can log in".format(
        source_ssh["Port"], source_ssh["HostName"]
    )
    run_shell_command_on_node(
        add_key_command,
        node_ssh=destination_ssh,
        failure_message="Failed to connect to the source from the destination",
    )


def authorize_destination_on_proxy():
    print("Creating destination proxy key")
    create_destination_proxy_key()
    print("Getting destination proxy public key")
    public_proxy_key = get_destination_proxy_public_key()
    print("Authorizing proxy key on bounce host")
    authorize_key_on_source(public_proxy_key, fixture="bouncer")
    print("Adding proxy key to known hosts on destination")
    add_key_to_known_hosts(fixture="bouncer")


def authorize_destination_on_source(fixture=None):
    print("Creating destination key")
    create_destination_key()
    print("Getting destination public key")
    public_key = get_destination_public_key()
    print("Authorizing public key on source")
    authorize_key_on_source(public_key, fixture=fixture)
    print("Adding source key to known hosts on destination")
    add_key_to_known_hosts(fixture=fixture)


def attempt_import(
    fixture=None, single_threaded=False, set_maintenance_on_source=False
):
    source_ssh = get_vagrant_ssh_config(fixture=fixture)
    destination_ssh = get_vagrant_ssh_config(fixture="clean")
    attempt_import_command = (
        "{} --user app --port {} --host {} --path /data/web/public --set-default-url --verbose "
        "{} {}"
        "".format(
            "hypernode-importer",
            source_ssh["Port"],
            source_ssh["HostName"],
            "--single-threaded" if single_threaded else "",
            "--set-maintenance-on-source" if set_maintenance_on_source else "",
        )
    )
    print(attempt_import_command)
    run_shell_command_on_node(
        attempt_import_command,
        node_ssh=destination_ssh,
        failure_message="Importer exited nonzero! Import failed",
    )


def check_site_output(ssh_info, string_to_check):
    check_call(
        "ssh -q -oStrictHostKeyChecking=no -o UserKnownHostsFile=/dev/null "
        "-i {} -p {} app@{} 'curl -s "
        '-H "Host: {}.hypernode.local" localhost | '
        'grep -q "{}"\''.format(
            ssh_info["IdentityFile"],
            ssh_info["Port"],
            ssh_info["HostName"],
            ssh_info["Fixture"],
            string_to_check,
        ),
        shell=True,
    )


def check_imported():
    destination_ssh = get_vagrant_ssh_config(fixture="clean")
    check_site_output(destination_ssh, "shopping")


def check_maintenance(fixture=None):
    print("Checking if the source shop was set to maintenance mode")
    source_ssh = get_vagrant_ssh_config(fixture=fixture)
    check_site_output(source_ssh, "maintenance")


def install_current_revision():
    """
    Install the importer project in which this process is running
    in the 'clean' destination vagrant. Purges any installed version
    of the importer first.
    :return None:
    """
    destination_ssh = get_vagrant_ssh_config(fixture="clean")
    print("Copying current revision to the shared folder")
    with suppress(FileNotFoundError):
        rmtree("tasks/fixtures/clean/hypernode-vagrant/hypernode-kamikaze3")
    copytree(
        "./",
        "tasks/fixtures/clean/hypernode-vagrant/hypernode-kamikaze3",
        ignore=ignore_patterns("tasks"),
    )

    install_command = "bash /data/web/install.sh"
    run_shell_command_on_node(
        install_command,
        user="root",
        node_ssh=destination_ssh,
        failure_message="Failed to build current revision",
    )


def block_destination_on_source(fixture=None):
    """
    Block the destination node from connecting directly
    to the source node. Must go through proxy.
    :param str fixture: The vagrant to import from
    :return None:
    """
    source_ssh = get_vagrant_ssh_config(fixture=fixture)
    destination_ssh = get_vagrant_ssh_config(fixture="clean")
    block_ip_command = "iptables -A INPUT -s {} -j DROP".format(
        destination_ssh["HostName"]
    )
    run_shell_command_on_node(
        block_ip_command,
        user="root",
        node_ssh=source_ssh,
        failure_message="Failed to block IP of destination node on source node",
    )


def configure_proxy_on_destination():
    """
    Configure the SSH config of the destination node to use
    the proxy host for all SSH traffic
    :return None:
    """
    proxy_ssh = get_vagrant_ssh_config(fixture="bouncer")
    destination_ssh = get_vagrant_ssh_config(fixture="clean")
    configure_proxy_command = "bash /data/web/configure_proxy.sh " "-b {}".format(
        proxy_ssh["HostName"]
    )
    run_shell_command_on_node(
        configure_proxy_command,
        user="app",
        node_ssh=destination_ssh,
        failure_message="Failed to write SSH config on destination node to "
        "go through the proxy with all SSH traffic",
    )


def pretend_behind_firewall(fixture=None):
    """
    Pretend the importing node needs to go through a proxy host to
    import the shop so we make sure to test the service panel
    situation where people can whitelist an IP of a server that's
    actually a proxyhost through which the hypernodes and hyperform
    test imports are performed.
    :param str fixture: The vagrant to import from
    :return None:
    """
    block_destination_on_source(fixture=fixture)
    configure_proxy_on_destination()


@task(
    help={
        "fixture": "The type of shop to use in the importing test",
        "keep_nodes": "Do not kill Vagrants after test. Also reuse running existing ones.",
        "proxy": "Firewall the remote host and import through a proxy server",
        "single_threaded": "Do not run the database and filesystem import at the same time",
        "set_maintenance_on_source": "Set the original shop to maintenance, but not the new one",
    }
)
def test_hypernode_importer(
    ctx,
    fixture="magento1",
    keep_nodes=False,
    proxy=False,
    single_threaded=False,
    set_maintenance_on_source=False,
):
    if fixture not in available_fixtures:
        raise RuntimeError("No such fixture. Options: {}".format(available_fixtures))

    if not keep_nodes:
        for fixture_to_clean_up in available_fixtures:
            destroy_vagrants(fixture=fixture_to_clean_up)

    try:
        run_vagrants(fixture=fixture, proxy=proxy)
        authorize_destination_on_source(fixture=fixture)
        install_current_revision()

        if proxy:
            authorize_destination_on_proxy()
            pretend_behind_firewall(fixture=fixture)
        attempt_import(
            fixture=fixture,
            single_threaded=single_threaded,
            set_maintenance_on_source=set_maintenance_on_source,
        )
        print("\n")

        if set_maintenance_on_source:
            check_maintenance(fixture=fixture)
        else:
            check_imported()
    finally:
        if not keep_nodes:
            destroy_vagrants(fixture=fixture, proxy=proxy)
